module basicIO                                          ! Basic in-/output

! Routines:
! 
! err                    Displays error message and stops
! testAllocErr           If allocation has failed, displays error message and stops
! pathExists             Returns: whether a path (to file or directory) exists
! tryOpenFilRW           Tries to open a file for read/write or other action and outputs an error message if it does not exist
! sendMailFin            Sends final notification email

  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !=====================================================[ TYPE DEFINITIONS ]========================================================

  type tyStdOutUnits
    integer :: n,cons,log,err
  end type


  !========================================================[ CONSTANTS ]============================================================

  type(tyStdOutUnits),parameter :: &
   I_UNIT=tyStdOutUnits(3,0,1,2), &
   UNIT=tyStdOutUnits(I_UNIT%n,2**I_UNIT%cons,2**I_UNIT%log,2**I_UNIT%err)

  integer,parameter :: &
   FORT_UNIT(0:I_UNIT%n-1)=[6,2**I_UNIT%log,0]


  !========================================================[ VARIABLES ]============================================================

  character(1000) :: &
   mailAddrFin
  
  logical :: &
   filOutput, &
   usrQuiet


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine err(msg)
!
!   displays error message and stops
!
!   Arguments:
!     msg                            message string (should start with minuscle)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*),parameter :: &
     PREFX="BASE: fatal error: "
    character(*) :: &
     msg
    character(len_trim(adjustl(msg))+len(PREFX)) :: &
     cplMsg
    intent(in) :: &
     msg

    cplMsg=PREFX//adjustl(msg)

    write(FORT_UNIT(I_UNIT%err),*)
    write(FORT_UNIT(I_UNIT%err),*)
    write(FORT_UNIT(I_UNIT%err),"(A)") trim(cplMsg)

    if (filOutput) then
      write(FORT_UNIT(I_UNIT%log),*)
      write(FORT_UNIT(I_UNIT%log),"(A)") trim(cplMsg)
    end if

    if (len_trim(mailAddrFin) > 0) call sendMailFin(errMsg=trim(cplMsg))

    close(FORT_UNIT(I_UNIT%log))
    error stop

  end subroutine err




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine testAllocErr(allocStat,code)
!
!   if allocation has failed, displays error message and stops
!
!   Arguments:
!     allocStat                      allocation status
!     code                           unique error code
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     allocStat, &
     code
    character(128) :: &
     errMsg
    intent(in) :: &
     allocStat, &
     code

    write(errMsg,"(A,I4,A,I10,A)") "allocation of memory failed (code ",code,", ",allocStat,")"

  end subroutine testAllocErr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function pathExists(path)
!
!   Returns:                         whether a path (to file or directory) exists
!
!   Arguments:
!     path                           path in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     pathExists
    character(*) :: &
     path
    intent(in) :: &
     path

    inquire(file=path,exist=pathExists)

  end function pathExists




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine tryOpenFilRW(desc,path,filUnit,stat,action,access)
!
!   tries to open a file for read/write or other action and outputs an error message if it does not exist
!
!   Arguments:
!     desc                           file-type description
!     path                           file path
!     filUnit                        file unit
!     stat                           file status
!     action                         action for the file to be opened for
!     access                         access mode
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     filUnit
    logical :: &
     exists
    character(*) :: &
     path, &
     desc, &
     stat, &
     action, &
     access
    character(128) :: &
     act, &
     acc, &
     st
    optional :: &
     stat, &
     action, &
     access
    intent(in) :: &
     desc, &
     path, &
     filUnit, &
     stat, &
     action, &
     access

    if (present(action)) then
      act=action
    else
      act="readwrite"
    end if
    if (present(access)) then
      acc=access
    else
      acc="sequential"
    end if
    if (present(stat)) then
      st=stat
    else
      st="unknown"
    end if

    inquire(file=trim(path),exist=exists)

    if (.NOT. exists .AND. (trim(st) == "old" .OR. trim(st) == "OLD")) then
      call err("tried to access non-existing"//trim(" "//desc)//" file: "//path)
    else
      open(filUnit,file=trim(path),status=trim(st),action=trim(act),access=trim(acc))
    end if

  end subroutine tryOpenFilRW




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine sendMailFin(errMsg,hostName,runIdent)
!
!   sends final notification email
!
!   Arguments:
!     errMsg                         optional error message
!     hostName                       name of host BASE has run on
!     runIdent                       run identifier
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     exitStat
    character(*) :: &
     errMsg, &
     hostName, &
     runIdent
    character(1000) :: &
     subj, &
     body
    optional :: &
     errMsg, &
     hostName, &
     runIdent
    intent(in) :: &
     errMsg, &
     hostName, &
     runIdent

    if (present(errMsg)) then
      subj="BASE exited with error"
      if (present(hostName)) subj=trim(subj)//" on "//hostName
      body=errMsg
    else
      subj="BASE finished"
      if (present(hostName)) subj=trim(subj)//" on "//hostName
      if (present(runIdent)) then
        body="Run identifier: "//trim(runIdent)
      else
        body=""
      end if
    end if

    call execute_command_line("echo '"//trim(body)//"' | mail -s '"//trim(subj)//"' "//trim(mailAddrFin),exitstat=exitStat)
    if (exitStat == 0) then
      if (.NOT. usrQuiet) write(FORT_UNIT(I_UNIT%cons),"(A)") "Email sent to "//trim(mailAddrFin)
      if (filOutput) write(FORT_UNIT(I_UNIT%log),"(A)") "Email sent to "//trim(mailAddrFin)
    else
      if (.NOT. usrQuiet) write(FORT_UNIT(I_UNIT%err),"(A)") "Sending mail to "//trim(mailAddrFin)//" failed"
      if (filOutput) write(FORT_UNIT(I_UNIT%log),"(A)") "Sending mail to "//trim(mailAddrFin)//" failed"
    end if

  end subroutine sendMailFin




end module
