module random                                            ! Random numbers

! Routines:
! 
! iniRNG                 Initialises random-number generator
! rndInt4                Returns: pseudo-random integer within a given range
! rndInt8                Returns: pseudo-random integer within a given range
! rndNorm                Returns: sample from Normal(0,1)
! fillRndNormNonZero     Fills array with samples from Normal(0,1)
! rndUni                 Returns: sample from a uniform distribution over (0,1)

  use basicTypes
  use io
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !========================================================[ VARIABLES ]============================================================

  logical :: &
   keepSeed


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  interface rndInt
    module procedure rndInt4
    module procedure rndInt8
  end interface

  interface rndPermute
    module procedure rndPermuteInt
    module procedure rndPermuteReal
    module procedure rndPermuteDbl
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !=====================================================[ INITIALISATION ]==========================================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine iniRNG
!
!   initialises random-number generator
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer,parameter :: &
     SEED_SIZ=12, &                  ! builtin proc. requires this number
     INT_SIZ=4
    integer :: &
     seed(SEED_SIZ), &
     readStat, &
     i, &
     exitStat, &
     j, &
     seedSiz
    character(INT_SIZ*SEED_SIZ) :: &
     lin
    character(STR_LEN_STD) :: &
     cErrMsg

    if (.NOT. keepSeed) then
      exitStat=unlink(FIL_NM_RND_SEED)
      call execute_command_line("dd if=/dev/urandom of="//FIL_NM_RND_SEED//" bs=1 count="//trim(niceFmt(INT_SIZ*SEED_SIZ))// &
       " 1>&2 2>/dev/null",exitstat=exitStat)
      if (exitStat /= 0) call err("executing dd failed")
    end if
    call tryOpenFilRW("",FIL_NM_RND_SEED,UN_RND,"old","read")
    read(UN_RND,"(A)",iostat=readStat,err=999,iomsg=cErrMsg) lin
    close(UN_RND)
    seed=-huge(seed)
    do j=1,SEED_SIZ
      do i=1,INT_SIZ
        seed(j)=seed(j)+iachar(lin((j-1)*INT_SIZ+i:(j-1)*INT_SIZ+i))*256**(i-1)
      end do
    end do

    seedSiz=SEED_SIZ
    call random_seed(size=seedSiz)   ! actual arg. needs to be variable
    call random_seed(put=seed)

    return

999 call err("file could not be read ("//trim(cErrMsg)//"): "//FIL_NM_RND_SEED)

  end subroutine iniRNG




  !=====================================================[ RANDOM NUMBERS ]==========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function rndInt4(min,max)
!
!   Returns:                         pseudo-random integer within a given range
!
!   Arguments:
!     min                            minimum return value
!     max                            maximum return value
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     min, &
     max, &
     rndInt4
    real(dp) :: &
     rndTmp
    intent(in) :: &
     min, &
     max

    rndTmp=rndUni()

    rndInt4=min+floor((max-min+1)*rndTmp)

  end function rndInt4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function rndInt8(min,max)
!
!   Returns:                         pseudo-random integer within a given range
!
!   Arguments:
!     min                            minimum return value
!     max                            maximum return value
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     min, &
     max, &
     rndInt8
    real(dp) :: &
     rndTmp
    intent(in) :: &
     min, &
     max

    rndTmp=rndUni()

    rndInt8=min+floor((max-min+1)*rndTmp)

  end function rndInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function rndNorm()
!
!   Returns:                         sample from Normal(0,1)
!
!   from [324]
!   CAUTION: not thread-safe!
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iset=0
    real(dp) :: &
     rndNorm, &
     fac, &
     gset, &
     rsq, &
     v1, &
     v2, &
     rndTmp1, &
     rndTmp2
    save :: &
     iset, &
     gset

    if (iset == 0) then
      do
        call random_number(rndTmp1)
        call random_number(rndTmp2)
        v1=2*rndTmp1-1._dp
        v2=2*rndTmp2-1._dp
        rsq=v1**2+v2**2
        if (.NOT.(rsq >= 1 .OR. rsq == 0)) exit
      end do
      fac=sqrt(-2*log(rsq)/rsq)
      gset=v1*fac
      rndNorm=v2*fac
      iset=1
    else
      rndNorm=gset
      iset=0
    end if

  end function rndNorm




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine fillRndNormNonZero(v)
!
!   fills array with samples from Normal(0,1)
!
!   Arguments:
!     v                              array
!
!   from [324]
!   result represents a sample from a uniform density over the unit sphere of the given dimensionality
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i, &
     N
    real(dp) :: &
     fac, &
     rsq, &
     v1, &
     v2, &
     rndTmp1, &
     rndTmp2, &
     v(:), &
     rsqInv
    intent(inout) :: &
     v

    N=size(v)
    do i=1,ceiling(N/2.)
      do
        call random_number(rndTmp1)
        call random_number(rndTmp2)
        v1=2._dp*rndTmp1-1._dp
        v2=2._dp*rndTmp2-1._dp
        rsq=v1**2+v2**2
        if (.NOT.(rsq >= 1 .OR. v1 == 0 .OR. v2 == 0)) exit
      end do
      rsqInv=1._dp/rsq
      fac=sqrt(2._dp*log(rsqInv)*rsqInv)
      v(2*i-1)=v1*fac
      v(min(2*i,N))=v2*fac
    end do

  end subroutine fillRndNormNonZero




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function rndUni()
!
!   Returns:                         sample from a uniform distribution over (0,1)
!
!   thread-safe
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     rndUni

    call random_number(rndUni)       ! thread-safe

  end function rndUni




  !===================================================[ RANDOM PERMUTATION ]========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    recursive function rndPermuteInt(arr,isIndxArr)
!
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     arr(1:), &
     N, &
     i, &
     j, &
     rndPermuteInt(size(arr))
    logical :: &
     isIndxArr, &
     empty(size(arr))
    optional :: &
     isIndxArr
    intent(in) :: &
     arr, &
     isIndxArr

    N=size(arr,1)

    if (present(isIndxArr) .AND. isIndxArr) then
      empty=.TRUE.
      do i=1,N
        do
          j=rndInt(1,N)
          if (empty(j)) then
            rndPermuteInt(j)=arr(i)
            empty(j)=.FALSE.
            exit
          end if
        end do
      end do
    else
      block
        integer :: &
         iArr(size(arr))

        do i=1,N
          iArr(i)=i
        end do
        iArr=rndPermute(iArr,.TRUE.)
        rndPermuteInt=arr(iArr)
      end block
    end if

  end function rndPermuteInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    recursive function rndPermuteReal(arr,isIndxArr)
!
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     arr(1:), &
     rndPermuteReal(size(arr))
    integer :: &
     N, &
     i, &
     j
    logical :: &
     isIndxArr, &
     empty(size(arr))
    optional :: &
     isIndxArr
    intent(in) :: &
     arr, &
     isIndxArr

    N=size(arr,1)

    if (present(isIndxArr) .AND. isIndxArr) then
      empty=.TRUE.
      do i=1,N
        do
          j=rndInt(1,N)
          if (empty(j)) then
            rndPermuteReal(j)=arr(i)
            empty(j)=.FALSE.
            exit
          end if
        end do
      end do
    else
      block
        integer :: &
         iArr(size(arr))

        do i=1,N
          iArr(i)=i
        end do
        iArr=rndPermute(iArr,.TRUE.)
        rndPermuteReal=arr(iArr)
      end block
    end if

  end function rndPermuteReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    recursive function rndPermuteDbl(arr,isIndxArr)
!
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     arr(1:), &
     rndPermuteDbl(size(arr))
    integer :: &
     N, &
     i, &
     j
    logical :: &
     isIndxArr, &
     empty(size(arr))
    optional :: &
     isIndxArr
    intent(in) :: &
     arr, &
     isIndxArr

    N=size(arr,1)

    if (present(isIndxArr) .AND. isIndxArr) then
      empty=.TRUE.
      do i=1,N
        do
          j=rndInt(1,N)
          if (empty(j)) then
            rndPermuteDbl(j)=arr(i)
            empty(j)=.FALSE.
            exit
          end if
        end do
      end do
    else
      block
        integer :: &
         iArr(size(arr))

        do i=1,N
          iArr(i)=i
        end do
        iArr=rndPermute(iArr,.TRUE.)
        rndPermuteDbl=arr(iArr)
      end block
    end if

  end function rndPermuteDbl




end module
