module dataModelsNoise                              ! Data, models, and noise

! Routines:
! 
! undefParsInt           Sets a var. undefined
! undefDQsInt            Sets a var. undefined
! isUndefParsInt         Returns: whether argument is undefined
! isUndefDQsInt          Returns: whether argument is undefined
! isDefParsInt           Returns: whether a var. is defined
! isDefDQsInt            Returns: whether a var. is defined
! setIfUndefParsInt      Sets variable to a given value if variable is undefined
! setIfUndefDQsInt       Sets variable to a given value if variable is undefined
! parseData              Parses observational data
! groupData              Arranges data into groups
! groupDataSet           Arranges data into groups (see groupData)
! warnDataSet            Displays warning message referring to specific data set
! errDataSet             Displays error message referring to specific data set and stops
! generateData           Generates synthetic data
! sampTimes              Returns: samples a set of observing times for synthetic data
! conformsWithWindows    Returns: whether given t conforms with all data-generation time windows referring to t1
! fmtDtSet               Returns: formatted string identifying a data set
! fmtObsvble             Returns: formatted string with name and unit of an observable
! printDtMdlProps        Prints relevant properties of data and (stellar) model
! preCalcEA              Pre-calculates values of E for e in [eMin,eMin+eDiff) and M in [0,pi)
! ecAn                   Returns: eccentric anomalies for given pars. and times
! auxMdlFctAAngArcsec    Returns: astrometric angular separation [arcsec]
! auxMdlFctDel           Returns: astrometric model component
! auxMdlFctRV            Returns: radial velocity [m/s], including only the non-systemic components
! mdlFctNMdRV            Returns: values of the RV model function in normal mode
! mdlFctNMdAMHipN        Returns: values of the AMHipN model function [mas] in normal mode
! mdlFctNMdAMHipNOnly    Returns: values of the AMHipN model function [mas] in normal mode
! mdlFctBMdAMAngPos      Returns: values of the AMAngPos model function (delta, alpha·cos(delta)) [mas] in binary mode
! mdlFctBMdAMAngPosOnly  Returns: values of the AMAngPos model function (delta, alpha·cos(delta)) [mas] in binary mode
! mdlFctBMdRV            Returns: values of the RV model function (binary mode)
! calcMdlFctWErr         Samples model function at given times and adds errors if the corresponding uncertainties are given
! setMdlFctPtrs          Sets pointers to relevant model functions
! setParsDQs             Sets important properties of pars. and DQs (called once)
! plStr                  Returns: formatted string identifying #-pl. mode
! fmtPar                 Returns: formatted string with par. and its unit
! fmtDQ                  Returns: formatted string with DQ and its unit
! printParsDQsTbl        Prints a table listing the pars. and DQs for all allowed combinations of data and normal/binary mode
! setPri                 Sets the support and other properties of the prior of a par.
! setPriFromFil          Sets normalised prior points from file, approximating the prior by trapezoids between the points
! setPriDiffFromFil      Sets differences between adjacent prior points from a file
! modJefPriKnee          Returns: knee of the modified Jeffreys prior for a given upper prior bound
! parOutsdPriSupp        Returns: whether parameter is outside its prior support (i.e. its prior is zero)
! priorIsNonZero         Returns: whether the prior of a par. vector is non-zero
! singlePrior            Returns: the prior of a single parameter
! prior                  Returns: prior of a parameter vector
! logPrior               Returns: natural logarithm of prior of a parameter vector
! disprsOvrPriIntvl      Returns: samples dispersed over the prior interval using rejection sampling
! fmtPrior               Returns: formatted string for prior of a parameter
! pltRge                 Returns: modification of given range of a parameter or DQ suitable for plotting
! printPriRges           Prints prior ranges of all pars.
! logLikh                Returns: ln(likelihood) minus const., chi², residuals, or normalised residuals

  use basicTypes
  use etc
  use userOpts
  use random
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !======================================================[ CONSTANTS (1) ]==========================================================

  integer,parameter :: &
   N_PL_MAX=9, &
   N_N_PL_MODES_MAX=1+N_PL_MAX, &
   EA_PRECALC_SIZE=5000, &           ! size in each dimension of array of precalcd. eccentric anomaly values (dp: 8 bytes)
   N_DT_TYPES=3, &
   N_DT_SETS_PER_TY_MAX=20, &
   N_DT_SETS_MAX=N_DT_TYPES*N_DT_SETS_PER_TY_MAX, &
   N_REC_PER_SET_MAX=10000, &
   N_REC_MAX=N_REC_PER_SET_MAX*N_DT_SETS_MAX, &
   N_PRI_TY=6, &
   N_GROUP_THRLDS(3)=[2,1,3]         ! AMAngPos, RV, AMHipN


  ! GENERATING DATA

  integer,parameter :: &
   GEN_D_N_USR_CMDS_MAX=50, &
   GEN_D_N_CMDS_MAX=50, &
   GEN_D_N_WINDOWS_MAX=9

  character(*),parameter :: &
   GEN_D_NM_STD="virtual system"


  !=====================================================[ TYPE DEFINITIONS ]========================================================

  type tySummJitt
    real(dp) :: woJitt,wJitt
  end type


  ! DATA PROPERTIES

  type tyDtPropsGeneric
    integer :: &
     nRec                                               ! # data
    real(dp) :: &
     obsStdDev
    type(tyIntvlDbl) :: &
     tRge
    type(tySummJitt) :: &
     normResid(N_DT_TYPES*N_DT_SETS_PER_TY_MAX*N_REC_PER_SET_MAX), &
     normResidMean, &
     normResidStdDev, &
     normResidPVal
  end type

  type,extends(tyDtPropsGeneric) :: tyDtProps1D
    real(dp) :: &
     obsDiff2SgmPerSetMax
    type(tyIntvlDbl) :: &
     obsRge, &
     obsRge2Sgm, &
     obsRge2SgmSect, &
     obsRgeCompo(2), &
     obsRge2SgmCompo(2)
  end type

  type,extends(tyDtPropsGeneric) :: tyDtProps2D
  end type

    ! props. of all data

  type,extends(tyDtPropsGeneric) :: tyDtPropsAll
    integer :: &
     nSets, &
     nSetsOfTy(N_DT_TYPES), &
     dtTyOfSets(N_DT_SETS_MAX), &
     onlyDtTy, &
     binCompo(N_DT_SETS_MAX), &
     nBObs, &
     nBObsWDtTy(N_DT_TYPES)
    logical :: &
     refIsCompa(N_DT_SETS_MAX), &    ! assumed false where information missing
     polCoord(N_DT_SETS_MAX), &      ! assumed false where information missing
     onlyExtRef, &
     onlyRV, &
     timesAreJD
    character(MAX_LEN_PATH) :: &
     filNms(N_DT_SETS_MAX)
    character(STR_LEN_STD) :: &
     sysNm
    type(tyIntvlDbl) :: &
     sngTimeSpansRge
  end type

    ! specific data types

  type,extends(tyDtProps2D) :: tyDtPropsAMAngPos
    type(tyIntvlDbl) :: &
     sepRge
  end type

  type,extends(tyDtProps1D) :: tyDtPropsAMHipN
  end type

  type,extends(tyDtProps1D) :: tyDtPropsRV
  end type


  ! BINARY DATA FILE FORMATS

  ! AMHipN data (files from IntermediateData/absrec/ directory on DVD accompanying [456])
  type tyDtFilHipNHdr
    integer(i4b) :: &
     hipNr, &            ! Hipparcos object #
     nData, &            ! # data
     dRA, &              ! decimal part of alpha [.01 mas]
     dDEC, &             ! decimal part of delta [.01 mas]
     plx, &              ! ref. parallax [.01 mas]
     muAlCosDel, &       ! ref. PM in alpha* [.01 mas/yr]
     muDel               ! ref. PM in delta [.01 mas/yr]
    integer(i2b) :: &
     iRA, &              ! integer part of alpha [°]
     iDEC                ! integer part of (delta+90°) [°]
  end type

  type tyDtFilHipNDt
    integer(i4b) :: &
     da, &               ! abscissa residual [.01 mas]
     dmy1(2), &          ! (dummy)
     BJD, &              ! barycentric JD [1E-5 d]
     dmy2, &             ! (dummy)
     psi, &              ! scan-orientation angle [1E-5 rad]
     dmy3                ! (dummy)
    integer(i2b) :: &
     gamma, &            ! parallax factor [1E-4]
     sgmDa, &            ! "formal error" on abscissa residual [.01 mas]
     dmy4(16)            ! (dummy)
  end type

  type tyDtFilHipN
    type(tyDtFilHipNHdr) :: hdr
    type(tyDtFilHipNDt) :: dt
  end type


  ! DATA TYPES
  
  ! when changing data types:
  ! * increase N_DT_TYPES
  ! * extend/change procs.:
  !   - logLikh
  !   - setParsDQs
  !   - calcDtCRC
  !   - calc[All]Predict
  !   - printHelp
  !   - groupData
  ! * extend tyMdlProps
  ! * extend N_GROUP_THRLDS
  ! * mind that earlier posterior samples saved with this set of data types!
  ! * …
  type tyDtTyInt
    integer :: AMAngPos,RV,AMHipN
  end type

  type tyDtTyReal
    real :: AMAngPos,RV,AMHipN
  end type

  type tyDtTyDbl
    real(dp) :: AMAngPos,RV,AMHipN
  end type

  type tyDtTyIntvlDbl
    type(tyIntvlDbl) :: AMAngPos,RV,AMHipN
  end type

  type tyDtTyBool
    logical :: AMAngPos,RV,AMHipN
  end type

  type tyDtTyStr
    character(STR_LEN_STD) :: AMAngPos,RV,AMHipN
  end type


  ! NON-BINARY DATA SETS

    ! data commands
  
  type tyDtCmdsStr
    character(STR_LEN_STD) :: sysNm,dtTy,polCoord,binCompo,mS,refIsCompa,nBObs,uniTm,withAMAngPos,withRV,window,times
  end type

    ! model-specific additional info required by model functions (besides tyParamsMdl, which is not model-specific)

  type tyDtParamsAMHipN
    real(dp) :: &
     psi(N_REC_PER_SET_MAX), &       ! scan-orientation angle [rad]
     cosPsi(N_REC_PER_SET_MAX), &    ! cos(psi)
     sinPsi(N_REC_PER_SET_MAX), &    ! sin(psi)
     gamma(N_REC_PER_SET_MAX)        ! parallax factor
    type(tyCoord2DDbl) :: &
     pos0, &                         ! reference position (delta_0, alpha_0·cos(delta_0)) as given in data file
                                     ! pos0%x = delta_0 [deg];
                                     ! pos0%y = alpha_0·cos(delta_0) [deg]
     mu0, &                          ! reference proper motion (mu_{delta,0}, mu_{alpha,0}·cos(delta)) as given in data file
                                     ! mu0%x = mu_{delta,0} [deg/d]
                                     ! mu0%y = mu_{alpha,0}·cos(delta) [deg/d]
     pos0AtT(N_REC_PER_SET_MAX)      ! ref. position (delta_0, alpha_0·cos(delta_0)) [deg] at time of measurement according to data file
    real(dp) :: &
     pi0                             ! reference parallax pi_0 [arcsec] as given in data file
  end type
  
    ! data sets

  type tyDtSetGeneric
    integer :: &
     N, &                            ! # data records
     binCompo                        ! # binary companion observed (1: primary; 2: secondary; or undefined)
    integer(i8b) :: &
     iUniqT(N_REC_PER_SET_MAX)       ! indexes at which timestamps can be found in uniqT array
    real(dp) :: &
     t(N_REC_PER_SET_MAX), &         ! time (Julian Date without offset)
     tMnEpoch(N_REC_PER_SET_MAX), &  ! t - epoch
     groupThrld(maxval(N_GROUP_THRLDS))
    type(tyIntvlDbl) :: &
     tRge
  end type

  type,extends(tyDtSetGeneric) :: tyDtSet1D
    real(dp) :: &
     obs(N_REC_PER_SET_MAX), &       ! values of observable
     sgm(N_REC_PER_SET_MAX), &       ! standard uncertainties of observable
     sgmSq(N_REC_PER_SET_MAX), &     ! (standard uncertainties of observable)²
     resid(N_N_PL_MODES_MAX,N_REC_PER_SET_MAX)   ! observable residuals
    type(tyIntvlDbl) :: &
     obsRge, &                       ! range of values of observable
     obsRge2Sgm                      ! range of values of observable, extended by ±sgm
                                     ! (using the sgm's corresponding to the extremal values of observable)
  end type

  type,extends(tyDtSetGeneric) :: tyDtSet2D
    type(tyCoord2DDbl) :: &
     obs(N_REC_PER_SET_MAX), &                 ! values of observable
     resid(N_N_PL_MODES_MAX,N_REC_PER_SET_MAX) ! the residuals using the best par. summaries
    type(tyUncEllDbl) :: &
     unc(N_REC_PER_SET_MAX)                    ! position uncertainty ellipses:
                                               ! posUnc%a = semi-major axis of uncertainty ellipse [mas]
                                               ! posUnc%b = semi-minor axis of uncertainty ellipse [mas]
                                               ! posUnc%phi = (counter-clockwise) angle of semi-major axis w.r.t. north celestial pole [rad]; range: [0,pi)
  end type

    ! sets of specific data types

  type,extends(tyDtSet2D) :: tyDtSetAMAngPos ! AM: angular positions: rectangular and polar
                                             ! obs%x = 1st component = (pol=.FALSE.: Delta(delta) [mas]; else: theta [°] (range: [0,360)))
                                             ! obs%y = 2nd component = (pol=.FALSE.: Delta(alpha·cos(delta)) [mas]; else: rho [mas])
                                             ! where alpha = right ascension; delta = declination; rho: distance;
                                             ! theta: (counter-clockwise) angle w.r.t. north celestial pole
    logical :: &
     pol, &                                  ! true if polar coord.
     refIsCompa                              ! true if reference point is companion (duplicated from dtProps)
    real(dp) :: &
     sep(N_REC_PER_SET_MAX)                  ! angular separation between components [mas] (identical to rho for polar coords.)
    type(tyIntvlDbl) :: &
     sepRge                                  ! angular-separation range
  end type

  type,extends(tyDtSet1D) :: tyDtSetAMHipN ! AM: Hipparcos new-reduction abscissa residuals
                                           ! obs, sgm, resid [mas]
    integer :: &
     hipNr                                 ! Hipparcos no.
    type(tyDtParamsAMHipN) :: &
     params                                ! additional parameters specific to AMHipN data
  end type

  type,extends(tyDtSet1D) :: tyDtSetRV ! RV
                                       ! obs, sgm, resid [m/s]
  end type


  ! GENERATING DATA

  type,extends(tyDtPropsAll) :: tyGenDtProps
    integer :: &
     dtTy, &                         ! type of data to be generated (component tyDtPropsAll::nSetsOfTy only used to determine model pars.)
     nWin
    real(dp) :: &
     win(GEN_D_N_WINDOWS_MAX), &
     winFrac(GEN_D_N_WINDOWS_MAX), &
     epoch
    real(dp),allocatable :: &
     times(:)
    logical :: &
     uniTm
  end type


  ! MODEL PROPERTIES

  type tyMdlPropsDtTy
    procedure(mdlFctNMdRV),pointer,nopass :: mdlFct
  end type
  type tyMdlProps
    real(dp) :: &
     mS, &                           ! mass of star [M_J]
     mSinIMax(N_N_PL_MODES_MAX), &   ! max. planet mass times sin(i) [M_J]
     epoch
    logical :: &
     bin, &
     RVAccel
    type(tyMdlPropsDtTy) :: AMAngPos,AMHipN,RV
  end type
  type tyParamsMdl
    integer :: &
     psdNPl, &
     iSet, &
     binCompo=VAL_UNDEF_INT
  end type


  ! PARS./DQS

  ! CAUTION:
  ! when adding a parameter, do the following:
  ! * increase N_G_PARS
  ! * extend G_PRI_TYPES array
  ! * extend tyParsInt type
  ! * extend VAL_UNDEF_TY_PARS_INT
  ! * extend G_I_PAR_ROLE and many other arrays
  ! * extend G_PAR_SCP_BY_COMB by appending a set of 16 lines pertaining to the new par.
  ! * extend procs.:
  !   - isUndefParsInt
  !   - setParRole (internal to setParsDQs)
  ! when adding a DQ, do the following:
  ! * increase N_G_DQS
  ! * extend tyDQsInt type
  ! * extend VAL_UNDEF_TY_DQS_INT
  ! * extend G_I_DQ_ROLE and many other arrays
  ! * extend G_DQ_SCP_BY_COMB by appending a set of 16 lines pertaining to the new DQ
  ! * extend procs.:
  !   - isUndefDQsInt
  !   - setDQRole (internal to setParsDQs)
  !   - maxDQRge
  type tyQtyScps
    integer :: no,pl,sD,s1,sC        ! none, planet, system (per data set), system (once), system (if there are companions)
  end type
  type tyParsInt                     ! CAUTION: DON'T CHANGE ORDER of existing components (would destroy, e.g., G_PAR_SCP_BY_COMB)
    integer :: e,f,chi,omegaPeri,K,V,sgmPlus,omega2,K1,K2,i,OmegaNode,varpi,aRelAng,aAng,alphaR,deltaR,muAlCosDel,muDel,tauPlus,aV
  end type
  type tyDQsInt                      ! CAUTION: DON'T CHANGE ORDER of existing components (would destroy, e.g., G_DQ_SCP_BY_COMB and calc. of DQs)
    integer :: P,T,mMin,aSinI,aPMax,KAlt,KAlt1,KAlt2,rho,m1,m2,d,aRel,a1,a2,thetaR,thetaSecR,tT1,tT2,omSum,omDiff,om2Sum,om2Diff,m
  end type

  ! CAUTION: when adding/removing prior types, change N_PRI_TY and arrays whose upper bound depends on this
  type tyPriTyInt
    integer :: uni,jef,modJef,fil,sgnModJef,norm
  end type


  !======================================================[ CONSTANTS (2) ]==========================================================

  real(dp),parameter :: &
   LOG_BOUNDS_RATIO_LOG_SCL_MIN=2._dp, &
   STD_EA_UNC_GOAL=PI*DIFF_EPS_DP, & ! 1e-9 ! cf. [01.05.2010/1]
   HIP_N_MEAN_EPOCH=2448349._dp, &   ! Julian Date, equiv. to Julian year J1991.25 [456]
   JITT_DEFAULT_UBOUND_STDEV_RATIO=3._dp

  character(*),parameter :: &
   SYS_NM_PREFX_HIP="HIP ", &
   DT_TY_CMDS(N_DT_TYPES)=[character(7) :: "ang_pos","RV",repeat(VAL_UNDEF_CHR,7)], &
   DT_TY_NMS(N_DT_TYPES)=[character(10) :: "AM_ang_pos","RV","AM_Hip_new"], &
   DT_TY_DESC(N_DT_TYPES)=[character(55) :: "astrometry (angular positions)","radial velocities", &
   "astrometry (Hipparcos new-reduction abscissa residuals)"], &
   OBSVBLE_NM(N_DT_TYPES,0:2)=reshape([character(17) :: "angular position","v_r","abscissa residual","delta","v_r","Delta(a)", &
   "alpha*cos(delta)",repeat(VAL_UNDEF_CHR,17),repeat(VAL_UNDEF_CHR,17)],[N_DT_TYPES,3]), &
   OBSVBLE_NM_TEX(N_DT_TYPES,0:2)=reshape([character(26) :: "\mathrm{angular position}","v_\mathrm{r}", &
   "\mathrm{abscissa residual}","\delta","v_\mathrm{r}","\Delta a","\alpha\cos\delta",repeat(VAL_UNDEF_CHR,26), &
   repeat(VAL_UNDEF_CHR,26)],[N_DT_TYPES,3]), &
   OBSVBLE_UNIT(N_DT_TYPES,0:2)=reshape([character(3) :: "mas","m/s","mas","mas","m/s","mas","mas",repeat(VAL_UNDEF_CHR,3), &
   repeat(VAL_UNDEF_CHR,3)], [N_DT_TYPES,3]), &
   OBSVBLE_UNIT_TEX(N_DT_TYPES,0:2)=reshape([character(27) :: "\mathrm{mas}","\frac{\mathrm m}{\mathrm s}", &
   "\mathrm{mas}","\mathrm{mas}","\frac{\mathrm m}{\mathrm s}","\mathrm{mas}","\mathrm{mas}",repeat(VAL_UNDEF_CHR,27), &
   repeat(VAL_UNDEF_CHR,27)], [N_DT_TYPES,3]), &
   PRI_TY_ID(0:N_PRI_TY-1)=["u","j","m",VAL_UNDEF_CHR,"s","n"]


  ! DATA TYPES

  type(tyDtTyInt),parameter :: &
   DT_TY=tyDtTyInt(1,2,3)        ! AMAngPos, RV, AMHipN

  type(tyDtTyStr),parameter :: &
   OBSVBLE_DESC=tyDtTyStr("angular position","radial velocity","abscissa residual")


  ! NON-BINARY DATA SETS

    ! data commands
  
  type(tyDtCmdsStr),parameter :: &
   DT_CMDS=tyDtCmdsStr("name=","type=","coord_sys=polar","bin_comp=","M=","ref_pt=companion","n_b_obs=","unif_time", &
   "with_ang_pos","with_RV","window","times")


  ! PARS./DQS
  
  type(tyParsInt),parameter :: &
   G_I_PAR_ROLE=tyParsInt(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21), &
   VAL_UNDEF_TY_PARS_INT=tyParsInt(VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT, &
   VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT, &
   VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT)

  type(tyDQsInt),parameter :: &
   G_I_DQ_ROLE=tyDQsInt(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24), &
   VAL_UNDEF_TY_DQS_INT=tyDQsInt(VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT, &
   VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT, &
   VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT,VAL_UNDEF_INT)

  type(tyPriTyInt),parameter :: &
   PRI_TY=tyPriTyInt(0,1,2,3,4,5)

  type(tyQtyScps),parameter :: &
   QTY_SCP=tyQtyScps(0,1,2,3,4)
  
  integer,parameter :: &
   N_G_PARS=21, &                    ! # global parameters (the set of global pars. contains each par. like e, f, chi once)
   N_L_PARS_MAX=N_G_PARS*(1+N_DT_SETS_MAX+N_PL_MAX), &
   BIN_COMPO_AM_ANG_POS_STD=2, &
   N_CALC_PROB_MAX=50, &
   MOD_JEF_PRI_KNEE_SCL_DIST=3, &    ! distance between upper bound and knee in logarithmic (10) scale for mod. Jef. priors
   DISPRS_N_TRIES=100, &
   PRI_N_PTS_MAX=1001, &
   PRI_N_PTS_MIN=499, &
   N_G_DQS=24, &
   N_L_DQ_MAX=0+6*N_PL_MAX, &
   N_L_PARS_OR_DQ_MAX=max(N_L_PARS_MAX,N_L_DQ_MAX), &

   G_PRI_TYPES(N_G_PARS)=[ &         ! types of priors
   PRI_TY%uni, &                     ! e
   PRI_TY%jef, &                     ! f
   PRI_TY%uni, &                     ! chi
   PRI_TY%uni, &                     ! omega
   PRI_TY%modJef, &                  ! K
   PRI_TY%uni, &                     ! V
   PRI_TY%modJef, &                  ! sigma_+
   PRI_TY%uni, &                     ! omega_2
   PRI_TY%modJef, &                  ! K_1
   PRI_TY%modJef, &                  ! K_2
   PRI_TY%uni, &                     ! i
   PRI_TY%uni, &                     ! Omega
   PRI_TY%jef, &                     ! pi
   PRI_TY%jef, &                     ! a'_rel
   PRI_TY%jef, &                     ! a'
   PRI_TY%uni, &                     ! alpha_r
   PRI_TY%uni, &                     ! delta_r
   PRI_TY%sgnModJef, &               ! mu_acd
   PRI_TY%sgnModJef, &               ! mu_d
   PRI_TY%modJef, &                  ! tau_+
   PRI_TY%sgnModJef], &              ! a_v

   G_I_PAR_ROLE_ORDERED(N_G_PARS)=[ &
   G_I_PAR_ROLE%V, &
   G_I_PAR_ROLE%aV, &
   G_I_PAR_ROLE%sgmPlus, &
   G_I_PAR_ROLE%varpi, &
   G_I_PAR_ROLE%alphaR, &
   G_I_PAR_ROLE%deltaR, &
   G_I_PAR_ROLE%muAlCosDel, &
   G_I_PAR_ROLE%muDel, &
   G_I_PAR_ROLE%tauPlus, &
   G_I_PAR_ROLE%e, &
   G_I_PAR_ROLE%f, &
   G_I_PAR_ROLE%chi, &
   G_I_PAR_ROLE%omegaPeri, &
   G_I_PAR_ROLE%omega2, &
   G_I_PAR_ROLE%i, &
   G_I_PAR_ROLE%OmegaNode, &
   G_I_PAR_ROLE%K, &
   G_I_PAR_ROLE%K1, &
   G_I_PAR_ROLE%K2, &
   G_I_PAR_ROLE%aRelAng, &
   G_I_PAR_ROLE%aAng], &

   G_I_DQ_ROLE_ORDERED(N_G_DQS)=[ &
   G_I_DQ_ROLE%P, &
   G_I_DQ_ROLE%T, &
   G_I_DQ_ROLE%thetaR, &
   G_I_DQ_ROLE%thetaSecR, &
   G_I_DQ_ROLE%omSum, &
   G_I_DQ_ROLE%omDiff, &
   G_I_DQ_ROLE%om2Sum, &
   G_I_DQ_ROLE%om2Diff, &
   G_I_DQ_ROLE%tT1, &
   G_I_DQ_ROLE%tT2, &
   G_I_DQ_ROLE%mMin, &
   G_I_DQ_ROLE%m, &
   G_I_DQ_ROLE%m1, &
   G_I_DQ_ROLE%m2, &
   G_I_DQ_ROLE%rho, &
   G_I_DQ_ROLE%KAlt, &
   G_I_DQ_ROLE%KAlt1, &
   G_I_DQ_ROLE%KAlt2, &
   G_I_DQ_ROLE%aSinI, &
   G_I_DQ_ROLE%aPMax, &
   G_I_DQ_ROLE%a1, &
   G_I_DQ_ROLE%a2, &
   G_I_DQ_ROLE%aRel, &
   G_I_DQ_ROLE%d], &

   G_PAR_SCP_BY_COMB(0:1,0:1,0:1,0:1,N_G_PARS)=reshape([ & ! use 10+DT_TY%… for per-data-set scope w.r.t. respective data type
                 ! m|1|2|3|p (m: mode [Normal/Binary]; 1…3: data of type #1…3 [AMAngPos/RV/AMHipN] given?; p: global par. #)
   QTY_SCP%no, & ! N|-|-|-|1   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|1   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|1   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|1
   QTY_SCP%pl, & ! N|-|×|-|1
   QTY_SCP%s1, & ! B|-|×|-|1
   QTY_SCP%no, & ! N|×|×|-|1   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|1
   QTY_SCP%pl, & ! N|-|-|×|1
   QTY_SCP%no, & ! B|-|-|×|1   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|1   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|1   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|1
   QTY_SCP%no, & ! B|-|×|×|1   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|1   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|1   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|2   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|2   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|2   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|2
   QTY_SCP%pl, & ! N|-|×|-|2
   QTY_SCP%s1, & ! B|-|×|-|2
   QTY_SCP%no, & ! N|×|×|-|2   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|2
   QTY_SCP%pl, & ! N|-|-|×|2
   QTY_SCP%no, & ! B|-|-|×|2   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|2   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|2   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|2
   QTY_SCP%no, & ! B|-|×|×|2   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|2   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|2   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|3   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|3   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|3   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|3
   QTY_SCP%pl, & ! N|-|×|-|3
   QTY_SCP%s1, & ! B|-|×|-|3
   QTY_SCP%no, & ! N|×|×|-|3   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|3
   QTY_SCP%pl, & ! N|-|-|×|3
   QTY_SCP%no, & ! B|-|-|×|3   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|3   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|3   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|3
   QTY_SCP%no, & ! B|-|×|×|3   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|3   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|3   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|4   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|4   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|4   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|4
   QTY_SCP%pl, & ! N|-|×|-|4
   QTY_SCP%no, & ! B|-|×|-|4
   QTY_SCP%no, & ! N|×|×|-|4   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|4
   QTY_SCP%pl, & ! N|-|-|×|4
   QTY_SCP%no, & ! B|-|-|×|4   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|4   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|4   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|4
   QTY_SCP%no, & ! B|-|×|×|4   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|4   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|4   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|5   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|5   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|5   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|5
   QTY_SCP%pl, & ! N|-|×|-|5
   QTY_SCP%no, & ! B|-|×|-|5
   QTY_SCP%no, & ! N|×|×|-|5   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|5
   QTY_SCP%no, & ! N|-|-|×|5
   QTY_SCP%no, & ! B|-|-|×|5   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|5   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|5   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|5
   QTY_SCP%no, & ! B|-|×|×|5   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|5   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|5   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|6   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|6   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|6   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|6
   10+DT_TY%RV,& ! N|-|×|-|6
   QTY_SCP%s1, & ! B|-|×|-|6
   QTY_SCP%no, & ! N|×|×|-|6   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|6
   QTY_SCP%no, & ! N|-|-|×|6
   QTY_SCP%no, & ! B|-|-|×|6   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|6   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|6   (not impl.)
   10+DT_TY%RV,& ! N|-|×|×|6
   QTY_SCP%no, & ! B|-|×|×|6   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|6   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|6   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|7   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|7   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|7   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|7
   QTY_SCP%s1, & ! N|-|×|-|7
   QTY_SCP%s1, & ! B|-|×|-|7
   QTY_SCP%no, & ! N|×|×|-|7   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|7
   QTY_SCP%no, & ! N|-|-|×|7
   QTY_SCP%no, & ! B|-|-|×|7   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|7   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|7   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|7
   QTY_SCP%no, & ! B|-|×|×|7   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|7   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|7   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|8   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|8   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|8   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|8
   QTY_SCP%no, & ! N|-|×|-|8
   QTY_SCP%s1, & ! B|-|×|-|8
   QTY_SCP%no, & ! N|×|×|-|8   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|8
   QTY_SCP%no, & ! N|-|-|×|8
   QTY_SCP%no, & ! B|-|-|×|8   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|8   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|8   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|8
   QTY_SCP%no, & ! B|-|×|×|8   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|8   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|8   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|9   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|9   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|9   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|9
   QTY_SCP%no, & ! N|-|×|-|9
   QTY_SCP%s1, & ! B|-|×|-|9
   QTY_SCP%no, & ! N|×|×|-|9   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|9
   QTY_SCP%no, & ! N|-|-|×|9
   QTY_SCP%no, & ! B|-|-|×|9   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|9   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|9   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|9
   QTY_SCP%no, & ! B|-|×|×|9   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|9   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|9   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|10   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|10   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|10   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|10
   QTY_SCP%no, & ! N|-|×|-|10
   QTY_SCP%s1, & ! B|-|×|-|10
   QTY_SCP%no, & ! N|×|×|-|10   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|10
   QTY_SCP%no, & ! N|-|-|×|10
   QTY_SCP%no, & ! B|-|-|×|10   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|10   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|10   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|10
   QTY_SCP%no, & ! B|-|×|×|10   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|10   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|10   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|11   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|11   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|11   (not impl.)
   QTY_SCP%sC, & ! B|×|-|-|11
   QTY_SCP%no, & ! N|-|×|-|11
   QTY_SCP%no, & ! B|-|×|-|11
   QTY_SCP%no, & ! N|×|×|-|11   (not impl.)
   QTY_SCP%sC, & ! B|×|×|-|11
   QTY_SCP%sC, & ! N|-|-|×|11
   QTY_SCP%no, & ! B|-|-|×|11   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|11   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|11   (not impl.)
   QTY_SCP%sC, & ! N|-|×|×|11
   QTY_SCP%no, & ! B|-|×|×|11   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|11   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|11   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|12   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|12   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|12   (not impl.)
   QTY_SCP%sC, & ! B|×|-|-|12
   QTY_SCP%no, & ! N|-|×|-|12
   QTY_SCP%no, & ! B|-|×|-|12
   QTY_SCP%no, & ! N|×|×|-|12   (not impl.)
   QTY_SCP%sC, & ! B|×|×|-|12
   QTY_SCP%sC, & ! N|-|-|×|12
   QTY_SCP%no, & ! B|-|-|×|12   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|12   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|12   (not impl.)
   QTY_SCP%sC, & ! N|-|×|×|12
   QTY_SCP%no, & ! B|-|×|×|12   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|12   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|12   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|13   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|13   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|13   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|13
   QTY_SCP%no, & ! N|-|×|-|13
   QTY_SCP%no, & ! B|-|×|-|13
   QTY_SCP%no, & ! N|×|×|-|13   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|13
   QTY_SCP%s1, & ! N|-|-|×|13
   QTY_SCP%no, & ! B|-|-|×|13   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|13   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|13   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|13
   QTY_SCP%no, & ! B|-|×|×|13   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|13   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|13   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|14   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|14   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|14   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|14
   QTY_SCP%no, & ! N|-|×|-|14
   QTY_SCP%no, & ! B|-|×|-|14
   QTY_SCP%no, & ! N|×|×|-|14   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|14
   QTY_SCP%no, & ! N|-|-|×|14
   QTY_SCP%no, & ! B|-|-|×|14   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|14   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|14   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|14
   QTY_SCP%no, & ! B|-|×|×|14   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|14   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|14   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|15   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|15   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|15   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|15
   QTY_SCP%no, & ! N|-|×|-|15
   QTY_SCP%no, & ! B|-|×|-|15
   QTY_SCP%no, & ! N|×|×|-|15   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|15
   QTY_SCP%pl, & ! N|-|-|×|15
   QTY_SCP%no, & ! B|-|-|×|15   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|15   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|15   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|15
   QTY_SCP%no, & ! B|-|×|×|15   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|15   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|15   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|16   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|16   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|16   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|16
   QTY_SCP%no, & ! N|-|×|-|16
   QTY_SCP%no, & ! B|-|×|-|16
   QTY_SCP%no, & ! N|×|×|-|16   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|16
   QTY_SCP%s1, & ! N|-|-|×|16
   QTY_SCP%no, & ! B|-|-|×|16   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|16   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|16   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|16
   QTY_SCP%no, & ! B|-|×|×|16   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|16   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|16   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|17   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|17   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|17   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|17
   QTY_SCP%no, & ! N|-|×|-|17
   QTY_SCP%no, & ! B|-|×|-|17
   QTY_SCP%no, & ! N|×|×|-|17   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|17
   QTY_SCP%s1, & ! N|-|-|×|17
   QTY_SCP%no, & ! B|-|-|×|17   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|17   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|17   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|17
   QTY_SCP%no, & ! B|-|×|×|17   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|17   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|17   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|18   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|18   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|18   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|18
   QTY_SCP%no, & ! N|-|×|-|18
   QTY_SCP%no, & ! B|-|×|-|18
   QTY_SCP%no, & ! N|×|×|-|18   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|18
   QTY_SCP%s1, & ! N|-|-|×|18
   QTY_SCP%no, & ! B|-|-|×|18   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|18   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|18   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|18
   QTY_SCP%no, & ! B|-|×|×|18   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|18   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|18   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|19   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|19   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|19   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|19
   QTY_SCP%no, & ! N|-|×|-|19
   QTY_SCP%no, & ! B|-|×|-|19
   QTY_SCP%no, & ! N|×|×|-|19   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|19
   QTY_SCP%s1, & ! N|-|-|×|19
   QTY_SCP%no, & ! B|-|-|×|19   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|19   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|19   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|19
   QTY_SCP%no, & ! B|-|×|×|19   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|19   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|19   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|20   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|20   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|20   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|20
   QTY_SCP%no, & ! N|-|×|-|20
   QTY_SCP%no, & ! B|-|×|-|20
   QTY_SCP%no, & ! N|×|×|-|20   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|20
   QTY_SCP%s1, & ! N|-|-|×|20
   QTY_SCP%no, & ! B|-|-|×|20   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|20   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|20   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|20
   QTY_SCP%no, & ! B|-|×|×|20   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|20   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|20   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|21   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|21   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|21   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|21
   QTY_SCP%s1, & ! N|-|×|-|21
   QTY_SCP%s1, & ! B|-|×|-|21
   QTY_SCP%no, & ! N|×|×|-|21   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|21
   QTY_SCP%no, & ! N|-|-|×|21
   QTY_SCP%no, & ! B|-|-|×|21   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|21   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|21   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|21
   QTY_SCP%no, & ! B|-|×|×|21   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|21   (not impl.)
   QTY_SCP%no], &! B|×|×|×|21   (not impl.)
   [2,2,2,2,N_G_PARS]), &

   G_DQ_SCP_BY_COMB(0:1,0:1,0:1,0:1,N_G_DQS)=reshape([ & ! use 10+DT_TY%… for per-data-set scope w.r.t. respective data type
                 ! m|1|2|3|q (m: mode [Normal/Binary]; 1…3: data of type #1…3 [AMAngPos/RV/AMHipN] given?; q: global DQ #)
   QTY_SCP%no, & ! N|-|-|-|1   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|1   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|1   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|1
   QTY_SCP%pl, & ! N|-|×|-|1
   QTY_SCP%s1, & ! B|-|×|-|1
   QTY_SCP%no, & ! N|×|×|-|1   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|1
   QTY_SCP%pl, & ! N|-|-|×|1
   QTY_SCP%no, & ! B|-|-|×|1   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|1   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|1   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|1
   QTY_SCP%no, & ! B|-|×|×|1   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|1   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|1   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|2   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|2   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|2   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|2
   QTY_SCP%pl, & ! N|-|×|-|2
   QTY_SCP%s1, & ! B|-|×|-|2
   QTY_SCP%no, & ! N|×|×|-|2   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|2
   QTY_SCP%pl, & ! N|-|-|×|2
   QTY_SCP%no, & ! B|-|-|×|2   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|2   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|2   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|2
   QTY_SCP%no, & ! B|-|×|×|2   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|2   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|2   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|3   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|3   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|3   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|3
   QTY_SCP%pl, & ! N|-|×|-|3
   QTY_SCP%no, & ! B|-|×|-|3
   QTY_SCP%no, & ! N|×|×|-|3   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|3
   QTY_SCP%no, & ! N|-|-|×|3
   QTY_SCP%no, & ! B|-|-|×|3   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|3   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|3   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|3
   QTY_SCP%no, & ! B|-|×|×|3   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|3   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|3   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|4   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|4   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|4   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|4
   QTY_SCP%pl, & ! N|-|×|-|4
   QTY_SCP%no, & ! B|-|×|-|4
   QTY_SCP%no, & ! N|×|×|-|4   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|4
   QTY_SCP%no, & ! N|-|-|×|4
   QTY_SCP%no, & ! B|-|-|×|4   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|4   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|4   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|4
   QTY_SCP%no, & ! B|-|×|×|4   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|4   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|4   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|5   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|5   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|5   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|5
   QTY_SCP%pl, & ! N|-|×|-|5
   QTY_SCP%no, & ! B|-|×|-|5
   QTY_SCP%no, & ! N|×|×|-|5   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|5
   QTY_SCP%no, & ! N|-|-|×|5
   QTY_SCP%no, & ! B|-|-|×|5   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|5   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|5   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|5
   QTY_SCP%no, & ! B|-|×|×|5   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|5   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|5   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|6   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|6   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|6   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|6
   QTY_SCP%pl, & ! N|-|×|-|6
   QTY_SCP%no, & ! B|-|×|-|6
   QTY_SCP%no, & ! N|×|×|-|6   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|6
   QTY_SCP%no, & ! N|-|-|×|6
   QTY_SCP%no, & ! B|-|-|×|6   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|6   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|6   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|6
   QTY_SCP%no, & ! B|-|×|×|6   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|6   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|6   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|7   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|7   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|7   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|7
   QTY_SCP%no, & ! N|-|×|-|7
   QTY_SCP%s1, & ! B|-|×|-|7
   QTY_SCP%no, & ! N|×|×|-|7   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|7
   QTY_SCP%no, & ! N|-|-|×|7
   QTY_SCP%no, & ! B|-|-|×|7   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|7   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|7   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|7
   QTY_SCP%no, & ! B|-|×|×|7   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|7   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|7   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|8   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|8   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|8   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|8
   QTY_SCP%no, & ! N|-|×|-|8
   QTY_SCP%s1, & ! B|-|×|-|8
   QTY_SCP%no, & ! N|×|×|-|8   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|8
   QTY_SCP%no, & ! N|-|-|×|8
   QTY_SCP%no, & ! B|-|-|×|8   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|8   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|8   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|8
   QTY_SCP%no, & ! B|-|×|×|8   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|8   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|8   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|9   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|9   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|9   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|9
   QTY_SCP%no, & ! N|-|×|-|9
   QTY_SCP%s1, & ! B|-|×|-|9
   QTY_SCP%no, & ! N|×|×|-|9   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|9
   QTY_SCP%no, & ! N|-|-|×|9
   QTY_SCP%no, & ! B|-|-|×|9   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|9   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|9   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|9
   QTY_SCP%no, & ! B|-|×|×|9   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|9   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|9   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|10   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|10   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|10   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|10
   QTY_SCP%no, & ! N|-|×|-|10
   QTY_SCP%no, & ! B|-|×|-|10
   QTY_SCP%no, & ! N|×|×|-|10   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|10
   QTY_SCP%no, & ! N|-|-|×|10
   QTY_SCP%no, & ! B|-|-|×|10   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|10   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|10   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|10
   QTY_SCP%no, & ! B|-|×|×|10   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|10   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|10   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|11   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|11   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|11   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|11
   QTY_SCP%no, & ! N|-|×|-|11
   QTY_SCP%no, & ! B|-|×|-|11
   QTY_SCP%no, & ! N|×|×|-|11   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|11
   QTY_SCP%no, & ! N|-|-|×|11
   QTY_SCP%no, & ! B|-|-|×|11   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|11   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|11   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|11
   QTY_SCP%no, & ! B|-|×|×|11   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|11   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|11   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|12   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|12   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|12   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|12
   QTY_SCP%no, & ! N|-|×|-|12
   QTY_SCP%no, & ! B|-|×|-|12
   QTY_SCP%no, & ! N|×|×|-|12   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|12
   QTY_SCP%s1, & ! N|-|-|×|12
   QTY_SCP%no, & ! B|-|-|×|12   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|12   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|12   (not impl.)
   QTY_SCP%s1, & ! N|-|×|×|12
   QTY_SCP%no, & ! B|-|×|×|12   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|12   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|12   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|13   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|13   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|13   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|13
   QTY_SCP%no, & ! N|-|×|-|13
   QTY_SCP%no, & ! B|-|×|-|13
   QTY_SCP%no, & ! N|×|×|-|13   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|13
   QTY_SCP%no, & ! N|-|-|×|13
   QTY_SCP%no, & ! B|-|-|×|13   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|13   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|13   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|13
   QTY_SCP%no, & ! B|-|×|×|13   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|13   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|13   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|14   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|14   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|14   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|14
   QTY_SCP%no, & ! N|-|×|-|14
   QTY_SCP%no, & ! B|-|×|-|14
   QTY_SCP%no, & ! N|×|×|-|14   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|14
   QTY_SCP%no, & ! N|-|-|×|14
   QTY_SCP%no, & ! B|-|-|×|14   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|14   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|14   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|14
   QTY_SCP%no, & ! B|-|×|×|14   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|14   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|14   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|15   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|15   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|15   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|15
   QTY_SCP%no, & ! N|-|×|-|15
   QTY_SCP%no, & ! B|-|×|-|15
   QTY_SCP%no, & ! N|×|×|-|15   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|15
   QTY_SCP%no, & ! N|-|-|×|15
   QTY_SCP%no, & ! B|-|-|×|15   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|15   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|15   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|15
   QTY_SCP%no, & ! B|-|×|×|15   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|15   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|15   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|16   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|16   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|16   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|16
   QTY_SCP%pl, & ! N|-|×|-|16
   QTY_SCP%no, & ! B|-|×|-|16
   QTY_SCP%no, & ! N|×|×|-|16   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|16
   QTY_SCP%pl, & ! N|-|-|×|16
   QTY_SCP%no, & ! B|-|-|×|16   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|16   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|16   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|16
   QTY_SCP%no, & ! B|-|×|×|16   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|16   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|16   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|17   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|17   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|17   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|17
   QTY_SCP%no, & ! N|-|×|-|17
   QTY_SCP%s1, & ! B|-|×|-|17
   QTY_SCP%no, & ! N|×|×|-|17   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|17
   QTY_SCP%no, & ! N|-|-|×|17
   QTY_SCP%no, & ! B|-|-|×|17   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|17   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|17   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|17
   QTY_SCP%no, & ! B|-|×|×|17   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|17   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|17   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|18   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|18   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|18   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|18
   QTY_SCP%pl, & ! N|-|×|-|18
   QTY_SCP%no, & ! B|-|×|-|18
   QTY_SCP%no, & ! N|×|×|-|18   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|18
   QTY_SCP%pl, & ! N|-|-|×|18
   QTY_SCP%no, & ! B|-|-|×|18   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|18   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|18   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|18
   QTY_SCP%no, & ! B|-|×|×|18   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|18   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|18   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|19   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|19   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|19   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|19
   QTY_SCP%pl, & ! N|-|×|-|19
   QTY_SCP%no, & ! B|-|×|-|19
   QTY_SCP%no, & ! N|×|×|-|19   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|19
   QTY_SCP%pl, & ! N|-|-|×|19
   QTY_SCP%no, & ! B|-|-|×|19   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|19   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|19   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|19
   QTY_SCP%no, & ! B|-|×|×|19   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|19   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|19   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|20   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|20   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|20   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|20
   QTY_SCP%no, & ! N|-|×|-|20
   QTY_SCP%no, & ! B|-|×|-|20
   QTY_SCP%no, & ! N|×|×|-|20   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|20
   QTY_SCP%pl, & ! N|-|-|×|20
   QTY_SCP%no, & ! B|-|-|×|20   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|20   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|20   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|20
   QTY_SCP%no, & ! B|-|×|×|20   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|20   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|20   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|21   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|21   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|21   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|21
   QTY_SCP%no, & ! N|-|×|-|21
   QTY_SCP%no, & ! B|-|×|-|21
   QTY_SCP%no, & ! N|×|×|-|21   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|21
   QTY_SCP%pl, & ! N|-|-|×|21
   QTY_SCP%no, & ! B|-|-|×|21   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|21   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|21   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|21
   QTY_SCP%no, & ! B|-|×|×|21   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|21   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|21   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|22   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|22   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|22   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|22
   QTY_SCP%no, & ! N|-|×|-|22
   QTY_SCP%no, & ! B|-|×|-|22
   QTY_SCP%no, & ! N|×|×|-|22   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|22
   QTY_SCP%no, & ! N|-|-|×|22
   QTY_SCP%no, & ! B|-|-|×|22   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|22   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|22   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|22
   QTY_SCP%no, & ! B|-|×|×|22   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|22   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|22   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|23   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|23   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|23   (not impl.)
   QTY_SCP%s1, & ! B|×|-|-|23
   QTY_SCP%no, & ! N|-|×|-|23
   QTY_SCP%no, & ! B|-|×|-|23
   QTY_SCP%no, & ! N|×|×|-|23   (not impl.)
   QTY_SCP%s1, & ! B|×|×|-|23
   QTY_SCP%no, & ! N|-|-|×|23
   QTY_SCP%no, & ! B|-|-|×|23   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|23   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|23   (not impl.)
   QTY_SCP%no, & ! N|-|×|×|23
   QTY_SCP%no, & ! B|-|×|×|23   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|23   (not impl.)
   QTY_SCP%no, & ! B|×|×|×|23   (not impl.)

   QTY_SCP%no, & ! N|-|-|-|24   (not impl.)
   QTY_SCP%no, & ! B|-|-|-|24   (not impl.)
   QTY_SCP%no, & ! N|×|-|-|24   (not impl.)
   QTY_SCP%no, & ! B|×|-|-|24
   QTY_SCP%no, & ! N|-|×|-|24
   QTY_SCP%no, & ! B|-|×|-|24
   QTY_SCP%no, & ! N|×|×|-|24   (not impl.)
   QTY_SCP%no, & ! B|×|×|-|24
   QTY_SCP%no, & ! N|-|-|×|24
   QTY_SCP%no, & ! B|-|-|×|24   (not impl.)
   QTY_SCP%no, & ! N|×|-|×|24   (not impl.)
   QTY_SCP%no, & ! B|×|-|×|24   (not impl.)
   QTY_SCP%pl, & ! N|-|×|×|24
   QTY_SCP%no, & ! B|-|×|×|24   (not impl.)
   QTY_SCP%no, & ! N|×|×|×|24   (not impl.)
   QTY_SCP%no], &! B|×|×|×|24   (not impl.)
   [2,2,2,2,N_G_DQS])


  real,parameter :: &
   PRI_REG_SPC_DEVN_RATIO_MAX=1e-4, &
   LOG_SCL_L_BND_KNEE_RATIO=.1

  real(dp),parameter :: &
   G_PAR_PERIOD(N_G_PARS)=[real(dp) :: & ! =(upper bound) for angle-like pars. with wrap-around behaviour, =0 otherwise
   0, &                              ! e
   0, &                              ! f
   1, &                              ! chi
   TWO_PI, &                         ! omega
   0, &                              ! K
   0, &                              ! V
   0, &                              ! sigma_+
   TWO_PI, &                         ! omega_2
   0, &                              ! K_1
   0, &                              ! K_2
   0, &                              ! i
   TWO_PI, &                         ! Omega
   0, &                              ! pi
   0, &                              ! a'_rel
   0, &                              ! a'
   360, &                            ! alpha_r
   0, &                              ! delta_r
   0, &                              ! mu_acd
   0, &                              ! mu_d
   0, &                              ! tau_+
   0]                                ! a_v

  logical,parameter :: &
   G_PAR_LOG_SCL(N_G_PARS)=[ &       ! marg. postr. abscissa in log scale?
   .FALSE., &                        ! e
   .FALSE., &                        ! f
   .FALSE., &                        ! chi
   .FALSE., &                        ! omega
   .TRUE., &                         ! K
   .FALSE., &                        ! V
   .TRUE., &                         ! sigma_+
   .FALSE., &                        ! omega_2
   .TRUE., &                         ! K_1
   .TRUE., &                         ! K_2
   .FALSE., &                        ! i
   .FALSE., &                        ! Omega
   .TRUE., &                         ! pi
   .TRUE., &                         ! a'_rel
   .TRUE., &                         ! a'
   .FALSE., &                        ! alpha_r
   .FALSE., &                        ! delta_r
   .TRUE., &                         ! mu_acd
   .TRUE., &                         ! mu_d
   .TRUE., &                         ! tau_+
   .TRUE.], &                        ! a_v
   G_PAR_IS_CYC(N_G_PARS)=[ &
   .FALSE., &                        ! e
   .FALSE., &                        ! f
   .TRUE., &                         ! chi
   .TRUE., &                         ! omega
   .FALSE., &                        ! K
   .FALSE., &                        ! V
   .FALSE., &                        ! sigma_+
   .TRUE., &                         ! omega_2
   .FALSE., &                        ! K_1
   .FALSE., &                        ! K_2
   .FALSE., &                        ! i
   .TRUE., &                         ! Omega
   .FALSE., &                        ! pi
   .FALSE., &                        ! a'_rel
   .FALSE., &                        ! a'
   .TRUE., &                         ! alpha_r
   .FALSE., &                        ! delta_r
   .FALSE., &                        ! mu_acd
   .FALSE., &                        ! mu_d
   .FALSE., &                        ! tau_+
   .FALSE.], &                       ! a_v
   G_PAR_HAS_EPOCH(N_G_PARS)=[ &
   .FALSE., &                        ! e
   .FALSE., &                        ! f
   .TRUE., &                         ! chi
   .FALSE., &                        ! omega
   .FALSE., &                        ! K
   .TRUE., &                         ! V
   .FALSE., &                        ! sigma_+
   .FALSE., &                        ! omega_2
   .FALSE., &                        ! K_1
   .FALSE., &                        ! K_2
   .FALSE., &                        ! i
   .FALSE., &                        ! Omega
   .FALSE., &                        ! pi
   .FALSE., &                        ! a'_rel
   .FALSE., &                        ! a'
   .TRUE., &                         ! alpha_r
   .TRUE., &                         ! delta_r
   .FALSE., &                        ! mu_acd
   .FALSE., &                        ! mu_d
   .FALSE., &                        ! tau_+
   .FALSE.], &                       ! a_v
   G_DQ_LOG_SCL(N_G_DQS)=[ &         ! marg. postr. abscissa in log scale?
   .TRUE., &                         ! P
   .FALSE., &                        ! T
   .TRUE., &                         ! m_min
   .TRUE., &                         ! a sin i
   .TRUE., &                         ! a_p,max
   .TRUE., &                         ! K_alt
   .TRUE., &                         ! K_alt,1
   .TRUE., &                         ! K_alt,2
   .TRUE., &                         ! rho
   .TRUE., &                         ! m_1
   .TRUE., &                         ! m_2
   .TRUE., &                         ! d
   .TRUE., &                         ! a_rel
   .TRUE., &                         ! a_1
   .TRUE., &                         ! a_1
   .FALSE., &                        ! theta_r
   .FALSE., &                        ! theta_r (secondary)
   .FALSE., &                        ! t_t1
   .FALSE., &                        ! t_t2
   .FALSE., &                        ! Omega+omega
   .FALSE., &                        ! Omega-omega
   .FALSE., &                        ! Omega+omega_2
   .FALSE., &                        ! Omega-omega_2
   .TRUE.]                           ! m

  character(*),parameter :: &
   G_PAR_NMS(N_G_PARS)=[character(7) :: & ! the names of the model parameters
    "e","f","chi","omega","K","V","sigma_+","omega_2","K_1","K_2","i","Omega","pi","a`_rel","a`","alpha_r","delta_r","mu_acd", &
    "mu_d","tau_+","a_v"], &
   G_PAR_NMS_TEX(N_G_PARS)=[character(24) :: & ! … in TeX format
    "e","f","\chi","\omega","K","V","\sigma_+","\omega_2","K_1","K_2","i","\Omega","\varpi","a^\prime_{\mathrm{rel}}","a^\prime", &
    "\alpha_{\mathrm r}","\delta_{\mathrm r}","\mu_\alpha\,\cos(\delta)","\mu_\delta","\tau_+","a_v"], &
   G_PAR_DESC(N_G_PARS)=[character(STR_LEN_STD) :: &
   "eccentricity", &                                                                    ! e
   "orbital frequency", &                                                               ! f
   "mean anomaly at t_r over (2*pi)", &                                                 ! chi
   "argument of periapsis of planet's orbit around centre of mass", &                   ! omega
   "radial-velocity semi-amplitude", &                                                  ! K
   "reference radial velocity", &                                                       ! V
   "magnitude of additional Gaussian radial-velocity noise", &                          ! sigma_+
   "argument of periapsis of orbit of secondary around primary", &                      ! omega_2
   "radial-velocity semi-amplitude of primary", &                                       ! K_1
   "radial-velocity semi-amplitude of secondary", &                                     ! K_2
   "inclination", &                                                                     ! i
   "position angle of ascending node (if no radial-velocity data: first node)", &       ! Omega
   "parallax", &                                                                        ! pi
   "semi-major axis of orbit of secondary around primary over distance", &              ! a'_rel
   "semi-major axis of stellar barycentric orbit over distance", &                      ! a'
   "position at time t_r, coordinate alpha", &                                          ! alpha_r
   "position at time t_r, coordinate delta", &                                          ! delta_r
   "proper motion in coordinate alpha times cos(delta)", &                              ! mu_acd
   "proper motion in coordinate delta", &                                               ! mu_d
   "magnitude of additional Gaussian astrometric noise", &                              ! tau_+
   "radial-velocity acceleration"], &                                                   ! a_v
   G_PAR_UNITS(N_G_PARS)=[character(6) :: & ! the units of the model parameters
   "", &                                    ! e
   "1/d", &                                 ! f
   "", &                                    ! chi
   "rad", &                                 ! omega
   "m/s", &                                 ! K
   "m/s", &                                 ! V
   "m/s", &                                 ! sigma_+
   "rad", &                                 ! omega_2
   "m/s", &                                 ! K_1
   "m/s", &                                 ! K_2
   "rad", &                                 ! i
   "rad", &                                 ! Omega
   "arcsec", &                              ! pi
   "mas", &                                 ! a'_rel
   "mas", &                                 ! a'
   "deg", &                                 ! alpha_r
   "deg", &                                 ! delta_r
   "mas/yr", &                              ! mu_acd
   "mas/yr", &                              ! mu_d
   "mas", &                                 ! tau_+
   "m/s/yr"], &                             ! a_v
   G_PAR_UNITS_TEX(N_G_PARS)=[character(36) :: & ! the units of the model parameters in TeX format
   "", &                                         ! e
   "\frac{1}{\mathrm d}", &                      ! f
   "", &                                         ! chi
   "\mathrm{rad}", &                             ! omega
   "\frac{\mathrm m}{\mathrm s}", &              ! K
   "\frac{\mathrm m}{\mathrm s}", &              ! V
   "\frac{\mathrm m}{\mathrm s}", &              ! sigma_+
   "\mathrm{rad}", &                             ! omega_2
   "\frac{\mathrm m}{\mathrm s}", &              ! K_1
   "\frac{\mathrm m}{\mathrm s}", &              ! K_2
   "\mathrm{rad}", &                             ! i
   "\mathrm{rad}", &                             ! Omega
   "\mathrm{arcsec}", &                          ! pi
   "\mathrm{mas}", &                             ! a'_rel
   "\mathrm{mas}", &                             ! a'
   "{}^{\circ}", &                               ! alpha_r
   "{}^{\circ}", &                               ! delta_r
   "\frac{\mathrm{mas}}{\mathrm{yr}}", &         ! mu_acd
   "\frac{\mathrm{mas}}{\mathrm{yr}}", &         ! mu_d
   "\mathrm{mas}", &                             ! tau_+
   "\frac{\mathrm m}{\mathrm{s\cdot yr}}"], &    ! a_v
   QTY_NM_SUFFX_DELIM_STD="_", &
   QTY_NM_SUFFX_DELIM_ALT=",", &
   PRI_TY_DESC(0:N_PRI_TY-1)=[character(24) :: "uniform","Jeffreys","modified Jeffreys","user-defined","signed modified Jeffreys", &
   "normal"], &
   G_DQ_NMS(N_G_DQS)=[character(13) :: & ! CAUTION: changing order would destroy calc. of DQs!
    "P","T","m_p,min","a sin i","a_p_max","K_alt","K_alt_1","K_alt_2","rho","m_1","m_2","d","a_rel","a_1","a_2","theta_r", &
    "theta_r","t_t1","t_t2","Omega+omega","Omega-omega","Omega+omega_2","Omega-omega_2","m_p"], &
   G_DQ_NMS_TEX(N_G_DQS)=[character(18) :: & ! … in TeX format
   "P", &                                   ! P
   "T", &                                   ! T
   "m_{\mathrm{p,min}}", &                  ! m_min
   "a \sin i", &                            ! a sin i
   "a_{\mathrm{p,max}}", &                  ! a_p,max
   "K_{\mathrm{real}}", &                   ! K_alt
   "K_{\mathrm{alt,1}}", &                  ! K_alt,1
   "K_{\mathrm{alt,2}}", &                  ! K_alt,2
   "\rho", &                                ! rho
   "m_1", &                                 ! m_1
   "m_2", &                                 ! m_1
   "d", &                                   ! d
   "a_{\mathrm{rel}}", &                    ! a_rel
   "a_1", &                                 ! a_1
   "a_2", &                                 ! a_2
   "\vartheta_{\mathrm{r}}", &              ! theta_r
   "\vartheta_{\mathrm{r}}", &              ! theta_r (secondary)
   "t_{\mathrm{t,1}}", &                    ! t_t1
   "t_{\mathrm{t,2}}", &                    ! t_t2
   "Omega+omega", &                         ! Omega+omega
   "Omega-omega", &                         ! Omega-omega
   "Omega+omega_2", &                       ! Omega+omega_2
   "Omega-omega_2", &                       ! Omega-omega_2
   "m_{\mathrm p}"], &                      ! m
   G_DQ_DESC(N_G_DQS)=[character(STR_LEN_STD) :: &
   "period", &
   "time of periapsis", &
   "minimum planetary mass", &
   "semi-major axis of stellar orbit times sine of inclination", &
   "maximum semi-major axis of planet orbit", &
   "radial-velocity semi-amplitude common in the literature", &
   "radial-velocity semi-amplitude of primary common in the literature", &
   "radial-velocity semi-amplitude of secondary common in the literature", &
   "ratio of binary masses", &
   "mass of primary binary component", &
   "mass of secondary binary component", &
   "system distance", &
   "semi-major axis of orbit of secondary around primary", &
   "semi-major axis of orbit of primary around centre of mass", &
   "semi-major axis of orbit of secondary around centre of mass", &
   "position angle of star in orbital plane w.r.t. ascending node, at t_r", &
   "position angle of secondary in orbital plane w.r.t. ascending (or first) node, at t_r", &
   "time of first potential transit after given time", &
   "time of second potential transit after given time", &
   "sum of position angle of ascending (or first) node and argument of periapsis", &
   "difference between position angle of ascending (or first) node and argument of periapsis", & 
   "sum of position angle of ascending (or first) node and argument of periapsis", & 
   "difference between position angle of ascending (or first) node and argument of periapsis", &
   "planetary mass"], & 
   G_DQ_UNITS(N_G_DQS)=[character(5) :: & ! the units of the DQs
   "d", &                                ! P
   "d", &                                ! T
   "M_J", &                              ! m_min
   "AU", &                               ! a sin i
   "AU", &                               ! a_p,max
   "m/s", &                              ! K_alt
   "m/s", &                              ! K_alt,1
   "m/s", &                              ! K_alt,2
   "", &                                 ! rho
   "M_Sun", &                            ! m_1
   "M_Sun", &                            ! m_1
   "pc", &                               ! d
   "AU", &                               ! a_rel
   "AU", &                               ! a_1
   "AU", &                               ! a_2
   "rad", &                              ! theta_r
   "rad", &                              ! theta_r (secondary)
   "d", &                                ! t_t1
   "d", &                                ! t_t2
   "rad", &                              ! Omega+omega
   "rad", &                              ! Omega-omega
   "rad", &                              ! Omega+omega_2
   "rad", &                              ! Omega-omega_2
   "M_J"], &                             ! m

   G_DQ_UNITS_TEX(N_G_DQS)=[character(27) :: "\mathrm d", & ! P
   "\mathrm d", &                                          ! T
   "M_\mathrm{J}", &                                       ! m_min
   "\mathrm{AU}", &                                        ! a sin i
   "\mathrm{AU}", &                                        ! a_p,max
   "\frac{\mathrm m}{\mathrm s}", &                        ! K_alt
   "\frac{\mathrm m}{\mathrm s}", &                        ! K_alt,1
   "\frac{\mathrm m}{\mathrm s}", &                        ! K_alt,2
   "", &                                                   ! rho
   "M_{\mathrm{Sun}}", &                                   ! m_1
   "M_{\mathrm{Sun}}", &                                   ! m_1
   "\mathrm{pc}", &                                        ! d
   "\mathrm{AU}", &                                        ! a_rel
   "\mathrm{AU}", &                                        ! a_1
   "\mathrm{AU}", &                                        ! a_1
   "\mathrm{rad}", &                                       ! theta_r
   "\mathrm{rad}", &                                       ! theta_r (secondary)
   "\mathrm d", &                                          ! t_t1
   "\mathrm d", &                                          ! t_t2
   "\mathrm{rad}", &                                       ! Omega+omega
   "\mathrm{rad}", &                                       ! Omega-omega
   "\mathrm{rad}", &                                       ! Omega+omega_2
   "\mathrm{rad}", &                                       ! Omega-omega_2
   "M_\mathrm{J}"]                                         ! m

  type(tyDtTyInt),parameter :: &
   MDL_FCT_DIM=tyDtTyInt(2,1,1)

  type(tyIntvlDbl),parameter :: &
   G_STRICT_PRI_RGE(N_G_PARS)=[ &    ! largest allowed prior ranges
                                     ! CAUTION: be very careful with changing these bounds (esp. e_max, f_min)
                                     ! since this can undermine the whole analysis! [21.11.2009]
   tyIntvlDbl(0._dp,1._dp), &                    ! e (< 1)
   tyIntvlDbl(EPS_DP,10._dp), &                  ! f (> 0, <~ 10 [10.06.2011/1])
   tyIntvlDbl(0._dp,1._dp), &                    ! chi (< 1)
   tyIntvlDbl(0._dp,TWO_PI), &                   ! omega (< 2 pi)
   tyIntvlDbl(0._dp,+HUGE_DP), &                 ! K
   tyIntvlDbl(-HUGE_DP,+HUGE_DP), &              ! V
   tyIntvlDbl(0._dp,+HUGE_DP), &                 ! sigma_+
   tyIntvlDbl(0._dp,TWO_PI), &                   ! omega_2
   tyIntvlDbl(0._dp,+HUGE_DP), &                 ! K_1
   tyIntvlDbl(0._dp,+HUGE_DP), &                 ! K_2
   tyIntvlDbl(0._dp,PI), &                       ! i (< pi)
   tyIntvlDbl(0._dp,TWO_PI), &                   ! Omega (< 2 pi; < pi if only AMAngPos data)
   tyIntvlDbl(1._dp/(5e4_dp*LY_IN_PC),.77_dp), & ! pi (ubound due to parallax of nearest star alpha Cen, cf. [05.04.2010/3])
   tyIntvlDbl(1e-3_dp,1e5_dp), &                 ! a'_rel
   tyIntvlDbl(1e-3_dp,1e5_dp), &                 ! a'
   tyIntvlDbl(0._dp,360._dp), &                  ! alpha_r
   tyIntvlDbl(-90._dp,90._dp), &                 ! delta_r
   tyIntvlDbl(-HUGE_DP,+HUGE_DP), &              ! mu_acd
   tyIntvlDbl(-HUGE_DP,+HUGE_DP), &              ! mu_d
   tyIntvlDbl(0._dp,+HUGE_DP), &                 ! tau_+
   tyIntvlDbl(-HUGE_DP,+HUGE_DP)], &             ! a_v

   G_STD_PRI_RGE(N_G_PARS)=[tyIntvlDbl(0._dp,1._dp), & ! e
   tyIntvlDbl(1._dp/(1000*YR_IN_D),10._dp), &          ! f
   tyIntvlDbl(0._dp,1._dp), &                          ! chi
   tyIntvlDbl(0._dp,TWO_PI), &                         ! omega
   tyIntvlDbl(0._dp,+HUGE_DP), &                       ! K
   tyIntvlDbl(-HUGE_DP,+HUGE_DP), &                    ! V
   tyIntvlDbl(0._dp,+HUGE_DP), &                       ! sigma_+
   tyIntvlDbl(0._dp,TWO_PI), &                         ! omega_2
   tyIntvlDbl(0._dp,+HUGE_DP), &                       ! K_1
   tyIntvlDbl(0._dp,+HUGE_DP), &                       ! K_2
   tyIntvlDbl(0._dp,PI), &                             ! i
   tyIntvlDbl(0._dp,TWO_PI), &                         ! Omega
   tyIntvlDbl(1._dp/(5e4_dp*LY_IN_PC),.77_dp), &       ! pi
   tyIntvlDbl(1e-3_dp,1e5_dp), &                       ! a'_rel
   tyIntvlDbl(1e-3_dp,1e5_dp), &                       ! a'
   tyIntvlDbl(0._dp,360._dp), &                        ! alpha_r
   tyIntvlDbl(-90._dp,90._dp), &                       ! delta_r
   tyIntvlDbl(-20000._dp,20000._dp), &                 ! mu_acd
   tyIntvlDbl(-20000._dp,20000._dp), &                 ! mu_d
   tyIntvlDbl(0._dp,10000._dp), &                      ! tau_+
   tyIntvlDbl(-HUGE_DP,+HUGE_DP)]                      ! a_v


  !========================================================[ VARIABLES ]============================================================

  integer :: &
   nNPlMds, &                        ! number of #-planets modes (each #-planets mode refers to a certain number of planets assumed)
   maxNPl, &                         ! the maximum # planets
   curNPl, &                         ! the current # planets
   psdNPl, &                         ! the current # "pseudo-planets" (i.e. 1 for binary mode and the # planets otherwise)
   psdNPlUndef, &                    ! a "pseudo-planets" number ((undefined) for binary mode and the # planets otherwise)
   iNPlMd, &                         ! the current index of #-planet mode
   curNPars, &                       ! the current # local parameters (changes with the #-planets mode)
   curNParsFree, &                   ! the current # fixed local parameters (changes with the #-planets mode)
   curNParsInfc, &
   nParsSys, &                       ! # used global pars. for the whole system (total)
   nParsSysC, &                      ! # used global pars. for the whole system (total, only with companion(s))
   nGParsPl, &                       ! # used global pars. per planet
   curNDQs, &                        ! the current # local derived quantities
   gIPars(N_L_PARS_MAX), &           ! maps local to global par. indexes
   iPlByPar(N_L_PARS_MAX), &         ! maps local par. index to index of planet which the par. refers to
   nPl(:), &                         ! numbers of planets in each of the #-planets modes
   nParsMax, &
   nPars(N_N_PL_MODES_MAX), &
   parScp(N_L_PARS_MAX), &
   iDtSetByPar(N_L_PARS_MAX), &
   iDtSetByDQ(N_L_DQ_MAX), &
   priTypes(N_L_PARS_MAX), &         ! local prior types
   tmpNPriRges(N_N_PL_MODES_MAX), &
   tmpNPriFromFil(N_N_PL_MODES_MAX), &
   priNPts(N_L_PARS_MAX), &
   tmpNCalcProb(N_N_PL_MODES_MAX), &
   tmpNCalcProbAllPars(N_N_PL_MODES_MAX), &
   tmpCalcProbAllParsDim(N_N_PL_MODES_MAX,N_CALC_PROB_MAX), &
   gIDQs(N_L_DQ_MAX), &
   iPlByDQ(N_L_DQ_MAX), &            ! maps local DQ index to index of planet which the DQ refers to
   nDQs(N_N_PL_MODES_MAX), &
   nGDQsPl, &
   nDQsSys, &
   nDQsSysC, &
   DQScp(N_L_DQ_MAX), &
   tmpPriTy(N_N_PL_MODES_MAX,N_L_PARS_MAX), &
   tmpNGroup, &
   tmpGroupDtSets(N_DT_SETS_MAX)

  integer(i8b) :: &
   nUniqT                            ! # unique timestamps of all data

  real(dp) :: &
   uniqT(N_DT_SETS_MAX*N_REC_PER_SET_MAX), & ! unique timestamps of all data
   uniqTMnEpoch(N_DT_SETS_MAX*N_REC_PER_SET_MAX), & ! unique timestamps of all data minus earliest time
   tOfs, &                           ! offset applied to times (0 for general time scale, -RJD_OFS to transform from JD to RJD scale)
   eAUncFact, &
   eAPrecalcd(0:EA_PRECALC_SIZE,0:EA_PRECALC_SIZE), & ! pre-calculated eccentric anomaly values
   eMin, &
   eDiff, &
   parPeriod(N_L_PARS_MAX), &
   parInvPeriod(N_L_PARS_MAX), &
   priKnees(N_L_PARS_MAX), &         ! local modified Jeffreys prior knees
   priLBnd(N_L_PARS_MAX), &          ! local prior ranges; as array for quick access to only lower/upper bound
   priUBnd(N_L_PARS_MAX), &
   priW(N_L_PARS_MAX), &
   priHalfW(N_L_PARS_MAX), &
   priCtr(N_L_PARS_MAX), &
   priVce(N_L_PARS_MAX), &           ! CAUTION: this is NOT the prior variance but, for normal priors, the variance of the underlying normal distribution
   priPsdLBnd(N_L_PARS_MAX), &
   priLBndPlt(N_L_PARS_MAX), &
   priFac(N_L_PARS_MAX), &
   priPtsPsdX(N_L_PARS_MAX,PRI_N_PTS_MAX), &
   priPtsY(N_L_PARS_MAX,PRI_N_PTS_MAX), &
   priDiffY(N_L_PARS_MAX,PRI_N_PTS_MAX), &
   priInvDelta(N_L_PARS_MAX), &
   likhFact, &                      ! factor omitted from likelihood, constant for given data (cf. [07.05.2012/1])
   tmpPriVce(N_N_PL_MODES_MAX,N_L_PARS_MAX), &
   tmpGroupThrlds(N_DT_SETS_MAX,3)

  logical :: &
   residCalcd, &
   parInfc(N_L_PARS_MAX), &
   parIsCyc(N_L_PARS_MAX), &
   cycParUsMaxRge(N_L_PARS_MAX), &
   parFixed(N_L_PARS_MAX), &
   priAvail, &
   usrKMaxFromVel(N_N_PL_MODES_MAX), &
   priLogScl(N_L_PARS_MAX), &
   maPoLogScl(N_L_PARS_MAX), &
   DQSkipped(N_L_DQ_MAX), &
   maPoLogSclDQ(N_L_PARS_MAX), &
   parHasEpoch(N_L_PARS_MAX)

  character(STR_LEN_STD) :: &
   tmpPriRgeNms(N_N_PL_MODES_MAX,N_L_PARS_MAX), &
   tmpCalcProbNms(N_N_PL_MODES_MAX,N_CALC_PROB_MAX), &
   parNms(N_L_PARS_MAX), &
   parNmsPlt(N_L_PARS_MAX), &
   tmpPriNmsFromFil(N_N_PL_MODES_MAX,N_L_PARS_MAX), &
   DQNms(N_L_DQ_MAX), &
   DQNmsPlt(N_L_DQ_MAX)

  character(MAX_LEN_PATH) :: &
   tmpPriFilNms(N_N_PL_MODES_MAX,N_L_PARS_MAX)

  type(tyGenDtProps) :: &
   genDtProps

  type(tyParsInt) :: &
   pRole, &                          ! local par. index by role
   pRoleByPl(N_PL_MAX), &            ! local par. index by planet and role
   pRoleBySet(N_DT_SETS_MAX)         ! local par. index by data set and role

  type(tyIntvlDbl) :: &
   gStdPriRge(N_G_PARS), &
   priRge(N_L_PARS_MAX), &           ! local prior ranges; as interval array for simple access to interval
   tmpPriRge(N_N_PL_MODES_MAX,N_L_PARS_MAX), &
   tmpCalcProbRge(N_N_PL_MODES_MAX,N_CALC_PROB_MAX), &
   tmpCalcProbAllParsRegion(N_N_PL_MODES_MAX,N_CALC_PROB_MAX,N_L_PARS_MAX), &
   zmRgeAllPars(N_N_PL_MODES_MAX,N_L_PARS_MAX), &
   nPeriodsRge(N_N_PL_MODES_MAX), &
   priRgeFByNPeriods(N_N_PL_MODES_MAX), &
   DQZmRge(N_L_DQ_MAX)

  type(tyDQsInt) :: &
   qRole, &
   qRoleByPl(N_PL_MAX), &
   qRoleBySet(N_DT_SETS_MAX)

  type(tyMdlProps) :: &
   mdlProps
   
  allocatable :: &
   nPl


  ! USER OPTIONS

  logical :: &
   usrDQs(N_N_PL_MODES_MAX), &
   usrGenerateData


  ! DATA PROPERTIES

  type(tyDtPropsAMAngPos) :: &
   dtPropsAMAngPos

  type(tyDtPropsAMHipN) :: &
   dtPropsAMHipN

  type(tyDtPropsRV) :: &
   dtPropsRV

  type(tyDtPropsAll) :: &
   dtProps

  ! BINARY DATA SETS

  type(tyDtFilHipN) :: &
   dtFilHipN


  ! NON-BINARY DATA SETS

    ! sets of specific data types

  type(tyDtSetAMAngPos) :: &
   dtSetsAMAngPos(N_DT_SETS_PER_TY_MAX)

  type(tyDtSetAMHipN) :: &
   dtSetsAMHipN(N_DT_SETS_PER_TY_MAX)

  type(tyDtSetRV) :: &
   dtSetsRV(N_DT_SETS_PER_TY_MAX)

  target :: &
   dtSetsAMHipN


  ! GENERATING DATA

  integer :: &
   genDtPsdNPl, &
   genDtNSgm, &
   genDtNPars

  real(dp) :: &
   genDtPars(:), &
   genDtSgm(:)

  type(tyUncEllDbl) :: &
   genDtUncEll(:)

  allocatable :: &
   genDtPars, &
   genDtSgm, &
   genDtUncEll

  type(tyMdlProps) :: &
   genDtMdlProps


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––



  !====================================================[ UNDEFINED VALUES ]=========================================================

  interface undef
    module procedure undefParsInt
    module procedure undefDQsInt
  end interface

  interface isUndef
    module procedure isUndefParsInt
    module procedure isUndefDQsInt
  end interface

  interface isDef
    module procedure isDefParsInt
    module procedure isDefDQsInt
  end interface
  
  interface setIfUndef
    module procedure setIfUndefParsInt
    module procedure setIfUndefDQsInt
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !====================================================[ UNDEFINED VALUES ]=========================================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefParsInt(var)
!
!   sets a var. undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyParsInt) :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_TY_PARS_INT

  end subroutine undefParsInt


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefDQsInt(var)
!
!   sets a var. undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyDQsInt) :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_TY_DQS_INT

  end subroutine undefDQsInt


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefParsInt(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyParsInt) :: &
     var
    logical :: &
     isUndefParsInt
    intent(in) :: &
     var

    isUndefParsInt= &
     var%e == VAL_UNDEF_INT .OR. &
     var%f == VAL_UNDEF_INT .OR. &
     var%chi == VAL_UNDEF_INT .OR. &
     var%omegaPeri == VAL_UNDEF_INT .OR. &
     var%K == VAL_UNDEF_INT .OR. &
     var%V == VAL_UNDEF_INT .OR. &
     var%sgmPlus == VAL_UNDEF_INT .OR. &
     var%omega2 == VAL_UNDEF_INT .OR. &
     var%K1 == VAL_UNDEF_INT .OR. &
     var%K2 == VAL_UNDEF_INT .OR. &
     var%i == VAL_UNDEF_INT .OR. &
     var%OmegaNode == VAL_UNDEF_INT .OR. &
     var%varpi == VAL_UNDEF_INT .OR. &
     var%aRelAng == VAL_UNDEF_INT .OR. &
     var%aAng == VAL_UNDEF_INT .OR. &
     var%alphaR == VAL_UNDEF_INT .OR. &
     var%deltaR == VAL_UNDEF_INT .OR. &
     var%muAlCosDel == VAL_UNDEF_INT .OR. &
     var%muDel == VAL_UNDEF_INT .OR. &
     var%tauPlus == VAL_UNDEF_INT .OR. &
     var%aV == VAL_UNDEF_INT

  end function isUndefParsInt


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefDQsInt(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyDQsInt) :: &
     var
    logical :: &
     isUndefDQsInt
    intent(in) :: &
     var

    isUndefDQsInt= &
     var%P == VAL_UNDEF_INT .OR. &
     var%T == VAL_UNDEF_INT .OR. &
     var%mMin == VAL_UNDEF_INT .OR. &
     var%aSinI == VAL_UNDEF_INT .OR. &
     var%aPMax == VAL_UNDEF_INT .OR. &
     var%KAlt == VAL_UNDEF_INT .OR. &
     var%KAlt1 == VAL_UNDEF_INT .OR. &
     var%KAlt2 == VAL_UNDEF_INT .OR. &
     var%rho == VAL_UNDEF_INT .OR. &
     var%m1 == VAL_UNDEF_INT .OR. &
     var%m2 == VAL_UNDEF_INT .OR. &
     var%d == VAL_UNDEF_INT .OR. &
     var%aRel == VAL_UNDEF_INT .OR. &
     var%a1 == VAL_UNDEF_INT .OR. &
     var%a2 == VAL_UNDEF_INT .OR. &
     var%thetaR == VAL_UNDEF_INT .OR. &
     var%thetaSecR == VAL_UNDEF_INT .OR. &
     var%tT1 == VAL_UNDEF_INT .OR. &
     var%tT2 == VAL_UNDEF_INT .OR. &
     var%omSum == VAL_UNDEF_INT .OR. &
     var%omDiff == VAL_UNDEF_INT .OR. &
     var%om2Sum == VAL_UNDEF_INT .OR. &
     var%om2Diff == VAL_UNDEF_INT .OR. &
     var%m == VAL_UNDEF_INT

  end function isUndefDQsInt


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefParsInt(var)
!
!   Returns:                         whether a var. is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyParsInt) :: &
     var
    logical :: &
     isDefParsInt
    intent(in) :: &
     var

    isDefParsInt=.NOT.isUndef(var)

  end function isDefParsInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefDQsInt(var)
!
!   Returns:                         whether a var. is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyDQsInt) :: &
     var
    logical :: &
     isDefDQsInt
    intent(in) :: &
     var

    isDefDQsInt=.NOT.isUndef(var)

  end function isDefDQsInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefParsInt(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyParsInt) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefParsInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefDQsInt(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyDQsInt) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefDQsInt




  !==========================================================[ DATA ]===============================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine parseData
!
!   parses observational data
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     readStat, &
     iSet, &
     pos, &
     iRec, &
     iTmp, &
     i, &
     iSetOfTy, &
     curDtTy, &
     iGroup
    real(dp) :: &
     valDbl
    logical :: &
     isHipN
    character(STR_LEN_STD) :: &
     lin, &
     linL, &
     tmpStr                          ! for commandline reading



    ! INITIALIZATIONS

    call undef(dtProps%dtTyOfSets)   ! array for all sets
    call undef(dtProps%binCompo)     ! array for all sets
    call undef(dtProps%nBObsWDtTy)
    dtProps%nSetsOfTy=0              ! array for diff. data types
    dtProps%refIsCompa=.FALSE.       ! array for all sets
    dtProps%polCoord=.FALSE.         ! array for all sets

    dtSetsAMAngPos%N=0               ! array for all sets
    dtSetsAMHipN%N=0                 ! array for all sets
    dtSetsRV%N=0                     ! array for all sets
    dtPropsAMAngPos%nRec=0
    dtPropsAMHipN%nRec=0
    dtPropsRV%nRec=0

    dtProps%nRec=0

    dtPropsRV%tRge%lBnd=huge(dtPropsRV%tRge%lBnd)
    dtPropsRV%tRge%uBnd=-huge(dtPropsRV%tRge%uBnd)
    dtPropsAMAngPos%tRge%lBnd=huge(dtPropsAMAngPos%tRge%lBnd)
    dtPropsAMAngPos%tRge%uBnd=-huge(dtPropsAMAngPos%tRge%uBnd)
    dtPropsAMHipN%tRge%lBnd=huge(dtPropsAMHipN%tRge%lBnd)
    dtPropsAMHipN%tRge%uBnd=-huge(dtPropsAMHipN%tRge%uBnd)
    dtProps%sngTimeSpansRge%lBnd=huge(dtProps%sngTimeSpansRge%lBnd)
    dtProps%sngTimeSpansRge%uBnd=-huge(dtProps%sngTimeSpansRge%uBnd)

    call undef(priRgeFByNPeriods)
    call undef(iGroup)


    ! LOOP OVER ALL DATA SETS

    do iSet=1,dtProps%nSets

      iRec=0
      isHipN=isDtFilHipN(iSet)
      if (tmpNGroup > 0) iGroup=posInArr(tmpGroupDtSets(:tmpNGroup),iSet)

      ! open current data file
      if (isHipN) then
        call tryOpenFilRW("data",trim(dataDir)//dtProps%filNms(iSet),UN_DT,"old","read","stream")
        call parseHipNHdr
      else
        call tryOpenFilRW("data",trim(dataDir)//dtProps%filNms(iSet),UN_DT,"old","read")
        call parseDtCmds
      end if

      if (isUndef(dtProps%dtTyOfSets(iSet))) call errDataSet(iSet,"type of data undefined")
      curDtTy=dtProps%dtTyOfSets(iSet)
      if (curDtTy == DT_TY%RV) then
        if (dtProps%refIsCompa(iSet)) &
         call errDataSet(iSet,"using the companion as reference point is not valid for "//trim(DT_TY_DESC(DT_TY%RV)))
        if (dtProps%polCoord(iSet)) &
         call errDataSet(iSet,"using polar coordinates is not valid for "//trim(DT_TY_DESC(DT_TY%RV)))
      end if

      call inc(dtProps%nSetsOfTy(curDtTy))
      iSetOfTy=dtProps%nSetsOfTy(curDtTy)
      if (iSetOfTy > N_DT_SETS_PER_TY_MAX) &
       call errDataSet(iSet,"too many (>"//trim(niceFmt(N_DT_SETS_PER_TY_MAX))//") data sets of type "// &
       trim(DT_TY_DESC(curDtTy)))
      select case (curDtTy)
      case (DT_TY%AMAngPos)
        dtSetsAMAngPos(iSetOfTy)%pol=dtProps%polCoord(iSet)
        dtSetsAMAngPos(iSetOfTy)%refIsCompa=dtProps%refIsCompa(iSet)
        dtSetsAMAngPos(iSetOfTy)%binCompo=dtProps%binCompo(iSet)
        if (isDef(iGroup)) then
          dtSetsAMAngPos(iSetOfTy)%groupThrld=tmpGroupThrlds(iGroup,1)
        else
          call undef(dtSetsAMAngPos(iSetOfTy)%groupThrld)
        end if
      case (DT_TY%RV)
        dtSetsRV(iSetOfTy)%binCompo=dtProps%binCompo(iSet)
        if (isDef(iGroup)) then
          dtSetsRV(iSetOfTy)%groupThrld=tmpGroupThrlds(iGroup,1)
        else
          call undef(dtSetsRV(iSetOfTy)%groupThrld)
        end if
      case (DT_TY%AMHipN)
        dtSetsAMHipN(iSetOfTy)%hipNr=dtFilHipN%hdr%hipNr
        if (.NOT. allEq(dtSetsAMHipN(1:iSetOfTy)%hipNr)) call err(trim(DT_TY_DESC(DT_TY%AMHipN))// &
         " data files contain different system names")
        call setIfUndef(dtProps%sysNm,SYS_NM_PREFX_HIP//trim(niceFmt(dtFilHipN%hdr%hipNr)))
        dtSetsAMHipN(iSetOfTy)%N=dtFilHipN%hdr%nData
        dtPropsAMHipN%nRec=dtPropsAMHipN%nRec+dtSetsAMHipN(iSetOfTy)%N

        dtSetsAMHipN(iSetOfTy)%params%pos0%x=dtFilHipN%hdr%iDEC+dtFilHipN%hdr%dDEC/3.6e8_dp-90._dp ! delta_0 [deg]
        dtSetsAMHipN(iSetOfTy)%params%pos0%y=(dtFilHipN%hdr%iRA+dtFilHipN%hdr%dRA/3.6e8_dp)* &     ! alpha_0·cos(delta_0) [deg]
         cos(dtSetsAMHipN(iSetOfTy)%params%pos0%x*DEG_IN_RAD)
        dtSetsAMHipN(iSetOfTy)%params%mu0%x=.01*MAS_IN_DEG/YR_IN_D*dtFilHipN%hdr%muDel             ! mu_{delta,0} [deg/d]
        dtSetsAMHipN(iSetOfTy)%params%mu0%y=.01*MAS_IN_DEG/YR_IN_D*dtFilHipN%hdr%muAlCosDel        ! mu_{alpha,0}·cos(delta) [deg/d]
        dtSetsAMHipN(iSetOfTy)%params%pi0=1e-5_dp*dtFilHipN%hdr%plx                                ! pi_0 [arcsec]
        if (isDef(iGroup)) then
          dtSetsAMHipN(iSetOfTy)%groupThrld=tmpGroupThrlds(iGroup,:)
        else
          call undef(dtSetsAMHipN(iSetOfTy)%groupThrld)
        end if
      end select

      ! read records
      do
        if ((isHipN .AND. iRec == dtSetsAMHipN(iSetOfTy)%N) .OR. (.NOT. isHipN .AND. readStat /= 0)) exit ! parseDtCmds casts an error if data commands occur after 1st data record
        call inc(iRec)
        call inc(dtProps%nRec)
        if (iRec > N_REC_PER_SET_MAX) call errDataSet(iSet,"too many (>"//trim(niceFmt(N_REC_PER_SET_MAX))// &
         ") records in data set no. "//trim(niceFmt(iSet)))
        call readRec
      end do

      select case (curDtTy)
      case (DT_TY%AMAngPos)
        ! sort by time
        associate(N=>dtSetsAMAngPos(iSetOfTy)%N)
          block
            integer(i8b) :: &
             iArr(N)
            call indxArr(dtSetsAMAngPos(iSetOfTy)%t(:N),iArr)
            dtSetsAMAngPos(iSetOfTy)%t(:N)=dtSetsAMAngPos(iSetOfTy)%t(iArr)
            dtSetsAMAngPos(iSetOfTy)%obs(:N)=dtSetsAMAngPos(iSetOfTy)%obs(iArr)
            dtSetsAMAngPos(iSetOfTy)%unc(:N)=dtSetsAMAngPos(iSetOfTy)%unc(iArr)
          end block
        end associate
      case (DT_TY%AMHipN)
        ! sort by time
        associate(N=>dtSetsAMHipN(iSetOfTy)%N)
          block
            integer(i8b) :: &
             iArr(N)
            call indxArr(dtSetsAMHipN(iSetOfTy)%t(:N),iArr)
            dtSetsAMHipN(iSetOfTy)%t(:N)=dtSetsAMHipN(iSetOfTy)%t(iArr)
            dtSetsAMHipN(iSetOfTy)%obs(:N)=dtSetsAMHipN(iSetOfTy)%obs(iArr)
            dtSetsAMHipN(iSetOfTy)%params%psi(:N)=dtSetsAMHipN(iSetOfTy)%params%psi(iArr)
            dtSetsAMHipN(iSetOfTy)%params%cosPsi(:N)=dtSetsAMHipN(iSetOfTy)%params%cosPsi(iArr)
            dtSetsAMHipN(iSetOfTy)%params%sinPsi(:N)=dtSetsAMHipN(iSetOfTy)%params%sinPsi(iArr)
            dtSetsAMHipN(iSetOfTy)%params%gamma(:N)=dtSetsAMHipN(iSetOfTy)%params%gamma(iArr)
            dtSetsAMHipN(iSetOfTy)%sgm(:N)=dtSetsAMHipN(iSetOfTy)%sgm(iArr)
          end block
        end associate
      case (DT_TY%RV)
        ! sort by time
        associate(N=>dtSetsRV(iSetOfTy)%N)
          block
            integer(i8b) :: &
             iArr(N)
            call indxArr(dtSetsRV(iSetOfTy)%t(:N),iArr)
            dtSetsRV(iSetOfTy)%t(:N)=dtSetsRV(iSetOfTy)%t(iArr)
            dtSetsRV(iSetOfTy)%obs(:N)=dtSetsRV(iSetOfTy)%obs(iArr)
            dtSetsRV(iSetOfTy)%sgm(:N)=dtSetsRV(iSetOfTy)%sgm(iArr)
          end block
        end associate
      end select

      close(UN_DT)

    end do                           ! loop over all data sets
    

    contains

    subroutine parseHipNHdr          ! parses header of AMHipN data file

      implicit none

      dtProps%dtTyOfSets(iSet)=DT_TY%AMHipN

      read(UN_DT) dtFilHipN%hdr%hipNr,dtFilHipN%hdr%nData,dtFilHipN%hdr%dRA,dtFilHipN%hdr%dDEC,dtFilHipN%hdr%plx, &
       dtFilHipN%hdr%muAlCosDel,dtFilHipN%hdr%muDel,dtFilHipN%hdr%iRA,dtFilHipN%hdr%iDEC

      if (dtFilHipN%hdr%nData > N_REC_PER_SET_MAX) call errDataSet(iSet,"too many (>"// &
       trim(niceFmt(N_REC_PER_SET_MAX))//") records in data set no. "//trim(niceFmt(iSet)))

    end subroutine


    subroutine parseDtCmds           ! parses commands in regular data file, returns .FALSE. if final position reached in current file

      implicit none

      do                             ! loop for parsing commands (error if occurring after 1st data record), ignoring comments

        read (UN_DT,'(A)',iostat=readStat) lin
        linL=adjustl(lin)

        if (iRec > 0 .AND. beginsWith(linL,CMD_PREFX_HEADER,pos)) then
          call errDataSet(iSet,"no commands allowed in data file after first data record")

        else if (beginsWith(linL,CMD_PREFX_HEADER//trim(DT_CMDS%sysNm),pos,OPT_LONGER)) then
          tmpStr=trim(adjustl(linL(pos:)))
          if (isUndef(dtProps%sysNm)) then ! system name not yet set
            dtProps%sysNm=tmpStr
          else
            if (trim(tmpStr) /= trim(dtProps%sysNm)) call warnDataSet(iSet,"data files contain different system names")
          end if

        else if (beginsWith(linL,CMD_PREFX_HEADER//trim(DT_CMDS%dtTy),pos,OPT_LONGER)) then
          if (isDef(dtProps%dtTyOfSets(iSet))) call errDataSet(iSet,"data type is set several times in current file")
          tmpStr=trim(adjustl(linL(pos:)))
          iTmp=posInArr(DT_TY_CMDS,tmpStr)
          select case (iTmp)
          case (VAL_UNDEF_INT)
            call errDataSet(iSet,"data type '"//trim(tmpStr)//"' unknown")
          case default
            dtProps%dtTyOfSets(iSet)=iTmp
          end select

        else if (trim(linL) == CMD_PREFX_HEADER//trim(DT_CMDS%polCoord)) then
          dtProps%polCoord(iSet)=.TRUE.

        else if (beginsWith(linL,CMD_PREFX_HEADER//trim(DT_CMDS%binCompo),pos,OPT_LONGER)) then
          if (isDef(dtProps%binCompo(iSet))) call errDataSet(iSet, &
           "binary component number is set several times in current file")
          iTmp=intFromStr(trim(adjustl(linL(pos:))))
          if (iTmp == 1 .OR. iTmp == 2) then
            dtProps%binCompo(iSet)=iTmp
          else
            call errDataSet(iSet,"binary-component number '"//trim(tmpStr)//"' not allowed")
          end if

        else if (beginsWith(linL,CMD_PREFX_HEADER//trim(DT_CMDS%mS),pos,OPT_LONGER)) then
          read(linL(pos:),*) valDbl
          if (isDef(mdlProps%mS) .AND. mdlProps%mS /= valDbl) call errDataSet(iSet, &
           "stellar mass is set several times to different values")
          if (valDbl <= 0) call errDataSet(iSet,"stellar mass must be positive")
          mdlProps%mS=valDbl

        else if (trim(linL) == CMD_PREFX_HEADER//trim(DT_CMDS%refIsCompa)) then
          dtProps%refIsCompa(iSet)=.TRUE.

        else if (beginsWith(linL,CMD_PREFX_HEADER,pos)) then
          call errDataSet(iSet,"invalid command in data file: '"//trim(linL(pos:))//"'")

        else if (readStat /= 0) then
          exit

        else if (trim(linL) /= "" .AND. linL(1:1) /= COMMENT_PREFX) then
          exit

        end if

      end do                         ! loop for parsing commands

    end subroutine


    subroutine readRec               ! reads next record from data file

      implicit none

      if (isHipN) then

        read(UN_DT) dtFilHipN%dt%da,dtFilHipN%dt%dmy1,dtFilHipN%dt%BJD,dtFilHipN%dt%dmy2,dtFilHipN%dt%psi,dtFilHipN%dt%dmy3, &
         dtFilHipN%dt%gamma,dtFilHipN%dt%sgmDa,dtFilHipN%dt%dmy4
        dtSetsAMHipN(iSetOfTy)%obs(iRec)=1e-2_dp*dtFilHipN%dt%da                                ! [mas]
        dtSetsAMHipN(iSetOfTy)%t(iRec)=2440000._dp+1e-5_dp*dtFilHipN%dt%BJD                     ! BJD [d]
        dtSetsAMHipN(iSetOfTy)%params%psi(iRec)=1e-5_dp*dtFilHipN%dt%psi                        ! [rad]
        dtSetsAMHipN(iSetOfTy)%params%gamma(iRec)=1e-4_dp*dtFilHipN%dt%gamma                    ! unitless
        dtSetsAMHipN(iSetOfTy)%sgm(iRec)=1e-2_dp*dtFilHipN%dt%sgmDa                             ! [mas]

      else

        linL=trim(linL)//"/"

        select case (curDtTy)
        case (DT_TY%AMAngPos)
          call inc(dtSetsAMAngPos(iSetOfTy)%N)
          call inc(dtPropsAMAngPos%nRec)

          read (linL,*) dtSetsAMAngPos(iSetOfTy)%t(iRec), &
           dtSetsAMAngPos(iSetOfTy)%obs(iRec)%y,dtSetsAMAngPos(iSetOfTy)%obs(iRec)%x, &
           dtSetsAMAngPos(iSetOfTy)%unc(iRec)%a,dtSetsAMAngPos(iSetOfTy)%unc(iRec)%b,dtSetsAMAngPos(iSetOfTy)%unc(iRec)%phi
          ! CAUTION: rectangular system is right-handed: pos%x=delta, pos%y=alpha·cos(delta)

            ! check if data are ok

          ! position (for polar coords.)
          if (dtSetsAMAngPos(iSetOfTy)%pol) then
            if (.NOT. isInIntvl(tyIntvlDbl(0._dp,360._dp),dtSetsAMAngPos(iSetOfTy)%obs(iRec)%x,openR=.TRUE.)) &
             call errDataSet(iSet,"position angles of "//trim(DT_TY_DESC(DT_TY%AMAngPos))//" must be >=0° and <360°")
            if (dtSetsAMAngPos(iSetOfTy)%obs(iRec)%y <= 0) call errDataSet(iSet, &
             "angular distances of "//trim(DT_TY_DESC(DT_TY%AMAngPos))//" must be positive")
          end if

          ! uncertainty ellipse
          if (dtSetsAMAngPos(iSetOfTy)%unc(iRec)%a <= 0 .OR. dtSetsAMAngPos(iSetOfTy)%unc(iRec)%b <= 0) &
           call errDataSet(iSet, &
           "lengths of axes of uncertainty ellipses of "//trim(DT_TY_DESC(DT_TY%AMAngPos))//" must be positive")
          if (.NOT. isInIntvl(tyIntvlDbl(0._dp,180._dp),dtSetsAMAngPos(iSetOfTy)%unc(iRec)%phi,openR=.TRUE.)) &
           call errDataSet(iSet, &
           "position angles of uncertainty ellipses of "//trim(DT_TY_DESC(DT_TY%AMAngPos))//" must be >=0° and <180°")

            ! transformations
          
          ! phi in radians
          dtSetsAMAngPos(iSetOfTy)%unc(iRec)%phi=dtSetsAMAngPos(iSetOfTy)%unc(iRec)%phi*DEG_IN_RAD

          ! from polar to rectangular coords.: (Delta(alpha cos(delta)),Delta(delta))=rho*(sin(theta),cos(theta))
          !  (cf. [06.04.2010/2])
          block
            type(tyCoord2DDbl) :: &
             tmpPos

            tmpPos=dtSetsAMAngPos(iSetOfTy)%obs(iRec)
            if (dtSetsAMAngPos(iSetOfTy)%pol) then
              dtSetsAMAngPos(iSetOfTy)%obs(iRec)%x=tmpPos%y*cos(tmpPos%x*DEG_IN_RAD)
              dtSetsAMAngPos(iSetOfTy)%obs(iRec)%y=tmpPos%y*sin(tmpPos%x*DEG_IN_RAD)
              dtSetsAMAngPos(iSetOfTy)%sep(iRec)=tmpPos%y
            else
              dtSetsAMAngPos(iSetOfTy)%sep(iRec)=sqrt(tmpPos%x**2+tmpPos%y**2)
            end if

          end block

        case (DT_TY%RV)
          call inc(dtSetsRV(iSetOfTy)%N)
          call inc(dtPropsRV%nRec)

          read (linL,*) dtSetsRV(iSetOfTy)%t(iRec),dtSetsRV(iSetOfTy)%obs(iRec),dtSetsRV(iSetOfTy)%sgm(iRec)

            ! check if data are ok

          ! time
          if (dtSetsRV(iSetOfTy)%t(iRec) < 0) call errDataSet(iSet,"times must be positive")

          ! uncertainties
          if (dtSetsRV(iSetOfTy)%sgm(iRec) <= 0) call errDataSet(iSet,"radial-velocity uncertainties must be positive")

        end select

        read (UN_DT,'(A)',iostat=readStat) lin
        linL=adjustl(lin)

      end if                         ! (isHipN)

    end subroutine


    function isDtFilHipN(iSet)       ! returns whether a non-opened file is (likely) a AMHipN data file

      implicit none
      integer :: &
       iSet, &
       status(13)
      logical :: &
       isDtFilHipN
      intent(in) :: &
       iSet

      call tryOpenFilRW("data",trim(dataDir)//dtProps%filNms(iSet),UN_DT,"old","read","stream")
      call fstat(UN_DT,status)
      read(UN_DT) dtFilHipN%hdr%hipNr,dtFilHipN%hdr%nData,dtFilHipN%hdr%dRA,dtFilHipN%hdr%dDEC,dtFilHipN%hdr%plx, &
       dtFilHipN%hdr%muAlCosDel,dtFilHipN%hdr%muDel,dtFilHipN%hdr%iRA,dtFilHipN%hdr%iDEC
      close(UN_DT)

      ! heuristic check cf. [02.02.2012/1]
      isDtFilHipN=(dtFilHipN%hdr%hipNr >= 1 .AND. dtFilHipN%hdr%hipNr <= 120404 .AND. status(8) == 32+dtFilHipN%hdr%nData*64 .AND. &
      dtFilHipN%hdr%iRA >= 0 .AND. dtFilHipN%hdr%iRA <= 360 .AND. dtFilHipN%hdr%iDEC >= 0 .AND. dtFilHipN%hdr%iDEC <= 180)

    end function

  end subroutine parseData




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine groupData
!
!   arranges data into groups
!
!   as many successive data as possible are included in a given group without exceeding any of the relevant thresholds
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    
    do i=1,dtProps%nSetsOfTy(DT_TY%AMAngPos)
      if (isDef(dtSetsAMAngPos(i)%groupThrld(1))) then
        call startSubsect(UNIT%cons+UNIT%log,"Grouping data set #"//trim(niceFmt(i))//" of type "// &
         trim(DT_TY_DESC(DT_TY%AMAngPos)))
        if (all(isDef(dtSetsAMAngPos(i)%groupThrld(:N_GROUP_THRLDS(DT_TY%AMAngPos))))) then
          call groupDataSet(dtSetsAMAngPos(i))
        else
          call err("incomplete set of thresholds")
        end if
        call endSubsect(UNIT%cons+UNIT%log)
      end if
    end do
    do i=1,dtProps%nSetsOfTy(DT_TY%RV)
      if (isDef(dtSetsRV(i)%groupThrld(1))) then
        call startSubsect(UNIT%cons+UNIT%log,"Grouping data set #"//trim(niceFmt(i))//" of type "// &
         trim(DT_TY_DESC(DT_TY%RV)))
        if (all(isDef(dtSetsRV(i)%groupThrld(:N_GROUP_THRLDS(DT_TY%RV))))) then
          call groupDataSet(dtSetsRV(i))
        else
          call err("incomplete set of thresholds")
        end if
        call endSubsect(UNIT%cons+UNIT%log)
      end if
    end do
    do i=1,dtProps%nSetsOfTy(DT_TY%AMHipN)
      if (isDef(dtSetsAMHipN(i)%groupThrld(1))) then
        call startSubsect(UNIT%cons+UNIT%log,"Grouping data set #"//trim(niceFmt(i))//" of type "// &
         trim(DT_TY_DESC(DT_TY%AMHipN)))
        if (all(isDef(dtSetsAMHipN(i)%groupThrld(:N_GROUP_THRLDS(DT_TY%AMHipN))))) then
          call groupDataSet(dtSetsAMHipN(i))
        else
          call err("incomplete set of thresholds")
        end if
        call endSubsect(UNIT%cons+UNIT%log)
      end if
    end do

  end subroutine groupData




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine groupDataSet(dtSet)
!
!   arranges data into groups (see groupData)
!
!   Arguments:
!     dtSet                          data set whose records are to be grouped
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type tyGroup
      integer :: &
       st, &
       siz, &
       constr
    end type
    class(tyDtSetGeneric) :: &
     dtSet
    integer :: &
     st, &
     siz, &
     nGrp, &
     N, &
     i, &
     constr
    character(STR_LEN_LONGEST) :: &
     cGrp, &
     cConstr
    type(tyGroup) :: &
     grp(dtSet%N)
    intent(in) :: &
     dtSet
    
    N=dtSet%N
    cGrp=""
    cConstr=""
    call undef(grp%constr)

    st=1
    nGrp=0
    do
      siz=2
      do
        if (st+siz-1 > N) exit
        constr=constrThrld()
        if (constr > 0) exit
        call inc(siz)
      end do
      call dec(siz)                  ! needed in both cases, i.e. constr > 0 or siz > N
      call inc(nGrp)
      grp(nGrp)%st=st
      grp(nGrp)%siz=siz
      grp(nGrp)%constr=constr
      cGrp=trim(cGrp)//trim(niceFmt(siz))//","
      cConstr=trim(cConstr)//trim(mrg(STR_NONE,niceFmt(constr),constr == 0))//","
      st=st+siz
      if (st > N) exit
    end do
    cGrp=rmTrail(trim(cGrp),",")
    cConstr=rmTrail(trim(cConstr),",")

    if (any(grp(:nGrp)%siz > 1)) then
      do i=nGrp,1,-1
        if (grp(i)%siz > 1) call makeGroup(i)
      end do
      call startSubsect(UNIT%cons+UNIT%log,"Resulting groups")
      call startSubsect(UNIT%cons+UNIT%log,"Sizes")
      call printMsg(UNIT%cons+UNIT%log,cGrp)
      call endSubsect(UNIT%cons+UNIT%log)
      call startSubsect(UNIT%cons+UNIT%log,"Constrained first by thresholds number")
      call printMsg(UNIT%cons+UNIT%log,cConstr)
      call endSubsect(UNIT%cons+UNIT%log)
      call endSubsect(UNIT%cons+UNIT%log) ! Resulting groups
      call startSubsect(UNIT%cons+UNIT%log,"# records")
      call printQty(UNIT%cons+UNIT%log,"Original",N)
      call printQty(UNIT%cons+UNIT%log,"New",dtSet%N)
      call endSubsect(UNIT%cons+UNIT%log)
      call warn("This run will be based on modified data")
      call warn("To read samples of this run using the "//trim(CL_OPTS(I_CL_OPTS%readSamp)%short)//" option, the same "// &
       trim(CL_OPTS(I_CL_OPTS%group)%short)//" option(s) must be used")
    else
      call printMsg(UNIT%cons+UNIT%log,"(no groups below given thresholds)")
    end if


    contains
    
    function constrThrld()           ! returns the constraining threshold
      implicit none
      integer :: &
       constrThrld
      
      constrThrld=0
      associate(fin=>st+siz-1)
        select type (dtSet)
        type is (tyDtSetAMAngPos)
          if (arrRge(dtSet%t(st:fin)) > dtSet%groupThrld(1)) then
            constrThrld=1
          else if (arrRge(dtSet%unc(st:fin)%phi) > dtSet%groupThrld(2)) then
            constrThrld=2
          end if
        type is (tyDtSetRV)
          if (arrRge(dtSet%t(st:fin)) > dtSet%groupThrld(1)) constrThrld=1
        type is (tyDtSetAMHipN)
          if (arrRge(dtSet%t(st:fin)) > dtSet%groupThrld(1)) then
            constrThrld=1
          else if (arrRge(dtSet%params%psi(st:fin)) > dtSet%groupThrld(2)) then
            constrThrld=2
          else if (arrRge(dtSet%params%gamma(st:fin))/arrMean(dtSet%params%gamma(st:fin)) > dtSet%groupThrld(3)) then
            constrThrld=3
          end if
        end select
      end associate
    end function
    

    subroutine makeGroup(iGroup)     ! creates a new group from a subset of original data
      implicit none
      integer :: &
       iGroup, &
       st, &
       siz, &
       fin, &
       nDel
      intent(in) :: &
       iGroup
      
      st=grp(iGroup)%st
      siz=grp(iGroup)%siz
      fin=st+siz-1
      nDel=siz-1

      dtProps%nRec=dtProps%nRec-nDel
      dtSet%N=dtSet%N-nDel

      block
        real(dp) :: &
         weights(st:fin), &
         xMod(st:fin), &
         yMod(st:fin), &
         xModMean, &
         yModMean

        select type (dtSet)
        type is (tyDtSetAMAngPos)
          dtPropsAMAngPos%nRec=dtPropsAMAngPos%nRec-nDel
          
          weights=1._dp/(dtSet%unc(st:fin)%a**2+dtSet%unc(st:fin)%b**2)

          dtSet%t(st)=arrMean(dtSet%t(st:fin),weights)
          dtSet%unc(st)%phi=arrMean(dtSet%unc(st:fin)%phi,weights)

          associate(phi=>dtSet%unc(st)%phi)
            ! rotate coord. system s.t. new x'-axis is along semi-major axis of "mean" ellipse determined by phi
            xMod=dtSet%obs(st:fin)%x*cos(phi)+dtSet%obs(st:fin)%y*sin(phi)
            yMod=-dtSet%obs(st:fin)%x*sin(phi)+dtSet%obs(st:fin)%y*cos(phi)

            ! take weighted mean according to individual uncertainty-ellipse axes, independently in (x',y')
            xModMean=arrMean(xMod,1._dp/(dtSet%unc(st:fin)%a**2))
            yModMean=arrMean(yMod,1._dp/(dtSet%unc(st:fin)%b**2))

            ! rotate back
            dtSet%obs(st)%x=xModMean*cos(phi)-yModMean*sin(phi)
            dtSet%obs(st)%y=xModMean*sin(phi)+yModMean*cos(phi)
          end associate

          ! new uncertainty is weighted mean independently in (x',y')
          dtSet%unc(st)%a=sqrt(1._dp/sum(1._dp/dtSet%unc(st:fin)%a**2))
          dtSet%unc(st)%b=sqrt(1._dp/sum(1._dp/dtSet%unc(st:fin)%b**2))

          dtSet%t=delSubset(dtSet%t,st+1,nDel)
          dtSet%obs%x=delSubset(dtSet%obs%x,st+1,nDel)
          dtSet%obs%y=delSubset(dtSet%obs%y,st+1,nDel)
          dtSet%unc%a=delSubset(dtSet%unc%a,st+1,nDel)
          dtSet%unc%b=delSubset(dtSet%unc%b,st+1,nDel)
          dtSet%unc%phi=delSubset(dtSet%unc%phi,st+1,nDel)
        type is (tyDtSetRV)
          weights=1._dp/dtSet%sgm(st:fin)**2

          dtPropsRV%nRec=dtPropsRV%nRec-nDel

          dtSet%t(st)=arrMean(dtSet%t(st:fin),weights)
          dtSet%obs(st)=arrMean(dtSet%obs(st:fin),weights)
          dtSet%sgm(st)=sqrt(1._dp/sum(weights))

          dtSet%t=delSubset(dtSet%t,st+1,nDel)
          dtSet%obs=delSubset(dtSet%obs,st+1,nDel)
          dtSet%sgm=delSubset(dtSet%sgm,st+1,nDel)
        type is (tyDtSetAMHipN)
          weights=1._dp/dtSet%sgm(st:fin)**2

          dtPropsAMHipN%nRec=dtPropsAMHipN%nRec-nDel

          dtSet%t(st)=arrMean(dtSet%t(st:fin),weights)
          dtSet%obs(st)=arrMean(dtSet%obs(st:fin),weights)
          dtSet%sgm(st)=sqrt(1._dp/sum(weights))
          dtSet%params%psi(st)=arrMean(dtSet%params%psi(st:fin),weights)
          dtSet%params%gamma(st)=arrMean(dtSet%params%gamma(st:fin),weights)

          dtSet%t=delSubset(dtSet%t,st+1,nDel)
          dtSet%obs=delSubset(dtSet%obs,st+1,nDel)
          dtSet%sgm=delSubset(dtSet%sgm,st+1,nDel)
          dtSet%params%psi=delSubset(dtSet%params%psi,st+1,nDel)
          dtSet%params%gamma=delSubset(dtSet%params%gamma,st+1,nDel)
        end select
      end block
    end subroutine
 
  end subroutine groupDataSet




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine warnDataSet(i,msg)
!
!   displays warning message referring to specific data set
!
!   Arguments:
!     i                              data-set index
!     msg                            message string (should start with minuscle)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    character(*) :: &
     msg
    intent(in) :: &
     msg, &
     i

    call warn("data set #"//trim(niceFmt(i))//": "//trim(adjustl(msg)))

  end subroutine warnDataSet




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine errDataSet(i,msg)
!
!   displays error message referring to specific data set and stops
!
!   Arguments:
!     i                              data-set index
!     msg                            message string (should start with minuscle)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    character(*) :: &
     msg
    intent(in) :: &
     msg, &
     i

    call err("data set #"//trim(niceFmt(i))//": "//trim(adjustl(msg)))

  end subroutine errDataSet




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine generateData
!
!   generates synthetic data
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iCmd, &
     i
    real(dp) :: &
     curPtsX(genDtProps%nRec), &
     curPtsY(genDtProps%nRec), &
     curPtsY2(genDtProps%nRec), &
     curPts2D(genDtProps%nRec,genDtProps%nRec), &
     curPts2DTmp(genDtProps%nRec,genDtProps%nRec), &
     sgmVArr(genDtProps%nRec), &
     sgmVArr2(genDtProps%nRec), &
     CSVArr(genDtProps%nRec,6)
    logical :: &
     sampTmUnc
    character(STR_LEN_LONG) :: &
     dtCmds(GEN_D_N_CMDS_MAX), &
     dtCmds2(GEN_D_N_CMDS_MAX)
    type(tyIntvlDbl) :: &
     tRge
    type(tyUncEllDbl) :: &
     allUncEll(genDtProps%nRec)


    call setIfUndef(genDtProps%epoch,intvlCentre(tRge))
    mdlProps%RVAccel=.TRUE.
    call setMdlFctPtrs(genDtProps,genDtMdlProps)

    eMin=0.
    eDiff=1.
    curNPars=genDtNPars
    tRge=genDtProps%tRge
    sampTmUnc=.NOT.allocated(genDtProps%times)

    if (genDtMdlProps%bin .OR. genDtPsdNPl > 0) call preCalcEA

    ! times
    if (sampTmUnc) then
      call startSubsect(UNIT%cons+UNIT%log,"Sampling observing times")
      curPtsX=sampTimes(genDtProps%nRec,tRge,.NOT. genDtProps%uniTm,OPT_WINDOW)
      call endSubsect(UNIT%cons+UNIT%log) ! Sampling observing times
    else
      curPtsX=genDtProps%times
    end if

    ! commands
    iCmd=1
    dtCmds(iCmd)=COMMENT_PREFX//"data generated with: "//trim(cmdLin)
    call inc(iCmd)
    dtCmds(iCmd)=COMMENT_PREFX//"epoch="//trim(niceFmt(genDtProps%epoch))
    call inc(iCmd)
    dtCmds(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%sysNm)//trim(genDtProps%sysNm)
    call inc(iCmd)
    if (.NOT. genDtMdlProps%bin) then
      dtCmds(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%mS)//trim(niceFmt(genDtMdlProps%mS)) ! stellar mass
      call inc(iCmd)
    end if


    select case (genDtProps%dtTy)
    case (DT_TY%AMAngPos)

      ! set par. roles
      call setParsDQs(.TRUE.,genDtProps,genDtPsdNPl,OPT_NO_DQS)

      ! commands
      dtCmds(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%dtTy)//DT_TY_CMDS(DT_TY%AMAngPos) ! data type
      call inc(iCmd)
      dtCmds(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%binCompo)//trim(niceFmt(BIN_COMPO_AM_ANG_POS_STD)) ! binary component 2 only so far
      call inc(iCmd)
      dtCmds(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%refIsCompa) ! reference is companion
      call inc(iCmd)
      if (genDtProps%polCoord(1)) then
        dtCmds(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%polCoord) ! polar coordinates
        call inc(iCmd)
      end if

      ! sample function
      call calcMdlFctWErr(int(genDtProps%nRec,i8b),MDL_FCT_DIM%AMAngPos,genDtMdlProps%AMAngPos%mdlFct, &
       reshape(genDtPars,[1,curNPars]),identInt8,curPtsX-genDtProps%epoch,nxtInt8,curPts2D, &
       tyParamsMdl(genDtPsdNPl,1,BIN_COMPO_AM_ANG_POS_STD),sampUnc=sampTmUnc,uncIn2=genDtUncEll,uncSamp2=allUncEll) ! pars IS CONTIGUOUS

      ! coordinate system, cf. [11.10.2010/1]
      if (genDtProps%polCoord(1)) then
        curPts2DTmp=curPts2D
        curPts2D(:,1)=modulo(atan2(curPts2DTmp(:,2),curPts2DTmp(:,1))+TWO_PI,TWO_PI)/DEG_IN_RAD
        curPts2D(:,2)=sqrt(curPts2DTmp(:,2)**2+curPts2DTmp(:,1)**2)
      end if

      CSVArr(:,1)=curPtsX
      CSVArr(:,2)=curPts2D(:,2)
      CSVArr(:,3)=curPts2D(:,1)
      CSVArr(:,4)=allUncEll(:)%a
      CSVArr(:,5)=allUncEll(:)%b
      CSVArr(:,6)=allUncEll(:)%phi

      ! write CSV output
      call writeArrCSV(trim(dataDir)//trim(dtProps%filNms(1)),CSVArr,dtCmds(:iCmd-1)) ! arr IS CONTIGUOUS


    case (DT_TY%RV)

      ! set par. roles
      if (genDtMdlProps%bin) then
        call setParsDQs(.TRUE.,genDtProps,genDtPsdNPl,OPT_NO_DQS)
      else
        call setParsDQs(.FALSE.,genDtProps,genDtPsdNPl,OPT_NO_DQS)
      end if

      ! commands
      dtCmds(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%dtTy)//DT_TY_CMDS(DT_TY%RV) ! data type
      call inc(iCmd)

      ! sample function
      if (genDtMdlProps%bin) then
        !commands
        dtCmds2=dtCmds
          ! binary components
        dtCmds(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%binCompo)//"1"
        dtCmds2(iCmd)=CMD_PREFX_HEADER//trim(DT_CMDS%binCompo)//"2"
        call inc(iCmd)

        ! 1st component
        call calcMdlFctWErr(int(genDtProps%nRec,i8b),MDL_FCT_DIM%RV,mdlFctBMdRV,reshape(genDtPars,[1,genDtNPars]),identInt8, &
         curPtsX-genDtProps%epoch,nxtInt8,curPtsY,tyParamsMdl(1,1,1),sampUnc=sampTmUnc,uncIn1=genDtSgm,uncSamp1=sgmVArr) ! pars IS CONTIGUOUS
        ! 2nd component
        call calcMdlFctWErr(int(genDtProps%nRec,i8b),MDL_FCT_DIM%RV,mdlFctBMdRV,reshape(genDtPars,[1,genDtNPars]),identInt8, &
         curPtsX-genDtProps%epoch,nxtInt8,curPtsY2,tyParamsMdl(1,1,2),sampUnc=sampTmUnc,uncIn1=genDtSgm,uncSamp1=sgmVArr2) ! pars IS CONTIGUOUS

        ! write CSV output
        call writeCSV(trim(dataDir)//trim(dtProps%filNms(1)),curPtsX,curPtsY,sgmVArr,dtCmds(:iCmd-1))
        call writeCSV(trim(dataDir)//trim(dtProps%filNms(2)),curPtsX,curPtsY2,sgmVArr2,dtCmds2(:iCmd-1))
      else
        call calcMdlFctWErr(int(genDtProps%nRec,i8b),MDL_FCT_DIM%RV,mdlFctNMdRV,reshape(genDtPars,[1,genDtNPars]),identInt8, &
         curPtsX-genDtProps%epoch,nxtInt8,curPtsY,tyParamsMdl(genDtPsdNPl,1),sampUnc=sampTmUnc,uncIn1=genDtSgm,uncSamp1=sgmVArr) ! pars IS CONTIGUOUS

        ! write CSV output
        call writeCSV(trim(dataDir)//trim(dtProps%filNms(1)),curPtsX,curPtsY,sgmVArr,dtCmds(:iCmd-1))
      end if

    case (DT_TY%AMHipN)

      call err("generating "//trim(DT_TY_DESC(DT_TY%AMHipN))//" data not implemented")

    end select

  end subroutine generateData




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function sampTimes(N,tRge,scatter,window)
!
!   Returns:                         samples a set of observing times for synthetic data
!
!   Arguments:
!     N                              # times
!     tRge                           time range
!     scatter                        whether the times should be scattered (i.e. randomly sampled)
!     window                         whether observing windows must be obeyed
!
!   CAUTION: not thread-safe!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     N, &
     i
    real(dp) :: &
     sampTimes(N), &
     w
    logical :: &
     scatter, &
     window
    type(tyIntvlDbl) :: &
     tRge
    intent(in) :: &
     N, &
     tRge, &
     scatter, &
     window

    w=tRge%width()

    if (scatter) then
      sampTimes(1)=tRge%lBnd
      sampTimes(N)=tRge%uBnd
      do i=2,N-1
        do
          sampTimes(i)=tRge%lBnd+rndUni()*w
          if (.NOT. window .OR. conformsWithWindows(sampTimes(i),tRge%lBnd)) exit
        end do
        if (window) call printPrgrs([i],[2],[N-1])
      end do
      call sort(sampTimes)
    else
      do i=1,N
        sampTimes(i)=tRge%lBnd+(i-1)*w/(N-1)
      end do
    end if

  end function sampTimes




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function conformsWithWindows(t,t1)
!
!   Returns:                         whether given t conforms with all data-generation time windows referring to t1,
!                                    where 1st half of period is forbidden
!
!   Arguments:
!     t                              time in question
!     t1                             reference time for windows
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     t, &
     t1
    logical :: &
     conformsWithWindows
    intent(in) :: &
     t, &
     t1

    conformsWithWindows=all(merge(modulo(t-t1,genDtProps%win(:genDtProps%nWin))/genDtProps%win(:genDtProps%nWin) <=  &
     genDtProps%winFrac(:genDtProps%nWin),modulo(t-t1,genDtProps%win(:genDtProps%nWin))/genDtProps%win(:genDtProps%nWin) >= &
     1._dp+genDtProps%winFrac(:genDtProps%nWin),genDtProps%winFrac(:genDtProps%nWin) > 0))

  end function conformsWithWindows




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtDtSet(dtTy,iSet)
!
!   Returns:                         formatted string identifying a data set
!
!   Arguments:
!     dtTy                           data type of set in question
!     iSet                           data-set index
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     dtTy, &
     iSet
    character(STR_LEN_STD) :: &
     fmtDtSet
    intent(in) :: &
     dtTy, &
     iSet

    fmtDtSet=mrg(mrg("","_set_"//trim(niceFmt(iSet)),dtProps%nSetsOfTy(dtTy) == 1),"_all-sets",isDef(iSet))

  end function fmtDtSet




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtObsvble(dtTy,dimArg,nameArg,forPlt)
!
!   Returns:                         formatted string with name and unit of an observable
!
!   Arguments:
!     dtTy                           data type which the observable corresponds to
!     dimArg                         dimension for 2D observables
!     nameArg                        optionally specified name of observable
!     forPlt                         whether the result is used for plotting
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     dtTy, &
     dimArg, &
     dim
    logical :: &
     forPlt, &
     tex
    character :: &
     mathEsc
    character(STR_LEN_STD) :: &
     fmtObsvble, &
     name, &
     unit
    character(*) :: &
     nameArg
    optional :: &
     nameArg, &
     forPlt
    intent(in) :: &
     dtTy, &
     dimArg, &
     forPlt, &
     nameArg

    dim=merge(0,dimArg,isUndef(dimArg))
    tex=pltPSTricks .AND. present(forPlt) .AND. forPlt
    if (present(nameArg)) then
      name=nameArg
    else
      name=mrg(OBSVBLE_NM_TEX(dtTy,dim),OBSVBLE_NM(dtTy,dim),tex)
    end if
    mathEsc=merge("$"," ",tex)
    unit=mrg(OBSVBLE_UNIT_TEX(dtTy,dim),OBSVBLE_UNIT(dtTy,dim),tex)

    fmtObsvble=trim(mathEsc)//trim(name)//" ["//trim(unit)//"]"//trim(mathEsc)

  end function fmtObsvble




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printDtMdlProps
!
!   prints relevant properties of data and (stellar) model
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i

    call startSubsect(UNIT%cons+UNIT%log,"Target properties")
    call printQty(UNIT%cons+UNIT%log,"Identifier",dtProps%sysNm)
    if (.NOT. mdlProps%bin) call printQty(UNIT%cons+UNIT%log,"Mass of star [M_Sun]",mdlProps%mS)
    call endSubsect(UNIT%cons+UNIT%log) ! Stellar properties


    call startSubsect(UNIT%cons+UNIT%log,"Data properties")

    call printQty(UNIT%cons+UNIT%log,"# observed targets",dtProps%nBObs)
    call printQty(UNIT%cons+UNIT%log,"# data sets",dtProps%nSets)
    call printQty(UNIT%cons+UNIT%log,"# data records",dtProps%nRec)
    call printQty(UNIT%cons+UNIT%log,"# unique timestamps",nUniqT)
    call printQty(UNIT%cons+UNIT%log,"Total timespan [d]",dtProps%tRge,OPT_WITH_RGE)


    call startSubsect(UNIT%cons+UNIT%log,"By data type")

    ! Astrometry (all sub-types)

    if (dtProps%nSetsOfTy(DT_TY%AMAngPos) > 0) then

      call startSubsect(UNIT%cons+UNIT%log,"Astrometry (all sub-types)")
      call printQty(UNIT%cons+UNIT%log,"# observed targets",dtProps%nBObsWDtTy(DT_TY%AMAngPos))
      call endSubsect(UNIT%cons+UNIT%log) ! Astrometry (all sub-types)

      ! Astrometry (angular positions)

      call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(DT_TY%AMAngPos))))

      call startSubsect(UNIT%cons+UNIT%log,"General")
      call printQty(UNIT%cons+UNIT%log,"# data records",dtPropsAMAngPos%nRec)
      call printQty(UNIT%cons+UNIT%log,"Timespan [d]",dtPropsAMAngPos%tRge,OPT_WITH_RGE)
      call printQty(UNIT%cons+UNIT%log,"Separation range [mas]",dtPropsAMAngPos%sepRge)
      call endSubsect(UNIT%cons+UNIT%log) ! General

      call startSubsect(UNIT%cons+UNIT%log,"By data sets")
      do i=1,dtProps%nSetsOfTy(DT_TY%AMAngPos)
        call startSubsect(UNIT%cons+UNIT%log,"Data set "//niceFmt(i))
        if (isDef(dtSetsAMAngPos(i)%binCompo)) &
         call printQty(UNIT%cons+UNIT%log,"Binary component",dtSetsAMAngPos(i)%binCompo)
        call printQty(UNIT%cons+UNIT%log,"Reference point is companion",dtSetsAMAngPos(i)%refIsCompa)
        call printQty(UNIT%cons+UNIT%log,"Polar coordinates",dtSetsAMAngPos(i)%pol)
        call printQty(UNIT%cons+UNIT%log,"# data records",dtSetsAMAngPos(i)%N)
        call printQty(UNIT%cons+UNIT%log,"Timespan [d]",dtSetsAMAngPos(i)%tRge,OPT_WITH_RGE)
        call printQty(UNIT%cons+UNIT%log,"Separation range [mas]",dtSetsAMAngPos(i)%sepRge)
        call endSubsect(UNIT%cons+UNIT%log) ! Data set …
      end do
      call endSubsect(UNIT%cons+UNIT%log) ! By data set

      call endSubsect(UNIT%cons+UNIT%log) ! DT_TY_DESC(DT_TY%AMAngPos)

    end if                           ! (dtProps%nSetsOfTy(DT_TY%AMAngPos) > 0)


    ! RV

    if (dtProps%nSetsOfTy(DT_TY%RV) > 0) then

      call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(DT_TY%RV))))

      call startSubsect(UNIT%cons+UNIT%log,"General")
      call printQty(UNIT%cons+UNIT%log,"# observed targets",dtProps%nBObsWDtTy(DT_TY%RV))
      call printQty(UNIT%cons+UNIT%log,"# data records",dtPropsRV%nRec)
      call printQty(UNIT%cons+UNIT%log,"Timespan [d]",dtPropsRV%tRge,OPT_WITH_RGE)
      call printQty(UNIT%cons+UNIT%log,"Range of "//trim(fmtObsvble(DT_TY%RV,VAL_UNDEF_INT,forPlt=.FALSE.)), &
       dtPropsRV%obsRge,OPT_WITH_RGE)
      call endSubsect(UNIT%cons+UNIT%log) ! General

      call startSubsect(UNIT%cons+UNIT%log,"By data sets")
      do i=1,dtProps%nSetsOfTy(DT_TY%RV)
        call startSubsect(UNIT%cons+UNIT%log,"Data set "//niceFmt(i))
        if (isDef(dtSetsRV(i)%binCompo)) &
        call printQty(UNIT%cons+UNIT%log,"Binary component",dtSetsRV(i)%binCompo)
        call printQty(UNIT%cons+UNIT%log,"# data records",dtSetsRV(i)%N)
        call printQty(UNIT%cons+UNIT%log,"Timespan [d]",dtSetsRV(i)%tRge,OPT_WITH_RGE)
        call printQty(UNIT%cons+UNIT%log,"Range of "//trim(fmtObsvble(DT_TY%RV,VAL_UNDEF_INT,forPlt=.FALSE.)),dtSetsRV(i)%obsRge, &
         OPT_WITH_RGE)
        call endSubsect(UNIT%cons+UNIT%log) ! Data set …
      end do
      call endSubsect(UNIT%cons+UNIT%log) ! By data set

      call endSubsect(UNIT%cons+UNIT%log) ! DT_TY_DESC(DT_TY%RV)

    end if                           ! (dtProps%nSetsOfTy(DT_TY%RV) > 0)


    ! AMHipN

    if (dtProps%nSetsOfTy(DT_TY%AMHipN) > 0) then

      call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(DT_TY%AMHipN))))

      call startSubsect(UNIT%cons+UNIT%log,"General")
      call printQty(UNIT%cons+UNIT%log,"# observed targets",dtProps%nBObsWDtTy(DT_TY%AMHipN))
      call printQty(UNIT%cons+UNIT%log,"# data records",dtPropsAMHipN%nRec)
      call printQty(UNIT%cons+UNIT%log,"Timespan [d]",dtPropsAMHipN%tRge,OPT_WITH_RGE)
      call printQty(UNIT%cons+UNIT%log,"Range of "//trim(fmtObsvble(DT_TY%AMHipN,VAL_UNDEF_INT,forPlt=.FALSE.)), &
       dtPropsAMHipN%obsRge,OPT_WITH_RGE)
      call endSubsect(UNIT%cons+UNIT%log) ! General

      call startSubsect(UNIT%cons+UNIT%log,"By data sets")
      do i=1,dtProps%nSetsOfTy(DT_TY%AMHipN)
        call startSubsect(UNIT%cons+UNIT%log,"Data set "//niceFmt(i))
        if (isDef(dtSetsAMHipN(i)%binCompo)) &
        call printQty(UNIT%cons+UNIT%log,"# data records",dtSetsAMHipN(i)%N)
        call printQty(UNIT%cons+UNIT%log,"Timespan [d]",dtSetsAMHipN(i)%tRge,OPT_WITH_RGE)
        call printQty(UNIT%cons+UNIT%log,"Range of "//trim(fmtObsvble(DT_TY%AMHipN,VAL_UNDEF_INT,forPlt=.FALSE.)), &
         dtSetsAMHipN(i)%obsRge,OPT_WITH_RGE)
        call endSubsect(UNIT%cons+UNIT%log) ! Data set …
      end do
      call endSubsect(UNIT%cons+UNIT%log) ! By data set

      call endSubsect(UNIT%cons+UNIT%log) ! DT_TY_DESC(DT_TY%AMHipN)

    end if                           ! (dtProps%nSetsOfTy(DT_TY%AMHipN) > 0)

    call endSubsect(UNIT%cons+UNIT%log) ! By data type

    call endSubsect(UNIT%cons+UNIT%log) ! Data properties


  end subroutine printDtMdlProps




  !==================================================[ MODEL AND PARAMETERS ]=======================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine preCalcEA
!
!   pre-calculates values of E for e in [eMin,eMin+eDiff) and M in [0,pi)
!
!   finds root of a function using the algorithm "Newton-Raphson combined with bisection" ([156], chapter 9.4; [324], function 'rtsafe')
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer,parameter :: &
     N_IT_MAX=100                    ! maximum # iterations
    integer :: &
     j, &
     iE, &
     iM
    real(dp) :: &
     eA, &
     df, &
     dEA, &
     dEAOld, &
     f, &
     eAOld, &
     eAH, &
     eAL, &
     mA, &
     ec

    call startSubsect(UNIT%cons+UNIT%log,"Pre-calculating eccentric-anomaly values")

    !$OMP parallel do private(mA,eA,iE,ec,eAL,eAH,dEAOld,dEA,f,df,j)
    do iM=0,EA_PRECALC_SIZE-1        ! M loop

      mA=iM*PI/EA_PRECALC_SIZE
      eA=mA                          ! for use as lower bound eAL

      do iE=0,EA_PRECALC_SIZE-1      ! e loop

        ! initialisations

        ec=eMin+iE*eDiff/EA_PRECALC_SIZE
        eAL=eA                       ! use recently calculated value as lower bound since E is increasing with e for fixed M in [0,pi)
        eAH=min(PI,mA+ec+STD_EA_UNC_GOAL)
        eA=.5_dp*(eAL+eAH)
        dEAOld=eAH-eAL
        dEA=dEAOld
        f=fct(eA,sin(eA),ec,mA)
        df=dFct(eA,cos(eA),ec)


        ! refinement loop

        do j=1,N_IT_MAX
          if (((eA-eAH)*df-f)*((eA-eAL)*df-f) > 0 .or. abs(2._dp*f) > abs(dEAOld*df)) then
            dEAOld=dEA
            dEA=.5_dp*(eAH-eAL)
            eA=eAL+dEA
          else
            dEAOld=dEA
            dEA=f/df
            eAOld=eA
            eA=eA-dEA
          end if

          if (abs(dEA) < STD_EA_UNC_GOAL) exit

          f=fct(eA,sin(eA),ec,mA)
          df=dFct(eA,cos(eA),ec)
          if (f < 0) then
            eAL=eA
          else
            eAH=eA
          end if
        end do

        eAPrecalcd(iE,iM)=eA

      end do                         ! iE=0,EA_PRECALC_SIZE-1

      ! for very high e; cf. [17.04.2010/6]
      eAPrecalcd(EA_PRECALC_SIZE,iM+1)=nearest(mA+1._dp,-1._dp)

    end do                           ! iM=0,EA_PRECALC_SIZE-1
    !$OMP end parallel do

    ! upper bound of E for very high M (cf. [23.11.2009/3])
    eAPrecalcd(0:EA_PRECALC_SIZE-1,EA_PRECALC_SIZE)=nearest(PI,-1._dp)

    call printPrgrs([1],[1],[1])
    call endSubsect(UNIT%cons+UNIT%log) ! Pre-calculating eccentric-anomaly values


    contains

    function fct(eA,sinEA,ec,mA)     ! the function whose root is the ecentric anomaly sought
      implicit none
      real(dp) :: &
       fct, &
       eA, &
       sinEA, &
       ec, &
       mA
      intent(in) :: &
       eA, &
       sinEA, &
       ec, &
       mA

      fct=eA-ec*sinEA-mA
    end function


    function dFct(eA,cosEA,ec)       ! derivative of fct w.r.t. ecc. anom.
      implicit none
      real(dp) :: &
       dFct, &
       eA, &
       cosEA, &
       ec
      intent(in) :: &
       eA, &
       cosEA, &
       ec

      dFct=1._dp-ec*cosEA
    end function

  end subroutine preCalcEA




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function ecAn(tMnEpochArr,N,psdNPl,pars,fctT,fctP)
!
!   Returns:                         eccentric anomalies for given pars. and times
!
!   Arguments:
!     tMnEpochArr                    values of (time - epoch)
!     N                              number of values
!     psdNPl                         # pseudo-planets (i.e. planets or binary companions)
!     pars                           parameters
!     fctT                           function to alter the time index after each ioteration
!     fctP                           function to alter the par. index after each iteration
!
!   uses pre-calculated values and refines them, if necessary,
!   using the algorithm "Newton-Raphson combined with bisection" (see [156], chapter 9.4; [324], function 'rtsafe')
!
!   actual args. of pars MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    intent(in) :: &
     pars, &
     tMnEpochArr, &
     N, &
     psdNPl

    interface
      function fctT(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctT
        intent(in) :: &
         x
      end function
      function fctP(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctP
        intent(in) :: &
         x
      end function
    end interface

    integer,parameter :: &
     N_IT_MAX=100                    ! maximum # iterations
    integer :: &
     ofs, &
     iPl, &
     j, &
     psdNPl, &
     iE, &
     iM
    integer(i8b) :: &
     N, &
     i, &
     iT, &
     iP
    real(dp) :: &
     tMnEpochArr(:), &
     tMnEpoch, &
     eA, &
     df, &
     dEA, &
     dEAOld, &
     f, &
     eAOld, &
     ec, &
     mA, &
     eAH, &
     eAL, &
     pars(:,:), &
     ecAn(N,psdNPl,3), &
     sinEA, &
     cosEA
    logical :: &
     highM


    iT=1
    iP=1

    do i=1,N                         ! loop over times or links

      tMnEpoch=tMnEpochArr(iT)
      associate(p=>pars(iP,:))

        do iPl=1,psdNPl              ! loop over planets

          ! determine e, M, and the indexes for the array of pre-calculated vals.

          ec=p(pRoleByPl(iPl)%e)
          iE=min(EA_PRECALC_SIZE-1,max(0,floor((ec-eMin)/eDiff*EA_PRECALC_SIZE)))
          mA=TWO_PI*modulo(p(pRoleByPl(iPl)%chi)+p(pRoleByPl(iPl)%f)*tMnEpoch,1._dp)
          ! only calc. for M in [0,pi) (cf. [17.04.2010/6])
          if (mA > PI) then
            highM=.TRUE.
            mA=TWO_PI-mA
          else
            highM=.FALSE.
          end if
          iM=min(EA_PRECALC_SIZE-1,floor(mA/PI*EA_PRECALC_SIZE))

          ! calc. E from pre-calculated vals.

          eAL=eAPrecalcd(iE,iM)
          eAH=eAPrecalcd(iE+1,iM+1)
          eA=.5_dp*(eAL+eAH)


          ! refine E

          ! initialisations
          dEAOld=eAH-eAL
          dEA=dEAOld
          f=fct(eA,sin(eA),ec,mA)
          df=dFct(eA,cos(eA),ec)

          ! refinement loop
          do j=1,N_IT_MAX
            if (abs(2._dp*f) > abs(dEAOld*df) .OR. ((eA-eAH)*df-f)*((eA-eAL)*df-f) > 0) then
              dEAOld=dEA
              dEA=.5_dp*(eAH-eAL)
              eA=eAL+dEA
            else
              dEAOld=dEA
              dEA=f/df
              eAOld=eA
              eA=eA-dEA
            end if
            sinEA=sin(eA)
            cosEA=cos(eA)

            if (abs(dEA) < STD_EA_UNC_GOAL) exit

            f=fct(eA,sinEA,ec,mA)
            df=dFct(eA,cosEA,ec)
            if (f < 0) then
              eAL=eA
            else
              eAH=eA
            end if
          end do


          if (highM) then
            ecAn(i,iPl,1)=TWO_PI-eA
            ecAn(i,iPl,2)=-sinEA       ! sin(E), cf. [18.04.2010/2]
            ecAn(i,iPl,3)=cosEA
          else
            ecAn(i,iPl,1)=eA
            ecAn(i,iPl,2)=sinEA        ! sin(E), cf. [18.04.2010/2]
            ecAn(i,iPl,3)=cosEA
          end if

          ofs=ofs+nGParsPl

        end do                         ! iPl=1,psdNPl

        ofs=ofs-nGParsPl*psdNPl

      end associate
      
      iT=fctT(iT)
      iP=fctP(iP)

    end do                           ! i=1,N


    contains

    function fct(eA,sinEA,ec,mA)     ! the function whose root is the ecentric anomaly sought
      implicit none
      real(dp) :: &
       fct, &
       eA, &
       sinEA, &
       ec, &
       mA
      intent(in) :: &
       eA, &
       sinEA, &
       ec, &
       mA

      fct=eA-ec*sinEA-mA
    end function


    function dFct(eA,cosEA,ec)       ! derivative of fct w.r.t. ecc. anom.
      implicit none
      real(dp) :: &
       dFct, &
       eA, &
       cosEA, &
       ec
      intent(in) :: &
       eA, &
       cosEA, &
       ec

      dFct=1._dp-ec*cosEA
    end function

  end function ecAn



!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function auxMdlFctAAngArcsec(K,plx,frq,inc)
!
!   Returns:                         astrometric angular separation [arcsec], according to [06.04.2010/3]
!
!   Arguments:
!     K                              RV semi-amplitude
!     plx                            parallax
!     frq                            orbital frequency
!     inc                            inclination
!
!   definition of the angular separation:
!     normal mode: between CM and star
!     binary mode: between the components
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     auxMdlFctAAngArcsec, &
     K, &                            ! binary mode: K := K_1+K_2; normal mode: K refers to the star (standard definition)
     plx, &
     frq, &
     inc
    intent(in) :: &
     K, &
     plx, &
     frq, &
     inc

    auxMdlFctAAngArcsec=K*plx/(TWO_PI*frq*sin(inc))*D_IN_S/AU_IN_M

  end function auxMdlFctAAngArcsec




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function auxMdlFctDel(aRelAng,cosEAMinEc,cos_omCosOm,cosInc,sin_omSinOm,sqrtOneMinEcSq,sinEA,sin_omCosOm,cos_omSinOm)
!
!   Returns:                         astrometric model component, according to [06.04.2010/4] and [22.05.2011/2]:
!                                      * normal mode: -(component Delta(delta) of star), in the same unit as aRelAng
!                                      * binary mode: component Delta(delta) of secondary, in the same unit as aRelAng
!
!   Arguments:
!     aRelAng                        relative angular separation
!     cosEAMinEc                     cos(E-e), where E is the eccentric anomaly and e is the eccentricity
!     cos_omCosOm                    cos(omega)·cos(Omega), where omega refers to:
!                                      * orbit of planet (normal mode)
!                                      * orbit of secondary binary component (binary mode)
!                                    and Omega is the position angle of the ascending/first node
!     cosInc                         cos(i), where i is the inclination
!     sin_omSinOm                    sin(omega)·sin(Omega)
!     sqrtOneMinEcSq                 sqrt(1-e²)
!     sinEA                          sin(E)
!     sin_omCosOm                    sin(omega)·cos(Omega)
!     cos_omSinOm                    cos(omega)·sin(Omega)
!   
!   does not account for proper motion nor parallax
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     auxMdlFctDel, &
     aRelAng, &
     cosEAMinEc, &
     cos_omCosOm, &
     cosInc, &
     sin_omSinOm, &
     sqrtOneMinEcSq, &
     sinEA, &
     sin_omCosOm, &
     cos_omSinOm
    intent(in) :: &
     aRelAng, &
     cosEAMinEc, &
     cos_omCosOm, &
     cosInc, &
     sin_omSinOm, &
     sqrtOneMinEcSq, &
     sinEA, &
     sin_omCosOm, &
     cos_omSinOm

    auxMdlFctDel=aRelAng*(cosEAMinEc*(-cos_omCosOm+cosInc*sin_omSinOm)+sqrtOneMinEcSq*sinEA*(sin_omCosOm+cosInc*cos_omSinOm))

  end function auxMdlFctDel




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function auxMdlFctAlCosDel(aRelAng,cosEAMinEc,cos_omSinOm,cosInc,sin_omCosOm,sqrtOneMinEcSq,sinEA,sin_omSinOm, &
     cos_omCosOm)
!
!   result:
!     normal mode: -(astrometric-position component Delta(alpha·cos(delta)) of star), in the same unit as aRelAng
!     binary mode: astrometric-position component Delta(alpha·cos(delta)) of secondary, in the same unit as aRelAng
!   according to [06.04.2010/4] and [22.05.2011/2]
!   does not account for proper motion or parallax
!   definition of omega:
!     normal mode: refers to orbit of planet
!     binary mode: refers to orbit of secondary binary component
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     auxMdlFctAlCosDel, &
     aRelAng, &
     cosEAMinEc, &
     cos_omSinOm, &
     cosInc, &
     sin_omCosOm, &
     sqrtOneMinEcSq, &
     sinEA, &
     sin_omSinOm, &
     cos_omCosOm

    auxMdlFctAlCosDel=aRelAng* &
     (-cosEAMinEc*(cos_omSinOm+cosInc*sin_omCosOm)+sqrtOneMinEcSq*sinEA*(sin_omSinOm-cosInc*cos_omCosOm))

  end function auxMdlFctAlCosDel




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function auxMdlFctRV(ec,om,K,sinEA,cosEA)
!
!   Returns:                         radial velocity [m/s], including only the non-systemic components
!                                    (ignores offset V and acceleration a_v)
!
!   Arguments:
!     ec                             eccentricity
!     om                             argument of periapsis:
!                                      normal mode: refers to orbit of planet
!                                      binary mode: may refer to orbit of primary or secondary binary component;
!                                                   result refers to the same body
!     K                              RV semi-amplitude
!     sinEA                          sin(E), where E is the eccentric anomaly
!     cosEA                          cos(E), where E is the eccentric anomaly
!   
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     auxMdlFctRV, &
     ec, &
     om, &
     K, &
     sinEA, &
     cosEA
    intent(in) :: &
     ec, &
     om, &
     K, &
     sinEA, &
     cosEA

    auxMdlFctRV=K*(sin(om)*sinEA-sqrt(1.-ec**2)*cos(om)*cosEA)/(1.-ec*cosEA)
    ! – changed sign to correct def. of omega and i (cf. [06.04.2010/4])
    ! – sign correct according to [15.04.2010/3]
    ! – flipped sign due to changed def. of omega
    !   (now refers to planetary orbit in normal mode; cf. [22.05.2011/1])

  end function auxMdlFctRV




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function mdlFctNMdRV(N,M,pars,fctT,fctP,eAArr,paramsMdl,paramsDt,tMnEpochArr)
!
!   Returns:                         values of the RV model function in normal mode
!
!   Arguments:
!     N                              # values
!     M                              dimensionality of result
!     pars                           parameters
!     fctT                           function to alter the time index after each ioteration
!     fctP                           function to alter the par. index after each iteration
!     eAArr                          values of eccentric anomaly
!     paramsMdl                      additional parameters of the model
!     paramsDt                       additional parameters of the data
!     tMnEpochArr                    values of (time - epoch)
!
!   actual args. of pars MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     M, &
     iPl
    integer(i8b) :: &
     N, &
     iP, &
     iT, &
     i
    real(dp) :: &
     pars(:,:), &
     mdlFctNMdRV(N,M), &
     eAArr(:,:,:), &
     tMnEpochArr(:)
    type(tyParamsMdl) :: &
     paramsMdl
    type(tyDtParamsAMHipN) :: &
     paramsDt
    optional :: &
     tMnEpochArr, &
     paramsDt
    intent(in) :: &
     N, &
     M, &
     pars, &
     eAArr, &
     paramsMdl, &
     paramsDt, &
     tMnEpochArr

    interface
      function fctT(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctT
        intent(in) :: &
         x
      end function
      function fctP(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctP
        intent(in) :: &
         x
      end function
    end interface

    iT=1
    iP=1

    do i=1,N
      associate(p=>pars(iP,:))
        mdlFctNMdRV(i,1)=p(pRoleBySet(paramsMdl%iSet)%V)+merge(tMnEpochArr(iT)*p(pRole%aV)/YR_IN_D,0._dp,mdlProps%RVAccel)
        do iPl=1,paramsMdl%psdNPl
          mdlFctNMdRV(i,1)=mdlFctNMdRV(i,1)+auxMdlFctRV(p(pRoleByPl(iPl)%e),p(pRoleByPl(iPl)%omegaPeri),p(pRoleByPl(iPl)%K), &
           eAArr(i,iPl,2),eAArr(i,iPl,3))
        end do
      end associate
      iT=fctT(iT)
      iP=fctP(iP)
    end do

  end function mdlFctNMdRV




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function mdlFctNMdAMHipN(N,M,pars,fctT,fctP,eAArr,paramsMdl,paramsDt,tMnEpochArr)
!
!   Returns:                         values of the AMHipN model function [mas] in normal mode
!
!   Arguments:
!     N                              # values
!     M                              dimensionality of result
!     pars                           parameters
!     fctT                           function to alter the time index after each ioteration
!     fctP                           function to alter the par. index after each iteration
!     eAArr                          values of eccentric anomaly
!     paramsMdl                      additional parameters of the model
!     paramsDt                       additional parameters of the data
!     tMnEpochArr                    values of (time - epoch)
!
!   Requirements:
!     - AMHipN and RV data given
!
!   actual args. of pars MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     M, &
     iPl
    integer(i8b) :: &
     i, &
     iP, &
     iT, &
     N
    real(dp) :: &
     pars(:,:), &
     mdlFctNMdAMHipN(N,M), &
     eAArr(:,:,:), &
     tMnEpochArr(:), &
     sqrtOneMinEcSq, &
     cos_om2, &
     sin_om2, &
     cosOm, &
     sinOm, &
     cosInc, &
     sin_omSinOm, &
     sin_omCosOm, &
     cos_omSinOm, &
     cos_omCosOm, &
     aAng, &
     cosEAMinEc, &
     del, &
     alCosDel
    type(tyParamsMdl) :: &
     paramsMdl
    type(tyDtParamsAMHipN) :: &
     paramsDt
    optional :: &
     tMnEpochArr, &
     paramsDt
    intent(in) :: &
     N, &
     M, &
     pars, &
     eAArr, &
     paramsMdl, &
     paramsDt, &
     tMnEpochArr

    interface
      function fctT(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctT
        intent(in) :: &
         x
      end function
      function fctP(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctP
        intent(in) :: &
         x
      end function
    end interface

    iT=1
    iP=1

    do i=1,N

      associate(p=>pars(iP,:))

        if (paramsMdl%psdNPl > 0) then
          cosOm=cos(p(pRole%OmegaNode))
          sinOm=sin(p(pRole%OmegaNode))
          cosInc=cos(p(pRole%i))
        end if

        ! all results in deg:
        del=tMnEpochArr(iT)*p(pRole%muDel)*MAS_IN_DEG/YR_IN_D
        alCosDel=tMnEpochArr(iT)*p(pRole%muAlCosDel)*MAS_IN_DEG/YR_IN_D

        do iPl=1,paramsMdl%psdNPl

          sqrtOneMinEcSq=sqrt(1-p(pRoleByPl(iPl)%e)**2)
          cos_om2=cos(p(pRoleByPl(iPl)%omegaPeri))
          sin_om2=sin(p(pRoleByPl(iPl)%omegaPeri))
          sin_omSinOm=sin_om2*sinOm
          sin_omCosOm=sin_om2*cosOm
          cos_omSinOm=cos_om2*sinOm
          cos_omCosOm=cos_om2*cosOm
          cosEAMinEc=eAArr(i,1,3)-p(pRoleByPl(iPl)%e)

          ! all results in deg:
          aAng=auxMdlFctAAngArcsec(p(pRoleByPl(iPl)%K),p(pRole%varpi),p(pRoleByPl(iPl)%f),p(pRole%i))*ARCSEC_IN_DEG
          del=del-auxMdlFctDel(aAng,cosEAMinEc,cos_omCosOm,cosInc,sin_omSinOm,sqrtOneMinEcSq,eAArr(i,1,2),sin_omCosOm, &
           cos_omSinOm)
          alCosDel=alCosDel-auxMdlFctAlCosDel(aAng,cosEAMinEc,cos_omSinOm,cosInc,sin_omCosOm,sqrtOneMinEcSq,eAArr(i,1,2), &
           sin_omSinOm,cos_omCosOm)

        end do

        if (M == 1) then             ! return abscissa residual [mas]
          del=del+p(pRole%deltaR)
          alCosDel=alCosDel+p(pRole%alphaR)*cos(p(pRole%deltaR)*DEG_IN_RAD)
          mdlFctNMdAMHipN(i,1)=(paramsDt%cosPsi(i)*(alCosDel-paramsDt%pos0AtT(i)%y)+ &
           paramsDt%sinPsi(i)*(del-paramsDt%pos0AtT(i)%x))/MAS_IN_DEG+paramsDt%gamma(i)*(p(pRole%varpi)-paramsDt%pi0)*1000._dp
        else                         ! return 2-d pos. [mas] w.r.t. ref. pos. (delta_r, alpha_r), excluding parallactic displacement
          mdlFctNMdAMHipN(i,1)=del/MAS_IN_DEG
          mdlFctNMdAMHipN(i,2)=alCosDel/MAS_IN_DEG
        end if

      end associate

      iT=fctT(iT)
      iP=fctP(iP)

    end do

  end function mdlFctNMdAMHipN




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function mdlFctNMdAMHipNOnly(N,M,pars,fctT,fctP,eAArr,paramsMdl,paramsDt,tMnEpochArr)
!
!   Returns:                         values of the AMHipN model function [mas] in normal mode
!
!   Arguments:
!     N                              # values
!     M                              dimensionality of result
!     pars                           parameters
!     fctT                           function to alter the time index after each ioteration
!     fctP                           function to alter the par. index after each iteration
!     eAArr                          values of eccentric anomaly
!     paramsMdl                      additional parameters of the model
!     paramsDt                       additional parameters of the data
!     tMnEpochArr                    values of (time - epoch)
!
!   Requirements:
!     - only AMHipN data given
!
!   actual args. of pars MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     M, &
     iPl
    integer(i8b) :: &
     N, &
     i, &
     iP, &
     iT
    real(dp) :: &
     pars(:,:), &
     mdlFctNMdAMHipNOnly(N,M), &
     eAArr(:,:,:), &
     tMnEpochArr(:), &
     sqrtOneMinEcSq, &
     cos_om2, &
     sin_om2, &
     cosOm, &
     sinOm, &
     cosInc, &
     sin_omSinOm, &
     sin_omCosOm, &
     cos_omSinOm, &
     cos_omCosOm, &
     aAng, &
     cosEAMinEc, &
     del, &
     alCosDel
    type(tyParamsMdl) :: &
     paramsMdl
    type(tyDtParamsAMHipN) :: &
     paramsDt
    optional :: &
     tMnEpochArr, &
     paramsDt
    intent(in) :: &
     N, &
     M, &
     pars, &
     eAArr, &
     paramsMdl, &
     paramsDt, &
     tMnEpochArr

    interface
      function fctT(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctT
        intent(in) :: &
         x
      end function
      function fctP(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctP
        intent(in) :: &
         x
      end function
    end interface

    iT=1
    iP=1

    do i=1,N

      associate(p=>pars(iP,:))

        if (paramsMdl%psdNPl > 0) then
          cosOm=cos(p(pRole%OmegaNode))
          sinOm=sin(p(pRole%OmegaNode))
          cosInc=cos(p(pRole%i))
        end if

        ! all results in deg:
        del=tMnEpochArr(iT)*p(pRole%muDel)*MAS_IN_DEG/YR_IN_D
        alCosDel=tMnEpochArr(iT)*p(pRole%muAlCosDel)*MAS_IN_DEG/YR_IN_D

        do iPl=1,paramsMdl%psdNPl

          sqrtOneMinEcSq=sqrt(1-p(pRoleByPl(iPl)%e)**2)
          cos_om2=cos(p(pRoleByPl(iPl)%omegaPeri))
          sin_om2=sin(p(pRoleByPl(iPl)%omegaPeri))
          sin_omSinOm=sin_om2*sinOm
          sin_omCosOm=sin_om2*cosOm
          cos_omSinOm=cos_om2*sinOm
          cos_omCosOm=cos_om2*cosOm
          cosEAMinEc=eAArr(i,1,3)-p(pRoleByPl(iPl)%e)

          ! all results in deg:
          aAng=p(pRoleByPl(iPl)%aAng)*MAS_IN_DEG
          del=del-auxMdlFctDel(aAng,cosEAMinEc,cos_omCosOm,cosInc,sin_omSinOm,sqrtOneMinEcSq,eAArr(i,1,2),sin_omCosOm, &
           cos_omSinOm)
          alCosDel=alCosDel-auxMdlFctAlCosDel(aAng,cosEAMinEc,cos_omSinOm,cosInc,sin_omCosOm,sqrtOneMinEcSq,eAArr(i,1,2), &
           sin_omSinOm,cos_omCosOm)

        end do

        if (M == 1) then             ! return abscissa residual [mas]
          del=del+p(pRole%deltaR)
          alCosDel=alCosDel+p(pRole%alphaR)*cos(p(pRole%deltaR)*DEG_IN_RAD)
          mdlFctNMdAMHipNOnly(i,1)=(paramsDt%cosPsi(i)*(alCosDel-paramsDt%pos0AtT(i)%y)+ &
           paramsDt%sinPsi(i)*(del-paramsDt%pos0AtT(i)%x))/MAS_IN_DEG+paramsDt%gamma(i)*(p(pRole%varpi)-paramsDt%pi0)*1000._dp
        else                         ! return 2-d pos. [mas] w.r.t. ref. pos. (delta_r, alpha_r), excluding parallactic displacement
          mdlFctNMdAMHipNOnly(i,1)=del/MAS_IN_DEG
          mdlFctNMdAMHipNOnly(i,2)=alCosDel/MAS_IN_DEG
        end if

      end associate

      iT=fctT(iT)
      iP=fctP(iP)

    end do

  end function mdlFctNMdAMHipNOnly




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function mdlFctBMdAMAngPos(N,M,pars,fctT,fctP,eAArr,paramsMdl,paramsDt,tMnEpochArr)
!
!   Returns:                         values of the AMAngPos model function (delta, alpha·cos(delta)) [mas] in binary mode
!
!   Arguments:
!     N                              # values
!     M                              dimensionality of result
!     pars                           parameters
!     fctT                           function to alter the time index after each ioteration
!     fctP                           function to alter the par. index after each iteration
!     eAArr                          values of eccentric anomaly
!     paramsMdl                      additional parameters of the model
!     paramsDt                       additional parameters of the data
!     tMnEpochArr                    values of (time - epoch)
!
!   Requirements:
!     - AMAngPos (secondary w.r.t. primary) and RV (both components observed) data given
!
!   actual args. of pars MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     M
    integer(i8b) :: &
     N, &
     i, &
     iP, &
     iT
    real(dp) :: &
     pars(:,:), &
     mdlFctBMdAMAngPos(N,M), &
     eAArr(:,:,:), &
     tMnEpochArr(:), &
     sqrtOneMinEcSq, &
     cos_om2, &
     sin_om2, &
     cosOm, &
     sinOm, &
     cosInc, &
     sin_omSinOm, &
     sin_omCosOm, &
     cos_omSinOm, &
     cos_omCosOm, &
     aRelAng, &
     cosEAMinEc
    type(tyParamsMdl) :: &
     paramsMdl
    type(tyDtParamsAMHipN) :: &
     paramsDt
    optional :: &
     tMnEpochArr, &
     paramsDt
    intent(in) :: &
     N, &
     M, &
     pars, &
     eAArr, &
     paramsMdl, &
     paramsDt, &
     tMnEpochArr

    interface
      function fctT(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctT
        intent(in) :: &
         x
      end function
      function fctP(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctP
        intent(in) :: &
         x
      end function
    end interface

    iT=1
    iP=1
    do i=1,N

      associate(p=>pars(iP,:))

        sqrtOneMinEcSq=sqrt(1-p(pRole%e)**2)
        cos_om2=cos(p(pRole%omega2))
        sin_om2=sin(p(pRole%omega2))
        cosOm=cos(p(pRole%OmegaNode))
        sinOm=sin(p(pRole%OmegaNode))
        cosInc=cos(p(pRole%i))
        sin_omSinOm=sin_om2*sinOm
        sin_omCosOm=sin_om2*cosOm
        cos_omSinOm=cos_om2*sinOm
        cos_omCosOm=cos_om2*cosOm
        cosEAMinEc=eAArr(i,1,3)-p(pRole%e)

        ! all results in mas:
        aRelAng=auxMdlFctAAngArcsec(p(pRole%K1)+p(pRole%K2),p(pRole%varpi),p(pRole%f),p(pRole%i))*1000._dp
        mdlFctBMdAMAngPos(i,1)=auxMdlFctDel(aRelAng,cosEAMinEc,cos_omCosOm,cosInc,sin_omSinOm,sqrtOneMinEcSq,eAArr(i,1,2), &
         sin_omCosOm,cos_omSinOm)
        mdlFctBMdAMAngPos(i,2)=auxMdlFctAlCosDel(aRelAng,cosEAMinEc,cos_omSinOm,cosInc,sin_omCosOm,sqrtOneMinEcSq,eAArr(i,1,2), &
         sin_omSinOm,cos_omCosOm)

      end associate
      
      iT=fctT(iT)
      iP=fctP(iP)

    end do

  end function mdlFctBMdAMAngPos




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function mdlFctBMdAMAngPosOnly(N,M,pars,fctT,fctP,eAArr,paramsMdl,paramsDt,tMnEpochArr)
!
!   Returns:                         values of the AMAngPos model function (delta, alpha·cos(delta)) [mas] in binary mode
!
!   Arguments:
!     N                              # values
!     M                              dimensionality of result
!     pars                           parameters
!     fctT                           function to alter the time index after each ioteration
!     fctP                           function to alter the par. index after each iteration
!     eAArr                          values of eccentric anomaly
!     paramsMdl                      additional parameters of the model
!     paramsDt                       additional parameters of the data
!     tMnEpochArr                    values of (time - epoch)
!
!   Requirements:
!     - only AMAngPos (secondary w.r.t. primary) data given
!
!   actual args. of pars MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


    implicit none
    integer :: &
     M
    integer(i8b) :: &
     N, &
     i, &
     iP, &
     iT
    real(dp) :: &
     pars(:,:), &
     mdlFctBMdAMAngPosOnly(N,M), &
     eAArr(:,:,:), &
     tMnEpochArr(:), &
     sqrtOneMinEcSq, &
     cos_om2, &
     sin_om2, &
     cosOm, &
     sinOm, &
     cosInc, &
     sin_omSinOm, &
     sin_omCosOm, &
     cos_omSinOm, &
     cos_omCosOm, &
     cosEAMinEc
    type(tyParamsMdl) :: &
     paramsMdl
    type(tyDtParamsAMHipN) :: &
     paramsDt
    optional :: &
     tMnEpochArr, &
     paramsDt
    intent(in) :: &
     N, &
     M, &
     pars, &
     eAArr, &
     paramsMdl, &
     paramsDt, &
     tMnEpochArr

    interface
      function fctT(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctT
        intent(in) :: &
         x
      end function
      function fctP(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctP
        intent(in) :: &
         x
      end function
    end interface


    iT=1
    iP=1
    do i=1,N

      associate(p=>pars(iP,:))

        sqrtOneMinEcSq=sqrt(1-p(pRole%e)**2)
        cos_om2=cos(p(pRole%omega2))
        sin_om2=sin(p(pRole%omega2))
        cosOm=cos(p(pRole%OmegaNode))
        sinOm=sin(p(pRole%OmegaNode))
        cosInc=cos(p(pRole%i))
        sin_omSinOm=sin_om2*sinOm
        sin_omCosOm=sin_om2*cosOm
        cos_omSinOm=cos_om2*sinOm
        cos_omCosOm=cos_om2*cosOm
        cosEAMinEc=eAArr(i,1,3)-p(pRole%e)

        ! all results in mas:
        mdlFctBMdAMAngPosOnly(i,1)=auxMdlFctDel(p(pRole%aRelAng),cosEAMinEc,cos_omCosOm,cosInc,sin_omSinOm,sqrtOneMinEcSq, &
         eAArr(i,1,2),sin_omCosOm,cos_omSinOm)
        mdlFctBMdAMAngPosOnly(i,2)=auxMdlFctAlCosDel(p(pRole%aRelAng),cosEAMinEc,cos_omSinOm,cosInc,sin_omCosOm, &
         sqrtOneMinEcSq,eAArr(i,1,2),sin_omSinOm,cos_omCosOm)

      end associate
      
      iT=fctT(iT)
      iP=fctP(iP)

    end do

  end function mdlFctBMdAMAngPosOnly




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function mdlFctBMdRV(N,M,pars,fctT,fctP,eAArr,paramsMdl,paramsDt,tMnEpochArr)
!
!   Returns:                         values of the RV model function (binary mode)
!
!   Arguments:
!     N                              # values
!     M                              dimensionality of result
!     pars                           parameters
!     fctT                           function to alter the time index after each ioteration
!     fctP                           function to alter the par. index after each iteration
!     eAArr                          values of eccentric anomaly
!     paramsMdl                      additional parameters of the model
!     paramsDt                       additional parameters of the data
!     tMnEpochArr                    values of (time - epoch)
!
!   only one RV offset V is used for all data sets
!   actual args. of pars MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     M, &
     iK
    integer(i8b) :: &
     N, &
     i, &
     iP, &
     iT
    real(dp) :: &
     pars(:,:), &
     mdlFctBMdRV(N,M), &
     eAArr(:,:,:), &
     tMnEpochArr(:), &
     omOfs
    type(tyParamsMdl) :: &
     paramsMdl
    type(tyDtParamsAMHipN) :: &
     paramsDt
    optional :: &
     tMnEpochArr, &
     paramsDt
    intent(in) :: &
     N, &
     M, &
     pars, &
     eAArr, &
     paramsMdl, &
     paramsDt, &
     tMnEpochArr

    interface
      function fctT(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctT
        intent(in) :: &
         x
      end function
      function fctP(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctP
        intent(in) :: &
         x
      end function
    end interface

    omOfs=(paramsMdl%binCompo-1)*PI  ! for primary, use omega_2; for secondary, shift it by pi to get omega_1 (cf. [22.05.2011/1])
    if (paramsMdl%binCompo == 1) then
      iK=pRole%K1
    else
      iK=pRole%K2
    end if

    iT=1
    iP=1
    do i=1,N
      associate(p=>pars(iP,:))
        mdlFctBMdRV(i,1)=p(pRole%V)+auxMdlFctRV(p(pRole%e),p(pRole%omega2)+omOfs,p(iK),eAArr(i,1,2),eAArr(i,1,3))+ &
         merge(tMnEpochArr(iT)*p(pRole%aV)/YR_IN_D,0._dp,mdlProps%RVAccel) ! first addend is V for current data set (currently the same for all data sets)
      end associate
      iT=fctT(iT)
      iP=fctP(iP)
    end do

  end function mdlFctBMdRV




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcMdlFctWErr(N,M,fct,pars,fctP,tMnEpochArr,fctT,fSamp,paramsMdl,paramsDt,sampUnc,uncIn1,uncSamp1,uncIn2,uncSamp2)
!
!   samples model function at given times and adds errors if the corresponding uncertainties are given
!
!   Arguments:
!     N                              # values
!     M                              dimensionality of result
!     fct                            model function to sample
!     pars                           parameters
!     fctP                           function to alter the par. index after each iteration
!     tMnEpochArr                    values of (time - epoch)
!     fctT                           function to alter the time index after each ioteration
!     fSamp                          resulting samples
!     paramsMdl                      additional parameters of the model
!     paramsDt                       additional parameters of the data
!     sampUnc                        whether the uncertainties should be sampled from (instead of being used as given)
!     uncIn1                         given 1D uncertainties
!     uncSamp1                       sampled 1D uncertainties
!     uncIn2                         given 2D uncertainties
!     uncSamp2                       sampled 2D uncertainties
!
!   CAUTION: not thread-safe!
!   actual args. of pars MAY NOT BE CONTIGUOUS
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     M
    integer(i8b) :: &
     N, &
     i, &
     nUnc
    real(dp) :: &
     errAng, &
     errTmp(2), &
     rotAng, &
     pars(:,:), &
     tMnEpochArr(:), &
     fSamp(N,M), &
     uncIn1(:), &
     uncSamp1(N), &
     jitt
    logical :: &
     sampUnc
    type(tyUncEllDbl) :: &
     uncIn2(:), &
     uncSamp2(N)
    type(tyParamsMdl) :: &
     paramsMdl
    type(tyDtParamsAMHipN) :: &
     paramsDt
    optional :: &
     sampUnc, &
     uncIn1, &
     uncSamp1, &
     uncIn2, &
     uncSamp2, &
     paramsDt
    intent(in) :: &
     N, &
     M, &
     pars, &
     tMnEpochArr, &
     paramsMdl, &
     paramsDt, &
     sampUnc, &
     uncIn1, &
     uncIn2
    intent(out) :: &
     fSamp, &
     uncSamp1, &
     uncSamp2

    interface
      function fct(N,M,pars,fctT,fctP,eAArr,paramsMdl,paramsDt,tMnEpochArr)
        import
        implicit none
        integer :: &
         M
        integer(i8b) :: &
         N
        real(dp) :: &
         pars(:,:), &
         fct(N,M), &
         eAArr(:,:,:), &
         tMnEpochArr(:)
        type(tyParamsMdl) :: &
         paramsMdl
        type(tyDtParamsAMHipN) :: &
         paramsDt
        optional :: &
         tMnEpochArr, &
         paramsDt
        intent(in) :: &
         N, &
         M, &
         pars, &
         eAArr, &
         paramsMdl, &
         paramsDt, &
         tMnEpochArr
        interface
          function fctT(x)
            import
            implicit none
            integer(i8b) :: &
             x, &
             fctT
            intent(in) :: &
             x
          end function
          function fctP(x)
            import
            implicit none
            integer(i8b) :: &
             x, &
             fctP
            intent(in) :: &
             x
          end function
        end interface
      end function
      function fctT(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctT
        intent(in) :: &
         x
      end function
      function fctP(x)
        import
        implicit none
        integer(i8b) :: &
         x, &
         fctP
        intent(in) :: &
         x
      end function
    end interface


    if (present(paramsDt)) then
      fSamp=fct(N,M,pars,fctT,fctP,ecAn(tMnEpochArr,N,paramsMdl%psdNPl,pars,fctT,fctP),paramsMdl,paramsDt,tMnEpochArr) ! pars MAY NOT BE CONTIGUOUS
    else
      fSamp=fct(N,M,pars,fctT,fctP,ecAn(tMnEpochArr,N,paramsMdl%psdNPl,pars,fctT,fctP),paramsMdl,tMnEpochArr=tMnEpochArr) ! pars MAY NOT BE CONTIGUOUS
    end if

    jitt=merge(pars(1,pRole%sgmPlus),merge(pars(1,pRole%tauPlus),0._dp,isDef(pRole%tauPlus)),isDef(pRole%sgmPlus))**2
    if (M == 1 .AND. present(uncIn1)) then
      nUnc=size(uncIn1)
      do i=1,N
        uncSamp1(i)=uncIn1(merge(rndInt(1_i8b,nUnc),i,sampUnc))
        fSamp(i,1)=fSamp(i,1)+sqrt(uncSamp1(i)**2+jitt)*rndNorm()
      end do
    else if (M == 2 .AND. present(uncIn2)) then
      nUnc=size(uncIn2)
      do i=1,N
        uncSamp2(i)=uncIn2(merge(rndInt(1_i8b,nUnc),i,sampUnc))
        errAng=rndUni()*TWO_PI
        errTmp=[sqrt(uncSamp2(i)%a**2+jitt)*cos(errAng),sqrt(uncSamp2(i)%b**2+jitt)*sin(errAng)]
        rotAng=rndUni()*PI         ! shall be in range [0,pi)
        uncSamp2(i)%phi=rotAng/DEG_IN_RAD
        fSamp(i,:)=fSamp(i,:)+[cos(rotAng)*errTmp(1)-sin(rotAng)*errTmp(2),sin(rotAng)*errTmp(1)+cos(rotAng)*errTmp(2)]
      end do
    end if

  end subroutine calcMdlFctWErr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setMdlFctPtrs(dtProps,mdlProps)
!
!   sets pointers to relevant model functions
!
!   Arguments:
!     dtProps                        data properties
!     mdlProps                       model properties
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyMdlProps) :: &
     mdlProps
    class(tyDtPropsAll) :: &
     dtProps
    intent(in) :: &
     dtProps
    intent(inout) :: &
     mdlProps

    select case (dtProps%onlyDtTy)
    case (DT_TY%AMAngPos)
      mdlProps%AMAngPos%mdlFct=>mdlFctBMdAMAngPosOnly
    case (DT_TY%AMHipN)
      mdlProps%AMHipN%mdlFct=>mdlFctNMdAMHipNOnly
    case (DT_TY%RV)
      if (mdlProps%bin) then
        mdlProps%RV%mdlFct=>mdlFctBMdRV
      else
        mdlProps%RV%mdlFct=>mdlFctNMdRV
      end if
    case (VAL_UNDEF_INT)
      mdlProps%AMAngPos%mdlFct=>mdlFctBMdAMAngPos
      mdlProps%AMHipN%mdlFct=>mdlFctNMdAMHipN
      if (mdlProps%bin) then
        mdlProps%RV%mdlFct=>mdlFctBMdRV
      else
        mdlProps%RV%mdlFct=>mdlFctNMdRV
      end if
    end select

  end subroutine setMdlFctPtrs




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setParsDQs(bin,dtProps,maxNPl,noDQs,nPars,periodogram)
!
!   sets important properties of pars. and DQs (called once)
!
!   Arguments:
!     bin                            whether action should refer to binary mode
!     dtProps                        data properties
!     maxNPl                         max. # planets
!     noDQs                          whether DQs should be skipped
!     nPars                          # pars.
!     periodogram                    whether action should refer to periodogram mode
!
!   CAUTION: when changing order of local pars. or DQs, change printParsDQsTbl as well!
!   CAUTION: when changing the order of categories (per system [data set], per system [once], per planet),
!   also change par./DQ help screen!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     maxNPl, &
     i, &
     gP1, &
     gP, &
     gQ1, &
     gQ, &
     p, &
     q, &
     s, &
     t, &
     iPl, &
     pPerPl, &
     qPerPl, &
     iBin, &
     iDtTy(N_DT_TYPES), &
     scp, &
     nPars, &                        ! # pars. for # planets == 'maxNPl'
     nPSys, &
     nPSysC, &
     nGPPl, &
     nQSys, &
     nQSysC, &
     nGQPl
    logical :: &
     bin, &
     noDQs, &
     setQties, &
     periodogram
    class(tyDtPropsAll) :: &
     dtProps
    intent(in) :: &
     bin, &
     maxNPl, &
     periodogram
    intent(inout) :: &
     nPars
    optional :: &
     noDQs, &
     nPars, &
     periodogram

    setQties=.NOT.present(nPars)
    if (setQties) then
      call undef(iDtSetByPar)
      call undef(iPlByPar)
      call undef(pRole)
      call undef(pRoleByPl)
      call undef(pRoleBySet)
    end if
    nPSys=0
    nPSysC=0
    nQSys=0
    nQSysC=0

    iBin=merge(1,0,bin)
    iDtTy(1:N_DT_TYPES)=min(1,dtProps%nSetsOfTy(1:N_DT_TYPES))
    
    ! pars.

    p=0

    ! referring to system (per data set)
    do gP1=1,N_G_PARS
      gP=G_I_PAR_ROLE_ORDERED(gP1)
      scp=G_PAR_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),gP)
      if (scp >= 10) then
        t=scp-10
        do s=1,dtProps%nSetsOfTy(t)
          call inc(nPSys)
          call inc(p)
          if (setQties) then
            parScp(p)=QTY_SCP%sD
            gIPars(p)=gP
            call setParRole(pRoleBySet(s),gP,p)
            iDtSetByPar(p)=s
          end if
        end do
      end if
    end do

    ! referring to system (once)
    do gP1=1,N_G_PARS
      gP=G_I_PAR_ROLE_ORDERED(gP1)
      if (G_PAR_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),gP) == QTY_SCP%s1) then
        call inc(nPSys)
        call inc(p)
        if (setQties) then
          parScp(p)=QTY_SCP%s1
          gIPars(p)=gP
          call setParRole(pRole,gP,p)
        end if
      end if
    end do

    ! referring to system (once, if there are companion(s))
    do gP1=1,N_G_PARS
      gP=G_I_PAR_ROLE_ORDERED(gP1)
      if (G_PAR_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),gP) == QTY_SCP%sC .AND. (bin .OR. maxNPl > 0)) then
        call inc(nPSysC)
        call inc(p)
        if (setQties) then
          parScp(p)=QTY_SCP%sC
          gIPars(p)=gP
          call setParRole(pRole,gP,p)
        end if
      end if
    end do

    ! referring to planets
    if (bin) then
      nGPPl=0
      if (setQties) then
        pRoleByPl(1)=pRole           ! for ecAn etc.
        call setIfUndef(pRoleBySet(1),pRole) ! for pars. like V not referring to data set but system in binary mode
      end if
    else
      nGPPl=count(G_PAR_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),1:N_G_PARS) == QTY_SCP%pl)
      if (setQties) then
        pPerPl=0
        do gP1=1,N_G_PARS
          gP=G_I_PAR_ROLE_ORDERED(gP1)
          if (G_PAR_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),gP) == QTY_SCP%pl) then
            call inc(pPerPl)
            do iPl=1,maxNPl
              p=nPSys+nPSysC+(iPl-1)*nGPPl+pPerPl
              parScp(p)=QTY_SCP%pl
              gIPars(p)=gP
              iPlByPar(p)=iPl
              call setParRole(pRoleByPl(iPl),gP,p)
              if (maxNPl == 1) call setParRole(pRole,gP,p) ! e.g. for periodogram
            end do
          end if
        end do
      end if
    end if
    
    if (nPSys+nPSysC+maxNPl*nGPPl == 0) call err("analysing the given types of data in "//merge("binary","normal",bin)// &
     " mode not implemented")
  
  
    if (setQties .AND. .NOT. (present(noDQs) .AND. noDQs)) then
    
      ! DQs

      q=0

      if (present(periodogram) .AND. periodogram) then

        if (bin) then
          gQ=G_I_DQ_ROLE%P
          nQSys=1
          q=1
          DQScp(q)=QTY_SCP%s1
          gIDQs(q)=gQ
          call setDQRole(qRole,gQ,q)
        else
          nGQPl=1
          qPerPl=1
          gQ=G_I_DQ_ROLE%P
          iPl=1
          q=nQSys+nQSysC+(iPl-1)*nGQPl+qPerPl
          DQScp(q)=QTY_SCP%pl
          gIDQs(q)=gQ
          iPlByDQ(q)=iPl
          call setDQRole(qRoleByPl(iPl),gQ,q)
        end if

      else
      
        ! referring to system (per data set)
        do gQ1=1,N_G_DQS
          gQ=G_I_DQ_ROLE_ORDERED(gQ1)
          scp=G_DQ_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),gQ)
          if (scp > 10) then
            t=scp-10
            do s=1,dtProps%nSetsOfTy(t)
              call inc(nQSys)
              call inc(q)
              DQScp(q)=QTY_SCP%sD
              gIDQs(q)=gQ
              call setDQRole(qRoleBySet(s),gQ,q)
              iDtSetByDQ(q)=s
            end do
          end if
        end do

        ! referring to system (once)
        do gQ1=1,N_G_DQS
          gQ=G_I_DQ_ROLE_ORDERED(gQ1)
          if (G_DQ_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),gQ) == QTY_SCP%s1) then
            call inc(nQSys)
            call inc(q)
            DQScp(q)=QTY_SCP%s1
            gIDQs(q)=gQ
            call setDQRole(qRole,gQ,q)
          end if
        end do

        ! referring to system (once, if there are companion(s))
        do gQ1=1,N_G_DQS
          gQ=G_I_DQ_ROLE_ORDERED(gQ1)
          if (G_DQ_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),gQ) == QTY_SCP%sC .AND. (bin .OR. maxNPl > 0)) then
            call inc(nQSysC)
            call inc(q)
            DQScp(q)=QTY_SCP%sC
            gIDQs(q)=gQ
            call setDQRole(qRole,gQ,q)
          end if
        end do

        ! referring to planets
        if (.NOT. bin) then
          nGQPl=count(G_DQ_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),1:N_G_DQS) == QTY_SCP%pl)
          qPerPl=0
          do gQ1=1,N_G_DQS
            gQ=G_I_DQ_ROLE_ORDERED(gQ1)
            if (G_DQ_SCP_BY_COMB(iBin,iDtTy(1),iDtTy(2),iDtTy(3),gQ) == QTY_SCP%pl) then
              call inc(qPerPl)
              do iPl=1,maxNPl
                q=nQSys+nQSysC+(iPl-1)*nGQPl+qPerPl
                DQScp(q)=QTY_SCP%pl
                gIDQs(q)=gQ
                iPlByDQ(q)=iPl
                call setDQRole(qRoleByPl(iPl),gQ,q)
              end do
            end if
          end do
        end if

      end if                         ! (present(periodogram) .AND. periodogram)
    end if                           ! (setQties .AND. .NOT. (present(noDQs) .AND. noDQs)
    
    if (setQties) then
      nParsSys=nPSys
      nParsSysC=nPSysC
      nGParsPl=nGPPl
      nDQsSys=nQSys
      nDQsSysC=nQSysC
      nGDQsPl=nGQPl
    else
      nPars=nPSys+maxNPl*nGPPl
    end if
    
    
    contains
    
    subroutine setParRole(pRole,gP,p) ! sets par. role
      implicit none
      integer :: &
       gP, &
       p
      type(tyParsInt) :: &
       pRole
      intent(in) :: &
       gP, &
       p
      intent(inout) :: &
       pRole
      
      select case (gP)
      case (G_I_PAR_ROLE%e)
        pRole%e=p
      case (G_I_PAR_ROLE%f)
        pRole%f=p
      case (G_I_PAR_ROLE%chi)
        pRole%chi=p
      case (G_I_PAR_ROLE%omegaPeri)
        pRole%omegaPeri=p
      case (G_I_PAR_ROLE%K)
        pRole%K=p
      case (G_I_PAR_ROLE%V)
        pRole%V=p
      case (G_I_PAR_ROLE%sgmPlus)
        pRole%sgmPlus=p
      case (G_I_PAR_ROLE%omega2)
        pRole%omega2=p
      case (G_I_PAR_ROLE%K1)
        pRole%K1=p
      case (G_I_PAR_ROLE%K2)
        pRole%K2=p
      case (G_I_PAR_ROLE%i)
        pRole%i=p
      case (G_I_PAR_ROLE%OmegaNode)
        pRole%OmegaNode=p
      case (G_I_PAR_ROLE%varpi)
        pRole%varpi=p
      case (G_I_PAR_ROLE%aRelAng)
        pRole%aRelAng=p
      case (G_I_PAR_ROLE%aAng)
        pRole%aAng=p
      case (G_I_PAR_ROLE%alphaR)
        pRole%alphaR=p
      case (G_I_PAR_ROLE%deltaR)
        pRole%deltaR=p
      case (G_I_PAR_ROLE%muAlCosDel)
        pRole%muAlCosDel=p
      case (G_I_PAR_ROLE%muDel)
        pRole%muDel=p
      case (G_I_PAR_ROLE%tauPlus)
        pRole%tauPlus=p
      case (G_I_PAR_ROLE%aV)
        pRole%aV=p
      case default
        call err("bug 1 ("//trim(niceFmt(gP))//")")
      end select
    end subroutine
    

    subroutine setDQRole(qRole,gQ,q) ! sets DQ role
      implicit none
      integer :: &
       gQ, &
       q
      type(tyDQsInt) :: &
       qRole
      intent(in) :: &
       gQ, &
       q
      intent(inout) :: &
       qRole
      
      select case (gQ)
      case (G_I_DQ_ROLE%P)
        qRole%P=q
      case (G_I_DQ_ROLE%T)
        qRole%T=q
      case (G_I_DQ_ROLE%mMin)
        qRole%mMin=q
      case (G_I_DQ_ROLE%aSinI)
        qRole%aSinI=q
      case (G_I_DQ_ROLE%aPMax)
        qRole%aPMax=q
      case (G_I_DQ_ROLE%KAlt)
        qRole%KAlt=q
      case (G_I_DQ_ROLE%KAlt1)
        qRole%KAlt1=q
      case (G_I_DQ_ROLE%KAlt2)
        qRole%KAlt2=q
      case (G_I_DQ_ROLE%rho)
        qRole%rho=q
      case (G_I_DQ_ROLE%m1)
        qRole%m1=q
      case (G_I_DQ_ROLE%m2)
        qRole%m2=q
      case (G_I_DQ_ROLE%d)
        qRole%d=q
      case (G_I_DQ_ROLE%aRel)
        qRole%aRel=q
      case (G_I_DQ_ROLE%a1)
        qRole%a1=q
      case (G_I_DQ_ROLE%a2)
        qRole%a2=q
      case (G_I_DQ_ROLE%thetaR)
        qRole%thetaR=q
      case (G_I_DQ_ROLE%thetaSecR)
        qRole%thetaSecR=q
      case (G_I_DQ_ROLE%tT1)
        qRole%tT1=q
      case (G_I_DQ_ROLE%tT2)
        qRole%tT2=q
      case (G_I_DQ_ROLE%omSum)
        qRole%omSum=q
      case (G_I_DQ_ROLE%omDiff)
        qRole%omDiff=q
      case (G_I_DQ_ROLE%om2Sum)
        qRole%om2Sum=q
      case (G_I_DQ_ROLE%om2Diff)
        qRole%om2Diff=q
      case (G_I_DQ_ROLE%m)
        qRole%m=q
      case default
        call err("bug 2 ("//trim(niceFmt(gQ))//")")
      end select
    end subroutine

  end subroutine setParsDQs




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    recursive function maxDQRge(q,iPl,parLBnd,parUBnd,sampRge) result(maxDQRgeRes)
!
!   returns a range compatible with log-scale plotting
!
!   IMPORTANT: when calling itself recursively, it must only call with LOWER q, s.t. sampRge for that q is already set
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p, &
     q, &
     gIDQ, &
     p2, &
     p3, &
     p4, &
     iPl
    real(dp) :: &
     mi, &
     ma, &
     muMin, &
     muProd, &
     muDiff, &
     parLBnd(curNPars), &
     parUBnd(curNPars)
    type(tyIntvlDbl) :: &
     maxDQRgeRes, &
     tmpRge, &
     tmpRge2, &
     sampRge(curNDQs)
    intent(in) :: &
     q, &
     iPl, &
     sampRge, &
     parLBnd, &
     parUBnd

    gIDQ=gIDQs(q)

    select case (gIDQ)
    case (G_I_DQ_ROLE%P)
      p=merge(pRole%f,pRoleByPl(iPl)%f,mdlProps%bin) ! f
      mi=1._dp/parUBnd(p)
      ma=1._dp/parLBnd(p)
    case (G_I_DQ_ROLE%T)
      p=merge(pRole%f,pRoleByPl(iPl)%f,mdlProps%bin) ! f
      p2=merge(pRole%chi,pRoleByPl(iPl)%chi,mdlProps%bin) ! chi
      mi=tOfs+mdlProps%epoch-parUBnd(p2)/parLBnd(p)
      ma=tOfs+mdlProps%epoch-parLBnd(p2)/parUBnd(p)
    case (G_I_DQ_ROLE%mMin)          ! cf. [17.02.2011/3]
      p=pRoleByPl(iPl)%K             ! K
      p2=pRoleByPl(iPl)%f             ! f
      muMin=parLBnd(p)/(CUBE_ROOT_2_PI_GRAVC*(parUBnd(p2)/D_IN_S)**(1._dp/3))
      muProd=muMin*mdlProps%mS**(2._dp/3)
      muDiff=(2._dp/3*muMin)**3-mdlProps%mS
      mi=merge(2._dp*muDiff,muProd,muProd <= muDiff)
      ma=SUBSTELLAR_MASS_MAX_PHYS
    case (G_I_DQ_ROLE%aSinI)
      p=pRoleByPl(iPl)%K             ! K
      p2=pRoleByPl(iPl)%f             ! f
      mi=parLBnd(p)/(TWO_PI*parUBnd(p2))*D_IN_S/AU_IN_M
      ma=parUBnd(p)/(TWO_PI*parLBnd(p2))*D_IN_S/AU_IN_M
    case (G_I_DQ_ROLE%aPMax)
      tmpRge=maxDQRge(qRoleByPl(iPl)%mMin,iPl,parLBnd,parUBnd,sampRge) ! m_min
      tmpRge2=maxDQRge(qRoleByPl(iPl)%aSinI,iPl,parLBnd,parUBnd,sampRge) ! a·sin(p)
      mi=mdlProps%mS/tmpRge%uBnd*tmpRge2%lBnd
      ma=mdlProps%mS/tmpRge%lBnd*tmpRge2%uBnd
    case (G_I_DQ_ROLE%KAlt)
      p=pRoleByPl(iPl)%e             ! e
      p2=pRoleByPl(iPl)%K             ! K
      mi=parLBnd(p2)/sqrt(1._dp-parLBnd(p)**2)
      ma=parUBnd(p2)/sqrt(1._dp-parUBnd(p)**2)
    case (G_I_DQ_ROLE%KAlt1)
      p=pRole%e                      ! e
      p2=pRole%K1                     ! K_1
      mi=parLBnd(p2)/sqrt(1._dp-parLBnd(p)**2)
      ma=parUBnd(p2)/sqrt(1._dp-parUBnd(p)**2)
    case (G_I_DQ_ROLE%KAlt2)
      p=pRole%e                      ! e
      p2=pRole%K2                     ! K_2
      mi=parLBnd(p2)/sqrt(1._dp-parLBnd(p)**2)
      ma=parUBnd(p2)/sqrt(1._dp-parUBnd(p)**2)
    case (G_I_DQ_ROLE%rho)
      p=pRole%K1                     ! K_1
      p2=pRole%K2                     ! K_2
      mi=parLBnd(p)/parUBnd(p2)
      ma=parUBnd(p)/parLBnd(p2)
    case (G_I_DQ_ROLE%m1)
      p=pRole%K1                     ! K_1
      p2=pRole%K2                     ! K_2
      p3=pRole%f                      ! f
      p4=pRole%i                      ! p
      mi=parLBnd(p2)*(parLBnd(p)+parLBnd(p2))**2/(TWO_PI*GRAVC_MKS*parUBnd(p3)*maxSinZeroPi(parLBnd(p4),parUBnd(p4))**3)* &
       D_IN_S/M_SUN_IN_KG
      ma=parUBnd(p2)*(parUBnd(p)+parUBnd(p2))**2/(TWO_PI*GRAVC_MKS*parLBnd(p3)*minSinZeroPi(parLBnd(p4),parUBnd(p4))**3)* &
       D_IN_S/M_SUN_IN_KG
    case (G_I_DQ_ROLE%m2)
      tmpRge=maxDQRge(qRole%rho,iPl,parLBnd,parUBnd,sampRge) ! rho
      tmpRge2=maxDQRge(qRole%m1,iPl,parLBnd,parUBnd,sampRge) ! m_1
      mi=tmpRge%lBnd*tmpRge2%lBnd
      ma=tmpRge%uBnd*tmpRge2%uBnd
    case (G_I_DQ_ROLE%d)
      p=pRole%varpi                  ! varpi
      mi=1._dp/parUBnd(p)
      ma=1._dp/parLBnd(p)
    case (G_I_DQ_ROLE%aRel)
      p=pRole%K1                     ! K_1
      p2=pRole%K2                     ! K_2
      p3=pRole%f                      ! f
      p4=pRole%i                      ! p
      mi=(parLBnd(p)+parLBnd(p2))/(TWO_PI*parUBnd(p3)*maxSinZeroPi(parLBnd(p4),parUBnd(p4)))*D_IN_S/AU_IN_M
      ma=(parUBnd(p)+parUBnd(p2))/(TWO_PI*parLBnd(p3)*minSinZeroPi(parLBnd(p4),parUBnd(p4)))*D_IN_S/AU_IN_M
    case (G_I_DQ_ROLE%a1)
      p=pRole%K1                     ! K_1
      p2=pRole%K2                     ! K_2
      tmpRge=maxDQRge(qRole%aRel,iPl,parLBnd,parUBnd,sampRge) ! a_rel
      mi=parLBnd(p)/(parUBnd(p)+parUBnd(p2))*tmpRge%lBnd
      ma=parUBnd(p)/(parLBnd(p)+parLBnd(p2))*tmpRge%uBnd
    case (G_I_DQ_ROLE%a2)
      tmpRge=maxDQRge(qRole%a1,iPl,parLBnd,parUBnd,sampRge) ! a_1
      tmpRge2=maxDQRge(qRole%rho,iPl,parLBnd,parUBnd,sampRge) ! rho
      mi=tmpRge%lBnd/tmpRge2%uBnd
      ma=tmpRge%uBnd/tmpRge2%lBnd
    case (G_I_DQ_ROLE%thetaR)
      p=pRoleByPl(iPl)%omegaPeri     ! omega
      p2=pRoleByPl(iPl)%chi           ! chi
      mi=parLBnd(p)+TWO_PI*parLBnd(p2)
      ma=parUBnd(p)+TWO_PI*parUBnd(p2)
      if (ma-mi >= TWO_PI) then
        mi=0._dp
        ma=TWO_PI
      else
        mi=modulo(mi,TWO_PI)
        ma=modulo(ma,TWO_PI)
        if (ma <= mi) then           ! two distinct ranges possible: [0,ma], [mi,2·PI] => [0,PI] possible as single range
          mi=0._dp
          ma=TWO_PI
        end if
      end if
    case (G_I_DQ_ROLE%thetaSecR)
      p=pRole%omega2                 ! omega_2
      p2=pRole%chi                    ! chi
      mi=parLBnd(p)+TWO_PI*parLBnd(p2)
      ma=parUBnd(p)+TWO_PI*parUBnd(p2)
      if (ma-mi >= TWO_PI) then
        mi=0._dp
        ma=TWO_PI
      else
        mi=modulo(mi,TWO_PI)
        ma=modulo(ma,TWO_PI)
        if (ma <= mi) then           ! two distinct ranges possible: [0,ma], [mi,2·PI] => [0,PI] possible as single range
          mi=0._dp
          ma=TWO_PI
        end if
      end if
    case (G_I_DQ_ROLE%tT1)           ! widest possible interval; analytical determination depending on pars. prob. impossible due to ocurrence of E
      p=merge(pRole%f,pRoleByPl(iPl)%f,mdlProps%bin) ! f
      mi=transTmStart
      ma=transTmStart+1._dp/parLBnd(p)
    case (G_I_DQ_ROLE%tT2)           ! widest possible interval; analytical determination depending on pars. prob. impossible due to ocurrence of E
      p=merge(pRole%f,pRoleByPl(iPl)%f,mdlProps%bin) ! f
      mi=transTmStart+1._dp/parUBnd(p)
      ma=transTmStart+2._dp/parLBnd(p)
    case (G_I_DQ_ROLE%omSum)
      p=pRole%OmegaNode
      p2=pRoleByPl(iPl)%omegaPeri
      mi=parLBnd(p)+parLBnd(p2)
      ma=parUBnd(p)+parUBnd(p2)
    case (G_I_DQ_ROLE%omDiff)
      p=pRole%OmegaNode
      p2=pRoleByPl(iPl)%omegaPeri
      mi=parLBnd(p)-parUBnd(p2)
      ma=parUBnd(p)-parLBnd(p2)
    case (G_I_DQ_ROLE%om2Sum)
      p=pRole%OmegaNode
      p2=pRole%omega2
      mi=parLBnd(p)+parLBnd(p2)
      ma=parUBnd(p)+parUBnd(p2)
    case (G_I_DQ_ROLE%om2Diff)
      p=pRole%OmegaNode
      p2=pRole%omega2
      mi=parLBnd(p)-parUBnd(p2)
      ma=parUBnd(p)-parLBnd(p2)
    case (G_I_DQ_ROLE%m)             ! cf. [17.02.2011/3, 29.10.2012/4]
      p=pRoleByPl(iPl)%K
      p2=pRoleByPl(iPl)%f
      p3=pRole%i
      muMin=parLBnd(p)/(CUBE_ROOT_2_PI_GRAVC*(parUBnd(p2)/D_IN_S)**(1._dp/3)*maxSinZeroPi(parLBnd(p3),parUBnd(p3)))
      muProd=muMin*mdlProps%mS**(2._dp/3)
      muDiff=(2._dp/3*muMin)**3-mdlProps%mS
      mi=merge(2._dp*muDiff,muProd,muProd <= muDiff)
      ma=SUBSTELLAR_MASS_MAX_PHYS
    end select

    ! special treatment of sums/differences of angles (result in [0,2·pi))
    if (isDef(posInArr([G_I_DQ_ROLE%thetaR,G_I_DQ_ROLE%thetaSecR,G_I_DQ_ROLE%omSum,G_I_DQ_ROLE%omDiff,G_I_DQ_ROLE%om2Sum, &
     G_I_DQ_ROLE%om2Diff],gIDQ))) then
      if (ma-mi >= TWO_PI) then
        mi=0._dp
        ma=TWO_PI
      else
        mi=modulo(mi,TWO_PI)
        ma=modulo(ma,TWO_PI)
        if (ma <= mi) then           ! two distinct ranges possible: [0,ma], [mi,2·PI] => [0,PI] possible as single range
          mi=0._dp
          ma=TWO_PI
        end if
      end if
    end if

    if (maPoLogSclDQ(q)) mi=max(mi,LOG_SCL_PSD_MIN)
    ma=min(HUGE_DP,ma)

    maxDQRgeRes=sampRge(q)%sect(tyIntvlDbl(mi,ma))
    if (isUndef(maxDQRgeRes)) maxDQRgeRes=intvlUnion(tyIntvlDbl(mi,ma),sampRge(q)) ! in order not to end up with undefined result

  end function maxDQRge




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function plStr(iMd,n,sep,append,prepend)
!
!   Returns:                         formatted string identifying #-pl. mode
!
!   Arguments:
!     iMd                            #-pl. mode index
!     n                              # planets
!     sep                            separating string
!     append                         whether sep should be appended to result
!     prepend                        whether sep should be prepended to result
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iMd, &
     n
    logical :: &
     append, &
     prepend
    character(*) :: &
     sep
    character(18) :: &
     plStr
    optional :: &
     append, &
     prepend
    intent(in) :: &
     iMd, &
     n, &
     sep, &
     append, &
     prepend

    if (isUndef(n)) then
      plStr=""
    else
      if (isUndef(iMd)) then
        write(plStr,"(I1,'"//sep//trim(merge("planet ","planets",n == 1))//"')") n
      else
        write(plStr,"('md',I1,'"//sep//"',I1,'"//sep//trim(merge("planet ","planets",n == 1))//"')") iMd,n
      end if
      if (present(prepend) .AND. prepend) plStr=sep//plStr
      if (present(append) .AND. append) plStr=trim(plStr)//sep
    end if

  end function plStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtPar(p,forPlt)
!
!   Returns:                         formatted string with par. and its unit
!
!   Arguments:
!     p                              par. index
!     forPlt                         whether result is used for plotting
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p
    logical :: &
     forPlt
    character(STR_LEN_STD) :: &
     fmtPar, &
     unit
    intent(in) :: &
     p, &
     forPlt

    unit=mrg(G_PAR_UNITS_TEX(gIPars(p)),G_PAR_UNITS(gIPars(p)),forPlt .AND. pltPSTricks)
    if (len_trim(unit) > 0) then
      unit="["//trim(unit)//"]"
      if (forPlt .AND. pltPSTricks) unit="$"//trim(unit)//"$"
    end if
    fmtPar=trim(mrg(parNmsPlt(p),parNms(p),forPlt))//" "//unit

  end function fmtPar




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtDQ(q,forPlt)
!
!   Returns:                         formatted string with DQ and its unit
!
!   Arguments:
!     q                              DQ index
!     forPlt                         whether result is used for plotting
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     q
    logical :: &
     forPlt
    character(STR_LEN_STD) :: &
     fmtDQ, &
     unit
    intent(in) :: &
     q, &
     forPlt

    unit=mrg(G_DQ_UNITS_TEX(gIDQs(q)),G_DQ_UNITS(gIDQs(q)),forPlt .AND. pltPSTricks)
    if (len_trim(unit) > 0) then
      unit="["//trim(unit)//"]"
      if (forPlt .AND. pltPSTricks) unit="$"//trim(unit)//"$"
    end if
    fmtDQ=trim(mrg(DQNmsPlt(q),DQNms(q),forPlt))//" "//unit

  end function fmtDQ




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printParsDQsTbl
!
!   prints a table listing the pars. and DQs for all allowed combinations of data and normal/binary mode
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iBin, &
     iDtTy1, &
     iDtTy2, &
     iDtTy3, &
     nPars, &
     gP1, &
     gP, &
     gQ1, &
     gQ, &
     nRows, &
     nCols, &
     allocStat, &
     i, &
     scp
    character :: &
     cBr1, &
     cBr2
    character(STR_LEN_STD) :: &
     matrx1(:,:), &
     matrx2(:), &
     strPars
    allocatable :: &
     matrx1, &
     matrx2


    ! header

    call printMsg(UNIT%cons,"BASE - "//BASE_LONG//trim(niceFmt(BASE_REV_NR)))
    call printMsg(UNIT%cons,uCaseFst(AUTHOR_STR))
    call printMsg(UNIT%cons,"")


    ! par. tables

      ! by par.
    
    call printMsg(UNIT%cons,"")
    call startSubsect(UNIT%cons,"Model parameters",spcPrefx=.TRUE.)
    
    allocate(matrx1(N_G_PARS,2),matrx2(N_G_PARS+1),stat=allocStat)
    call testAllocErr(allocStat,57)
    
    do gP1=1,N_G_PARS
      gP=G_I_PAR_ROLE_ORDERED(gP1)
      matrx1(gP1,1)=G_PAR_NMS(gP)
      matrx1(gP1,2)=uCaseFst(G_PAR_DESC(gP))
    end do
    matrx2=fmtMatrix(N_G_PARS+1,matrx1,reshape([character(STR_LEN_STD) :: "Symbol","Description"],[1,2]))
    do i=1,N_G_PARS+1
      call printMsg(UNIT%cons,matrx2(i),setFlg(FLG%spcPrfx))
    end do
    call endSubsect(UNIT%cons) ! Model parameters

      ! by combination
    
    call printMsg(UNIT%cons,"")
    call startSubsect(UNIT%cons,"Combinations of modes and data and associated model parameters",spcPrefx=.TRUE.)
    call startSubsect(UNIT%cons,"Notation:",spcPrefx=.TRUE.)
    call printMsg(UNIT%cons,"(B)           binary mode",setFlg(FLG%spcPrfx))
    do i=1,N_DT_TYPES
      call printMsg(UNIT%cons,"("//trim(niceFmt(i))//")           "//DT_TY_DESC(i),setFlg(FLG%spcPrfx))
    end do
    call printMsg(UNIT%cons,"[<parameter>] system parameter is not present in 0-planets mode",setFlg(FLG%spcPrfx))
    call printMsg(UNIT%cons,"(<parameter>) parameter is part of subset repeated for each planet",setFlg(FLG%spcPrfx))
    call endSubsect(UNIT%cons,spcPrefx=.TRUE.) ! Notation
    
    nRows=2**(1+N_DT_TYPES)
    nCols=2+N_DT_TYPES
    deallocate(matrx1,matrx2)
    allocate(matrx1(nRows,nCols),matrx2(nRows+1),stat=allocStat)
    call testAllocErr(allocStat,58)
    
    i=0
    do iBin=0,1
      cBr1=merge("["," ",iBin == 0)
      cBr2=merge("]"," ",iBin == 0)
      do iDtTy1=0,1
        do iDtTy2=0,1
          do iDtTy3=0,1
            call inc(i)
            matrx1(i,1)=merge("-","x",iBin == 0)
            matrx1(i,2)=merge("-","x",iDtTy1 == 0)
            matrx1(i,3)=merge("-","x",iDtTy2 == 0)
            matrx1(i,4)=merge("-","x",iDtTy3 == 0)
            strPars=""
            ! pars. referring to system (per data set)
            do gP1=1,N_G_PARS
              gP=G_I_PAR_ROLE_ORDERED(gP1)
              if (G_PAR_SCP_BY_COMB(iBin,iDtTy1,iDtTy2,iDtTy3,gP) >= 10) strPars=trim(strPars)//"; "//trim(G_PAR_NMS(gP))
            end do
            ! pars. referring to system (once)
            do gP1=1,N_G_PARS
              gP=G_I_PAR_ROLE_ORDERED(gP1)
              if (G_PAR_SCP_BY_COMB(iBin,iDtTy1,iDtTy2,iDtTy3,gP) == QTY_SCP%s1) strPars=trim(strPars)//"; "//trim(G_PAR_NMS(gP))
            end do
            ! pars. referring to system (once, if there are companion(s))
            do gP1=1,N_G_PARS
              gP=G_I_PAR_ROLE_ORDERED(gP1)
              if (G_PAR_SCP_BY_COMB(iBin,iDtTy1,iDtTy2,iDtTy3,gP) == QTY_SCP%sC) strPars=trim(strPars)//"; "//trim(cBr1)// &
               trim(G_PAR_NMS(gP))//trim(cBr2)
            end do
            ! pars. referring to planet
            do gP1=1,N_G_PARS
              gP=G_I_PAR_ROLE_ORDERED(gP1)
              if (G_PAR_SCP_BY_COMB(iBin,iDtTy1,iDtTy2,iDtTy3,gP) == QTY_SCP%pl) strPars=trim(strPars)//"; ("// &
               trim(G_PAR_NMS(gP))//")"
            end do
            strPars=rmLead(strPars,"; ")
            if (len_trim(strPars) == 0) strPars=STR_UNDEF
            matrx1(i,5)=strPars
          end do
        end do
      end do
    end do
    
    matrx2=fmtMatrix(nRows+1,matrx1,reshape([character(STR_LEN_STD) :: "(B)","(1)","(2)","(3)","Parameters"],[1,5]))
    do i=1,nRows+1
      call printMsg(UNIT%cons,matrx2(i),setFlg(FLG%spcPrfx))
    end do
    call endSubsect(UNIT%cons) ! Combinations of modes and data and associated model parameters


    ! DQ tables

      ! by DQ
    
    call printMsg(UNIT%cons,"")
    call startSubsect(UNIT%cons,"Derived quantities",spcPrefx=.TRUE.)
    
    deallocate(matrx1,matrx2)
    allocate(matrx1(N_G_DQS,2),matrx2(N_G_DQS+1),stat=allocStat)
    call testAllocErr(allocStat,59)
    
    do gQ1=1,N_G_DQS
      gQ=G_I_DQ_ROLE_ORDERED(gQ1)
      matrx1(gQ1,1)=G_DQ_NMS(gQ)
      matrx1(gQ1,2)=uCaseFst(G_DQ_DESC(gQ))
    end do
    matrx2=fmtMatrix(N_G_DQS+1,matrx1,reshape([character(STR_LEN_STD) :: "Symbol","Description"],[1,2]))
    do i=1,N_G_DQS+1
      call printMsg(UNIT%cons,matrx2(i),setFlg(FLG%spcPrfx))
    end do
    call endSubsect(UNIT%cons) ! Derived quantities

      ! by combination
    
    call printMsg(UNIT%cons,"")
    call startSubsect(UNIT%cons,"Combinations of modes and data and associated derived quantities",spcPrefx=.TRUE.)
    call startSubsect(UNIT%cons,"Notation:",spcPrefx=.TRUE.)
    call printMsg(UNIT%cons,"(B)          binary mode",setFlg(FLG%spcPrfx))
    do i=1,N_DT_TYPES
      call printMsg(UNIT%cons,"("//trim(niceFmt(i))//")          "//DT_TY_DESC(i),setFlg(FLG%spcPrfx))
    end do
    call printMsg(UNIT%cons,"[<quantity>] system quantity is not present in 0-planets mode",setFlg(FLG%spcPrfx))
    call printMsg(UNIT%cons,"(<quantity>) quantity is part of subset repeated for each planet",setFlg(FLG%spcPrfx))
    call endSubsect(UNIT%cons,spcPrefx=.TRUE.) ! Notation
    
    nRows=2**(1+N_DT_TYPES)
    nCols=2+N_DT_TYPES
    deallocate(matrx1,matrx2)
    allocate(matrx1(nRows,nCols),matrx2(nRows+1),stat=allocStat)
    call testAllocErr(allocStat,60)
    
    i=0
    do iBin=0,1
      cBr1=merge("["," ",iBin == 0)
      cBr2=merge("]"," ",iBin == 0)
      do iDtTy1=0,1
        do iDtTy2=0,1
          do iDtTy3=0,1
            call inc(i)
            matrx1(i,1)=merge("-","x",iBin == 0)
            matrx1(i,2)=merge("-","x",iDtTy1 == 0)
            matrx1(i,3)=merge("-","x",iDtTy2 == 0)
            matrx1(i,4)=merge("-","x",iDtTy3 == 0)
            strPars=""
            ! DQs referring to system (per data set)
            do gQ1=1,N_G_DQS
              gQ=G_I_DQ_ROLE_ORDERED(gQ1)
              if (G_DQ_SCP_BY_COMB(iBin,iDtTy1,iDtTy2,iDtTy3,gQ) == QTY_SCP%sD) strPars=trim(strPars)//"; "//trim(G_DQ_NMS(gQ))
            end do
            ! DQs referring to system (once)
            do gQ1=1,N_G_DQS
              gQ=G_I_DQ_ROLE_ORDERED(gQ1)
              if (G_DQ_SCP_BY_COMB(iBin,iDtTy1,iDtTy2,iDtTy3,gQ) == QTY_SCP%s1) strPars=trim(strPars)//"; "//trim(G_DQ_NMS(gQ))
            end do
            ! DQs referring to system (once, if there are companion(s))
            do gQ1=1,N_G_DQS
              gQ=G_I_DQ_ROLE_ORDERED(gQ1)
              if (G_DQ_SCP_BY_COMB(iBin,iDtTy1,iDtTy2,iDtTy3,gQ) == QTY_SCP%sC) strPars=trim(strPars)//"; "//trim(cBr1)// &
               trim(G_DQ_NMS(gQ))//trim(cBr2)
            end do
            ! DQs referring to planet
            do gQ1=1,N_G_DQS
              gQ=G_I_DQ_ROLE_ORDERED(gQ1)
              if (G_DQ_SCP_BY_COMB(iBin,iDtTy1,iDtTy2,iDtTy3,gQ) == QTY_SCP%pl) strPars=trim(strPars)//"; ("//trim(G_DQ_NMS(gQ))// &
               ")"
            end do
            strPars=rmLead(strPars,"; ")
            if (len_trim(strPars) == 0) strPars=STR_UNDEF
            matrx1(i,5)=strPars
          end do
        end do
      end do
    end do
    
    matrx2=fmtMatrix(nRows+1,matrx1,reshape([character(STR_LEN_STD) :: "(B)","(1)","(2)","(3)","Derived quantities"],[1,5]))
    do i=1,nRows+1
      call printMsg(UNIT%cons,matrx2(i),setFlg(FLG%spcPrfx))
    end do
    call endSubsect(UNIT%cons) ! Combinations of modes and data and associated derived quantities
    
  end subroutine printParsDQsTbl




  !=========================================================[ PRIORS ]==============================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setPri(p,intvl,onlyIfUndef)
!
!   sets the support and other properties of the prior of a par.
!
!   Arguments:
!     p                              par. index
!     intvl                          prior support
!     onlyIfUndef                    whether prior support should only be set if it was undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p
    logical :: &
     onlyIfUndef
    type(tyIntvlDbl) :: &
     intvl
    optional :: &
     onlyIfUndef
    intent(in) :: &
     p, &
     intvl, &
     onlyIfUndef

    if (isUndef(p)) call err("bug 3 ("//trim(niceFmt(intvl%lBnd))//","//trim(niceFmt(intvl%uBnd))//")")
    
    if (present(onlyIfUndef) .AND. onlyIfUndef) then
      call setIfUndef(priLBnd(p),intvl%lBnd)
      call setIfUndef(priUBnd(p),intvl%uBnd)
    else
      priLBnd(p)=intvl%lBnd
      priUBnd(p)=intvl%uBnd
    end if
    priRge(p)%lBnd=priLBnd(p)
    priRge(p)%uBnd=priUBnd(p)
    priCtr(p)=intvlCentre(priRge(p))
    priW(p)=priRge(p)%width()
    priHalfW(p)=.5_dp*priW(p)

    if (priW(p) == 0) then

      parFixed(p)=.TRUE.

    else

      ! check prior type
      
      if (priLBnd(p) < 0) then
        if (priTypes(p) == PRI_TY%jef .OR. priTypes(p) == PRI_TY%modJef) then
          if (eql(-priLBnd(p),priUBnd(p))) then
            call warn("switching to signed modified Jeffreys prior for "//trim(parNms(p))//" because lower bound is negative")
            priTypes(p)=PRI_TY%sgnModJef
          else
            call warn("switching to uniform prior for "//trim(parNms(p))//" because lower bound is negative")
            priTypes(p)=PRI_TY%uni
          end if
        end if
      else if (priLBnd(p) == 0 .AND. priTypes(p) == PRI_TY%jef) then
        call warn("switching to modified Jeffreys prior for "//trim(parNms(p))//" because lower bound is zero")
        priTypes(p)=PRI_TY%modJef
      else if (priLBnd(p) > 0 .AND. (priTypes(p) == PRI_TY%modJef .OR. priTypes(p) == PRI_TY%sgnModJef)) then
        call warn("switching to Jeffreys prior for "//trim(parNms(p))//" because lower bound is positive")
        priTypes(p)=PRI_TY%jef
      end if
     
      if (priTypes(p) == PRI_TY%sgnModJef .AND. .NOT. eql(-priLBnd(p),priUBnd(p))) then
        if (priLBnd(p) < 0) then
          call warn("switching to uniform prior for "//trim(parNms(p))//" due to non-standard bounds")
          priTypes(p)=PRI_TY%uni
        else if (priLBnd(p) == 0) then
          call warn("switching to modified Jeffreys prior for "//trim(parNms(p))//" because lower bound is zero")
          priTypes(p)=PRI_TY%modJef
        else
          call warn("switching to Jeffreys prior for "//trim(parNms(p))//" due to non-standard bounds")
          priTypes(p)=PRI_TY%jef
        end if
      end if

      if (priLBnd(p) <= 0._dp .OR. priTypes(p) == PRI_TY%uni .OR. priTypes(p) == PRI_TY%norm) then
        priLogScl(p)=.FALSE.
        maPoLogScl(p)=.FALSE.
      end if
      priPsdLBnd(p)=merge(log10(priLBnd(p)),priLBnd(p),priLogScl(p))

      select case (priTypes(p))
      case (PRI_TY%uni)
        priFac(p)=1._dp/priW(p)
      case (PRI_TY%jef)
        priFac(p)=1._dp/log(priUBnd(p)/priLBnd(p))
      case (PRI_TY%modJef)
        priKnees(p)=modJefPriKnee(priUBnd(p))
        priFac(p)=1._dp/log((priUBnd(p)+priKnees(p))/priKnees(p))
      case (PRI_TY%sgnModJef)
        priKnees(p)=modJefPriKnee(priUBnd(p))
        priFac(p)=1._dp/(2._dp*log((priUBnd(p)+priKnees(p))/priKnees(p)))
      case (PRI_TY%norm)
        block
          real(dp) :: &
           tmp
          
          tmp=1._dp/(sqrt(2._dp*priVce(p)))
          priFac(p)=1._dp/erf(.5_dp*priW(p)*tmp)*tmp/sqrt(PI) ! cf. [02.07.2012/4]

        end block
      end select
    end if                           ! (priW(p) == 0)

  end subroutine setPri




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setPriFromFil(p,path)
!
!   sets normalised prior points from file, approximating the prior by trapezoids between the points;
!   if prior range is also set independently, uses only the corresponding subset of points
!
!   Arguments:
!     p                              par. index
!     path                           path to file
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     PRI_N_PTS_TMP_MAX=1000000
    integer :: &
     p, &
     i, &
     N, &
     i1, &
     i2, &
     allocStat                       ! ,iArr(PRI_N_PTS_MAX)
    integer(i8b) :: &
     iArr(:)
    real(dp) :: &
     delta, &
     mul, &
     priArr(:,:)                     ! needs to be allocatable to avoid insufficient memory
    logical :: &
     regSpc, &
     logScl
    character(*) :: &
     path
    type(tyIntvlDbl) :: &
     priRgeFil, &
     psdRge
    type(tyCoord2DDbl) :: &
     densEst(PRI_N_PTS_MAX)
    allocatable :: &
     priArr, &
     iArr
    intent(in) :: &
     p, &
     path

    allocate(priArr(PRI_N_PTS_TMP_MAX,2),iArr(PRI_N_PTS_TMP_MAX),stat=allocStat)
    call testAllocErr(allocStat,52)
    if (.NOT. readCSVArr(path,priArr,N)) call err("too many (> "//trim(niceFmt(PRI_N_PTS_TMP_MAX))//") prior points to read in") ! arr IS CONTIGUOUS


    ! sort

    call indxArr(priArr(1:N,1),iArr(1:N))
    priArr(1:N,:)=priArr(iArr(1:N),:)


    ! check for duplicates

    do i=1,N-1
      if (priArr(i,1) == priArr(i+1,1)) call err("prior file for "//trim(parNms(p))// &
       " contains conflicting points")
    end do


    ! check compatibility with potientially given prior range

    if (isDef(priRge(p)) .AND. .NOT. intvlIsInIntvl(priRge(p),tyIntvlDbl(priArr(1,1),priArr(N,1)))) &
     call err("prior ranges for "//trim(parNms(p))//" as given with "// &
     OPT_PREFX//trim(CL_OPTS(I_CL_OPTS%prior)%short)//" and "//OPT_PREFX//trim(CL_OPTS(I_CL_OPTS%range)%short)// &
     " options are incompatible")


    ! find positive prior points (within a potentially given prior range)

    psdRge=merge(priRge(p),G_STRICT_PRI_RGE(gIPars(p)),isDef(priRge(p)))
    call undef(i1)
    call undef(i2)
    do i=1,N
      if (isInIntvl(psdRge,priArr(i,1)) .AND. priArr(i,2) > 0) then
        i1=i
        exit
      end if
    end do
    if (isUndef(i1)) call err("insufficient prior points for "//trim(parNms(p)))
    do i=N,i1,-1
      if (isInIntvl(psdRge,priArr(i,1)) .AND. priArr(i,2) > 0) then
        i2=i
        exit
      end if
    end do
    N=i2-i1+1                        ! # points in range
    if (isUndef(i2) .OR. N <= 1) call err("insufficient prior points for "//trim(parNms(p)))
    delta=(priArr(i2,1)-priArr(i1,1))/(N-1)


    ! guess whether prior abscissa given in log scale
    ! (assume log scale unless first three points evenly spaced or leftmost point at non-pos. position)

    logScl=merge(.FALSE.,(priArr(i1+2,1)-priArr(i1+1,1)-(priArr(i1+1,1)-priArr(i1,1)))/(priArr(i1+1,1)-priArr(i1,1)) > &
     PRI_REG_SPC_DEVN_RATIO_MAX,N == 2 .OR. priArr(i1,1) <= 0)


    ! multiplier or delta; prior range from file

    if (logScl) mul=(priArr(i2,1)/priArr(i1,1))**(1._dp/(N-1))
    priRgeFil=G_STRICT_PRI_RGE(gIPars(p))%sect(tyIntvlDbl(priArr(i1,1),priArr(i2,1)))


    ! check compatibility of prior range from file with strict range

    if (isUndef(priRgeFil)) call err("prior range for "//trim(parNms(p))//" as given with "// &
     OPT_PREFX//trim(CL_OPTS(I_CL_OPTS%prior)%short)//" option must be a subset of ["// &
     trim(niceFmt(G_STRICT_PRI_RGE(gIPars(p))%lBnd,OPT_REMOVE_0))//","// &
     trim(niceFmt(G_STRICT_PRI_RGE(gIPars(p))%uBnd,OPT_REMOVE_0))//"]")


    ! final prior range

    if (isUndef(priRge(p))) call setPri(p,priRgeFil)


    ! check compatibility of prior lbound and log scale

    if (logScl) then
      if (priLBnd(p) == 0) then
        priKnees(p)=modJefPriKnee(priUBnd(p)) ! mod. Jef. prior
      else if (priLBnd(p) < 0) then
        logScl=.FALSE.
      end if
    end if


    ! check for regular spacing (log scale)
    ! (switch back to non-log scale if non-regular spacing)

    if (logScl) then

      regSpc=all(regSpc1Log(genSeq(i1,i2-1)))

      if (regSpc .AND. N >= PRI_N_PTS_MIN) then
        if (abs(priArr(i1,1)/priLBnd(p)-1._dp) > PRI_REG_SPC_DEVN_RATIO_MAX) priLBnd(p)=priArr(i1,1)
        if (abs(priUBnd(p)/priArr(i2,1)-1._dp) > PRI_REG_SPC_DEVN_RATIO_MAX) priUBnd(p)=priArr(i2,1)
        call setPri(p,tyIntvlDbl(priLBnd(p),priUBnd(p)))
      end if
      logScl=regSpc

    end if


    ! reduce # points if necessary

    if (N > PRI_N_PTS_MAX) then
      do i=0,PRI_N_PTS_MAX-1
        priArr(i+1,:)=priArr(1+nint(1._dp*i*(N-1)/(PRI_N_PTS_MAX-1)),:)
      end do
      N=PRI_N_PTS_MAX
    end if


    ! check for regular spacing (non-log scale)

    if (.NOT. logScl) then
      regSpc=all(regSpc1(genSeq(i1,i2-1)))
      if (regSpc .AND. N >= PRI_N_PTS_MIN) then
        if (abs((priArr(i1,1)-priLBnd(p))/delta-1._dp) > PRI_REG_SPC_DEVN_RATIO_MAX) priLBnd(p)=priArr(i1,1)
        if (abs((priUBnd(p)-priArr(i2,1))/delta-1._dp) > PRI_REG_SPC_DEVN_RATIO_MAX) priUBnd(p)=priArr(i2,1)
        call setPri(p,tyIntvlDbl(priLBnd(p),priUBnd(p)))
      end if
    end if


    ! set points either by KDE (if too few) or just using those from file

    if (N < PRI_N_PTS_MIN .OR. .NOT. regSpc) then

      N=PRI_N_PTS_MAX
      if (logScl) then
        mul=(priUBnd(p)/priLBnd(p))**(1._dp/(N-1))
      else
        delta=priW(p)/(N-1)
      end if
      call kerDensEst1D(dens1DNorm,I_KER_1D%norm,priArr(i1:i2,1),sqrt(arrVar(priArr(i1:i2,1))),pltRge(priRge(p),logScl, &
       priTypes(p)),int(N,i8b),densEst,setFlg([FLG%sorted,FLG%logScl],[.TRUE.,logScl]),wArrIn=priArr(i1:i2,2))

      priPtsPsdX(p,1:N)=densEst%x
      priPtsY(p,1:N)=densEst%y

      call warn("prior for "//trim(parNms(p))//" has been re-estimated due to "// &
       trim(mrg("small number of points","nonregularly spaced points",regSpc)))

    else

      do i=1,N
        priPtsPsdX(p,i)=priArr(i-1+i1,1)
        priPtsY(p,i)=priArr(i-1+i1,2)
      end do

    end if


    ! normalise

    if (logScl) then
      delta=log10(mul)
      priPtsPsdX(p,1:N)=log10(priPtsPsdX(p,1:N))
    end if

    priPtsY(p,1:N)=priPtsY(p,1:N)/(delta*(.5_dp*(priPtsY(p,1)+priPtsY(p,N))+sum(priPtsY(p,2:N-1))))


    ! bounds & width

    priInvDelta(p)=1._dp/delta

    if (priW(p) <= 0) then
      priLBnd(p)=priUBnd(p)
      parFixed(p)=.TRUE.
    else
      priLogScl(p)=logScl
    end if

    priNPts(p)=N


    call setPriDiffFromFil(p)


    contains

    elemental function regSpc1(i1)   ! returns whether spacing of two points corresponds to regular spacing
      implicit none
      integer :: &
       i1
      logical :: &
       regSpc1
      intent(in) :: &
       i1

      regSpc1=(abs((priArr(i1+1,1)-priArr(i1,1))/delta-1._dp) <= PRI_REG_SPC_DEVN_RATIO_MAX)
    end function


    elemental function regSpc1Log(i1)   ! returns whether spacing of two points corresponds to regular spacing in log scale
      implicit none
      integer :: &
       i1
      logical :: &
       regSpc1Log
      intent(in) :: &
       i1

      regSpc1Log=abs(priArr(i1+1,1)/priArr(i1,1)-mul) <= PRI_REG_SPC_DEVN_RATIO_MAX
    end function

  end subroutine setPriFromFil




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setPriDiffFromFil(p)
!
!   sets differences between adjacent prior points from a file
!
!   Arguments:
!     p                              par. index
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p, &
     i
    intent(in) :: &
     p


    do i=1,priNPts(p)-1
      priDiffY(p,i)=priPtsY(p,i+1)-priPtsY(p,i)
    end do

  end subroutine setPriDiffFromFil




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function modJefPriKnee(uBnd)
!
!   Returns:                         knee of the modified Jeffreys prior for a given upper prior bound
!
!   Arguments:
!     uBnd                           upper prior bound
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    real(dp) :: &
     modJefPriKnee, &
     uBnd
    intent(in) :: &
     uBnd

    modJefPriKnee=min(uBnd/10._dp**MOD_JEF_PRI_KNEE_SCL_DIST,1._dp)

  end function modJefPriKnee




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function parOutsdPriSupp(p,par)
!
!   Returns:                         whether parameter is outside its prior support (i.e. its prior is zero)
!
!   Arguments:
!     p                              par. index
!     par                            par. value
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p
    real(dp) :: &
     par
    logical :: &
     parOutsdPriSupp
    intent(in) :: &
     p, &
     par

    parOutsdPriSupp=(par-priLBnd(p))*(priUBnd(p)-par) <= 0

  end function parOutsdPriSupp




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function priorIsNonZero(pars)
!
!   Returns:                         whether the prior of a par. vector is non-zero
!
!   Arguments:
!     pars                           parameter vector
!
!   assumes that the prior support is continuous (only checks bounds of allowed range)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     pars(:)
    logical :: &
     priorIsNonZero
    intent(in) :: &
     pars

    priorIsNonZero=all(parFixed(1:curNPars) .OR. (pars-priLBnd(1:curNPars))*(priUBnd(1:curNPars)-pars) > 0)

  end function priorIsNonZero




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function singlePrior(p,par)
!
!   Returns:                         the prior of a single parameter
!
!   Arguments:
!     p                              par. index
!     par                            par. value
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p, &
     i
    real(dp) :: &
     par, &
     psdPar, &
     singlePrior
    intent(in) :: &
     p, &
     par

    if (parFixed(p)) then
      singlePrior=1._dp
    else if (parOutsdPriSupp(p,par)) then ! parameter out of prior range
      singlePrior=0
    else
      select case (priTypes(p))
      case (PRI_TY%uni)
        singlePrior=priFac(p)
      case (PRI_TY%jef)
        singlePrior=priFac(p)/par
      case (PRI_TY%modJef)
        singlePrior=priFac(p)/(par+priKnees(p))
      case (PRI_TY%sgnModJef)
        singlePrior=priFac(p)/(abs(par)+priKnees(p))
      case (PRI_TY%norm)
        singlePrior=priFac(p)*exp(-(par-priCtr(p))**2/(2._dp*priVce(p)))
      case (PRI_TY%fil)
        if (priLogScl(p)) then
          psdPar=log10(par)
        else
          psdPar=par
        end if
        i=ceiling((psdPar-priPsdLBnd(p))*priInvDelta(p))
        singlePrior=priPtsY(p,i)+(psdPar-priPtsPsdX(p,i))*priDiffY(p,i)*priInvDelta(p)
      end select
    end if

  end function singlePrior




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function prior(pars)
!
!   Returns:                         prior of a parameter vector
!
!   Arguments:
!     pars                           parameter vector
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    real(dp) :: &
     prior, &
     pars(curNPars)
    intent(in) :: &
     pars

    prior=1._dp
    do i=1,curNPars
      prior=prior*singlePrior(i,pars(i))
    end do

  end function prior




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function logPrior(pars)
!
!   Returns:                         natural logarithm of prior of a parameter vector
!
!   Arguments:
!     pars                           parameter vector
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    real(dp) :: &
     logPrior, &
     pars(curNPars)
    intent(in) :: &
     pars

    logPrior=0._dp
    do i=1,curNPars
      logPrior=logPrior+log(singlePrior(i,pars(i)))
    end do

  end function logPrior




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function disprsOvrPriIntvl(p,N,sorted)
!
!   Returns:                         samples dispersed over the prior interval using rejection sampling,
!                                    making sure that sample variance is greater than prior var.
!
!   Arguments:
!     p                              par. index
!     N                              # samples
!     sorted                         whether the result should be sorted
!
!   CAUTION: not thread-safe!
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p, &
     i, &
     j, &
     N
    real(dp) :: &
     disprsOvrPriIntvl(N), &
     setWMaxVar(N), &
     varCur, &
     varMax, &
     priVar
    logical :: &
     sorted
    intent(in) :: &
     p, &
     N, &
     sorted

    varMax=0._dp
    priVar=0._dp

    ! numerically estimate prior variance
    block
      integer,parameter :: &
       N_PTS=1000
      real(dp) :: &
       step, &
       parCur, &
       lBnd

      step=priW(p)/N_PTS
      lBnd=priLBnd(p)

      do i=1,N_PTS-1
        parCur=lBnd+i*step
        priVar=priVar+(parCur-expect(p))**2*singlePrior(p,parCur)
      end do
      priVar=priVar*step
    end block

    if (N == 1) then
      disprsOvrPriIntvl(1)=rejSample(p)
    else
      do i=1,DISPRS_N_TRIES
        do j=1,N
          disprsOvrPriIntvl(j)=rejSample(p)
        end do
        varCur=arrVar(disprsOvrPriIntvl)
        if (varCur > varMax) then
          varMax=varCur
          setWMaxVar=disprsOvrPriIntvl
        end if
        if (varCur > priVar) exit
      end do
      disprsOvrPriIntvl=setWMaxVar
      if (sorted) call sort(disprsOvrPriIntvl)
    end if


    contains

    function expect(p)               ! returns the prior expectation of a par.
      implicit none
      integer :: &
       p
      real(dp) :: &
       expect
      intent(in) :: &
       p

      select case (priTypes(p))
      case (PRI_TY%uni)
        expect=(priLBnd(p)+priUBnd(p))/2._dp
      case (PRI_TY%jef)
        expect=priW(p)/log(priUBnd(p)/priLBnd(p))
      case (PRI_TY%modJef)
        expect=priUBnd(p)/log(priUBnd(p)/priKnees(p)+1._dp)-priKnees(p)
      case (PRI_TY%sgnModJef)
        expect=0._dp
      case (PRI_TY%fil)
        if (priLogScl(p)) then
          expect=sum(10._dp**(priPtsPsdX(p,1:priNPts(p)))*priPtsY(p,1:priNPts(p)))
        else
          expect=sum(priPtsPsdX(p,1:priNPts(p))*priPtsY(p,1:priNPts(p)))
        end if
      end select
    end function


    function rejSample(p)            ! returns a rejection sample from the prior of a par.
      integer :: &
       p
      real(dp) :: &
       rejSample, &
       rndTmp, &
       rndTmp2
      intent(in) :: &
       p

      do
        call random_number(rndTmp)
        call random_number(rndTmp2)
        rejSample=priLBnd(p)+rndTmp*priW(p)
        if (rndTmp2 < singlePrior(p,rejSample)) exit
      end do
    end function

  end function disprsOvrPriIntvl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtPrior(parNm)
!
!   Returns:                         formatted string for prior of a parameter
!
!   Arguments:
!     parNm                          par. name
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     parNm
    character(STR_LEN_STD) :: &
     fmtPrior
    intent(in) :: &
     parNm

    if (pltPSTricks) then
      fmtPrior="$p($"//trim(parNm)//"$|\mathcal M,I)$"
    else
      fmtPrior="p("//trim(parNm)//"|M,I)"
    end if

  end function fmtPrior




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function pltRge(qtyRge,logScl,priTy)
!
!   Returns:                         modification of given range of a parameter or DQ suitable for plotting
!
!   Arguments:
!     qtyRge                         original range for qty.
!     logScl                         whether qty. is plotted in logarithmic scale
!     priTy                          optional prior type if qty. is a parameter
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     priTy
    real(dp) :: &
     k
    logical :: &
     logScl
    type(tyIntvlDbl) :: &
     pltRge, &
     qtyRge
    optional :: &
     priTy
    intent(in) :: &
     qtyRge, &
     logScl, &
     priTy


    if (logScl) then
      if (present(priTy) .AND. priTy == PRI_TY%modJef) then
        k=LOG_SCL_L_BND_KNEE_RATIO*modJefPriKnee(qtyRge%uBnd)
      else
        k=-huge(k)
      end if
      pltRge=tyIntvlDbl(merge(qtyRge%lBnd,max(LOG_SCL_PSD_MIN,k),qtyRge%lBnd > 0),qtyRge%uBnd)
    else
      pltRge=qtyRge
    end if

  end function pltRge




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printPriRges
!
!   prints prior ranges of all pars.
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p

    call startSubsect(UNIT%cons+UNIT%log,"Priors")

    do p=1,curNPars
      call startSubsect(UNIT%cons+UNIT%log,fmtPar(p,OPT_FOR_STDOUT))
      if (parFixed(p)) then
        call printQty(UNIT%cons+UNIT%log,"Fixed value",priLBnd(p))
      else
        call printQty(UNIT%cons+UNIT%log,"Type",PRI_TY_DESC(priTypes(p)))
        call printQty(UNIT%cons+UNIT%log,"Lower limit",priLBnd(p))
        call printQty(UNIT%cons+UNIT%log,"Upper limit",priUBnd(p))
        if (priTypes(p) == PRI_TY%modJef .OR. priTypes(p) == PRI_TY%sgnModJef) then
          call printQty(UNIT%cons+UNIT%log,"Knee",priKnees(p))
        else if (priTypes(p) == PRI_TY%norm) then
          call startSubsect(UNIT%cons+UNIT%log,"Underlying normal distribution")
          call printQty(UNIT%cons+UNIT%log,"Mean",priCtr(p))
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",sqrt(priVce(p)))
          call endSubsect(UNIT%cons+UNIT%log)
        end if
      end if
      call endSubsect(UNIT%cons+UNIT%log)
    end do

    call endSubsect(UNIT%cons+UNIT%log) ! Prior parameter ranges

  end subroutine printPriRges




  !=======================================================[ LIKELIHOOD ]============================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function logLikh(pars,flags,resid1D,resid2D,normResid,dtTy)
!
!   Returns:                         depending on arguments:
!                                      * ln(likelihood) minus const. (i.e. const. for given data; cf. [07.05.2012/1]);
!                                        const. can be subtracted since only difference of two log(likelihood) is used
!                                      * chi²; flags needs to be given; dtTy may be given to have chi² be calcd. only w.r.t. to dtTy
!                                      * residuals; flags, {resid1D or resid2D} and dtTy need to be given
!                                      * normalised residuals; flags, normResid and dtTy need to be given
!
!   Arguments:
!     pars                           parameters
!     flags                          flags:
!                                      * FLG%ignJitt: ignore jitter (only for chi² and normalised residuals)
!     resid1D                        1D residuals
!     resid2D                        2D residuals
!     normResid                      normalised residuals
!     dtTy                           data type (for (normalised) residuals, chi², or ln(likelihood))
!
!   OMP not advantageous for loops below
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     flags, &
     dtTy, &
     iSet
    integer(i8b) :: &
     i, &
     j, &
     N
    real(dp) :: &
     logLikh, &
     pars2D(1,curNPars), &
     eAArr(nUniqT,psdNPl,3), &
     mdlFct1D(N_REC_PER_SET_MAX,1), &
     mdlFct2D(N_REC_PER_SET_MAX,2), &
     logLikhAddend, &
     invUncSq1D, &

     ! args.
     pars(:), &
     resid1D(:), &
     normResid(:), &
     
     ! additional noise
     sgmPlusSq, &                    ! not initialised with zero if isUndef(pRole%sgmPlus)!
     tauPlusSq, &                    ! not initialised with zero if isUndef(pRole%tauplus)!
     
     ! AMAngPos
     normResidSqAMAngPos(dtPropsAMAngPos%nRec), &
     normResidSgnAMAngPos(dtPropsAMAngPos%nRec), &
     recprASq, &
     recprBSq, &
     phi, &
     sinPhi, &
     sinSqPhi, &
     cosPhi, &
     cosSqPhi, &
     dDel, &
     dDelSq, &
     dAlCosDel, &
     dAlCosDelSq, &
     residPosAng, &
     
     ! RV
     normResidSqRV(dtPropsRV%nRec), &
     residRV(dtPropsRV%nRec), &
     
     ! AMHipN
     normResidSqAMHipN(dtPropsAMHipN%nRec), &
     residAMHipN(dtPropsAMHipN%nRec), &
     chiSq(N_DT_TYPES)
    type(tyCoord2DDbl) :: &
     resid2D(:), &
     residAMAngPos(dtPropsAMAngPos%nRec)
    type(tyUncEllDbl) :: &
     uncAMAngPos
    type(tyParamsMdl) :: &
     paramsMdl
    optional :: &
     flags, &
     resid1D, &
     resid2D, &
     normResid, &
     dtTy
    intent(in) :: &
     pars, &
     flags, &
     dtTy
    intent(inout) :: &
     normResid



    pars2D=reshape(pars,[1,curNPars])
    paramsMdl%psdNPl=psdNPl
    chiSq=0._dp
    logLikhAddend=0._dp
    

    ! prepare non-/usage of jitter

    if (isDef(pRole%sgmPlus)) then
      if (present(flags) .AND. getFlg(flags,FLG%ignJitt)) pars2D(1,pRole%sgmPlus)=0._dp
      sgmPlusSq=pars2D(1,pRole%sgmPlus)**2
    end if
    if (isDef(pRole%tauPlus)) then
      if (present(flags) .AND. getFlg(flags,FLG%ignJitt)) pars2D(1,pRole%tauPlus)=0._dp
      tauPlusSq=pars2D(1,pRole%tauPlus)**2
    end if


    ! calc. eccentric anomaly

    eAArr=ecAn(uniqTMnEpoch(:nUniqT),nUniqT,psdNPl,pars2D,nxtInt8,identInt8) ! pars2D IS CONTIGUOUS


    ! main part of calculation
    
      ! AMAngPos

    j=0
    do iSet=1,dtProps%nSetsOfTy(DT_TY%AMAngPos)

      N=dtSetsAMAngPos(iSet)%N
      paramsMdl%iSet=iSet
      paramsMdl%binCompo=dtSetsAMAngPos(iSet)%binCompo
      mdlFct2D(1:N,:)= &
       mdlProps%AMAngPos%mdlFct(N,MDL_FCT_DIM%AMAngPos,pars2D,nxtInt8,identInt8,eAArr(dtSetsAMAngPos(iSet)%iUniqT(1:N),:,:), &
       paramsMdl)
      
      do i=1,N

        call inc(j)
        uncAMAngPos=dtSetsAMAngPos(iSet)%unc(i)
        recprASq=1._dp/(uncAMAngPos%a**2+tauPlusSq)
        recprBSq=1._dp/(uncAMAngPos%b**2+tauPlusSq)
        phi=uncAMAngPos%phi
        sinPhi=uncAMAngPos%sinPhi
        sinSqPhi=uncAMAngPos%sinSqPhi
        cosPhi=uncAMAngPos%cosPhi
        cosSqPhi=uncAMAngPos%cosSqPhi
        dDel=dtSetsAMAngPos(iSet)%obs(i)%x-mdlFct2D(i,1)
        dDelSq=dDel**2
        dAlCosDel=dtSetsAMAngPos(iSet)%obs(i)%y-mdlFct2D(i,2)
        dAlCosDelSq=dAlCosDel**2
        residAMAngPos(j)%x=dDel
        residAMAngPos(j)%y=dAlCosDel
        residPosAng=atan2(dAlCosDel,dDel)
        normResidSgnAMAngPos(j)=(residPosAng-(phi-PI))*(phi-residPosAng)

        normResidSqAMAngPos(j)= &
         sinSqPhi*(dDelSq*recprBSq+dAlCosDelSq*recprASq)+cosSqPhi*(dDelSq*recprASq+dAlCosDelSq*recprBSq)+ & ! cf. [18.11.2010/2]
         2._dp*sinPhi*cosPhi*dDel*dAlCosDel*(recprASq-recprBSq)
        
        logLikhAddend=logLikhAddend+log(recprASq*recprBSq)
        
      end do
      
      chiSq(DT_TY%AMAngPos)=sum(normResidSqAMAngPos)

    end do

      ! RV

    j=0
    do iSet=1,dtProps%nSetsOfTy(DT_TY%RV)

      N=dtSetsRV(iSet)%N
      paramsMdl%iSet=iSet
      paramsMdl%binCompo=dtSetsRV(iSet)%binCompo
      mdlFct1D(1:N,:)=mdlProps%RV%mdlFct(N,MDL_FCT_DIM%RV,pars2D,nxtInt8,identInt8,eAArr(dtSetsRV(iSet)%iUniqT(1:N),:,:), &
       paramsMdl,tMnEpochArr=dtSetsRV(iSet)%tMnEpoch)
      
      do i=1,N

        call inc(j)
        residRV(j)=dtSetsRV(iSet)%obs(i)-mdlFct1D(i,1)
        invUncSq1D=1._dp/(dtSetsRV(iSet)%sgmSq(i)+sgmPlusSq)
        normResidSqRV(j)=residRV(j)**2*invUncSq1D
        logLikhAddend=logLikhAddend+log(invUncSq1D)
      end do
      
      chiSq(DT_TY%RV)=sum(normResidSqRV)

    end do

      ! AMHipN


    j=0
    do iSet=1,dtProps%nSetsOfTy(DT_TY%AMHipN)

      N=dtSetsAMHipN(iSet)%N
      paramsMdl%iSet=iSet
      paramsMdl%binCompo=dtSetsAMHipN(iSet)%binCompo
      mdlFct1D(1:N,:)=mdlProps%AMHipN%mdlFct(N,MDL_FCT_DIM%AMHipN,pars2D,nxtInt8,identInt8, &
       eAArr(dtSetsAMHipN(iSet)%iUniqT(1:N),:,:),paramsMdl,dtSetsAMHipN(iSet)%params,dtSetsAMHipN(iSet)%tMnEpoch)
      
      do i=1,N

        call inc(j)
        residAMHipN(j)=dtSetsAMHipN(iSet)%obs(i)-mdlFct1D(i,1)
        invUncSq1D=1._dp/(dtSetsAMHipN(iSet)%sgmSq(i)+tauPlusSq)
        normResidSqAMHipN(j)=residAMHipN(j)**2*invUncSq1D
        logLikhAddend=logLikhAddend+log(invUncSq1D)

      end do
      
      chiSq(DT_TY%AMHipN)=sum(normResidSqAMHipN)

    end do


    ! finish calculation depending on requested type of result

    if (present(flags)) then            ! return either chi² or normalised residuals
      if (present(resid1D)) then        ! "return" 1-d residuals
        select case (dtTy)
        case (DT_TY%RV)
          resid1D(1:dtPropsRV%nRec)=residRV
        case (DT_TY%AMHipN)
          resid1D(1:dtPropsAMHipN%nRec)=residAMHipN
        end select
      else if (present(resid2D)) then   ! "return" 2-d residuals
        select case (dtTy)
        case (DT_TY%AMAngPos)
          resid2D(1:dtPropsAMAngPos%nRec)=residAMAngPos
        end select
      else if (present(normResid)) then ! "return" normalised residuals
        select case (dtTy)
        case (DT_TY%AMAngPos)
          normResid(1:dtPropsAMAngPos%nRec)=sign(sqrt(normResidSqAMAngPos),normResidSgnAMAngPos)
        case (DT_TY%RV)
          normResid(1:dtPropsRV%nRec)=sign(sqrt(normResidSqRV),residRV)
        case (DT_TY%AMHipN)
          normResid(1:dtPropsAMHipN%nRec)=sign(sqrt(normResidSqAMHipN),residAMHipN)
        end select
      else                              ! return chi²
        if (present(dtTy)) then
          logLikh=chiSq(dtTy)
        else
          logLikh=sum(chiSq)
        end if
      end if
    else                                ! return ln(likelihood)-const.
      logLikh=.5_dp*(logLikhAddend-sum(chiSq))
    end if

  end function logLikh




end module
