module inference                                      ! Posterior inference

! Routines:
! 
! chkPostrMode           Checks whether posterior mode is defined, else warns/casts an error
! findBestEst            Determines best summary of pars.
! calcQuantMaPoSort      Calculates quantiles, marg. postrs. and residuals, sorting the post-burn-in chain links
! calcMedHPDI            Calculates highest-posterior-density intervals (HPDIs), median, and (optionally) range of posterior samples
! calcHDIsDbl            Calculates highest-density intervals (HDIs) based on samples
! calcHDIsQP             Calculates highest-density intervals (HDIs) based on samples
! calcProb               Calculates probability over given intervals
! calcMargPostr          Calculates (zoomed) marginal posteriors
! estimFrqModeStdDev     Returns: estimated standard deviation of marginal mode of orbital frequency
! fmtHPDIProbCont        Returns: formatted string with HPDI probability content
! calcEvVT               Calculates evidence with Volume-Tesselation Algorithm (VTA)
! calcResid              Calculates residuals using best summaries for all pars.
! calcChiSq              Calculates chi²
! calcAllPred            Calculates all uncertainty predictions
! plotDataAndModel       Plots data and model
! plotDistrNormResid     Plots distributions of normalised residuals
! calcSummaries          Calculates various posterior summaries
! calcProbAllPars        Calculates posterior probability over given hypercubes
! printSummaries         Prints posterior summaries
! doPlotting             Performs plotting
! makePostrPlots         Produces posterior-plot files
! fmtPostr               Returns: formatted string, depending on args., for marginal posterior of a parameter or posterior times evidence
! fmtFilNmPlt            Returns: formatted string to be used in file name for plotting

  use basicTypes
  use strings
  use intervals
  use io
  use kdTree
  use dataModelsNoise
  use sampling
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !======================================================[ CONSTANTS (1) ]==========================================================

  integer,parameter :: &
   N_PTS_PERIODOGRAM_PER_STDEV=7, &
   N_STD_HPDI=6, &                   ! # HPDIs
   N_USR_HPDI_MAX=50, &              ! max. # user-requested HPDIs
   I_STD_HPDI_MA_PO_ZM=5, &          ! index of STD_HPDI_PROB_CONT for zoomed marg. postrs.
   HPDI_N_DIGS_MAX=4, &              ! max. # counting digits of STD_HPDI_PROB_CONT
   N_CONTOURS=3                      ! # contours for joint marg. postrs.


  !=====================================================[ TYPE DEFINITIONS ]========================================================

  type tyPostrSumm
    type(tySummJitt) :: &
     maxPostrChiSq, &                ! MAP chi²
     maxPostrChiSqByDtTy(N_DT_TYPES) ! MAP chi² by data type
  end type

  
  ! PARAMETER/UNCERTAINTY ESTIMATION
  
  type tyParEst
    real(dp) :: median,mode,bestSumm,sampStdDev,cov(N_L_PARS_OR_DQ_MAX),corrCoeff(N_L_PARS_OR_DQ_MAX), &
     KDEBandw
    integer :: nIndepSamp,bestSummTy
    integer(i8b) :: nInHPDI(N_STD_HPDI+N_USR_HPDI_MAX)
    type(tyIntvlDbl) :: HPDI(N_STD_HPDI+N_USR_HPDI_MAX)
    type(tyUncQtyDbl) :: mean,mMode
  end type

  type tyParEstSummInt
    integer :: mean,median,mMode,mode
  end type

  type tyParEstSummStr
    character(STR_LEN_STD) :: mean,median,mMode,mode
  end type

  type tyParEstStr
    character(STR_LEN_STD) :: scalar,mean,median,mMode,mode,bestSumm,sampStdDev,stdDevMean,nIndepSamp,HPDIL,HPDIU,matrix,covM, &
     corrM,KDEBandw
  end type
  
  type tyMdlSelStr
    character(STR_LEN_STD) :: ev,bayesFac
  end type


  ! PREDICTION

  type tyPredGeneric
    integer :: &
     nPts
    logical,allocatable :: &
     skipSet(:,:)
    type(tyIntvlDbl),allocatable :: &
     tRge(:,:) ! data type, data set
    real(dp),allocatable :: &
     t(:,:,:), & ! data type, data set, prediction point
     minUnc(:,:), & ! data type, data set
     maxUnc(:,:), &
     tMinUnc(:,:), &
     tMaxUnc(:,:), &
     tResol(:,:)
  end type

  type,extends(tyPredGeneric) :: tyPred1D
    real(dp),allocatable :: &
     best(:,:,:), &
     unc(:,:,:)
  end type

  type,extends(tyPredGeneric) :: tyPred2D
    real(dp),allocatable :: &
     phiMinUnc(:,:), &
     phiMaxUnc(:,:)
    type(tyCoord2DDbl),allocatable :: &
     best(:,:,:)
    type(tyUncEllDbl),allocatable :: &
     unc(:,:,:)
  end type


  ! PLOTTING

  type tyPltResolInt
    integer :: std,quick
  end type

  type tyPltResolInt8
    integer(i8b) :: std,quick
  end type

  type tyPltResolReal
    integer :: std,quick
  end type

  type tyPltTypesStr
    character(STR_LEN_STD) :: data,resid,residNorm,residVect,trace,PSR,maPo,maPoDQ,scatter,jointMaPo,mdl,CPO,zoom,median,mean, &
     mode,fold,uncEll,distrNormResid,prior,predict,range,perCoord,uncert,lBnd,uBnd,a,b,phi,wJitt,woJitt,negLogPriLikh,autoCorr
  end type

  type tyPltSymInt
    integer :: line,point,dot,errBar
  end type

  type tyPltSymStr
    character(STR_LEN_STD) :: line,point,dot,errBar
  end type

  type tyTtlPartsStr
    character(STR_LEN_STD) :: data,mdl,distr,normResid,normResids,wJitt,woJitt,jointMaPo,maPo,sampNo,trace,predict, &
     uncEllAx,uncEllPosAng,prior,probDens,PSR,resid,scatter,negLogPriLikh,autoCorr,lag
  end type


  !======================================================[ CONSTANTS (2) ]==========================================================

  type(tyPltTypesStr),parameter :: &
   FIL_NM_PART=tyPltTypesStr("data","resid","resid-norm","resid-vect","trace","PSR","marg-posterior","marg-posterior-deriv", &
   "scatter","joint-marg-posterior","model","CPO","zoom","median","mean","mode","fold","uncert-ell","distr-norm-resid","prior", &
   "predict","range","per-coord","uncert","lbound","ubound","semi-maj","semi-min","pos-ang","with-jitter","without-jitter", &
   "ln-pri-likh","autocorr")


  ! PARAMETER/UNCERTAINTY ESTIMATION

  type(tyParEstSummInt),parameter :: &
   PAR_EST_SUMM_TY=tyParEstSummInt(1,2,3,4)

  type(tyParEstStr),parameter :: &
   PAR_EST_FIL_NM=tyParEstStr("scalar-summ","","","","","","","","","","","","cov-matrix","corr-matrix",""), &
   PAR_EST_DESC=tyParEstStr("scalar summaries","mean","median","marginal mode","mode","maximum-posterior summary", &
   "MCMC sample standard deviation","standard deviation of mean","estimated number of independent samples","HPDI lbound", &
   "HPDI ubound","matrix summaries","covariance matrix","correlation matrix","kernel density estimation bandwidth")
  
  type(tyMdlSelStr) :: &
   MDL_SEL_FIL_NM=tyMdlSelStr("evid","bayes-fac")

  real(dp),parameter :: &
   STD_HPDI_PROB_CONT(N_STD_HPDI)=[.5,.6827,.95,.9545,.99,.9973], &
   PROB_CONT_CONTOUR(N_CONTOURS)=[.5,.6827,.95], &
   M_MIN_REL_UNC=1e-6, &
   PDGM_ACCUR=1e-6_dp
  
  
  ! MODEL SELECTION
  
  integer,parameter :: &
   MDL_SEL_N_VT=5, &
   MDL_SEL_VT_N_PTS(MDL_SEL_N_VT)=[4,8,16,32,64], & ! [421]
   MDL_SEL_VT_N_SUBSAMP_MAX=100
  
  real,parameter :: &
   MDL_SEL_VT_USE_FRAC=.9, &
   MDL_SEL_VT_SUBSAMP_DEF_FRAC_MIN=.1, &
   MDL_SEL_VT_SUBSAMP_N_SAMP_FRAC=.01


  ! PLOTTING

  integer,parameter :: &
   PLT_UNC_ELL_N_PTS=30, &
   PLT_N_LBLS=5, &
   N_SAMP_TRIANG_KER_MAPO_MIN=100000
  
  real,parameter :: &
   AM_HIP_N_AS_ANG_POS_SEMIMAJ_RATIO=.1

  real(dp),parameter :: &
   TM_RGE_STRETCH_FACT=1.1_dp

  character(*),parameter :: &
   STD_HPDI_DESC(N_STD_HPDI)=[character(7) :: "","1 sigma","","2 sigma","","3 sigma"], &
   PAR_EST_SUMM_DESC(4)=[character(13) :: "mean","median","marginal mode","mode"], &
   REL_DIR_PLT_CSV="../", &
   SUFFX_PS=".ps", &
   SUFFX_TEX=".tex", &
   SUFFX_GNUPLT=".gnuplot", &
   GNUPLT_STD_CMDS_ONCE_NORM(1)=["set terminal postscript color"], &
   GNUPLT_STD_CMDS_ONCE_PSTRICKS(1)=["set terminal pstricks"], &
   GNUPLT_STD_CMDS_ONCE_ALL(1)=["unset key"], &
   GNUPLT_STD_CMDS(3)=[character(13) :: "set autoscale","set xtic auto","set ytic auto"], &
   GNUPLT_CMDS_SPLOT(4)=[character(32) :: "set palette rgbformulae 21,22,23","set pm3d map","unset ztic","set contour base"], &
   GNUPLT_CMDS_ONCE_SPLOT(2)=[character(12) :: "set xtic out","set ytic out"], & ! "set terminal postscript color"
   GNUPLT_CMDS_POST_SPLOT(2)=[character(11) :: "set xtic in","set ytic in"], & ! GNUPLT_STD_CMDS_ONCE_NORM(1)
   PLT_SYM_NMS(4)=[character(9) :: "lines","points","dots","errorbars"], &
   TEX_CMDS_PRE_ONCE(3)=[character(24) :: "\documentclass{scrreprt}","\usepackage{pst-pdf}","\begin{document}"], &
   TEX_CMDS_POST_ONCE(1)=["\end{document}"], &
   TEX_CMDS_PRE(1)=["  \begin{postscript}"], &
   TEX_CMDS_POST(1)=["  \end{postscript}"], &
   FIL_NM_PART_SAV_SAMP="postr-samp"

  type(tyTtlPartsStr),parameter :: &
   TTL_PART=tyTtlPartsStr("data","model","distribution","normalised residual","normalised residuals", &
   "considering additional jitter","ignoring additional jitter", &
   "joint marginal posterior","marginal posterior","MCMC sample number","MCMC trace","model-prediction uncertainty", &
   "semi-major and -minor axes of uncertainty ellipses","position angles of semi-major axes of uncertainty ellipses","prior", &
   "probability density","PSR","residuals","scatter plot","negative logarithm of (posterior times evidence)", &
   "autocorrelation","lag")

  type(tyPltResolInt8),parameter :: &
   PLT_N_PTS=tyPltResolInt8(500_i8b,100_i8b), &
   PLT_N_PTS_MDL=tyPltResolInt8(10000_i8b,1000_i8b), & ! CAUTION: must be even values!
   PLT_N_PTS_PREDICT=tyPltResolInt8(1000_i8b,100_i8b), &
   PLT_N_PTS_SCATTER=tyPltResolInt8(30000_i8b,1000_i8b), & ! # points for scatter plots; standard should lead to CSV file <~ 55 MB for 11 pars., cf. [22.05.2010/1]
   PLT_N_PTS_KDE=tyPltResolInt8(100_i8b,50_i8b)   ! # points for KDE plots

  type(tyPltSymInt),parameter :: &
   PLT_SYM=tyPltSymInt(1,2,3,4)


  !========================================================[ VARIABLES ]============================================================

  integer :: &
   pltNPts, &
   pltNPtsMdl, &
   pltNPtsScatter, &
   pltNPtsKDE, &
   nPtsPeriodogram, &
   nUsrHPDI, &
   nHPDI, &
   iStdHPDI(N_STD_HPDI), &           ! iStdHPDI(j) contains index of HPDIProbCont() for j-th standard HPDI
   HPDIRole(N_STD_HPDI+N_USR_HPDI_MAX), &
   iMaPoKer, &
   mdlSelVTNSub

  integer(i8b) :: &
   mdlSelVTNUsed, &
   mdlSelVTSubNUsed

  real :: &
   mdlSelVTUseFrac

  real(dp) :: &
   logPostrEv, &                      ! ln(posterior×evidence)=ln(prior×likelihood), at posterior mode
   prob1stTransIsPri, &
   maPoShft(N_L_PARS_MAX), &
   HPDIProbCont(N_STD_HPDI+N_USR_HPDI_MAX)

  real(qp) :: &
   mdlSelEvVT(N_N_PL_MODES_MAX,MDL_SEL_N_VT), &
   mdlSelBFVT(2:N_N_PL_MODES_MAX,1:N_N_PL_MODES_MAX-1,MDL_SEL_N_VT), &
   logPostrEvConst, &
   calcdPostrProb(N_N_PL_MODES_MAX,N_CALC_PROB_MAX), &
   calcdPostrProbAllPars(N_N_PL_MODES_MAX,N_CALC_PROB_MAX)

  logical :: &
   onlyPlotData, &
   execGnuplt, &
   execPS4PDF, &
   singlePlts, &
   usrPred, &
   usrNoInfc, &
   usrPeriodogram, &
   usrJointMaPo(N_N_PL_MODES_MAX), &
   usrMaPo(N_N_PL_MODES_MAX), &
   usrScatterPlts(N_N_PL_MODES_MAX), &
   usrZmRge(N_N_PL_MODES_MAX), &
   cycParMaPoShifted(N_L_PARS_MAX)

  character(MAX_LEN_PATH) :: &
   filNmGnupltCmds, &
   filNmGnupltCmdsNoDirs

  character(4) :: &
   pltFilExt

  type(tyPostrSumm) :: &
   postrSumm(N_N_PL_MODES_MAX)

  type(tyPred1D) :: &
   pred1D

  type(tyPred2D) :: &
   pred2D

  type(tyIntvlDbl) :: &
   DQSampRge(N_L_DQ_MAX)             ! the DQ ranges for the ("zoomed-in") DQ marginal posteriors

  type(tyCoord2DDbl) :: &
   maPoDQPts(N_L_DQ_MAX,PLT_N_PTS%std), &
   maPoDQPtsZm(N_L_DQ_MAX,PLT_N_PTS%std), &
   maPoPts(N_L_PARS_MAX,PLT_N_PTS%std), &
   maPoPtsZm(N_L_PARS_MAX,PLT_N_PTS%std), &
   periodogramPts(:), &
   periodogramPtsLog(:)

  type(tyParEst) :: &
   parEst(N_N_PL_MODES_MAX,N_L_PARS_MAX), &
   parEstDQ(N_N_PL_MODES_MAX,N_L_DQ_MAX)
  
  type(tyUncQtyQP) :: &
   mdlSelEvSubVT(N_N_PL_MODES_MAX,MDL_SEL_N_VT), &
   mdlSelBFSubVT(2:N_N_PL_MODES_MAX,1:N_N_PL_MODES_MAX-1,MDL_SEL_N_VT), &
   mdlSelEv(N_N_PL_MODES_MAX), &
   mdlSelBF(2:N_N_PL_MODES_MAX,1:N_N_PL_MODES_MAX-1)

  allocatable :: &
   periodogramPts, &
   periodogramPtsLog
  
  procedure(dens1DUni),pointer :: &
   maPoKer


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  interface calcHDIs
    module procedure calcHDIsDbl
    module procedure calcHDIsQP
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !===========================================[ PARAMETER AND UNCERTAINTY ESTIMATION ]==============================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine chkPostrMode(chain)
!
!   checks whether posterior mode is defined, else warns/casts an error
!
!   Arguments:
!     chain                          MCMC chain

!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyMCMCChain) :: &
     chain
    intent(inout) :: &
     chain

    if (any(isUndef(parEst(iNPlMd,:curNPars)%mode))) then
      if (usrPeriodogram) then
        call err("posterior mode, indispensable to calculate periodogram, could not be found"// &
         trim(mrg(mrg(" because no data are given","; please try again with more samples",dtProps%nRec == 0), &
         " because prior is unknown",priAvail)))
      else
        call warn("posterior mode could not be found")
      end if
    else
      logPostrEv=logPrior(parEst(iNPlMd,:curNPars)%mode)+logLikh(parEst(iNPlMd,:curNPars)%mode)
    end if


  end subroutine chkPostrMode




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine findBestEst
!
!   determines best summary of pars.:
!     * mode, if defined
!     * otherwise that combination of either medians or marginal modes (or means where marg. modes undefined) which maximises posterior
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p, &
     iC, &
     curSummTy(curNPars), &
     bestSummTy(curNPars), &
     i
    real(dp) :: &
     curPars(curNPars), &
     bestLogPostr, &
     curLogPostr

    if (all(isDef(parEst(iNPlMd,:curNPars)%mode))) then

      call undef(parEst(iNPlMd,:curNPars)%bestSummTy) ! "best summary" only defined if any comparison has been carried out
      parEst(iNPlMd,:curNPars)%bestSumm=parEst(iNPlMd,:curNPars)%mode

    else

      where (parFixed(:curNPars)) curSummTy=VAL_UNDEF_INT

      call undef(bestSummTy)
      bestLogPostr=-huge(bestLogPostr)

      do iC=0,2**curNPars-1        ! for all combinations of par. summary types

        ! set summary types for all pars.

        do p=1,curNPars
          if (.NOT. parFixed(p)) curSummTy(p)=merge(PAR_EST_SUMM_TY%median, &
           merge(PAR_EST_SUMM_TY%mMode,PAR_EST_SUMM_TY%mean,isDef(parEst(iNPlMd,p)%mMode)),ibits(iC,p-1,1) == 0)
          ! median is associated with bit == 0, mode (or mean, if mode undef.) with bit == 1
        end do


        ! set cur. pars.

        where (isUndef(curSummTy))
          curPars=priLBnd(:curNPars) ! fixed pars.
        elsewhere (curSummTy == PAR_EST_SUMM_TY%median)
          curPars=parEst(iNPlMd,:curNPars)%median
        elsewhere (isUndef(parEst(iNPlMd,:curNPars)%mMode))
          curPars=parEst(iNPlMd,:curNPars)%mean%val
        elsewhere
          curPars=parEst(iNPlMd,:curNPars)%mMode%val
        end where


        ! calc. posterior

        curLogPostr=logLikh(curPars)+logPrior(curPars)

        if (curLogPostr > bestLogPostr) then
          bestSummTy=curSummTy
          bestLogPostr=curLogPostr
        end if

      end do                       ! iC=0,2**curNPars


      ! save the info globally

      where (isUndef(bestSummTy))
        parEst(iNPlMd,:curNPars)%bestSummTy=VAL_UNDEF_INT
        parEst(iNPlMd,:curNPars)%bestSumm=parEst(iNPlMd,:curNPars)%median
      elsewhere
        parEst(iNPlMd,:curNPars)%bestSummTy=bestSummTy
        where (bestSummTy == PAR_EST_SUMM_TY%median)
          parEst(iNPlMd,:curNPars)%bestSumm=parEst(iNPlMd,:curNPars)%median
        elsewhere (bestSummTy == PAR_EST_SUMM_TY%mMode)
          parEst(iNPlMd,:curNPars)%bestSumm=parEst(iNPlMd,:curNPars)%mMode%val
        elsewhere (bestSummTy == PAR_EST_SUMM_TY%mean)
          parEst(iNPlMd,:curNPars)%bestSumm=parEst(iNPlMd,:curNPars)%mean%val
        end where
      end where

    end if                         ! (all(isDef(parEst(iNPlMd,:curNPars)%mode)))

  end subroutine findBestEst




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcQuantMaPoSort(chain)
!
!   calculates quantiles, marg. postrs. and residuals, sorting the post-burn-in chain links
!
!   Arguments:
!     chain                          MCMC chain
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p, &
     allocStat
    integer(i8b) :: &
     l, &
     lB, &
     uBTmp, &
     uBShift, &
     nSh, &
     nOri
    real(dp) :: &
     tmpArr(:)
    type(tyMCMCChain) :: &
     chain
    allocatable :: &
     tmpArr
    intent(inout) :: &
     chain


    call startSubsect(UNIT%cons+UNIT%log,"Calculating quantiles"//trim(mrg(" and marginal posteriors","",usrMaPo(iNPlMd))))
    l=chain%length
    lB=chain%burnInLen

    call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
    !$OMP parallel do schedule(dynamic) private(uBShift,tmpArr,nSh,nOri,uBTmp)
    do p=1,curNPars
      if (parInfc(p)) then
        call calcMedHPDI(chain%links(lB+1:l,p),chain%sorted(p),parEst(iNPlMd,p)%HPDI,parEst(iNPlMd,p)%nInHPDI, &
         parEst(iNPlMd,p)%median) ! also sorts

        cycParMaPoShifted(p)=(cycParUsMaxRge(p) .AND. &
         xor(parEst(iNPlMd,p)%HPDI(1)%lBnd == parEst(iNPlMd,p)%HPDI(2)%lBnd, &
         parEst(iNPlMd,p)%HPDI(1)%uBnd == parEst(iNPlMd,p)%HPDI(2)%uBnd))

        if (cycParMaPoShifted(p)) then
          maPoShft(p)=parPeriod(p)/2._dp
          uBTmp=posInSortedArrDbl(chain%links(lB+1:l,p),parPeriod(p)/2._dp,.FALSE.)
          if (isDef(uBTmp)) then
            uBShift=lB+uBTmp-1_i8b
          else
            uBShift=l
          end if
          nSh=uBShift-lB             ! # shifted samples
          nOri=l-uBShift             ! # unshifted samples

          !$OMP critical (OMPCrit2)
          ! … because of RAM
          if (nSh <= floor(.5*(l-lB))) then ! save shifted samples
            allocate(tmpArr(nSh),stat=allocStat)
            call testAllocErr(allocStat,47)
            tmpArr=chain%links(lB+1:uBShift,p)+parPeriod(p) ! save shifted samples
            chain%links(lB+1:lB+nOri,p)=chain%links(uBShift+1:l,p) ! move original samples (>= period/2) to the left
            chain%links(lB+nOri+1:l,p)=tmpArr ! insert shifted samples on the right
          else                       ! save original samples
            allocate(tmpArr(nOri),stat=allocStat)
            call testAllocErr(allocStat,48)
            tmpArr=chain%links(uBShift+1:l,p) ! save original samples
            chain%links(l-(nSh-1):l,p)=chain%links(lB+1:uBShift,p)+parPeriod(p) ! move shifted samples to the right
            chain%links(lB+1:lB+nOri,p)=tmpArr ! insert original samples on the left
          end if

          deallocate(tmpArr)
          !$OMP end critical (OMPCrit2)

          call calcMedHPDI(chain%links(lB+1:l,p),chain%sorted(p),parEst(iNPlMd,p)%HPDI,parEst(iNPlMd,p)%nInHPDI, &
           parEst(iNPlMd,p)%median) ! redo

        end if

        if (filOutputPlt) then
          call calcMargPostr(chain,p, &
           pltRge(shiftIntvl(priRge(p),maPoShft(p)),maPoLogScl(p),priTypes(p)), &
           parEst(iNPlMd,p)%HPDI(iStdHPDI(I_STD_HPDI_MA_PO_ZM)),parEst(iNPlMd,p),maPoPts(p,:pltNPts), &
           maPoPtsZm(p,:pltNPts),maPoLogScl(p))
        end if

        if (cycParMaPoShifted(p)) then
          !$OMP critical (OMPCrit3)
          if (nSh <= floor(.5*(l-lB))) then ! save shifted samples
            allocate(tmpArr(nSh),stat=allocStat)
            call testAllocErr(allocStat,49)
            tmpArr=chain%links(lB+nOri+1:l,p)-parPeriod(p) ! save shifted-back samples
            chain%links(uBShift+1:l,p)=chain%links(lB+1:lB+nOri,p) ! move original samples (>= period/2) to the right
            chain%links(lB+1:uBShift,p)=tmpArr ! insert shifted-back samples on the left
          else                       ! save original samples
            allocate(tmpArr(nOri),stat=allocStat)
            call testAllocErr(allocStat,50)
            tmpArr=chain%links(lB+1:lB+nOri,p) ! save original samples
            chain%links(lB+1:uBShift,p)=chain%links(l-(nSh-1):l,p)-parPeriod(p) ! move shifted-back samples to the left
            chain%links(uBShift+1:l,p)=tmpArr ! insert original samples on the right
          end if
          deallocate(tmpArr)
          !$OMP end critical (OMPCrit3)
        end if

        !$OMP critical (OMPCrit4)
        call printPrgrs([VAL_UNDEF_INT],nTot=curNParsInfc)
        !$OMP end critical (OMPCrit4)

      else
        call undef(parEst(iNPlMd,p)%HPDI)
        call undef(parEst(iNPlMd,p)%nInHPDI)
        call undef(parEst(iNPlMd,p)%median)
      end if                         ! (parInfc(p))

    end do
    !$OMP end parallel do

    call endSubsect(UNIT%cons+UNIT%log) ! Calculating quantiles (and marginal posteriors)

  end subroutine calcQuantMaPoSort




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcMedHPDI(arr,sorted,HPDI,nInHPDI,med,sampRge,iHPDI)
!
!   calculates highest-posterior-density intervals (HPDIs), median, and (optionally) range of posterior samples
!
!   Arguments:
!     arr                            array of samples
!     sorted                         whether array is already sorted
!     HPDI                           HPDIs
!     nInHPDI                        # samples in each HPDI (corresponding to the HPDI's prob. content)
!     med                            median
!     sampRge                        sample range
!     iHPDI                          index of single HPDI to be calcd.
!
!   sorts array if unsorted
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer :: &
     nLinks, &
     iHPDI
    integer(i8b) :: &
     nInHPDI(nHPDI)
    real(dp) :: &
     arr(:), &
     med, &
     lBnd, &
     uBnd
    logical :: &
     sorted
    type(tyIntvlDbl) :: &
     HPDI(:), &
     sampRge
    optional :: &
     iHPDI, &
     sampRge
    intent(in) :: &
     iHPDI
    intent(inout) :: &
     sorted, &
     arr
    intent(out) :: &
     HPDI, &
     nInHPDI, &
     med, &
     sampRge

    nLinks=size(arr,1)
    nInHPDI=floor(HPDIProbCont(1:nHPDI)*nLinks)

    ! sort

    if (.NOT. sorted) then
      call sort(arr(1:nLinks))
      sorted=.TRUE.
    end if


    ! extremal values of arr

    lBnd=arr(1)
    uBnd=arr(nLinks)

    if (present(sampRge)) then
      sampRge%lBnd=lBnd
      sampRge%uBnd=uBnd
    end if


    ! HPD intervals

    if (present(iHPDI)) then
      call calcHDIs(arr,nInHPDI(iHPDI:iHPDI),HPDI(1:1))
    else
      call calcHDIs(arr,nInHPDI(1:nHPDI),HPDI(1:nHPDI))
    end if


    ! median

    med=median(arr,setFlg(FLG%sorted))

  end subroutine calcMedHPDI




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcHDIsDbl(arr,nInHDI,HDI)
!
!   calculates highest-density intervals (HDIs) based on samples
!
!   Arguments:
!     arr                            array of samples
!     nInHDI                         # samples in each HDI
!     HDI                            highest-density intervals
!
!   Requirements:
!     - arr must be sorted
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i, &
     j, &
     nHDIs, &
     nVal, &
     jMin, &
     jMax, &
     jTmp, &
     curN
    integer(i8b) :: &
     nInHDI(:)
    real(dp) :: &
     arr(:), &
     tmpW
    type(tyIntvlDbl) :: &
     HDI(size(nInHDI,1))
    intent(in) :: &
     arr, &
     nInHDI
    intent(out) :: &
     HDI

    nHDIs=size(nInHDI,1)
    nVal=size(arr,1)

    jMin=1
    jMax=nVal
    jTmp=jMin

    do i=nHDIs,1,-1

      curN=nInHDI(i)
      tmpW=huge(tmpW)

      ! loop over values
      do j=jMin,jMax-(curN-1)
        if (arr(j+curN-1)-arr(j) < tmpW) then
          tmpW=arr(j+curN-1)-arr(j)
          jTmp=j
        end if
      end do
      jMin=jTmp
      jMax=jTmp+curN-1
      HDI(i)%lBnd=arr(jMin)
      HDI(i)%uBnd=arr(jMax)

    end do

  end subroutine calcHDIsDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcHDIsQP(arr,nInHDI,HDI)
!
!   calculates highest-density intervals (HDIs) based on samples
!
!   Arguments:
!     arr                            array of samples
!     nInHDI                         # samples in each HDI
!     HDI                            highest-density intervals
!
!   Requirements:
!     - arr must be sorted
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i, &
     j, &
     nHDIs, &
     nVal, &
     jMin, &
     jMax, &
     jTmp, &
     curN
    integer(i8b) :: &
     nInHDI(:)
    real(qp) :: &
     arr(:), &
     tmpW
    type(tyIntvlQP) :: &
     HDI(size(nInHDI,1))
    intent(in) :: &
     arr, &
     nInHDI
    intent(out) :: &
     HDI

    nHDIs=size(nInHDI,1)
    nVal=size(arr,1)

    jMin=1
    jMax=nVal
    jTmp=jMin

    do i=nHDIs,1,-1

      curN=nInHDI(i)
      tmpW=huge(tmpW)

      ! loop over values
      do j=jMin,jMax-(curN-1)
        if (arr(j+curN-1)-arr(j) < tmpW) then
          tmpW=arr(j+curN-1)-arr(j)
          jTmp=j
        end if
      end do
      jMin=jTmp
      jMax=jTmp+curN-1
      HDI(i)%lBnd=arr(jMin)
      HDI(i)%uBnd=arr(jMax)

    end do

  end subroutine calcHDIsQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcProb(chain)
!
!   calculates probability over given intervals
!
!   Arguments:
!     chain                          MCMC chain
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i, &
     p
    integer(i8b) :: &
     j, &
     k
    real(dp) :: &
     mi, &
     ma
    type(tyMCMCChain) :: &
     chain
    intent(in) :: &
     chain
    
    do i=1,tmpNCalcProb(iNPlMd)
      mi=tmpCalcProbRge(iNPlMd,i)%lBnd
      ma=tmpCalcProbRge(iNPlMd,i)%uBnd
      p=posInArr(parNms(:curNPars),trim(tmpCalcProbNms(iNPlMd,i)))
      if (isUndef(p)) then
        call warn("parameter "//trim(tmpCalcProbNms(iNPlMd,i))// &
         " for calculating posterior probability is unknown; skipping calculation")
        call undef(calcdPostrProb(iNPlMd,i))
      else
        j=minloc(abs(mi-chain%links(chain%burnInLen+1:chain%length,p)),1)+chain%burnInLen
        k=minloc(abs(ma-chain%links(chain%burnInLen+1:chain%length,p)),1)+chain%burnInLen
        do
          if (chain%links(j,p) >= mi .OR. j >= k) exit
          call inc(j)
        end do
        do
          if (chain%links(k,p) <= ma .OR. j >= k) exit
          call dec(k)
        end do
        calcdPostrProb(iNPlMd,i)=real(max(0,k-j+1),qp)/(chain%length-chain%burnInLen)
      end if
    end do

  end subroutine calcProb




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcMargPostr(chain,q,xRge,xRgeZm,parEstQty,pts,ptsZm,logScl)
!
!   calculates (zoomed) marginal posteriors
!
!   Arguments:
!     chain                          MCMC chain
!     q                              index of quantity in question
!     xRge                           x range for KDE
!     xRgeZm                         zoomed x range for KDE
!     parEstQty                      parameter-estimation summaries of quantity
!     pts                            marg. postr. points
!     ptsZm                          zoomed marg. postr. points
!     logScl                         whether abscissa should be in logarithmic scales
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     q, &
     allocStat, &
     iMax, &
     iA, &
     iC
    real(dp) :: &
     modeStdDev, &
     ax, &
     bx, &
     cx, &
     valMax                          ! unused
    logical :: &
     logScl
    type(tyMCMCChain) :: &
     chain
    type(tyIntvlDbl) :: &
     xRge, &
     xRgeZm
    type(tyCoord2DDbl) :: &
     pts(pltNPts), &
     ptsZm(pltNPts)
    type(tyParEst) :: &
     parEstQty
    intent(in) :: &
     q, &
     xRge, &
     xRgeZm, &
     logScl
    intent(inout) :: &
     chain, &
     parEstQty
    intent(out) :: &
     pts, &
     ptsZm


    if (usrPeriodogram) then         ! calc. marginal posterior of f

      call startSubsect(UNIT%cons+UNIT%log,"Periodogram ("//trim(TTL_PART%maPo)//" of frequency)")

      call startSubsect(UNIT%cons+UNIT%log,"Preliminary "//TTL_PART%maPo)
      call kerDensEst1D(dens1DTriang,I_KER_1D%triang,chain%links(chain%burnInLen+1:chain%length,q), &
       min(1._dp/dtProps%tRge%width(),parEstQty%sampStdDev),xRge,int(pltNPts,i8b),pts, &
       setFlg([FLG%sorted,FLG%logScl],[chain%sorted(q),logScl])) ! also sorts
      chain%sorted(q)=.TRUE.

      call printMsg(UNIT%cons+UNIT%log,"Done")
      call endSubsect(UNIT%cons+UNIT%log) ! Preliminary marginal posterior

      call startSubsect(UNIT%cons+UNIT%log,"Estimating kernel width for density estimation")
      modeStdDev=estimFrqModeStdDev(parEst(iNPlMd,:curNPars)%mode,parEstQty%sampStdDev,chain,parEstQty,pts)
      call printMsg(UNIT%cons+UNIT%log,"Done")
      call endSubsect(UNIT%cons+UNIT%log) ! Estimating kernel width for density estimation

      nPtsPeriodogram=nint(priW(pRole%f)/modeStdDev*N_PTS_PERIODOGRAM_PER_STDEV)
      pltNPts=nPtsPeriodogram
      allocate(periodogramPts(nPtsPeriodogram),periodogramPtsLog(nPtsPeriodogram),stat=allocStat)
      call testAllocErr(allocStat,46)

      call startSubsect(UNIT%cons+UNIT%log,"Final "//TTL_PART%maPo)
      call startSubsect(UNIT%cons+UNIT%log,"Linear abscissa")
      call kerDensEst1D(maPoKer,iMaPoKer,chain%links(chain%burnInLen+1:chain%length,q),modeStdDev,xRge, &
       int(nPtsPeriodogram,i8b),periodogramPts,setFlg(FLG%sorted,chain%sorted(q)),bandw=parEstQty%KDEBandw)
      chain%sorted(q)=.TRUE.
      call printMsg(UNIT%cons+UNIT%log,"Done")
      call endSubsect(UNIT%cons+UNIT%log)

      call startSubsect(UNIT%cons+UNIT%log,"Logarithmic abscissa")
      call kerDensEst1D(maPoKer,iMaPoKer,chain%links(chain%burnInLen+1:chain%length,q),modeStdDev,xRge, &
       int(nPtsPeriodogram,i8b),periodogramPtsLog,setFlg([FLG%sorted,FLG%logScl]),bandw=parEstQty%KDEBandw)
      call printMsg(UNIT%cons+UNIT%log,"Done")
      call endSubsect(UNIT%cons+UNIT%log)
      call endSubsect(UNIT%cons+UNIT%log) ! Final marginal posterior

      call endSubsect(UNIT%cons+UNIT%log) ! Periodogram (marginal posterior of frequency)


      ! marginal mode

      if (.NOT. chain%sorted(pRole%f)) then
        call sort(chain%links(chain%burnInLen+1:chain%length,pRole%f))
        chain%sorted(pRole%f)=.TRUE.
      end if

      call startSubsect(UNIT%cons+UNIT%log,"Refining marginal mode")
      iMax=maxloc(periodogramPts%y,1)
      parEstQty%mMode%val=periodogramPts(iMax)%x

      if (iMax == 1) then
        iA=1
        bx=periodogramPts(2)%x
        iC=min(nPtsPeriodogram,iA+N_PTS_PERIODOGRAM_PER_STDEV)
      else if (iMax == nPtsPeriodogram) then
        iC=nPtsPeriodogram
        bx=periodogramPts(nPtsPeriodogram-1)%x
        iA=max(1,iC-N_PTS_PERIODOGRAM_PER_STDEV)
      else
        iA=max(1,iMax-N_PTS_PERIODOGRAM_PER_STDEV)
        bx=parEstQty%mMode%val
        iC=min(nPtsPeriodogram,iMax+N_PTS_PERIODOGRAM_PER_STDEV)
      end if
      ax=periodogramPts(iA)%x
      cx=periodogramPts(iC)%x

      parEstQty%mMode%val=posMin(ax,bx,cx,PDGM_ACCUR,-1._dp,estMapoF,HUGE_DP,valMax,-HUGE_DP)
      parEstQty%mMode%unc=PDGM_ACCUR*parEstQty%mMode%val
      call printMsg(UNIT%cons+UNIT%log,"Done")
      call endSubsect(UNIT%cons+UNIT%log) ! Refining marginal mode

    else
      ! normal
      call kerDensEst1D(maPoKer,iMaPoKer,chain%links(chain%burnInLen+1:chain%length,q), &
       parEstQty%sampStdDev,xRge,int(pltNPts,i8b),pts,setFlg([FLG%sorted,FLG%logScl],[chain%sorted(q),logScl]), &
       bandw=parEstQty%KDEBandw)     ! also sorts
      chain%sorted(q)=.TRUE.

      ! zoomed
      call kerDensEst1D(maPoKer,iMaPoKer,chain%links(chain%burnInLen+1:chain%length,q),parEstQty%sampStdDev, &
       xRgeZm,int(pltNPts,i8b),ptsZm,setFlg([FLG%sorted,FLG%logScl],[chain%sorted(q),logScl]),xRge,bandw=parEstQty%KDEBandw)

      ! marginal mode
      parEstQty%mMode%val=ptsZm(maxloc(ptsZm%y,1))%x
      parEstQty%mMode%unc=.5_dp*(ptsZm(2)%x-ptsZm(1)%x)

    end if


    contains

    function estMapoF(f)             ! returns estimated marg. postr. of f
      implicit none
      real(dp) :: &
       estMapoF, &
       f
      type(tyCoord2DDbl) :: &
       pts(1)
      intent(in) :: &
       f

      if (parOutsdPriSupp(pRole%f,f)) then
        estMapoF=0
      else
        call kerDensEst1D(dens1DTriang,I_KER_1D%triang,chain%links(chain%burnInLen+1:chain%length,pRole%f), &
         modeStdDev,tyIntvlDbl(f,1.1_dp*f),1_i8b,pts,setFlg(FLG%sorted))
        estMapoF=pts(1)%y
      end if
    end function

  end subroutine calcMargPostr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function estimFrqModeStdDev(modePars,stdDev,chain,parEst,mapoPts)
!
!   Returns:                         estimated standard deviation of marginal mode of orbital frequency
!
!   Arguments:
!     modePars                       MAP parameter vector
!     stdDev                         initial standard deviation
!     chain                          MCMC chain
!     parEst                         parameter-estimation summaries
!     mapoPts                        initial marg.-postr. points
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     N_PTS_MAPO_VAR=50
    real(dp),parameter :: &
     NEAR_FRAC=1e-4_dp, &
     MIN_MAX_RATIO_THRLD=.5_dp
    integer :: &
     nMapoPts, &
     i, &
     iMapoMin, &
     iMapoMax
    real(dp) :: &
     modePars(curNPars), &
     estimFrqModeStdDev, &
     nearDist, &
     fMargMin, &
     fMargMax, &
     ax, &
     bx, &
     cx, &
     stdDev, &
     lstV, &
     curV, &
     minDist, &
     curDist, &
     mapoVarPtsX(N_PTS_MAPO_VAR), &
     mapoVarPtsY(N_PTS_MAPO_VAR), &
     fMargMinMaxDiff, &
     valMax, &
     valMin, &
     halfInvTRge
    logical :: &
     incr
    type(tyMCMCChain) :: &
     chain
    type(tyCoord2DDbl) :: &
     mapoPts(:)
    type(tyParEst) :: &
     parEst
    intent(in) :: &
     modePars, &
     stdDev, &
     mapoPts
    intent(inout) :: &
     chain, &
     parEst

    nMapoPts=size(mapoPts,1)
    estimFrqModeStdDev=stdDev
    halfInvTRge=1._dp/(2._dp*dtProps%tRge%width())


    ! step 1: find marginal max./min. near mode (using 1st prelim. mapo.)
    call startSubsect(UNIT%cons+UNIT%log,"Searching for marginal maximum and minimum near mode")

      ! find marg. max. nearest to mode

    lstV=mapoPts(1)%y
    minDist=huge(minDist)
    incr=.FALSE.
    do i=2,nMapoPts
      curV=mapoPts(i)%y
      if (curV >= lstV) then         ! value increasing
        incr=.TRUE.
      elseif (incr) then             ! local max. passed
        curDist=abs(modePars(pRole%f)-mapoPts(i-1)%x)
        if (curDist < minDist) then
          iMapoMax=i-1
          minDist=curDist
        end if
        incr=.FALSE.
      end if
      lstV=curV
    end do
    fMargMax=mapoPts(iMapoMax)%x

      ! find marg. min. nearest to marg. max.

    lstV=mapoPts(1)%y
    minDist=huge(minDist)
    incr=.TRUE.
    do i=2,nMapoPts
      curV=mapoPts(i)%y
      if (curV < lstV) then          ! value decreasing
        incr=.FALSE.
      elseif (.NOT. incr) then       ! local min. passed
        curDist=abs(fMargMax-mapoPts(i-1)%x)
        if (curDist < minDist) then
          iMapoMin=i-1
          minDist=curDist
        end if
        incr=.TRUE.
      end if
      lstV=curV
    end do
    fMargMin=mapoPts(iMapoMin)%x
    fMargMinMaxDiff=fMargMin-fMargMax


    call printMsg(UNIT%cons+UNIT%log,"Done")
    call endSubsect(UNIT%cons+UNIT%log) ! Searching for marginal maximum and minimum near mode


    ! step 2: calculate standard deviation over relevant marginal peak (used for calc. of subsequent marg. postr.)

    call sort(chain%links(chain%burnInLen+1:chain%length,pRole%f))
    chain%sorted(pRole%f)=.TRUE.
    
    call startSubsect(UNIT%cons+UNIT%log,"Sampling marginal peak (step 1/4)")
    call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
    !$OMP parallel do
    do i=1,N_PTS_MAPO_VAR
      mapoVarPtsX(i)=fMargMin+(fMargMax-fMargMin)*(i-1)/(N_PTS_MAPO_VAR-1)
      mapoVarPtsY(i)=estMapoF(mapoVarPtsX(i))
      !$OMP critical
      call printPrgrs([VAL_UNDEF_INT],nTot=N_PTS_MAPO_VAR)
      !$OMP end critical
    end do
    !$OMP end parallel do
    call printPrgrs([1],[1],[1])
    call endSubsect(UNIT%cons+UNIT%log) ! Sampling marginal peak
    estimFrqModeStdDev=min(stdDev,sqrt(2._dp*arrVar(mapoVarPtsX,weights=mapoVarPtsY)))
    if (halfInvTRge < estimFrqModeStdDev) estimFrqModeStdDev=sqrt(estimFrqModeStdDev*halfInvTRge)


    ! step 3: find marginal max./min. near mode (using 2nd prelim. mapo.)

    nearDist=NEAR_FRAC*min(halfInvTRge,abs(fMargMinMaxDiff))

      ! find marg. max. nearest to mode

    call startSubsect(UNIT%cons+UNIT%log,"Refining local marginal maximum (step 2/4)")
    ax=modePars(pRole%f)
    bx=ax+nearDist
    if (parOutsdPriSupp(pRole%f,bx)) then
      bx=ax-nearDist
      if (parOutsdPriSupp(pRole%f,bx)) bx=merge(priLBnd(pRole%f),priUBnd(pRole%f),ax <= priCtr(pRole%f))
    end if
    call brktMin(ax,bx,cx,-1._dp,estMapoF,HUGE_DP,-HUGE_DP)
    fMargMax=posMin(ax,bx,cx,PDGM_ACCUR,-1._dp,estMapoF,HUGE_DP,valMax,-HUGE_DP)
    valMax=-valMax
    call printMsg(UNIT%cons+UNIT%log,"Done")
    call endSubsect(UNIT%cons+UNIT%log) ! Refining local marginal maximum

      ! find marg. min. nearest to marg. max.

    call startSubsect(UNIT%cons+UNIT%log,"Refining local marginal minimum (step 3/4)")
    ax=fMargMax
    bx=ax+sign(nearDist,fMargMinMaxDiff)
    if (parOutsdPriSupp(pRole%f,bx)) then
      bx=ax-sign(nearDist,fMargMinMaxDiff)
      if (parOutsdPriSupp(pRole%f,bx)) bx=merge(priLBnd(pRole%f),priUBnd(pRole%f),ax <= priCtr(pRole%f))
    end if
    call brktMin(ax,bx,cx,1._dp,estMapoF,MIN_MAX_RATIO_THRLD*valMax,EPS_DP)
    fMargMin=posMin(ax,bx,cx,PDGM_ACCUR,1._dp,estMapoF,MIN_MAX_RATIO_THRLD*valMax,valMin,EPS_DP)
    call printMsg(UNIT%cons+UNIT%log,"Done")
    call endSubsect(UNIT%cons+UNIT%log) ! Refining local marginal minimum


    ! step 4: re-calculate standard deviation over relevant marginal peak between min. and max.

    call startSubsect(UNIT%cons+UNIT%log,"Sampling marginal peak (step 4/4)")
    call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
    !$OMP parallel do
    do i=1,N_PTS_MAPO_VAR
      mapoVarPtsX(i)=fMargMin+(fMargMax-fMargMin)*(i-1)/(N_PTS_MAPO_VAR-1)
      mapoVarPtsY(i)=estMapoF(mapoVarPtsX(i))
      !$OMP critical
      call printPrgrs([VAL_UNDEF_INT],nTot=N_PTS_MAPO_VAR)
      !$OMP end critical
    end do
    !$OMP end parallel do
    call printPrgrs([1],[1],[1])
    call endSubsect(UNIT%cons+UNIT%log) ! Sampling marginal peak

    estimFrqModeStdDev=min(stdDev,sqrt(2._dp*arrVar(mapoVarPtsX,weights=mapoVarPtsY)))


    contains

    function estMapoF(f)             ! returns estimated marg. postr. of f
      implicit none
      real(dp) :: &
       estMapoF, &
       f
      type(tyCoord2DDbl) :: &
       pts(1)
      intent(in) :: &
       f

      if (parOutsdPriSupp(pRole%f,f)) then
        estMapoF=0
      else
        call kerDensEst1D(dens1DTriang,I_KER_1D%triang,chain%links(chain%burnInLen+1:chain%length,pRole%f), &
         estimFrqModeStdDev,tyIntvlDbl(f,1.1_dp*f),1_i8b,pts,setFlg(FLG%sorted))
        estMapoF=pts(1)%y
      end if
    end function

  end function estimFrqModeStdDev




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtHPDIProbCont(iHPDI)
!
!   Returns:                         formatted string with HPDI probability content
!
!   Arguments:
!     iHPDI                          HPDI index
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iHPDI
    character(HPDI_N_DIGS_MAX+3) :: &
     fmtHPDIProbCont
    intent(in) :: &
     iHPDI

    fmtHPDIProbCont=trim(fmtNr(100._dp*HPDIProbCont(iHPDI), &
     "(F"//trim(niceFmt(HPDI_N_DIGS_MAX+1))//"."//trim(niceFmt(HPDI_N_DIGS_MAX-2))//")",HPDI_N_DIGS_MAX+1,setFlg(FLG%rm0)))//" %"

  end function fmtHPDIProbCont




  !=====================================================[ MODEL SELECTION ]=========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcEvVT(chn,ev,evSub,nLeaves)
!
!   calculates evidence with Volume-Tesselation Algorithm (VTA) [421r]
!
!   Arguments:
!     chn                            MCMC chain
!     ev                             evidence
!     evSub                          evidence from subsampling
!     nLeaves                        # leaves in each considered subvolume
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyMCMCChain) :: &
     chn
    integer :: &
     pFree(curNParsFree), &
     j, &
     k, &
     nLeaves
    integer(i8b) :: &
     i, &
     nTries, &
     nAftrBurn, &
     nUsedSub, &
     N, &
     iMax, &
     iCritUsed, &
     s(chn%burnInLen+1_i8b:chn%length)
    real(qp) :: &
     ev, &
     priLikh(chn%burnInLen+1_i8b:chn%length)
    type(tyUncQtyQP) :: &
     evSub
    class(tyKDTreeNode) :: &
     tree
    pointer :: &
     tree
    intent(inout) :: &
     ev
    intent(in) :: &
     chn, &
     nLeaves
    

    N=chn%length
    nAftrBurn=N-chn%burnInLen

    if (nAftrBurn < nLeaves) then ! integral cannot be estimated
      
      call undef(ev)
      call undef(evSub%val)
      call undef(evSub%unc)
      
    else

      evSub%val=0
      evSub%unc=0
      ! sort chn%links(s(:),:) by prior·likelihood in ascending oder
      do i=chn%burnInLen+1_i8b,N
        priLikh(i)=real(prior(chn%links(i,:)),qp)*exp(chn%logLikh(i))
      end do
      call indxArr(priLikh,s)
      s=s+chn%burnInLen
      mdlSelVTNUsed=ceiling(MDL_SEL_VT_USE_FRAC*nAftrBurn,i8b)
      mdlSelVTUseFrac=real(mdlSelVTNUsed,qp)/nAftrBurn
      iCritUsed=N-mdlSelVTNUsed
      
      ! find indexes of free pars.
      k=0
      do j=1,curNPars
        if (.NOT. parFixed(j)) then
          call inc(k)
          pFree(k)=j
        end if
      end do


      ! build tree
      tree=>newKDTree(chn%links(s(iCritUsed+1_i8b:),pFree),priLikh(s(iCritUsed+1_i8b:N)))

      ! estimate Z
      ev=tree%estimIntg(nLeaves)
      deallocate(tree)
      call setIfDef(ev,ev*likhFact/mdlSelVTUseFrac)
      
      ! subsampling

      mdlSelVTSubNUsed=10_i8b*ceiling(sqrt(MDL_SEL_VT_USE_FRAC*nAftrBurn),i8b)
      mdlSelVTNSub=min(MDL_SEL_VT_N_SUBSAMP_MAX,int(real(mdlSelVTSubNUsed)*MDL_SEL_VT_SUBSAMP_N_SAMP_FRAC))
      
      block
        integer :: &
         nEvSamp
        integer(i8b) :: &
         iSub(mdlSelVTSubNUsed)
        real(qp) :: &
         evTmp, &
         evSamp(mdlSelVTNSub)

        nEvSamp=0
        nTries=0
        do
          do i=1,mdlSelVTSubNUsed
            iSub(i)=rndInt(iCritUsed+1_i8b,N)
          end do
          
          tree=>newKDTree(chn%links(s(iSub),pFree),[(real(prior(chn%links(s(iSub(i)),:)),qp)*exp(chn%logLikh(s(iSub(i)))), &
           i=1,mdlSelVTSubNUsed)])
          evTmp=tree%estimIntg(nLeaves)
          deallocate(tree)
          call setIfDef(evTmp,evTmp*likhFact/mdlSelVTUseFrac)
          call inc(nTries)

          if (isDef(evTmp)) then
            call inc(nEvSamp)
            evSamp(nEvSamp)=evTmp
            if (nEvSamp == mdlSelVTNSub) exit
          end if
          if (nTries >= mdlSelVTNSub .AND. real(nEvSamp)/nTries < MDL_SEL_VT_SUBSAMP_DEF_FRAC_MIN) then
            call undef(evSub)
            exit
          end if
        end do
        
        call setIfDef(evSub%val,arrMean(evSamp))
        call setIfDef(evSub%unc,sqrt(arrVar(evSamp,evSub%val)))

      end block

    end if

  end subroutine calcEvVT




  !===============================================[ DATA, MODELS, AND RESIDUALS ]===================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcResid
!
!   calculates residuals using best summaries for all pars.
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iSet, &
     N, &
     nAllSets, &
     allocStat, &
     i, &
     iAllResid, &
     iAllResidOfTy
    real(dp) :: &
     curPtsX(:), &
     curPtsY(:), &
     curPts2D(:,:), &
     sgmSampArr(:), &
     bestPars(curNPars), &
     dmy
    allocatable :: &
     curPtsX, &
     curPtsY, &
     curPts2D, &
     sgmSampArr

    iAllResid=0

    if (all(isDef(parEst(iNPlMd,:curNPars)%bestSumm))) then

      bestPars=parEst(iNPlMd,:curNPars)%bestSumm


      ! AMAngPos
      
      if (dtPropsAMAngPos%nRec > 0) then
      
        block

          type(tyCoord2DDbl) :: &
           resid(dtPropsAMAngPos%nRec)

          nAllSets=dtPropsAMAngPos%nRec
          iAllResidOfTy=0
          
          dmy=logLikh(bestPars,NO_FLAGS,resid2D=resid(:nAllSets),dtTy=DT_TY%AMAngPos) ! residuals
          dmy=logLikh(bestPars,NO_FLAGS,normResid=dtPropsAMAngPos%normResid(:nAllSets)%wJitt,dtTy=DT_TY%AMAngPos) ! norm. residuals (w/ jitter)
          dmy=logLikh(bestPars,setFlg(FLG%ignJitt),normResid=dtPropsAMAngPos%normResid(:nAllSets)%woJitt,dtTy=DT_TY%AMAngPos) ! norm. residuals (w/o jitter)
        
          do iSet=1,dtProps%nSetsOfTy(DT_TY%AMAngPos)

            N=dtSetsAMAngPos(iSet)%N

            dtSetsAMAngPos(iSet)%resid(iNPlMd,:N)=resid(iAllResidOfTy+1:iAllResidOfTy+N)

            dtPropsAMAngPos%normResidMean%wJitt=arrMean(dtPropsAMAngPos%normResid(:nAllSets)%wJitt)
            dtPropsAMAngPos%normResidMean%woJitt=arrMean(dtPropsAMAngPos%normResid(:nAllSets)%woJitt)

            dtPropsAMAngPos%normResidStdDev%wJitt=sqrt(arrVar(dtPropsAMAngPos%normResid(:nAllSets)%wJitt, &
             dtPropsAMAngPos%normResidMean%wJitt))
            dtPropsAMAngPos%normResidStdDev%woJitt=sqrt(arrVar(dtPropsAMAngPos%normResid(:nAllSets)%woJitt, &
             dtPropsAMAngPos%normResidMean%woJitt))

            dtPropsAMAngPos%normResidPVal%wJitt=KSSignifLvl(KSStat(normalCDF,0._dp,1._dp, &
             dtPropsAMAngPos%normResid(:nAllSets)%wJitt)*sqrt(1._dp*nAllSets))
            dtPropsAMAngPos%normResidPVal%woJitt=KSSignifLvl(KSStat(normalCDF,0._dp,1._dp, &
             dtPropsAMAngPos%normResid(:nAllSets)%woJitt)*sqrt(1._dp*nAllSets))

            iAllResidOfTy=iAllResidOfTy+N

          end do

          dtProps%normResid(iAllResid+1:iAllResid+nAllSets)%wJitt=dtPropsAMAngPos%normResid(:nAllSets)%wJitt
          dtProps%normResid(iAllResid+1:iAllResid+nAllSets)%woJitt=dtPropsAMAngPos%normResid(:nAllSets)%woJitt

          iAllResid=iAllResid+nAllSets

        end block

      end if


      ! RV
      
      if (dtPropsRV%nRec > 0) then
      
        block

          real(dp) :: &
           resid(dtPropsRV%nRec)

          nAllSets=dtPropsRV%nRec
          iAllResidOfTy=0

          dmy=logLikh(bestPars,NO_FLAGS,resid1D=resid(:nAllSets),dtTy=DT_TY%RV) ! residuals
          dmy=logLikh(bestPars,NO_FLAGS,normResid=dtPropsRV%normResid(:nAllSets)%wJitt,dtTy=DT_TY%RV) ! norm. residuals (w/ jitter)
          dmy=logLikh(bestPars,setFlg(FLG%ignJitt),normResid=dtPropsRV%normResid(:nAllSets)%woJitt,dtTy=DT_TY%RV) ! norm. residuals (w/o jitter)
        
          do iSet=1,dtProps%nSetsOfTy(DT_TY%RV)

            N=dtSetsRV(iSet)%N

            dtSetsRV(iSet)%resid(iNPlMd,:N)=resid(iAllResidOfTy+1:iAllResidOfTy+N)

            dtPropsRV%normResidMean%wJitt=arrMean(dtPropsRV%normResid(:nAllSets)%wJitt)
            dtPropsRV%normResidMean%woJitt=arrMean(dtPropsRV%normResid(:nAllSets)%woJitt)

            dtPropsRV%normResidStdDev%wJitt=sqrt(arrVar(dtPropsRV%normResid(:nAllSets)%wJitt,dtPropsRV%normResidMean%wJitt))
            dtPropsRV%normResidStdDev%woJitt=sqrt(arrVar(dtPropsRV%normResid(:nAllSets)%woJitt,dtPropsRV%normResidMean%woJitt))

            dtPropsRV%normResidPVal%wJitt=KSSignifLvl(KSStat(normalCDF,0._dp,1._dp, &
             dtPropsRV%normResid(:nAllSets)%wJitt)*sqrt(1._dp*nAllSets))
            dtPropsRV%normResidPVal%woJitt=KSSignifLvl(KSStat(normalCDF,0._dp,1._dp, &
             dtPropsRV%normResid(:nAllSets)%woJitt)*sqrt(1._dp*nAllSets))

            iAllResidOfTy=iAllResidOfTy+N

          end do

          dtProps%normResid(iAllResid+1:iAllResid+nAllSets)%wJitt=dtPropsRV%normResid(:nAllSets)%wJitt
          dtProps%normResid(iAllResid+1:iAllResid+nAllSets)%woJitt=dtPropsRV%normResid(:nAllSets)%woJitt

          iAllResid=iAllResid+nAllSets

        end block

      end if


      ! AMHipN
      
      if (dtPropsAMHipN%nRec > 0) then
      
        block

          real(dp) :: &
           resid(dtPropsAMHipN%nRec)

          nAllSets=dtPropsAMHipN%nRec
          iAllResidOfTy=0

          dmy=logLikh(bestPars,NO_FLAGS,resid1D=resid(:nAllSets),dtTy=DT_TY%AMHipN) ! residuals
          dmy=logLikh(bestPars,NO_FLAGS,normResid=dtPropsAMHipN%normResid(:nAllSets)%wJitt,dtTy=DT_TY%AMHipN) ! norm. residuals (w/ jitter)
          dmy=logLikh(bestPars,setFlg(FLG%ignJitt),normResid=dtPropsAMHipN%normResid(:nAllSets)%woJitt,dtTy=DT_TY%AMHipN) ! norm. residuals (w/o jitter)
        
          do iSet=1,dtProps%nSetsOfTy(DT_TY%AMHipN)

            N=dtSetsAMHipN(iSet)%N

            dtSetsAMHipN(iSet)%resid(iNPlMd,:N)=resid(iAllResidOfTy+1:iAllResidOfTy+N)

            dtPropsAMHipN%normResidMean%wJitt=arrMean(dtPropsAMHipN%normResid(:nAllSets)%wJitt)
            dtPropsAMHipN%normResidMean%woJitt=arrMean(dtPropsAMHipN%normResid(:nAllSets)%woJitt)

            dtPropsAMHipN%normResidStdDev%wJitt=sqrt(arrVar(dtPropsAMHipN%normResid(:nAllSets)%wJitt, &
             dtPropsAMHipN%normResidMean%wJitt))
            dtPropsAMHipN%normResidStdDev%woJitt=sqrt(arrVar(dtPropsAMHipN%normResid(:nAllSets)%woJitt, &
             dtPropsAMHipN%normResidMean%woJitt))

            dtPropsAMHipN%normResidPVal%wJitt=KSSignifLvl(KSStat(normalCDF,0._dp,1._dp, &
             dtPropsAMHipN%normResid(:nAllSets)%wJitt)*sqrt(1._dp*nAllSets))
            dtPropsAMHipN%normResidPVal%woJitt=KSSignifLvl(KSStat(normalCDF,0._dp,1._dp, &
             dtPropsAMHipN%normResid(:nAllSets)%woJitt)*sqrt(1._dp*nAllSets))

            iAllResidOfTy=iAllResidOfTy+N

          end do

          dtProps%normResid(iAllResid+1:iAllResid+nAllSets)%wJitt=dtPropsAMHipN%normResid(:nAllSets)%wJitt
          dtProps%normResid(iAllResid+1:iAllResid+nAllSets)%woJitt=dtPropsAMHipN%normResid(:nAllSets)%woJitt

          iAllResid=iAllResid+nAllSets

        end block

      end if
      
      dtProps%normResidMean%wJitt=arrMean(dtProps%normResid(:dtProps%nRec)%wJitt)
      dtProps%normResidMean%woJitt=arrMean(dtProps%normResid(:dtProps%nRec)%woJitt)
      
      dtProps%normResidStdDev%wJitt=sqrt(arrVar(dtProps%normResid(:dtProps%nRec)%wJitt,dtProps%normResidMean%wJitt))
      dtProps%normResidStdDev%woJitt=sqrt(arrVar(dtProps%normResid(:dtProps%nRec)%woJitt,dtProps%normResidMean%woJitt))

      residCalcd=.TRUE.

    else

      residCalcd=.FALSE.

    end if

  end subroutine calcResid




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcChiSq
!
!   calculates chi²:
!     * considering jitter
!     * ignoring jitter (i.e. setting sigma_+ = tau_+ = 0 in the chi² calc.)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i

    call undef(postrSumm(iNPlMd)%maxPostrChiSqByDtTy%woJitt)
    call undef(postrSumm(iNPlMd)%maxPostrChiSqByDtTy%wJitt)

    postrSumm(iNPlMd)%maxPostrChiSq%wJitt=logLikh(parEst(iNPlMd,:curNPars)%bestSumm,NO_FLAGS)
    postrSumm(iNPlMd)%maxPostrChiSq%woJitt=logLikh(parEst(iNPlMd,:curNPars)%bestSumm,setFlg(FLG%ignJitt))
    do i=1,N_DT_TYPES
      if (dtProps%nSetsOfTy(i) > 0) then
        postrSumm(iNPlMd)%maxPostrChiSqByDtTy(i)%wJitt=logLikh(parEst(iNPlMd,:curNPars)%bestSumm,NO_FLAGS,dtTy=i)
        postrSumm(iNPlMd)%maxPostrChiSqByDtTy(i)%woJitt=logLikh(parEst(iNPlMd,:curNPars)%bestSumm,setFlg(FLG%ignJitt),dtTy=i)
      end if
    end do

  end subroutine calcChiSq




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcAllPred(chain)
!
!   calculates all uncertainty predictions
!
!   Arguments:
!     chain                          MCMC chain

!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyMCMCChain) :: &
     chain
    intent(in) :: &
     chain

    if (usrPred) then
      call startSubsect(UNIT%cons+UNIT%log,"Calculating predictions")
      if (dtPropsAMAngPos%nRec > 0) call calcPred(chain,DT_TY%AMAngPos,mdlProps%AMAngPos%mdlFct,pred2D=pred2D)
      if (dtPropsRV%nRec > 0) call calcPred(chain,DT_TY%RV,mdlProps%RV%mdlFct,pred1D=pred1D)
      call endSubsect(UNIT%cons+UNIT%log) ! Calculating predictions
    end if


    contains

    subroutine calcPred(chain,dtTy,fct,pred1D,pred2D) ! calculates uncertainty prediction for a given data type
      implicit none
      integer :: &
       iT, &
       iTMinUnc, &
       iTMaxUnc, &
       iSet, &
       dtTy
      integer(i8b) :: &
       nLinks
      type(tyMCMCChain) :: &
       chain
      real(dp) :: &
       mdlFct2D(chain%length-chain%burnInLen,2), &
       curUnc1D, &
       minUnc, &
       maxUnc, &
       s, &
       t, &
       u, &
       sqr
      logical :: &
       binCompoDone(2)
      type(tyUncEllDbl) :: &
       curUnc2D
      type(tyPred1D) :: &
       pred1D
      type(tyPred2D) :: &
       pred2D
      type(tyParamsMdl) :: &
       paramsMdl
      optional :: &
       pred1D, &
       pred2D
      intent(in) :: &
       chain, &
       dtTy
      intent(inout) :: &
       pred1D

      interface
        function fct(N,M,pars,fctT,fctP,eAArr,paramsMdl,paramsDt,tMnEpochArr)
          import
          implicit none
          integer :: &
           M
          integer(i8b) :: &
           N
          real(dp) :: &
           pars(:,:), &
           fct(N,M), &
           eAArr(:,:,:), &
           tMnEpochArr(:)
          type(tyParamsMdl) :: &
           paramsMdl
          type(tyDtParamsAMHipN) :: &
           paramsDt
          optional :: &
           tMnEpochArr, &
           paramsDt
          intent(in) :: &
           N, &
           M, &
           pars, &
           eAArr, &
           paramsMdl, &
           paramsDt, &
           tMnEpochArr
          interface
            function fctT(x)
              import
              implicit none
              integer(i8b) :: &
               x, &
               fctT
              intent(in) :: &
               x
            end function
            function fctP(x)
              import
              implicit none
              integer(i8b) :: &
               x, &
               fctP
              intent(in) :: &
               x
            end function
          end interface
        end function
      end interface

      nLinks=chain%length-chain%burnInLen
      paramsMdl%psdNPl=psdNPl

      binCompoDone=.FALSE.


      select case (dtTy)
      case (DT_TY%AMAngPos)

        do iSet=1,dtProps%nSetsOfTy(DT_TY%AMAngPos)
          if (binCompoDone(dtSetsAMAngPos(iSet)%binCompo)) then
            pred2D%skipSet(dtTy,iSet)=.TRUE.
          else
            binCompoDone(dtSetsAMAngPos(iSet)%binCompo)=.TRUE.
            minUnc=huge(minUnc)
            maxUnc=-huge(maxUnc)
            paramsMdl%binCompo=dtSetsAMAngPos(iSet)%binCompo
            paramsMdl%iSet=iSet
            do iT=1,pred2D%nPts
            
              ! time
              pred2D%t(dtTy,iSet,iT)=pred2D%tRge(dtTy,iSet)%lBnd+real(iT-1,dp)/(pred2D%nPts-1)*pred2D%tRge(dtTy,iSet)%width()

              ! model fct.
              call calcMdlFctWErr(nLinks,MDL_FCT_DIM%AMAngPos,fct,chain%links(chain%burnInLen+1:chain%length,:),nxtInt8, &
               pred2D%t(dtTy,iSet,iT:iT)-mdlProps%epoch,identInt8,mdlFct2D,paramsMdl) ! pars IS NOT CONTIGUOUS

              ! cov. matrix (cf. [16.02.2011/1])
              s=arrVar(mdlFct2D(:,1))
              t=arrCov(mdlFct2D(:,1),mdlFct2D(:,2)) ! calcd. w.r.t. model function of best estimates!
              u=arrVar(mdlFct2D(:,2))

              associate(trHalf=>(s+u)/2._dp)
                sqr=sqrt(trHalf**2-(s*u-t**2))
                curUnc2D%a=trHalf+sqr
                curUnc2D%b=trHalf-sqr
              end associate

              ! pos. angle: is supposed to be in [0, 180°], can be shifted by integer multiples of pi
              curUnc2D%phi=(t/sqrt(t**2+(curUnc2D%b-s)**2))/PI
              curUnc2D%phi=(curUnc2D%phi-floor(curUnc2D%phi))*PI/DEG_IN_RAD

              pred2D%unc(dtTy,iSet,iT)=curUnc2D

              if (curUnc2D%b < minUnc) then
                iTMinUnc=iT
                minUnc=curUnc2D%b
              end if
              if (curUnc2D%a > maxUnc) then
                iTMaxUnc=iT
                maxUnc=curUnc2D%a
              end if

            end do

            pred2D%minUnc(dtTy,iSet)=minUnc
            pred2D%maxUnc(dtTy,iSet)=maxUnc
            pred2D%phiMinUnc(dtTy,iSet)=modulo(pred2D%unc(dtTy,iSet,iTMinUnc)%phi+90._dp,180._dp) ! pos. angle of semi-minor axis of unc. ell.
            pred2D%phiMaxUnc(dtTy,iSet)=pred2D%unc(dtTy,iSet,iTMaxUnc)%phi
            pred2D%tMinUnc(dtTy,iSet)=pred2D%t(dtTy,iSet,iTMinUnc)
            pred2D%tMaxUnc(dtTy,iSet)=pred2D%t(dtTy,iSet,iTMaxUnc)

          end if                       ! (binCompoDone(dtSetsAMAngPos(iSet)%binCompo))
        end do


      case (DT_TY%RV)

        do iSet=1,dtProps%nSetsOfTy(DT_TY%RV)
          if (mdlProps%bin .AND. binCompoDone(dtSetsRV(iSet)%binCompo)) then
            pred1D%skipSet(dtTy,iSet)=.TRUE.
          else
            if (mdlProps%bin) binCompoDone(dtSetsRV(iSet)%binCompo)=.TRUE.

            minUnc=huge(minUnc)
            maxUnc=-huge(maxUnc)
            paramsMdl%binCompo=dtSetsRV(iSet)%binCompo
            paramsMdl%iSet=iSet
            do iT=1,pred1D%nPts
              pred1D%t(dtTy,iSet,iT)=pred1D%tRge(dtTy,iSet)%lBnd+real(iT-1,dp)/(pred1D%nPts-1)*pred1D%tRge(dtTy,iSet)%width()
              call calcMdlFctWErr(nLinks,MDL_FCT_DIM%RV,fct,chain%links(chain%burnInLen+1:chain%length,:),nxtInt8, &
               pred1D%t(dtTy,iSet,iT:iT)-mdlProps%epoch,identInt8,mdlFct2D(:,1),paramsMdl) ! pars IS NOT CONTIGUOUS

              curUnc1D=sqrt(arrVar(mdlFct2D(:,1))) ! calcd. w.r.t. model function of best estimates!
              pred1D%unc(dtTy,iSet,iT)=curUnc1D
              if (curUnc1D < minUnc) then
                iTMinUnc=iT
                minUnc=curUnc1D
              end if
              if (curUnc1D > maxUnc) then
                iTMaxUnc=iT
                maxUnc=curUnc1D
              end if
            end do

            pred1D%minUnc(dtTy,iSet)=minUnc
            pred1D%maxUnc(dtTy,iSet)=maxUnc
            pred1D%tMinUnc(dtTy,iSet)=pred1D%t(dtTy,iSet,iTMinUnc)
            pred1D%tMaxUnc(dtTy,iSet)=pred1D%t(dtTy,iSet,iTMaxUnc)

          end if                     ! (binCompoDone(dtSetsRV(iSet)%binCompo))

        end do

      end select
    end subroutine

  end subroutine calcAllPred




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine plotDataAndModel(flags,refPos)
!
!   plots data and model
!
!   Arguments:
!     flags                          set of flags
!     refPos                         ref. pos. (alpha·cos(delta), delta) [deg] for transforming AMHipN data into AMAngPos data
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     flags, &
     curNPts, &
     allocStat, &
     i, &
     j, &
     curNData, &
     iSet, &
     k, &
     l, &
     curPltSymArr(:), &
     curPltSymArr2(:)
    real(dp) :: &
     curPtsX(:), &
     curPtsY(:), &
     curPtsZ(:), &
     curPts2D(:,:), &
     curPtsUnc2D(:,:), &
     curPars(curNPars), &
     curPars2D(1,curNPars), &
     curP, &
     curTPeri
    logical :: &
     folded, &
     resid, &
     onlyDt
    character(STR_LEN_STD) :: &
     strFold, &
     strFoldTitle, &
     strCenter, &
     lblX, &
     lblY, &
     curModArr(:), &
     curStyModArr(:), &
     curModArr2(:), &
     curStyModArr2(:)
    character(MAX_LEN_PATH) :: &
     curFilNmCSVDt, &
     curFilNmCSVMdl, &
     curFilNmPlt, &
     curFilNmCSVDtUncEll(:), &
     curFilNmCSVMdlResidVect(:)
    type(tyIntvlDbl) :: &
     tRge
    type(tyCoord2DDbl) :: &
     curPos, &
     refPos
    type(tyDtSetAMAngPos) :: &
     allDt
    allocatable :: &
     curPltSymArr, &
     curPltSymArr2, &
     curPtsX, &
     curPtsY, &
     curPtsZ, &
     curPts2D, &
     curPtsUnc2D, &
     curFilNmCSVDtUncEll, &
     curFilNmCSVMdlResidVect, &
     curModArr, &
     curStyModArr, &
     curModArr2, &
     curStyModArr2
    intent(in) :: &
     flags, &
     refPos


    ! current drawbacks:
    !  – "single plots" ignored
    !  – folding only for 1-pl. or binary models
    !  – AMAngPos data: 1 plot for all data sets

    folded=getFlg(flags,FLG%fold) .AND. psdNPl == 1
    resid=getFlg(flags,FLG%resid)
    onlyDt=getFlg(flags,FLG%onlyDt)
    curPars=parEst(iNPlMd,:curNPars)%bestSumm
    curPars2D=reshape(curPars,[1,curNPars])
    strCenter=""
    if (psdNPl == 1) then
      curP=1._dp/curPars(pRoleByPl(1)%f)
    else
      call undef(curP)
    end if


    ! AMAngPos DATA

    if (dtPropsAMAngPos%nRec > 0) then

      strFold=""
      strFoldTitle=""
      
      i=0
      do j=1,dtProps%nSetsOfTy(DT_TY%AMAngPos)
        associate(dtSet=>dtSetsAMAngPos(j))
          curNData=dtSet%N
          allDt%t(i+1:i+curNData)=dtSet%t(1:curNData)
          allDt%obs(i+1:i+curNData)=dtSet%obs(1:curNData)
          allDt%resid(iNPlMd,i+1:i+curNData)=dtSet%resid(iNPlMd,1:curNData)
          allDt%unc(i+1:i+curNData)=dtSet%unc(1:curNData)
        end associate
        i=i+curNData
      end do
      allDt%N=dtPropsAMAngPos%nRec
      allDt%tRge=dtPropsAMAngPos%tRge

      if (.NOT. folded) call plotUncEll(allDt,DT_TY%AMAngPos,PLT_UNC_ELL_N_PTS,TWO_PI,setFlg([FLG%residVect,FLG%onlyDt,FLG%resid], &
       [.NOT. (onlyDt .OR. resid),onlyDt,resid]))
      call plotAMAngPosPerCoord(allDt,DT_TY%AMAngPos,setFlg([FLG%fold,FLG%resid,FLG%onlyDt],[folded,resid,onlyDt]))

    end if                           ! (dtPropsAMAngPos%nRec > 0)



    ! RV DATA

    if (dtPropsRV%nRec > 0) then

      if (folded) then
        strFold="_"//FIL_NM_PART%fold
        lblX="Orbital phase"
      else
        strFold=""
        strFoldTitle=""
        lblX=dateDescAndUnitPlt
      end if
      lblY=fmtObsvble(DT_TY%RV,1)



      if (resid) then                ! plot residuals

        do iSet=1,dtProps%nSetsOfTy(DT_TY%RV)

          curNPts=dtSetsRV(iSet)%N

          if (allocated(curPtsX)) deallocate(curPtsX)
          if (allocated(curPtsY)) deallocate(curPtsY)
          allocate(curPtsX(curNPts),curPtsY(curNPts),stat=allocStat)
          call testAllocErr(allocStat,29)

          curPtsX=tOfs+dtSetsRV(iSet)%t(:curNPts)
          curPtsY=dtSetsRV(iSet)%resid(iNPlMd,:curNPts)

          curFilNmCSVDt=trim(fmtFilNmPlt(trim(FIL_NM_PART%resid)//"_"//trim(DT_TY_NMS(DT_TY%RV))//trim(fmtDtSet(DT_TY%RV,iSet)))) &
           //FIL_EXT_CSV

          call writeCSV(trim(outDirCSV)//trim(curFilNmCSVDt),curPtsX,curPtsY,dtSetsRV(iSet)%sgm(:curNPts))

          curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%resid)//"_"//trim(DT_TY_NMS(DT_TY%RV))// &
           trim(fmtDtSet(DT_TY%RV,iSet))))//trim(pltFilExt)
          call addGnupltCmds(curFilNmCSVDt,curFilNmPlt,trim(lblX),trim(lblY),trim(uCaseFst(trim(TTL_PART%resid))),NO_FLAGS, &
           sym=PLT_SYM%errBar)

        end do


      else                           ! don't plot residuals


        block
          character(MAX_LEN_PATH) :: &
           allFilNmsCSVDt(dtProps%nSetsOfTy(DT_TY%RV))


          do iSet=1,dtProps%nSetsOfTy(DT_TY%RV)

            if (folded) then
              curTPeri=mdlProps%epoch-curPars(pRoleByPl(1)%chi)/curPars(pRoleByPl(1)%f)
              tRge=tyIntvlDbl(curTPeri,curTPeri+curP)
              strFoldTitle=","//trim(mrg(" acceleration cancelled,","",mdlProps%RVAccel))//" folded (T="// &
               trim(niceFmt(tOfs+curTPeri))//" d, P="//trim(niceFmt(curP))//" d)"
            else
              tRge=centralStretchIntvl(dtSetsRV(iSet)%tRge,TM_RGE_STRETCH_FACT)
            end if

              ! data

            curNData=dtSetsRV(iSet)%N
            curNPts=curNData*merge(2,1,folded)
            if (allocated(curPtsX)) deallocate(curPtsX)
            if (allocated(curPtsY)) deallocate(curPtsY)
            if (allocated(curPtsZ)) deallocate(curPtsZ)
            allocate(curPtsX(curNPts),curPtsY(curNPts),curPtsZ(curNPts),stat=allocStat)
            call testAllocErr(allocStat,10)


            ! set (& fold) points
            curPtsX(1:curNData)=dtSetsRV(iSet)%t(1:curNData)
            curPtsY(1:curNData)=dtSetsRV(iSet)%obs(1:curNData)
            curPtsZ(1:curNData)=dtSetsRV(iSet)%sgm(1:curNData)
            if (folded) then
              if (mdlProps%RVAccel) call subtrConstChg(curPtsX(:curNData),curPtsY(:curNData),mdlProps%epoch, &
               curPars(pRole%aV)/YR_IN_D)
              call foldArr(curPtsX(:curNData),curPtsY(:curNData),curTPeri,curP,.TRUE.,curPtsZ(:curNData))
              call duplicateArr(curPtsX,curPtsY,.5_dp,.TRUE.,curPtsZ)
            else
              curPtsX=tOfs+curPtsX
            end if

            ! write CSV output
            curFilNmCSVDt=trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"_"//trim(DT_TY_NMS(DT_TY%RV))//trim(strFold)// &
             trim(fmtDtSet(DT_TY%RV,iSet))))//FIL_EXT_CSV
            allFilNmsCSVDt(iSet)=curFilNmCSVDt
            call writeCSV(trim(outDirCSV)//trim(curFilNmCSVDt),curPtsX,curPtsY,curPtsZ)


            if (onlyDt) then

              ! add plotting commands
              curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"_"//trim(DT_TY_NMS(DT_TY%RV))// &
               trim(fmtDtSet(DT_TY%RV,iSet))))//trim(pltFilExt)
              call addGnupltCmds(curFilNmCSVDt,curFilNmPlt,trim(lblX),trim(lblY),uCaseFst(trim(TTL_PART%data))// &
               trim(strFoldTitle),NO_FLAGS,sym=PLT_SYM%errBar)

            else


                ! model

              curNPts=pltNPtsMdl*merge(2,1,folded)
              if (allocated(curPtsX)) deallocate(curPtsX)
              if (allocated(curPtsY)) deallocate(curPtsY)
              allocate(curPtsX(curNPts),curPtsY(curNPts),stat=allocStat)
              call testAllocErr(allocStat,11)

              curPtsX(:pltNPtsMdl)=sampTimes(pltNPtsMdl,tRge,.NOT. OPT_SCATTER,.NOT. OPT_WINDOW)

              ! sample function
              call calcMdlFctWErr(int(pltNPtsMdl,i8b),MDL_FCT_DIM%RV,mdlProps%RV%mdlFct,curPars2D,identInt8, &
               curPtsX(:pltNPtsMdl)-mdlProps%epoch,nxtInt8,curPtsY(:pltNPtsMdl),tyParamsMdl(psdNPl,iSet,dtSetsRV(iSet)%binCompo)) ! pars IS CONTIGUOUS

              ! (fold)
              if (folded) then
                if (mdlProps%RVAccel) call subtrConstChg(curPtsX(:pltNPtsMdl),curPtsY(:pltNPtsMdl),mdlProps%epoch, &
                 curPars(pRole%aV)/YR_IN_D)
                call foldArr(curPtsX(:pltNPtsMdl),curPtsY(:pltNPtsMdl),curTPeri,curP,.TRUE.)
                call duplicateArr(curPtsX,curPtsY,.5_dp,.TRUE.)
              else
                curPtsX=tOfs+curPtsX
              end if

              ! write model CSV output
              curFilNmCSVMdl=trim(fmtFilNmPlt(trim(FIL_NM_PART%mdl)//"_"//trim(DT_TY_NMS(DT_TY%RV))//trim(strFold)// &
               trim(fmtDtSet(DT_TY%RV,iSet))))//FIL_EXT_CSV
              call writeCSV(trim(outDirCSV)//trim(curFilNmCSVMdl),curPtsX,curPtsY)

              ! add plotting commands
              curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"+"//trim(FIL_NM_PART%mdl)//"_"// &
               trim(DT_TY_NMS(DT_TY%RV))//trim(fmtDtSet(DT_TY%RV,iSet))//trim(strFold)))//trim(pltFilExt)
              call addGnupltCmds("",curFilNmPlt,trim(lblX),trim(lblY),"Data and model"// &
               trim(strFoldTitle),NO_FLAGS,filNmInArr=[curFilNmCSVMdl,curFilNmCSVDt],symArr=[PLT_SYM%line,PLT_SYM%errBar])

              ! plot all data sets in one plot
              if (dtProps%nSetsOfTy(DT_TY%RV) > 1 .AND. iSet == dtProps%nSetsOfTy(DT_TY%RV)) then
                block
                  character(STR_LEN_STD) :: &
                   modArr(1+dtProps%nSetsOfTy(DT_TY%RV))
                  
                  if (.NOT. folded) then ! recalc. model over whole timespan

                    curPtsX(:pltNPtsMdl)=sampTimes(pltNPtsMdl,centralStretchIntvl(dtPropsRV%tRge,TM_RGE_STRETCH_FACT), &
                     .NOT. OPT_SCATTER,.NOT. OPT_WINDOW)
                    call calcMdlFctWErr(int(pltNPtsMdl,i8b),MDL_FCT_DIM%RV,mdlProps%RV%mdlFct,curPars2D,identInt8, &
                     curPtsX(:pltNPtsMdl)-mdlProps%epoch,nxtInt8,curPtsY(:pltNPtsMdl), &
                     tyParamsMdl(psdNPl,iSet,dtSetsRV(iSet)%binCompo)) ! pars IS CONTIGUOUS
                    curPtsX=tOfs+curPtsX

                    curFilNmCSVMdl=trim(fmtFilNmPlt(trim(FIL_NM_PART%mdl)//"_"//trim(DT_TY_NMS(DT_TY%RV))//trim(strFold)// &
                     trim(fmtDtSet(DT_TY%RV,VAL_UNDEF_INT))))//FIL_EXT_CSV
                    call writeCSV(trim(outDirCSV)//trim(curFilNmCSVMdl),curPtsX,curPtsY)

                  end if

                  curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"+"//trim(FIL_NM_PART%mdl)//"_"// &
                   trim(DT_TY_NMS(DT_TY%RV))//trim(fmtDtSet(DT_TY%RV,VAL_UNDEF_INT))//trim(strFold)))//trim(pltFilExt)
                  modArr(1)="us 1:($2-("//trim(niceFmt(curPars(pRoleBySet(dtProps%nSetsOfTy(DT_TY%RV))%V)))//"))"
                  do i=1,dtProps%nSetsOfTy(DT_TY%RV)
                    modArr(1+i)="us 1:($2-("//trim(niceFmt(curPars(pRoleBySet(i)%V)))//")):3"
                  end do
                  call addGnupltCmds("",curFilNmPlt,trim(lblX),trim(lblY),"Data (all sets) and model, relative offset cancelled"// &
                   trim(strFoldTitle), &
                   NO_FLAGS,filNmInArr=[curFilNmCSVMdl,(allFilNmsCSVDt(iSet), iSet=1,dtProps%nSetsOfTy(DT_TY%RV))], &
                   symArr=[PLT_SYM%line,(PLT_SYM%errBar, iSet=1,dtProps%nSetsOfTy(DT_TY%RV))],modArr=modArr)

                end block
              end if
            end if

          end do                       ! iSet=1,dtProps%nSetsOfTy(DT_TY%RV)

        end block

      end if                         ! (resid)

    end if                           ! (dtPropsRV%nRec > 0)



    ! AMHipN DATA

    if (dtPropsAMHipN%nRec > 0) then
      
        ! interpreted as AMAngPos data
      
      if (dtPropsAMHipN%nRec > N_REC_PER_SET_MAX) then
      
        call warn("plotting "//trim(DT_TY_DESC(DT_TY%AMHipN))//" as "//trim(DT_TY_DESC(DT_TY%AMAngPos))//" skipped (too many data)")
        
      else

        strFold=""
        strFoldTitle=""
      
        i=0
        do j=1,dtProps%nSetsOfTy(DT_TY%AMHipN)
          associate(dtSet=>dtSetsAMHipN(j))
            curNData=dtSet%N
            associate(params=>dtSet%params)
              ! transform AMHipN data into 2-d coordinates w.r.t. ref. pos.
              allDt%obs(i+1:i+curNData)%x=(params%pos0AtT(:curNData)%x-refPos%x)/MAS_IN_DEG+ &
               params%sinPsi(:curNData)*dtSet%obs(:curNData)

              allDt%obs(i+1:i+curNData)%y=(params%pos0AtT(:curNData)%y-refPos%y)/MAS_IN_DEG+ &
               params%cosPsi(:curNData)*dtSet%obs(:curNData)

              allDt%t(i+1:i+curNData)=dtSet%t(:curNData)
              allDt%resid(iNPlMd,i+1:i+curNData)%x=params%sinPsi(:curNData)*dtSet%resid(iNPlMd,:curNData)
              allDt%resid(iNPlMd,i+1:i+curNData)%y=params%cosPsi(:curNData)*dtSet%resid(iNPlMd,:curNData)
              allDt%unc(i+1:i+curNData)%b=dtSet%sgm(:curNData) ! used for per-coordinate plots (substituted by zero for 2-d plots)
              allDt%unc(i+1:i+curNData)%phi=-params%psi(:curNData) ! cf. [24.06.2012/2]
            end associate
          end associate
          i=i+curNData
        end do
        
        allDt%N=dtPropsAMHipN%nRec
        allDt%tRge=dtPropsAMHipN%tRge
        allDt%unc(:allDt%N)%sinPhi=sin(allDt%unc(:allDt%N)%phi)
        allDt%unc(:allDt%N)%sinSqPhi=allDt%unc(:allDt%N)%sinPhi**2
        allDt%unc(:allDt%N)%cosPhi=cos(allDt%unc(:allDt%N)%phi)
        allDt%unc(:allDt%N)%cosSqPhi=allDt%unc(:allDt%N)%cosPhi**2
        if (getFlg(flags,FLG%center)) then
          allDt%obs(:allDt%N)%x=allDt%obs(:allDt%N)%x-.5_dp*(minval(allDt%obs(:allDt%N)%x)+maxval(allDt%obs(:allDt%N)%x))
          allDt%obs(:allDt%N)%y=allDt%obs(:allDt%N)%y-.5_dp*(minval(allDt%obs(:allDt%N)%y)+maxval(allDt%obs(:allDt%N)%y))
          strCenter=" (centered at the origin)"
        end if

        if (.NOT. folded) call plotUncEll(allDt,DT_TY%AMHipN,2,PI,setFlg([FLG%residVect,FLG%onlyDt,FLG%resid], &
         [.NOT. (onlyDt .OR. resid),onlyDt,resid]))
        call plotAMAngPosPerCoord(allDt,DT_TY%AMHipN,setFlg([FLG%fold,FLG%resid,FLG%onlyDt],[folded,resid,onlyDt]))
        
      end if


        ! normal

      if (folded) then
        strFold="_"//FIL_NM_PART%fold
        lblX="Orbital phase"
      else
        strFold=""
        strFoldTitle=""
        lblX=dateDescAndUnitPlt
      end if
      lblY=fmtObsvble(DT_TY%AMHipN,1)


      if (resid) then                ! plot residuals

        do iSet=1,dtProps%nSetsOfTy(DT_TY%AMHipN)

          curNPts=dtSetsAMHipN(iSet)%N

          if (allocated(curPtsX)) deallocate(curPtsX)
          if (allocated(curPtsY)) deallocate(curPtsY)
          allocate(curPtsX(curNPts),curPtsY(curNPts),stat=allocStat)
          call testAllocErr(allocStat,29)

          curPtsX=tOfs+dtSetsAMHipN(iSet)%t(:curNPts)
          curPtsY=dtSetsAMHipN(iSet)%resid(iNPlMd,:curNPts)

          curFilNmCSVDt=trim(fmtFilNmPlt(trim(FIL_NM_PART%resid)//"_"//trim(DT_TY_NMS(DT_TY%AMHipN))// &
           trim(fmtDtSet(DT_TY%AMHipN,iSet))))//FIL_EXT_CSV

          call writeCSV(trim(outDirCSV)//trim(curFilNmCSVDt),curPtsX,curPtsY,dtSetsAMHipN(iSet)%sgm(:curNPts))

          curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%resid)//"_"//trim(DT_TY_NMS(DT_TY%AMHipN))// &
           trim(fmtDtSet(DT_TY%AMHipN,iSet))))//trim(pltFilExt)
          call addGnupltCmds(curFilNmCSVDt,curFilNmPlt,trim(lblX),trim(lblY),trim(uCaseFst(trim(TTL_PART%resid))),NO_FLAGS, &
           sym=PLT_SYM%errBar)

        end do


      else                           ! don't plot residuals


        do iSet=1,dtProps%nSetsOfTy(DT_TY%AMHipN)

          if (folded) then
            curTPeri=mdlProps%epoch-curPars(pRoleByPl(1)%chi)/curPars(pRoleByPl(1)%f)
            tRge=tyIntvlDbl(curTPeri,curTPeri+curP)
            strFoldTitle=", folded (T="//trim(niceFmt(tOfs+curTPeri))//" d, P="//trim(niceFmt(curP))//" d)"
          else
            tRge=centralStretchIntvl(dtSetsAMHipN(iSet)%tRge,TM_RGE_STRETCH_FACT)
          end if

            ! data

          curNData=dtSetsAMHipN(iSet)%N
          curNPts=curNData*merge(2,1,folded)
          if (allocated(curPtsX)) deallocate(curPtsX)
          if (allocated(curPtsY)) deallocate(curPtsY)
          if (allocated(curPtsZ)) deallocate(curPtsZ)
          allocate(curPtsX(curNPts),curPtsY(curNPts),curPtsZ(curNPts),stat=allocStat)
          call testAllocErr(allocStat,10)


          ! set (& fold) points
          curPtsX(1:curNData)=dtSetsAMHipN(iSet)%t(1:curNData)
          curPtsY(1:curNData)=dtSetsAMHipN(iSet)%obs(1:curNData)
          curPtsZ(1:curNData)=dtSetsAMHipN(iSet)%sgm(1:curNData)
          if (folded) then
            call foldArr(curPtsX(:curNData),curPtsY(:curNData),curTPeri,curP,.TRUE.,curPtsZ(:curNData))
            call duplicateArr(curPtsX,curPtsY,.5_dp,.TRUE.,curPtsZ)
          else
            curPtsX=tOfs+curPtsX
          end if

          ! write CSV output
          curFilNmCSVDt=trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"_"//trim(DT_TY_NMS(DT_TY%AMHipN))//trim(strFold)// &
           trim(fmtDtSet(DT_TY%AMHipN,iSet))))//FIL_EXT_CSV
          call writeCSV(trim(outDirCSV)//trim(curFilNmCSVDt),curPtsX,curPtsY,curPtsZ)



          if (onlyDt) then

            ! add plotting commands
            curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"_"//trim(DT_TY_NMS(DT_TY%AMHipN))// &
             trim(fmtDtSet(DT_TY%AMHipN,iSet))))//trim(pltFilExt)
            call addGnupltCmds(curFilNmCSVDt,curFilNmPlt,trim(lblX),trim(lblY),uCaseFst(trim(TTL_PART%data))//trim(strFoldTitle), &
             NO_FLAGS,sym=PLT_SYM%errBar)

          else

            block
              
              integer :: &
               pltNPtsMdl            ! we need a "local" version of this here


                ! model

              pltNPtsMdl=curNData
              curNPts=pltNPtsMdl*merge(2,1,folded)
              if (allocated(curPtsX)) deallocate(curPtsX)
              if (allocated(curPtsY)) deallocate(curPtsY)
              allocate(curPtsX(curNPts),curPtsY(curNPts),stat=allocStat)
              call testAllocErr(allocStat,11)

              curPtsX(1:curNPts)=dtSetsAMHipN(iSet)%t(1:curNPts)

              ! sample function
              call calcMdlFctWErr(int(pltNPtsMdl,i8b),MDL_FCT_DIM%AMHipN,mdlProps%AMHipN%mdlFct,curPars2D,identInt8, &
               curPtsX(:pltNPtsMdl)-mdlProps%epoch,nxtInt8,curPtsY(:pltNPtsMdl),tyParamsMdl(psdNPl,iSet, &
               dtSetsAMHipN(iSet)%binCompo),dtSetsAMHipN(iSet)%params) ! pars IS CONTIGUOUS

              ! (fold)
              if (folded) then
                call foldArr(curPtsX(:pltNPtsMdl),curPtsY(:pltNPtsMdl),curTPeri,curP,.TRUE.)
                call duplicateArr(curPtsX,curPtsY,.5_dp,.TRUE.)
              else
                curPtsX=tOfs+curPtsX
              end if

              ! write model CSV output
              curFilNmCSVMdl=trim(fmtFilNmPlt(trim(FIL_NM_PART%mdl)//"_"//trim(DT_TY_NMS(DT_TY%AMHipN))//trim(strFold)// &
               trim(fmtDtSet(DT_TY%AMHipN,iSet))))//FIL_EXT_CSV
              call writeCSV(trim(outDirCSV)//trim(curFilNmCSVMdl),curPtsX,curPtsY)

              ! add plotting commands
              curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"+"//trim(FIL_NM_PART%mdl)//"_"// &
               trim(DT_TY_NMS(DT_TY%AMHipN))//trim(fmtDtSet(DT_TY%AMHipN,iSet))//trim(strFold)))//trim(pltFilExt)
              call addGnupltCmds("",curFilNmPlt,trim(lblX),trim(lblY),"Data and model"//trim(strFoldTitle),NO_FLAGS, &
               filNmInArr=[curFilNmCSVMdl,curFilNmCSVDt],symArr=[PLT_SYM%point,PLT_SYM%errBar], &
               titleArr=[character(STR_LEN_STD) :: TTL_PART%mdl,TTL_PART%data]) ! CAUTION: using, e.g., uCaseFst() on components of titleArr will cause BASE to crash (gcc bug #53296)

            end block

          end if

        end do                       ! iSet=1,dtProps%nSetsOfTy(DT_TY%AMHipN)

      end if                         ! (resid)

    end if                           ! (dtPropsAMHipN%nRec > 0)


    contains

    subroutine subtrConstChg(arrX,arrY,epoch,rate) ! subtracts a change with constant rate
      implicit none
      real(dp) :: &
       arrX(:), &
       arrY(:), &
       epoch, &
       rate
      intent(inout) :: &
       arrX, &
       arrY
      intent(in) :: &
       epoch, &
       rate

      arrY=arrY-(arrX-epoch)*rate
    end subroutine


    subroutine plotUncEll(allDt,dtTy,nPts,angTot,flags) ! plots uncertainty ellipses
      implicit none
      integer :: &
       dtTy, &
       flags, &
       curNRec, &
       nPts
      real(dp) :: &
       angTot, &
       angStep, &
       theta
      logical :: &
       resid
      character(STR_LEN_STD) :: &
       dtTyDesc
      type(tyDtSetAMAngPos) :: &
       allDt
      type(tyUncEllDbl) :: &
       curUnc
      intent(in) :: &
       dtTy, &
       nPts, &
       angTot, &
       flags
      intent(inout) :: &
       allDt

      dtTyDesc=trim(DT_TY_NMS(dtTy))//trim(mrg("_as_"//trim(DT_TY_NMS(DT_TY%AMAngPos)),"",dtTy == DT_TY%AMHipN))
      angStep=angTot/(nPts-1)
      resid=getFlg(flags,FLG%resid)

      lblX=fmtObsvble(DT_TY%AMAngPos,2) ! alpha·cos(delta)
      lblY=fmtObsvble(DT_TY%AMAngPos,1) ! delta

      curNData=allDt%N
      if (allocated(curPtsX)) deallocate(curPtsX)
      if (allocated(curPtsY)) deallocate(curPtsY)
      if (allocated(curFilNmCSVDtUncEll)) deallocate(curFilNmCSVDtUncEll)
      if (allocated(curPltSymArr)) deallocate(curPltSymArr)
      if (allocated(curModArr)) deallocate(curModArr)
      if (allocated(curStyModArr)) deallocate(curStyModArr)
      allocate(curPtsX(curNData*nPts),curPtsY(curNData*nPts),curFilNmCSVDtUncEll(curNData), &
       curPltSymArr(curNData),curModArr(curNData+1),curStyModArr(curNData+1),stat=allocStat)
      call testAllocErr(allocStat,19)
      
      ! set unc.-ell. semi-major axis
      if (dtTy == DT_TY%AMHipN) then
        if (resid) then
          allDt%unc(:allDt%N)%a=AM_HIP_N_AS_ANG_POS_SEMIMAJ_RATIO* &
           sqrt(arrRge(allDt%resid(iNPlMd,:allDt%N)%x)*arrRge(allDt%resid(iNPlMd,:allDt%N)%y))
        else
          allDt%unc(:allDt%N)%a=AM_HIP_N_AS_ANG_POS_SEMIMAJ_RATIO*sqrt(arrRge(allDt%obs(:allDt%N)%x)*arrRge(allDt%obs(:allDt%N)%y))
        end if
      end if
        
      curModArr(1)=""                ! for model
      curStyModArr(1)=""             ! for model
      curStyModArr(2:curNData+1)="lt 1" ! for data
      do i=1,curNData
        if (resid) then
          curPos=allDt%resid(iNPlMd,i)
        else
          curPos=allDt%obs(i)
        end if
        curUnc=allDt%unc(i)
        if (dtTy == DT_TY%AMHipN) curUnc%b=0._dp
        do l=1,nPts
          ! cf. [book.p156]
          theta=(l-1)*angStep
          curPtsX((i-1)*nPts+l)=curPos%y+curUnc%a*curUnc%sinPhi*cos(theta)+curUnc%b*curUnc%cosPhi*sin(theta)
          curPtsY((i-1)*nPts+l)=curPos%x+curUnc%a*curUnc%cosPhi*cos(theta)-curUnc%b*curUnc%sinPhi*sin(theta)
        end do
        curModArr(1+i)="ev ::"//trim(niceFmt((i-1)*nPts))//"::"//trim(niceFmt(i*nPts-1)) ! 1st record has index 0 in Gnuplot
      end do
      curFilNmCSVDt=trim(fmtFilNmPlt(trim(mrg(FIL_NM_PART%resid,FIL_NM_PART%data,resid))//"_"//trim(dtTyDesc)// &
       trim(strFold)//"_"//FIL_NM_PART%uncEll))//FIL_EXT_CSV
      curFilNmCSVDtUncEll=curFilNmCSVDt
      curPltSymArr=PLT_SYM%line

      call writeCSV(trim(outDirCSV)//trim(curFilNmCSVDt),curPtsX,curPtsY)

      if (resid .OR. getFlg(flags,FLG%onlyDt)) then

        ! add plotting commands
        curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(mrg(FIL_NM_PART%resid,FIL_NM_PART%data,resid))//"_"// &
         trim(dtTyDesc)))//trim(pltFilExt)
        call addGnupltCmds("",curFilNmPlt,trim(lblX),trim(lblY),uCaseFst(trim(mrg(TTL_PART%resid,TTL_PART%data,resid)))// &
         trim(strCenter)//trim(strFoldTitle),setFlg([FLG%rvX,FLG%eqAxR]),filNmInArr=curFilNmCSVDtUncEll,symArr=curPltSymArr, &
         modArr=curModArr(2:),styModArr=curStyModArr(2:))

      else                           ! don't plot residuals; data+model

        if (allocated(curPts2D)) deallocate(curPts2D)
        if (allocated(curPtsX)) deallocate(curPtsX)
        allocate(curPts2D(pltNPtsMdl,2),curPtsX(pltNPtsMdl),stat=allocStat)
        call testAllocErr(allocStat,9)

        ! set pars.
        if (dtTy == DT_TY%AMAngPos) then
          if (isDef(curP)) then
            tRge=tyIntvlDbl(0._dp,curP)
          else
            tRge=centralStretchIntvl(dtPropsAMAngPos%tRge,TM_RGE_STRETCH_FACT)
          end if
        else
          tRge=centralStretchIntvl(dtPropsAMHipN%tRge,max(TM_RGE_STRETCH_FACT,merge(curP/dtPropsAMHipN%tRge%width(),-HUGE_DP, &
           isDef(curP))))
        end if
        curPtsX=sampTimes(pltNPtsMdl,tRge,.NOT. OPT_SCATTER,.NOT. OPT_WINDOW)

        ! sample function
        if (dtTy == DT_TY%AMAngPos) then
          call calcMdlFctWErr(int(pltNPtsMdl,i8b),MDL_FCT_DIM%AMAngPos,mdlProps%AMAngPos%mdlFct,curPars2D,identInt8, &
           curPtsX-mdlProps%epoch,nxtInt8,curPts2D,tyParamsMdl(psdNPl,1,BIN_COMPO_AM_ANG_POS_STD)) ! pars IS CONTIGUOUS
        else                         ! for AMHipN data
          call calcMdlFctWErr(int(pltNPtsMdl,i8b),MDL_FCT_DIM%AMAngPos,mdlProps%AMHipN%mdlFct,curPars2D,identInt8, & ! 2-d positions
           curPtsX-mdlProps%epoch,nxtInt8,curPts2D,tyParamsMdl(psdNPl,1,BIN_COMPO_AM_ANG_POS_STD)) ! pars IS CONTIGUOUS
        end if

        ! write CSV output
        curFilNmCSVMdl=trim(fmtFilNmPlt(trim(FIL_NM_PART%mdl)//"_"//trim(dtTyDesc)//trim(strFold)))//FIL_EXT_CSV
        call writeCSV(trim(outDirCSV)//trim(curFilNmCSVMdl),curPts2D(:,2),curPts2D(:,1))


        if (getFlg(flags,FLG%residVect)) then

          curNData=allDt%N
          if (allocated(curPts2D)) deallocate(curPts2D)
          if (allocated(curPtsX)) deallocate(curPtsX)
          if (allocated(curModArr2)) deallocate(curModArr2)
          if (allocated(curStyModArr2)) deallocate(curStyModArr2)
          if (allocated(curPltSymArr2)) deallocate(curPltSymArr2)
          if (allocated(curFilNmCSVMdlResidVect)) deallocate(curFilNmCSVMdlResidVect)
          allocate(curPts2D(2*curNData,2),curPtsX(curNData),curModArr2(curNData),curStyModArr2(curNData),curPltSymArr2(curNData), &
           curFilNmCSVMdlResidVect(curNData),stat=allocStat)
          call testAllocErr(allocStat,24)


          ! set data times and positions

          curPtsX=allDt%t(:curNData)
          curPts2D(1:2*curNData-1:2,1)=allDt%obs(:curNData)%x
          curPts2D(1:2*curNData-1:2,2)=allDt%obs(:curNData)%y


          ! sample function

          if (dtTy == DT_TY%AMAngPos) then
            call calcMdlFctWErr(int(curNData,i8b),MDL_FCT_DIM%AMAngPos,mdlProps%AMAngPos%mdlFct,curPars2D,identInt8, &
             curPtsX-mdlProps%epoch,nxtInt8,curPts2D(2:2*curNData:2,:),tyParamsMdl(psdNPl,1,BIN_COMPO_AM_ANG_POS_STD)) ! pars IS CONTIGUOUS
          else                         ! for AMHipN data
            call calcMdlFctWErr(int(curNData,i8b),MDL_FCT_DIM%AMAngPos,mdlProps%AMHipN%mdlFct,curPars2D,identInt8, & ! 2-d positions
             curPtsX-mdlProps%epoch,nxtInt8,curPts2D(2:2*curNData:2,:),tyParamsMdl(psdNPl,1,BIN_COMPO_AM_ANG_POS_STD)) ! pars IS CONTIGUOUS
          end if

          curFilNmCSVMdlResidVect=trim(fmtFilNmPlt(trim(FIL_NM_PART%residVect)//"_"//trim(dtTyDesc)))//FIL_EXT_CSV

          call writeCSV(trim(outDirCSV)//trim(curFilNmCSVMdlResidVect(1)),curPts2D(:,2),curPts2D(:,1))


          ! plot

          do i=1,curNData
            curModArr2(i)="ev ::"//trim(niceFmt(2*(i-1)))//"::"//trim(niceFmt(2*i-1))
          end do
          curStyModArr2="lt 2"
          curPltSymArr2=PLT_SYM%line

          curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"+"//trim(FIL_NM_PART%mdl)//"+"// &
           trim(FIL_NM_PART%residVect)//"_"//trim(dtTyDesc)//trim(strFold)))//trim(pltFilExt)

          call addGnupltCmds("",curFilNmPlt,trim(lblX),trim(lblY),"Data, model"// &
           trim(mrg(" (excluding parallatic displacement)","",dtTy == DT_TY%AMHipN))//", and residuals", &
           setFlg([FLG%rvX,FLG%eqAxR]),filNmInArr=[curFilNmCSVMdl,curFilNmCSVDtUncEll,curFilNmCSVMdlResidVect], &
           symArr=[PLT_SYM%line,curPltSymArr,curPltSymArr2], &
           modArr=[curModArr,curModArr2],styModArr=[curStyModArr,curStyModArr2])

        else

          ! add plotting commands
          curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"+"//trim(FIL_NM_PART%mdl)//"_"// &
           trim(dtTyDesc)//trim(strFold)))//trim(pltFilExt)
          call addGnupltCmds("",curFilNmPlt,trim(lblX),trim(lblY),"Data and model"//trim(strFoldTitle), &
           setFlg([FLG%rvX,FLG%eqAxR]),filNmInArr=[curFilNmCSVMdl,curFilNmCSVDtUncEll],symArr=[PLT_SYM%line,curPltSymArr], &
           modArr=curModArr,styModArr=curStyModArr)

        end if

      end if
    end subroutine


    subroutine plotAMAngPosPerCoord(allDt,dtTy,flags) ! plots AMAngPos data/model per coordinate
      implicit none
      integer :: &
       dtTy, &
       flags
      logical :: &
       onlyDt, &
       resid, &
       folded
      character(STR_LEN_STD) :: &
       dtTyDesc
      type(tyDtSetAMAngPos) :: &
       allDt
      intent(in) :: &
       allDt, &
       dtTy, &
       flags

      dtTyDesc=trim(DT_TY_NMS(dtTy))//trim(mrg("_as_"//trim(DT_TY_NMS(DT_TY%AMAngPos)),"",dtTy == DT_TY%AMHipN))
      onlyDt=getFlg(flags,FLG%onlyDt)
      resid=getFlg(flags,FLG%resid)
      folded=getFlg(flags,FLG%fold)

      if (folded) then
        curTPeri=mdlProps%epoch-curPars(pRoleByPl(1)%chi)/curPars(pRoleByPl(1)%f)
        tRge=tyIntvlDbl(curTPeri,curTPeri+curP)
        strFold="_"//FIL_NM_PART%fold
        strFoldTitle=", folded (T="//trim(niceFmt(tOfs+curTPeri))//" d, P="//trim(niceFmt(curP))//" d)"
        lblX="Orbital phase"
      else
        tRge=centralStretchIntvl(allDt%tRge,TM_RGE_STRETCH_FACT)
        strFold=""
        strFoldTitle=""
        lblX=dateDescAndUnitPlt
      end if


      curNData=allDt%N
      curNPts=curNData*merge(2,1,folded)
      if (allocated(curPtsX)) deallocate(curPtsX)
      if (allocated(curPts2D)) deallocate(curPts2D)
      if (allocated(curPtsUnc2D)) deallocate(curPtsUnc2D)
      allocate(curPtsX(curNPts),curPts2D(curNPts,2),curPtsUnc2D(curNPts,2),stat=allocStat) ! times, uncertainties, (delta, alpha·cos(delta))
      call testAllocErr(allocStat,22)

      if (resid) then
        curPts2D(:curNData,1)=allDt%resid(iNPlMd,:curNData)%x
        curPts2D(:curNData,2)=allDt%resid(iNPlMd,:curNData)%y
      else
        curPts2D(:curNData,1)=allDt%obs(:curNData)%x
        curPts2D(:curNData,2)=allDt%obs(:curNData)%y
      end if
      curPtsX(:curNData)=allDt%t(:curNData)
      if (dtTy == DT_TY%AMAngPos) then
        curPtsUnc2D(:curNData,1)=sqrt((allDt%unc(:curNData)%a*cos(allDt%unc(:curNData)%phi))**2+ &
         (allDt%unc(:curNData)%b*sin(allDt%unc(:curNData)%phi))**2)
        curPtsUnc2D(:curNData,2)=sqrt((allDt%unc(:curNData)%a*sin(allDt%unc(:curNData)%phi))**2+ &
         (allDt%unc(:curNData)%b*cos(allDt%unc(:curNData)%phi))**2)
      else
        curPtsUnc2D(:curNData,1)=-sin(allDt%unc(:curNData)%phi)*allDt%unc(:curNData)%b
        curPtsUnc2D(:curNData,2)=cos(allDt%unc(:curNData)%phi)*allDt%unc(:curNData)%b
      end if

      ! (fold)
      if (folded) then
        if (dtTy == DT_TY%AMHipN .AND. .NOT. resid) then
          call subtrConstChg(curPtsX(:curNData),curPts2D(:curNData,1),mdlProps%epoch,curPars(pRole%muDel)/YR_IN_D)
          call subtrConstChg(curPtsX(:curNData),curPts2D(:curNData,2),mdlProps%epoch,curPars(pRole%muAlCosDel)/YR_IN_D)
        end if
        call foldArr(curPtsX(:curNData),curPts2D(:curNData,1),curTPeri,curP,.FALSE.)
        call foldArr(curPtsX(:curNData),curPts2D(:curNData,2),curTPeri,curP,.FALSE.)
        call foldArr(curPtsX(:curNData),curPtsUnc2D(:curNData,1),curTPeri,curP,.FALSE.)
        call foldArr(curPtsX(:curNData),curPtsUnc2D(:curNData,2),curTPeri,curP,.TRUE.)
        call duplicateArr(curPtsX,curPts2D(:,1),.5_dp,.FALSE.)
        call duplicateArr(curPtsX,curPts2D(:,2),.5_dp,.FALSE.)
        call duplicateArr(curPtsX,curPtsUnc2D(:,1),.5_dp,.FALSE.)
        call duplicateArr(curPtsX,curPtsUnc2D(:,2),.5_dp,.TRUE.)
      else
        curPtsX=tOfs+curPtsX
      end if

      do i=1,2

        lblY=fmtObsvble(DT_TY%AMAngPos,i)
        curFilNmCSVDt=trim(fmtFilNmPlt(trim(mrg(FIL_NM_PART%resid,FIL_NM_PART%data,resid))//"_"//trim(dtTyDesc)// &
         trim(strFold)//"_"//OBSVBLE_NM(DT_TY%AMAngPos,i)))
        curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSVDt)//pltFilExt
        curFilNmCSVDt=trim(curFilNmCSVDt)//FIL_EXT_CSV
        call writeCSV(trim(outDirCSV)//trim(curFilNmCSVDt),curPtsX,curPts2D(:,i),curPtsUnc2D(:,i))
        if (resid .OR. onlyDt) call addGnupltCmds(curFilNmCSVDt,curFilNmPlt,lblX,lblY, &
         uCaseFst(trim(mrg(TTL_PART%resid,TTL_PART%data,resid)))//trim(strCenter)//trim(strFoldTitle),NO_FLAGS,sym=PLT_SYM%errBar)

      end do

      if (.NOT. (resid .OR. onlyDt)) then ! data+model (no residuals)

        curNPts=pltNPtsMdl*merge(2,1,folded)
        if (allocated(curPts2D)) deallocate(curPts2D)
        if (allocated(curPtsX)) deallocate(curPtsX)
        allocate(curPts2D(curNPts,2),curPtsX(curNPts),stat=allocStat)
        call testAllocErr(allocStat,23)

        curPtsX(:pltNPtsMdl)=sampTimes(pltNPtsMdl,tRge,.NOT. OPT_SCATTER,.NOT. OPT_WINDOW)

        ! sample function
        if (dtTy == DT_TY%AMAngPos) then
          call calcMdlFctWErr(int(pltNPtsMdl,i8b),MDL_FCT_DIM%AMAngPos,mdlProps%AMAngPos%mdlFct,curPars2D,identInt8, &
           curPtsX(:pltNPtsMdl)-mdlProps%epoch,nxtInt8,curPts2D(:pltNPtsMdl,:),tyParamsMdl(psdNPl,1,BIN_COMPO_AM_ANG_POS_STD)) ! pars IS CONTIGUOUS
        else                         ! for AMHipN data
          call calcMdlFctWErr(int(pltNPtsMdl,i8b),MDL_FCT_DIM%AMAngPos,mdlProps%AMHipN%mdlFct,curPars2D,identInt8, & ! 2-d positions
           curPtsX(:pltNPtsMdl)-mdlProps%epoch,nxtInt8,curPts2D(:pltNPtsMdl,:),tyParamsMdl(psdNPl,1,BIN_COMPO_AM_ANG_POS_STD)) ! pars IS CONTIGUOUS
        end if

        ! (fold)
        if (folded) then
          if (dtTy == DT_TY%AMHipN) then
            call subtrConstChg(curPtsX(:pltNPtsMdl),curPts2D(:pltNPtsMdl,1),mdlProps%epoch,curPars(pRole%muDel)/YR_IN_D)
            call subtrConstChg(curPtsX(:pltNPtsMdl),curPts2D(:pltNPtsMdl,2),mdlProps%epoch,curPars(pRole%muAlCosDel)/YR_IN_D)
          end if
          call foldArr(curPtsX(:pltNPtsMdl),curPts2D(:pltNPtsMdl,1),curTPeri,curP,.FALSE.)
          call foldArr(curPtsX(:pltNPtsMdl),curPts2D(:pltNPtsMdl,2),curTPeri,curP,.TRUE.)
          call duplicateArr(curPtsX,curPts2D(:,1),.5_dp,.FALSE.)
          call duplicateArr(curPtsX,curPts2D(:,2),.5_dp,.TRUE.)
        else
          curPtsX=tOfs+curPtsX
        end if

        ! write CSV output
        do i=1,2

          curFilNmCSVDt=trim(fmtFilNmPlt(trim(mrg(FIL_NM_PART%resid,FIL_NM_PART%data,resid))//"_"//trim(dtTyDesc) &
           //trim(strFold)//"_"//OBSVBLE_NM(DT_TY%AMAngPos,i)))//FIL_EXT_CSV
          curFilNmCSVMdl=trim(fmtFilNmPlt(trim(FIL_NM_PART%mdl)//"_"//trim(dtTyDesc)//trim(strFold)//"_"// &
           OBSVBLE_NM(DT_TY%AMAngPos,i)))//FIL_EXT_CSV
          curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%data)//"+"//trim(FIL_NM_PART%mdl)//"_"// &
           trim(dtTyDesc)//trim(strFold)//"_"//OBSVBLE_NM(DT_TY%AMAngPos,i)))//trim(pltFilExt)

          lblY=fmtObsvble(DT_TY%AMAngPos,i)
          call writeCSV(trim(outDirCSV)//trim(curFilNmCSVMdl),curPtsX,curPts2D(:,i))

          ! add plotting commands
          call addGnupltCmds("",curFilNmPlt,trim(lblX),trim(lblY),"Data and model"// &
           trim(mrg(" (excluding parallatic displacement)","",dtTy == DT_TY%AMHipN))//trim(strFoldTitle),NO_FLAGS, &
           filNmInArr=[curFilNmCSVMdl,curFilNmCSVDt],symArr=[PLT_SYM%line,PLT_SYM%errBar])
        end do

      end if
    end subroutine

  end subroutine plotDataAndModel




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine plotDistrNormResid
!
!   plots distributions of normalised residuals
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(MAX_LEN_PATH) :: &
     filNmCSV, &
     filNmPlt
    type(tyCoord2DDbl) :: &
     densEst(PLT_N_PTS%std)



    ! AMAngPos

    if (dtProps%nSetsOfTy(DT_TY%AMAngPos) > 0) then


      ! without jitter

      ! do KDE
      call kerDensEst1D(dens1DNorm,I_KER_1D%norm,dtPropsAMAngPos%normResid(:dtPropsAMAngPos%nRec)%woJitt, &
       dtPropsAMAngPos%normResidStdDev%woJitt, &
       tyIntvlDbl(minval(dtPropsAMAngPos%normResid(:dtPropsAMAngPos%nRec)%woJitt), &
       maxval(dtPropsAMAngPos%normResid(:dtPropsAMAngPos%nRec)%woJitt)),PLT_N_PTS%std,densEst,setFlg(FLG%xtndRgeX)) ! also sorts


      ! write CSV output
      filNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%woJitt)//"_"// &
       trim(DT_TY_NMS(DT_TY%AMAngPos))))//FIL_EXT_CSV
      call writeCSV(trim(outDirCSV)//trim(filNmCSV),densEst%x,densEst%y)


      ! add plotting commands
      filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%woJitt)//"_"// &
       trim(DT_TY_NMS(DT_TY%AMAngPos))))//trim(pltFilExt)
      call addGnupltCmds(filNmCSV,filNmPlt,uCaseFst(trim(TTL_PART%normResid))//" "//OBSVBLE_DESC%AMAngPos, &
       uCaseFst(trim(TTL_PART%probDens)),uCaseFst(trim(TTL_PART%distr))//" of "//trim(TTL_PART%normResid)//" "// &
       OBSVBLE_DESC%AMAngPos,NO_FLAGS)

      
        ! with jitter

      if (priUBnd(pRole%tauPlus) > 0) then

        ! do KDE
        call kerDensEst1D(dens1DNorm,I_KER_1D%norm,dtPropsAMAngPos%normResid(:dtPropsAMAngPos%nRec)%wJitt, &
         dtPropsAMAngPos%normResidStdDev%wJitt, &
         tyIntvlDbl(minval(dtPropsAMAngPos%normResid(:dtPropsAMAngPos%nRec)%wJitt), &
         maxval(dtPropsAMAngPos%normResid(:dtPropsAMAngPos%nRec)%wJitt)),PLT_N_PTS%std,densEst,setFlg(FLG%xtndRgeX)) ! also sorts


        ! write CSV output
        filNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%wJitt)//"_"// &
         trim(DT_TY_NMS(DT_TY%AMAngPos))))//FIL_EXT_CSV
        call writeCSV(trim(outDirCSV)//trim(filNmCSV),densEst%x,densEst%y)


        ! add plotting commands
        filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%wJitt)//"_"// &
         trim(DT_TY_NMS(DT_TY%AMAngPos))))//trim(pltFilExt)
        call addGnupltCmds(filNmCSV,filNmPlt,uCaseFst(trim(TTL_PART%normResid))//" "//OBSVBLE_DESC%AMAngPos, &
         uCaseFst(trim(TTL_PART%probDens)),uCaseFst(trim(TTL_PART%distr))//" of "//trim(TTL_PART%normResid)//" "// &
         OBSVBLE_DESC%AMAngPos,NO_FLAGS)

      end if

    end if


    ! RV

    if (dtProps%nSetsOfTy(DT_TY%RV) > 0) then


      ! without jitter

      ! do KDE
      call kerDensEst1D(dens1DNorm,I_KER_1D%norm,dtPropsRV%normResid(:dtPropsRV%nRec)%woJitt, &
       dtPropsRV%normResidStdDev%woJitt, &
       tyIntvlDbl(minval(dtPropsRV%normResid(:dtPropsRV%nRec)%woJitt),maxval(dtPropsRV%normResid(:dtPropsRV%nRec)%woJitt)), &
       PLT_N_PTS%std,densEst,setFlg(FLG%xtndRgeX)) ! also sorts


      ! write CSV output
      filNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%woJitt)//"_"// &
       trim(DT_TY_NMS(DT_TY%RV))))//FIL_EXT_CSV
      call writeCSV(trim(outDirCSV)//trim(filNmCSV),densEst%x,densEst%y)


      ! add plotting commands
      filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%woJitt)//"_"// &
       trim(DT_TY_NMS(DT_TY%RV))))//trim(pltFilExt)
      call addGnupltCmds(filNmCSV,filNmPlt,uCaseFst(trim(TTL_PART%normResid))//" "//OBSVBLE_DESC%RV, &
       uCaseFst(trim(TTL_PART%probDens)),uCaseFst(trim(TTL_PART%distr))//" of "//trim(TTL_PART%normResid)//" "//OBSVBLE_DESC%RV, &
       NO_FLAGS)


      ! with jitter

      if (priUBnd(pRole%sgmPlus) > 0) then
      
        ! do KDE
        call kerDensEst1D(dens1DNorm,I_KER_1D%norm,dtPropsRV%normResid(:dtPropsRV%nRec)%wJitt, &
        dtPropsRV%normResidStdDev%wJitt, &
        tyIntvlDbl(minval(dtPropsRV%normResid(:dtPropsRV%nRec)%wJitt),maxval(dtPropsRV%normResid(:dtPropsRV%nRec)%wJitt)), &
        PLT_N_PTS%std,densEst,setFlg(FLG%xtndRgeX)) ! also sorts


        ! write CSV output
        filNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%wJitt)//"_"// &
         trim(DT_TY_NMS(DT_TY%RV))))//FIL_EXT_CSV
        call writeCSV(trim(outDirCSV)//trim(filNmCSV),densEst%x,densEst%y)


        ! add plotting commands
        filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%wJitt)//"_"// &
         trim(DT_TY_NMS(DT_TY%RV))))//trim(pltFilExt)
        call addGnupltCmds(filNmCSV,filNmPlt,uCaseFst(trim(TTL_PART%normResid))//" "//OBSVBLE_DESC%RV, &
         uCaseFst(trim(TTL_PART%probDens)),uCaseFst(trim(TTL_PART%distr))//" of "//trim(TTL_PART%normResid)//" "// &
         trim(OBSVBLE_DESC%RV)//" ("//trim(TTL_PART%wJitt)//")",NO_FLAGS)

      end if

    end if


    ! AMHipN

    if (dtProps%nSetsOfTy(DT_TY%AMHipN) > 0) then


      ! without jitter

      ! do KDE
      call kerDensEst1D(dens1DNorm,I_KER_1D%norm,dtPropsAMHipN%normResid(:dtPropsAMHipN%nRec)%woJitt, &
       dtPropsAMHipN%normResidStdDev%woJitt, &
       tyIntvlDbl(minval(dtPropsAMHipN%normResid(:dtPropsAMHipN%nRec)%woJitt), &
       maxval(dtPropsAMHipN%normResid(:dtPropsAMHipN%nRec)%woJitt)),PLT_N_PTS%std,densEst,setFlg(FLG%xtndRgeX)) ! also sorts


      ! write CSV output
      filNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%woJitt)//"_"// &
       trim(DT_TY_NMS(DT_TY%AMHipN))))//FIL_EXT_CSV
      call writeCSV(trim(outDirCSV)//trim(filNmCSV),densEst%x,densEst%y)


      ! add plotting commands
      filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%woJitt)//"_"// &
       trim(DT_TY_NMS(DT_TY%AMHipN))))//trim(pltFilExt)
      call addGnupltCmds(filNmCSV,filNmPlt,uCaseFst(trim(TTL_PART%normResid))//" "//OBSVBLE_DESC%AMHipN, &
       uCaseFst(trim(TTL_PART%probDens)),uCaseFst(trim(TTL_PART%distr))//" of "//trim(TTL_PART%normResid)//" "// &
       OBSVBLE_DESC%AMHipN,NO_FLAGS)


      ! with jitter

      if (priUBnd(pRole%tauPlus) > 0) then

        ! do KDE
        call kerDensEst1D(dens1DNorm,I_KER_1D%norm,dtPropsAMHipN%normResid(:dtPropsAMHipN%nRec)%wJitt, &
        dtPropsAMHipN%normResidStdDev%wJitt, &
        tyIntvlDbl(minval(dtPropsAMHipN%normResid(:dtPropsAMHipN%nRec)%wJitt), &
        maxval(dtPropsAMHipN%normResid(:dtPropsAMHipN%nRec)%wJitt)),PLT_N_PTS%std,densEst,setFlg(FLG%xtndRgeX)) ! also sorts


        ! write CSV output
        filNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%wJitt)//"_"// &
         trim(DT_TY_NMS(DT_TY%AMHipN))))//FIL_EXT_CSV
        call writeCSV(trim(outDirCSV)//trim(filNmCSV),densEst%x,densEst%y)


        ! add plotting commands
        filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)//"_"//trim(FIL_NM_PART%wJitt)//"_"// &
         trim(DT_TY_NMS(DT_TY%AMHipN))))//trim(pltFilExt)
        call addGnupltCmds(filNmCSV,filNmPlt,uCaseFst(trim(TTL_PART%normResid))//" "//OBSVBLE_DESC%AMHipN, &
         uCaseFst(trim(TTL_PART%probDens)),uCaseFst(trim(TTL_PART%distr))//" of "//trim(TTL_PART%normResid)//" "// &
         trim(OBSVBLE_DESC%AMHipN)//" ("//trim(TTL_PART%wJitt)//")",NO_FLAGS)

      end if

    end if


    ! all data

    if (isUndef(dtProps%onlyDtTy)) then

        ! without jitter

      ! do KDE
      call kerDensEst1D(dens1DNorm,I_KER_1D%norm,dtProps%normResid(:dtProps%nRec)%woJitt,dtProps%normResidStdDev%woJitt, &
       tyIntvlDbl(minval(dtProps%normResid(:dtProps%nRec)%woJitt),maxval(dtProps%normResid(:dtProps%nRec)%woJitt)),PLT_N_PTS%std, &
       densEst,setFlg(FLG%xtndRgeX)) ! also sorts


      ! write CSV output
      filNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)))//"_"//trim(FIL_NM_PART%woJitt)//FIL_EXT_CSV
      call writeCSV(trim(outDirCSV)//trim(filNmCSV),densEst%x,densEst%y)


      ! add plotting commands
      filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)))//"_"//trim(FIL_NM_PART%woJitt)//trim(pltFilExt)
       call addGnupltCmds(filNmCSV,filNmPlt,uCaseFst(TTL_PART%normResids),uCaseFst(trim(TTL_PART%probDens)), &
        uCaseFst(trim(TTL_PART%distr))//" of "//TTL_PART%normResids,NO_FLAGS)

        ! with jitter

      if ((isDef(pRole%tauPlus) .AND. priUBnd(pRole%tauPlus) > 0) .OR. (isDef(pRole%sgmPlus) .AND. priUBnd(pRole%sgmPlus) > 0)) then
      
        ! do KDE
        call kerDensEst1D(dens1DNorm,I_KER_1D%norm,dtProps%normResid(:dtProps%nRec)%wJitt,dtProps%normResidStdDev%wJitt, &
         tyIntvlDbl(minval(dtProps%normResid(:dtProps%nRec)%wJitt),maxval(dtProps%normResid(:dtProps%nRec)%wJitt)),PLT_N_PTS%std, &
         densEst,setFlg(FLG%xtndRgeX)) ! also sorts


        ! write CSV output
        filNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)))//"_"//trim(FIL_NM_PART%wJitt)//FIL_EXT_CSV
        call writeCSV(trim(outDirCSV)//trim(filNmCSV),densEst%x,densEst%y)


        ! add plotting commands
        filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%distrNormResid)))//"_"//trim(FIL_NM_PART%wJitt)//trim(pltFilExt)
         call addGnupltCmds(filNmCSV,filNmPlt,uCaseFst(TTL_PART%normResids),uCaseFst(trim(TTL_PART%probDens)), &
          uCaseFst(trim(TTL_PART%distr))//" of "//TTL_PART%normResids,NO_FLAGS)

      end if

    end if


  end subroutine plotDistrNormResid




  !==============================================[ POSTERIOR INFERENCE IN GENERAL ]=================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcSummaries(chain)
!
!   calculates various posterior summaries
!
!   Arguments:
!     chain                          MCMC chain
!
!   does not sort chain
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p, &
     p2, &
     p3, &
     p4, &
     allocStat, &
     iPl, &
     q, &
     trN1st, &
     trS, &
     j
    integer(i8b) :: &
     i, &
     iMin, &
     iMax, &
     iMaxDQ, &
     iDQ, &
     nLinks, &
     burnInLen
    real(dp) :: &
     trParFrq, &
     trParEcc, &
     trEA, &
     trSinEA, &
     trCosEA, &
     trSinOm, &
     trCosOm, &
     trTmp, &
     trTmp2, &
     trTmp3, &
     trTmShift, &
     trTmpA(-1:1)
    logical :: &
     curASinISkipped, &
     curMMinSkipped, &
     curMSkipped, &
     curRhoSkipped, &
     curM1Skipped
    type(tyMCMCChain) :: &
     chain, &
     DQChainTmp, &
     DQChainTmp2
    intent(inout) :: &
     chain


    call startSubsect(UNIT%cons+UNIT%log,"Calculating summaries")

    nLinks=chain%length
    burnInLen=chain%burnInLen
    iMin=burnInLen+1

    call undef(parEst(iNPlMd,1:curNPars)%mMode)
    call undef(parEst(iNPlMd,1:curNPars)%KDEBandw)
    
    if (nLinks-burnInLen >= N_SAMP_TRIANG_KER_MAPO_MIN) then
      maPoKer=>dens1DTriang
      iMaPoKer=I_KER_1D%triang
    else
      maPoKer=>dens1DNorm
      iMaPoKer=I_KER_1D%norm
    end if


    call startSubsect(UNIT%cons+UNIT%log,"Parameters")

    call startSubsect(UNIT%cons+UNIT%log,"Scalar summaries")
    call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
    do p=1,curNPars
      if (parFixed(p)) then
        parEst(iNPlMd,p)%mean%val=priLBnd(p)
        parEst(iNPlMd,p)%mean%unc=0
        parEst(iNPlMd,p)%mMode%val=priLBnd(p)
        parEst(iNPlMd,p)%mMode%unc=0
        parEst(iNPlMd,p)%sampStdDev=0
        parEst(iNPlMd,p)%cov=0
        call undef(parEst(iNPlMd,p)%corrCoeff)
        call undef(parEst(iNPlMd,:curNPars)%corrCoeff(p))
      else if (parInfc(p)) then
        parEst(iNPlMd,p)%mean%val=arrMean(chain%links(burnInLen+1:nLinks,p))
        parEst(iNPlMd,p)%cov(p)=arrVar(chain%links(burnInLen+1:nLinks,p),parEst(iNPlMd,p)%mean%val)
        parEst(iNPlMd,p)%corrCoeff(p)=1._dp
        parEst(iNPlMd,p)%sampStdDev=sqrt(parEst(iNPlMd,p)%cov(p))
        if (filOutputPlt) call calcAutocorr(iNPlMd,chain,p,parEst(iNPlMd,p)%mean%val,parEst(iNPlMd,p)%cov(p))
      else
        call undef(parEst(iNPlMd,p)%mean)
        call undef(parEst(iNPlMd,p)%sampStdDev)
      end if
      if (isDef(fracIndep(p))) then
        parEst(iNPlMd,p)%nIndepSamp=int(round(fracIndep(p)*(nLinks-burnInLen),1._dp))
        if (isDef(parEst(iNPlMd,p)%sampStdDev)) then
          parEst(iNPlMd,p)%mean%unc=parEst(iNPlMd,p)%sampStdDev/sqrt(fracIndep(p)*(nLinks-burnInLen))
        else
          call undef(parEst(iNPlMd,p)%mean%unc)
        end if
      else
        call undef(parEst(iNPlMd,p)%nIndepSamp)
        call undef(parEst(iNPlMd,p)%mean%unc)
      end if
      if (parInfc(p)) call printPrgrs([VAL_UNDEF_INT],nTot=curNParsInfc)
    end do                           ! p=1,curNPars
    call endSubsect(UNIT%cons+UNIT%log) ! Scalar summaries

    call startSubsect(UNIT%cons+UNIT%log,"Covariances and correlations")
    if (curNParsInfc > 1) then
      call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
      do p=1,curNPars-1
        if (parInfc(p)) then
          do p2=p+1,curNPars
            if (parInfc(p2)) then
              parEst(iNPlMd,p)%cov(p2)=arrCov(chain%links(burnInLen+1:nLinks,p),chain%links(burnInLen+1:nLinks,p2), &
               parEst(iNPlMd,p)%mean%val,parEst(iNPlMd,p2)%mean%val)
              parEst(iNPlMd,p)%corrCoeff(p2)=parEst(iNPlMd,p)%cov(p2)/(parEst(iNPlMd,p)%sampStdDev*parEst(iNPlMd,p2)%sampStdDev)
              if (parEst(iNPlMd,p)%corrCoeff(p2) > 1._dp) call undef(parEst(iNPlMd,p)%corrCoeff(p2)) ! may happen due to numerics
              parEst(iNPlMd,p2)%cov(p)=parEst(iNPlMd,p)%cov(p2)
              parEst(iNPlMd,p2)%corrCoeff(p)=parEst(iNPlMd,p)%corrCoeff(p2)
              call printPrgrs([VAL_UNDEF_INT],nTot=int(curNParsInfc*(curNParsInfc-1)/2.))
            end if
          end do
        end if
      end do
    else
      call printMsg(UNIT%cons+UNIT%log,"Skipped (inference refers to 1 parameter only)")
    end if
    call endSubsect(UNIT%cons+UNIT%log) ! Covariances and correlations

    call endSubsect(UNIT%cons+UNIT%log) ! Parameters


    if (usrDQs(iNPlMd)) then

      call startSubsect(UNIT%cons+UNIT%log,"Derived quantities")

      allocate(DQChainTmp%links(nLinks-burnInLen,1),DQChainTmp2%links(nLinks-burnInLen,1),stat=allocStat)
      call testAllocErr(allocStat,7)

      ! use 1st after-burn-in sample to calc. DQs corresponding to mode
      chain%links(burnInLen+1,:curNPars)=parEst(iNPlMd,:curNPars)%mode
      DQChainTmp%burnInLen=1
      DQChainTmp%length=nLinks-burnInLen-1
      DQChainTmp2%burnInLen=1
      DQChainTmp2%length=nLinks-burnInLen-1

      call undef(parEstDQ(iNPlMd,1:curNDQs)%mean%unc)
      call undef(parEstDQ(iNPlMd,1:curNDQs)%mMode)
      call undef(parEstDQ(iNPlMd,1:curNDQs)%KDEBandw)


      call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
      do q=1,curNDQs
        iPl=iPlByDQ(q)
        iMax=nLinks
        iMaxDQ=nLinks-burnInLen
        DQChainTmp%sorted(1)=.FALSE.

        select case (gIDQs(q))
        case (G_I_DQ_ROLE%P)         ! P = 1/f
          p=merge(pRole%f,pRoleByPl(iPl)%f,mdlProps%bin) ! f[_iPl]
          call chkFixed(parFixed(p))
          DQChainTmp%links(1:iMaxDQ,1)=1._dp/chain%links(burnInLen+1:burnInLen+iMaxDQ,p)

        case (G_I_DQ_ROLE%T)         ! T = t_ref,chi-chi/f
          ! influences folding of data & model!
          p=merge(pRole%f,pRoleByPl(iPl)%f,mdlProps%bin) ! f[_iPl]
          p2=merge(pRole%chi,pRoleByPl(iPl)%chi,mdlProps%bin) ! chi[_iPl]
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          DQChainTmp%links(1:iMaxDQ,1)=tOfs+mdlProps%epoch- &
           chain%links(burnInLen+1:burnInLen+iMaxDQ,p2)/chain%links(burnInLen+1:burnInLen+iMaxDQ,p)

        case (G_I_DQ_ROLE%mMin)      ! m_min (numerically)
          p=pRoleByPl(iPl)%K
          p2=pRoleByPl(iPl)%f
          curMMinSkipped=parFixed(p) .AND. parFixed(p2)
          call chkFixed(curMMinSkipped)
          call calcMMin(chain,iPl,DQChainTmp,iMin,iMax) ! uses parallelism
          if (curMMinSkipped) DQChainTmp%links(:,1)=DQChainTmp%links(1,1)
          DQChainTmp2%links(:,1)=DQChainTmp%links(:,1)

        case (G_I_DQ_ROLE%aSinI)     ! a·sin(i) [AU] = K[m/s] / (2·pi· f [1/d]) · (d/s)/(AU/m)   [04.04.2010/1]
          p=pRoleByPl(iPl)%f
          p2=pRoleByPl(iPl)%K
          curASinISkipped=parFixed(p) .AND. parFixed(p2)
          call chkFixed(curASinISkipped)
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=chain%links(i,p2)/(TWO_PI*chain%links(i,p)*AU_IN_M)*D_IN_S
          end do
          !$OMP end parallel do
          if (curASinISkipped) DQChainTmp%links(:,1)=DQChainTmp%links(1,1)

        case (G_I_DQ_ROLE%aPMax)     ! a_p,max [AU] = M [M_J] / (m_min [M_J]) · a·sin(i) [AU]   [05.04.2010/2]
          call chkFixed(curASinISkipped .AND. curMMinSkipped)
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=mdlProps%mS/DQChainTmp2%links(iDQ,1)*DQChainTmp%links(iDQ,1) ! uses DQChainTmp from a·sin(i)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%KAlt)     ! K_alt = K/sqrt(1-e²)
          p=pRoleByPl(iPl)%e
          p2=pRoleByPl(iPl)%K
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=chain%links(i,p2)/sqrt(1._dp-chain%links(i,p)**2)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%KAlt1)    ! K_alt,1 = K_1/sqrt(1-e²)
          p=pRole%e
          p2=pRole%K1
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=chain%links(i,p2)/sqrt(1._dp-chain%links(i,p)**2)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%KAlt2)    ! K_alt,2 = K_2/sqrt(1-e²)
          p=pRole%e
          p2=pRole%K2
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=chain%links(i,p2)/sqrt(1._dp-chain%links(i,p)**2)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%rho)       ! rho = m_2/m_1 = K_1/K_2
          p=pRole%K1
          p2=pRole%K2
          curRhoSkipped=parFixed(p) .AND. parFixed(p2)
          call chkFixed(curRhoSkipped)
          DQChainTmp%links(1:iMaxDQ,1)=chain%links(burnInLen+1:burnInLen+iMaxDQ,p)/ &
           chain%links(burnInLen+1:burnInLen+iMaxDQ,p2)
          DQChainTmp2%links(1:iMaxDQ,1)=DQChainTmp%links(1:iMaxDQ,1) ! needed for m_2, a_2

        case (G_I_DQ_ROLE%m1)        ! m_1 = K_2 · (K_1 + K_2)² / (2·pi·G·f·sin³(i))
          p=pRole%K1
          p2=pRole%K2
          p3=pRole%f
          p4=pRole%i
          curM1Skipped=((parFixed(p) .AND. parFixed(p2) .AND. parFixed(p3) .AND. parFixed(p4)) .OR. &
           parFixed(p4) .AND. priLBnd(p4) == 0)
          call chkFixed(curM1Skipped)
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=chain%links(i,p2)*(chain%links(i,p)+chain%links(i,p2))**2/ &
             (TWO_PI*GRAVC_MKS*chain%links(i,p3)*sin(chain%links(i,p4))**3)*D_IN_S/M_SUN_IN_KG
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%m2)        ! m_2 = rho · m_1
          call chkFixed(curRhoSkipped .AND. curM1Skipped)
          DQChainTmp%links(1:iMaxDQ,1)=DQChainTmp2%links(1:iMaxDQ,1)*DQChainTmp%links(1:iMaxDQ,1) ! uses DQChainTmp from m_1

        case (G_I_DQ_ROLE%d)         ! d [pc] = 1 / (pi [arcsec])
          p=pRole%varpi
          call chkFixed(parFixed(p))
          DQChainTmp%links(1:iMaxDQ,1)=1._dp/chain%links(burnInLen+1:burnInLen+iMaxDQ,p)

        case (G_I_DQ_ROLE%aRel)      ! a_rel = (K_1+K_2)/(2·pi·f·sin(i))
          p=pRole%K1
          p2=pRole%K2
          p3=pRole%f
          p4=pRole%i
          call chkFixed(parFixed(p) .AND. parFixed(p2) .AND. parFixed(p3) .AND. parFixed(p4))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=(chain%links(i,p)+chain%links(i,p2))/(TWO_PI*chain%links(i,p3)* &
             sin(chain%links(i,p4)))*D_IN_S/AU_IN_M
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%a1)        ! a_1 = K_1/(K_1+K_2)·a_rel
          p=pRole%K1
          p2=pRole%K2
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=chain%links(i,p)/(chain%links(i,p)+chain%links(i,p2))*DQChainTmp%links(iDQ,1) ! uses DQChainTmp from a_rel
          end do
          !$OMP end parallel do


        case (G_I_DQ_ROLE%a2)        ! a_2 = a_1/rho
          p=pRole%K1
          p2=pRole%K2
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          DQChainTmp%links(1:iMaxDQ,1)=DQChainTmp%links(1:iMaxDQ,1)/DQChainTmp2%links(1:iMaxDQ,1) ! uses DQChainTmp from a_1

        case (G_I_DQ_ROLE%thetaR)    ! theta_r = modulo(omega+2·pi·chi,2·pi) [13.05.2011/1]
          p=pRoleByPl(iPl)%omegaPeri
          p2=pRoleByPl(iPl)%chi
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=modulo(chain%links(i,p)+TWO_PI*chain%links(i,p2),TWO_PI)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%thetaSecR) ! theta_r = modulo(omega_2+2·pi·chi,2·pi) [13.05.2011/1]
          p=pRole%omega2
          p2=pRole%chi
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=modulo(chain%links(i,p)+TWO_PI*chain%links(i,p2),TWO_PI)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%tT1)       ! 1st potential transit time, cf. [13.05.2011/6]
          p=pRoleByPl(1)%f
          p2=pRoleByPl(iPl)%e
          p3=pRoleByPl(iPl)%chi
          p4=pRoleByPl(iPl)%omegaPeri
          trN1st=0

          call chkFixed(parFixed(pRoleByPl(iPl)%f) .AND. parFixed(pRoleByPl(iPl)%e) .AND. parFixed(pRoleByPl(iPl)%chi) .AND. &
           parFixed(pRoleByPl(iPl)%omegaPeri))

          !$OMP parallel do &
           !$OMP private(i,trParFrq,trParEcc,trSinOm,trCosOm,trEA,trSinEA,trCosEA,trTmp,trTmp2,trTmp3,trTmShift,trTmpA,trS) &
           !$OMP reduction(+:trN1st)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ

            trParFrq=chain%links(i,p)
            trParEcc=chain%links(i,p2)
            trTmp=sqrt(1._dp-trParEcc**2)

            trSinOm=sin(chain%links(i,p4))
            trCosOm=cos(chain%links(i,p4))

            trEA=atan(trTmp*trCosOm/trSinOm)
            trSinEA=sin(trEA)
            trCosEA=cos(trEA)
            trTmp2=chain%links(i,p3)-(trEA-trParEcc*trSinEA)/TWO_PI ! chi-(E-e·sin(E))/(2·pi)

            trTmp3=trParEcc*trSinEA/PI ! e·sin(E)/pi

            ! temp. transit time
            trTmShift=(trTmp3-.5_dp)/trParFrq ! shift from original to earlier transit, cf. [14.05.2011/1]
            trTmpA(0)=mdlProps%epoch+( &    ! original time
             !
             ceiling((transTmStart-mdlProps%epoch)*trParFrq+trTmp2) & ! N
             !
             -trTmp2)/trParFrq
            trTmpA(-1)=trTmpA(0)+trTmShift ! shifted time
            trTmpA(1)=trTmpA(0)+(trTmp3+.5_dp)/trParFrq ! shifted from original to later transit

            trS=ceiling(sign(.5_dp,trTmpA(-1)/transTmStart-1._dp)) ! switch for original vs. shifted transit time (1: shifted; 0: original)
            DQChainTmp%links(iDQ,1)=tOfs+trTmpA(-trS)
            DQChainTmp2%links(iDQ,1)=tOfs+trTmpA(-trS+1) ! value of t_t2

            trN1st=trN1st+abs(trS- & ! correct for potential shift (i.e. don't count if trTmShift > 0), cf. [14.05.2011/1]
             ceiling(sign(.5_dp,trSinEA*((1._dp-trParEcc*trCosEA)*(trSinOm*trCosEA/trSinEA+trTmp*trCosOm)+ & ! decide between primary/secondary transit, cf. [13.05.2011/5]
             trParEcc*(trSinEA*trSinOm-trTmp*trCosEA*trCosOm)))))

          end do
          !$OMP end parallel do

          prob1stTransIsPri=1._dp*trN1st/iMaxDQ

        case (G_I_DQ_ROLE%tT2)       ! 2nd potential transit time, cf. [13.05.2011/6]
          p=pRoleByPl(iPl)%f
          p2=pRoleByPl(iPl)%e

          call chkFixed(parFixed(pRoleByPl(iPl)%f) .AND. parFixed(pRoleByPl(iPl)%e) .AND. parFixed(pRoleByPl(iPl)%chi) .AND. &
           parFixed(pRoleByPl(iPl)%omegaPeri))

          DQChainTmp%links(1:iMaxDQ,1)=DQChainTmp2%links(1:iMaxDQ,1)

        case (G_I_DQ_ROLE%omSum)     ! modulo(Omega+omega,2·pi) (important for nearly face-on orbits with i>~0)
          p=pRole%OmegaNode
          p2=pRoleByPl(iPl)%omegaPeri
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=mod(chain%links(i,p)+chain%links(i,p2),TWO_PI)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%omDiff)    ! modulo(Omega-omega+2·pi,2·pi) (important for nearly face-on orbits with i<~pi)
          p=pRole%OmegaNode
          p2=pRoleByPl(iPl)%omegaPeri
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=mod(chain%links(i,p)-chain%links(i,p2)+TWO_PI,TWO_PI)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%om2Sum)    ! modulo(Omega+omega_2,2·pi) (important for nearly face-on orbits with i>~0)
          p=pRole%OmegaNode
          p2=pRole%omega2
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=mod(chain%links(i,p)+chain%links(i,p2),TWO_PI)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%om2Diff)   ! modulo(Omega-omega_2+2·pi,2·pi) (important for nearly face-on orbits with i<~pi)
          p=pRole%OmegaNode
          p2=pRole%omega2
          call chkFixed(parFixed(p) .AND. parFixed(p2))
          !$OMP parallel do private(i)
          do iDQ=1,iMaxDQ
            i=burnInLen+iDQ
            DQChainTmp%links(iDQ,1)=mod(chain%links(i,p)-chain%links(i,p2)+TWO_PI,TWO_PI)
          end do
          !$OMP end parallel do

        case (G_I_DQ_ROLE%m)      ! m (numerically)
          p=pRoleByPl(iPl)%K
          p2=pRoleByPl(iPl)%f
          p3=pRole%i
          curMSkipped=parFixed(p) .AND. parFixed(p2) .AND. parFixed(p3)
          call chkFixed(curMSkipped)
          call calcMMin(chain,iPl,DQChainTmp,iMin,iMax,.TRUE.) ! uses parallelism
          if (curMSkipped) DQChainTmp%links(:,1)=DQChainTmp%links(1,1)
          DQChainTmp2%links(:,1)=DQChainTmp%links(:,1)

        end select


        parEstDQ(iNPlMd,q)%mode=DQChainTmp%links(1,1)

        if (DQSkipped(q)) then

          parEstDQ(iNPlMd,q)%HPDI%lBnd=DQChainTmp%links(1,1)
          parEstDQ(iNPlMd,q)%HPDI%uBnd=DQChainTmp%links(1,1)
          parEstDQ(iNPlMd,q)%median=DQChainTmp%links(1,1)
          parEstDQ(iNPlMd,q)%mean%val=DQChainTmp%links(1,1)
          parEstDQ(iNPlMd,q)%mMode%val=DQChainTmp%links(1,1)
          parEstDQ(iNPlMd,q)%sampStdDev=0

        else
          call calcMedHPDI(DQChainTmp%links(:,1),DQChainTmp%sorted(1),parEstDQ(iNPlMd,q)%HPDI,parEstDQ(iNPlMd,q)%nInHPDI, &
           parEstDQ(iNPlMd,q)%median,DQSampRge(q))             ! also sorts

          parEstDQ(iNPlMd,q)%mean%val=arrMean(DQChainTmp%links(:,1))
          parEstDQ(iNPlMd,q)%sampStdDev=sqrt(arrVar(DQChainTmp%links(:,1),parEstDQ(iNPlMd,q)%mean%val))
          maPoLogSclDQ(q)=G_DQ_LOG_SCL(gIDQs(q)) .AND. &
           (DQSampRge(q)%lBnd == 0._dp .OR. log10(intvlBoundsRatio(DQSampRge(q))) >= LOG_BOUNDS_RATIO_LOG_SCL_MIN)


          if (filOutputPlt .AND. .NOT. usrPeriodogram) then

            ! range for zoomed mapo.
            if (usrZmRge(iNPlMd)) then
              DQZmRge(q)=pltRge( &
               maxDQRge(q,iPl,zmRgeAllPars(iNPlMd,:curNPars)%lBnd,zmRgeAllPars(iNPlMd,:curNPars)%uBnd,DQSampRge(:curNDQs)), &
               maPoLogSclDQ(q))
            else
              DQZmRge(q)=parEstDQ(iNPlMd,q)%HPDI(iStdHPDI(I_STD_HPDI_MA_PO_ZM))
            end if

            call calcMargPostr(DQChainTmp,1,DQSampRge(q),DQZmRge(q),parEstDQ(iNPlMd,q), &
             maPoDQPts(q,:pltNPts),maPoDQPtsZm(q,:pltNPts),maPoLogSclDQ(q))

          end if                       ! (DQSkipped(q))

        end if                       ! (DQSkipped(q))

        call printPrgrs([q],[1],[curNDQs])

      end do                         ! q=1,curNDQs
      call endSubsect(UNIT%cons+UNIT%log) ! Derived quantities

    end if                           ! (usrDQs(iNPlMd))



    if (nNPlMds > 1) then
    
      block
      
        integer :: &
         i0, &
         i, &
         j, &
         k, &
         l, &
         nEvSamp
        real(qp) :: &
         allEvSamp(2*MDL_SEL_N_VT)
        

        call startSubsect(UNIT%cons+UNIT%log,"Evidence")

        call undef(mdlSelEvVT(iNPlMd,:))
        nEvSamp=0

        call startSubsect(UNIT%cons+UNIT%log,"VTA")
        
        do j=1,MDL_SEL_N_VT

          if (isUndef(mdlSelEvVT(iNPlMd,j))) call calcEvVT(chain,mdlSelEvVT(iNPlMd,j),mdlSelEvSubVT(iNPlMd,j),MDL_SEL_VT_N_PTS(j))
          if (isDef(mdlSelEvVT(iNPlMd,j))) then
            call inc(nEvSamp)
            allEvSamp(nEvSamp)=mdlSelEvVT(iNPlMd,j)
          end if
          if (isDef(mdlSelEvSubVT(iNPlMd,j))) then
            call inc(nEvSamp)
            allEvSamp(nEvSamp)=mdlSelEvSubVT(iNPlMd,j)%val
          end if

          call printPrgrs([j],[1],[MDL_SEL_N_VT])

        end do

        call endSubsect(UNIT%cons+UNIT%log) ! VTA
        
        if (nEvSamp > 1) then
          mdlSelEv(iNPlMd)%val=arrMean(allEvSamp(:nEvSamp))
          mdlSelEv(iNPlMd)%unc=sqrt(arrVar(allEvSamp(:nEvSamp),mdlSelEv(iNPlMd)%val))
        else
          call undef(mdlSelEv(iNPlMd))
        end if

        call endSubsect(UNIT%cons+UNIT%log) ! Evidence
      
      
        ! Bayes factors B_i,j = p(D|mode i) / p(D|mode j)

        if (iNPlMd == nNPlMds) then
          do i=2,nNPlMds
            do j=1,i-1

              where (isDef(mdlSelEvVT(i,:)) .AND. isDef(mdlSelEvVT(j,:)) .AND. mdlSelEvVT(j,:) > 0._qp)
                mdlSelBFVT(i,j,:)=mdlSelEvVT(i,:)/mdlSelEvVT(j,:)
              elsewhere
                mdlSelBFVT(i,j,:)=VAL_UNDEF_QP
              end where

              where (isDef(mdlSelEvSubVT(i,:)) .AND. isDef(mdlSelEvSubVT(j,:)) .AND. mdlSelEvSubVT(j,:)%val > 0._qp)
                mdlSelBFSubVT(i,j,:)%val=mdlSelEvSubVT(i,:)%val/mdlSelEvSubVT(j,:)%val
                mdlSelBFSubVT(i,j,:)%unc=mdlSelBFSubVT(i,j,:)%val* &
                 sqrt(mdlSelEvSubVT(i,:)%unc**2+mdlSelEvSubVT(j,:)%unc**2)
              elsewhere
                mdlSelBFSubVT(i,j,:)%val=VAL_UNDEF_QP
                mdlSelBFSubVT(i,j,:)%unc=VAL_UNDEF_QP
              end where
              
              if (isDef(mdlSelEv(i)) .AND. isDef(mdlSelEv(j)) .AND. mdlSelEv(j)%val > 0._qp) then
                mdlSelBF(i,j)%val=mdlSelEv(i)%val/mdlSelEv(j)%val
                mdlSelBF(i,j)%unc=mdlSelBF(i,j)%val*sqrt(mdlSelEv(i)%unc**2+mdlSelEv(j)%unc**2)
              else
                call undef(mdlSelBF(i,j))
              end if

            end do
          end do
        end if
              

      end block

    end if

    call endSubsect(UNIT%cons+UNIT%log) ! Calculating summaries


    contains

    function f(x,mu,sumPot)          ! function whose root w.r.t. x is the (minimum) planetary mass sought
      implicit none
      real(dp) :: &
       f, &
       x, &
       mu, &
       sumPot

      f=x-mu*sumPot**2
    end function


    function df(mu,sumPot)           ! derivative of f w.r.t. x (to see this, note that the value of sumPot depends on x)
      implicit none
      real(dp) :: &
       df, &
       mu, &
       sumPot

      df=1._dp-2._dp/3*mu/sumPot
    end function


    subroutine calcMMin(chain,iPl,DQChainTmp,iMin,iMax,actMass) ! calculates m_p,min [M_J]
      implicit none

      integer(i4b),parameter :: &
       MAXIT=100
      integer :: &
       iPl, &
       j, &
       p, &
       p2, &
       p3
      integer(i8b) :: &
       i, &
       iMin, &
       iMax
      real(dp) :: &
       mMin, &
       mSPot, &
       dx, &
       mMinAbsUnc, &
       xMin, &
       mu, &
       sumPot
      logical :: &
       actMass, &
       useI
      type(tyMCMCChain) :: &
       chain, &
       DQChainTmp
      optional :: &
       actMass
      intent(in) :: &
       chain, &
       iMin, &
       iMax, &
       iPl, &
       actMass
      intent(inout) :: &
       DQChainTmp


      p=pRoleByPl(iPl)%K
      p2=pRoleByPl(iPl)%f
      p3=pRole%i
      mSPot=mdlProps%mS**(2._dp/3)
      useI=present(actMass) .AND. actMass

      !$OMP parallel do private(mu,xMin,mMin,mMinAbsUnc,j,sumPot,dx)
      do i=iMin,iMax

        ! µ, cf. [02.05.2010/5,6]
        if (useI) then
          mu=chain%links(i,p)/(CUBE_ROOT_2_PI_GRAVC*(chain%links(i,p2)/D_IN_S)**(1._dp/3)*sin(chain%links(i,p3)))
        else
          mu=chain%links(i,p)/(CUBE_ROOT_2_PI_GRAVC*(chain%links(i,p2)/D_IN_S)**(1._dp/3))
        end if
        xMin=(2*mu/3)**3-mdlProps%mS
        mMin=mu*mSPot
        if (mMin <= xMin) mMin=2*xMin ! in order for mMin to start off in region where it is increasing, i.e. right of the minimum
        mMinAbsUnc=mMin*M_MIN_REL_UNC

        do j=1,MAXIT
          sumPot=(mMin+mdlProps%mS)**(1._dp/3)
          dx=f(mMin,mu,sumPot)/df(mu,sumPot)
          mMin=mMin-dx
          if (abs(dx) < mMinAbsUnc) exit
        end do

        DQChainTmp%links(i-burnInLen,1)=mMin

      end do
      !$OMP end parallel do
    end subroutine


    subroutine chkFixed(fixed)       ! checks for fixed qties.
      implicit none
      logical :: &
       fixed
      intent(in) :: &
       fixed

      if (fixed) then
        DQSkipped(q)=.TRUE.
        iMax=iMin
        iMaxDQ=1
      end if
    end subroutine

  end subroutine calcSummaries




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine calcProbAllPars(chain)
!
!   calculates posterior probability over given hypercubes
!
!   Arguments:
!     chain                          MCMC chain
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    integer(i8b) :: &
     j, &
     N
    real(dp) :: &
     curLBnd(curNPars), &
     curUBnd(curNPars)
    type(tyMCMCChain) :: &
     chain
    intent(in) :: &
     chain

    do i=1,tmpNCalcProbAllPars(iNPlMd)
      if (tmpCalcProbAllParsDim(iNPlMd,i) /= curNPars) then
        call warn("dimension of hypercube for calculating posterior probability is invalid; skipping calculation")
        call undef(calcdPostrProbAllPars(iNPlMd,i))
      else
        curLBnd=tmpCalcProbAllParsRegion(iNPlMd,i,1:curNPars)%lBnd
        curuBnd=tmpCalcProbAllParsRegion(iNPlMd,i,1:curNPars)%uBnd
        where (isUndef(curLBnd))
          curLBnd=-HUGE_dp
          curUBnd=HUGE_dp
        end where

        N=0
        do j=chain%burnInLen+1,chain%length
          if (all((chain%links(j,:curNPars)-curLBnd)*(curUBnd-chain%links(j,:curNPars)) >= 0._qp)) N=N+1
        end do
        calcdPostrProbAllPars(iNPlMd,i)=real(N,qp)/(chain%length-chain%burnInLen)
      end if
    end do

  end subroutine calcProbAllPars




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printSummaries(final)
!
!   prints posterior summaries
!
!   Arguments:
!     final                          whether the output is the final one of this kind
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i, &
     j, &
     prefxLen, &
     maxLen, &
     unMx, &
     iSet, &
     k, &
     l0, &
     l, &
     m
    real :: &
     corrMx(curNPars,curNPars)
    logical :: &
     final
    character(STR_LEN_LONG) :: &
     cCorrMx(2+curNPars)
    character(STR_LEN_STD) :: &
     hdr(2,curNPars), &
     curHPDIDesc
    optional :: &
     final
    intent(in) :: &
     final

    if (present(final) .AND. final) then
    
      if (nNPlMds > 1) then
        call startSubsect(UNIT%cons+UNIT%log,"Model selection based on final MCMC samples of all #-planets modes")
        call printMsg(UNIT%cons+UNIT%log,"All evidences and Bayes factors based on # planets and parameter priors in corresponding &
         &modes (see above)")
            
        call startSubsect(UNIT%cons+UNIT%log,"Calculated by VTA")
        do m=1,MDL_SEL_N_VT
          call startSubsect(UNIT%cons+UNIT%log,"Integral-estimation variant "//trim(niceFmt(m)))
          call printQty(UNIT%cons+UNIT%log,"# leaves",MDL_SEL_VT_N_PTS(m))

          call startSubsect(UNIT%cons+UNIT%log,"Using highest-posterior "//trim(niceFmt(100.*MDL_SEL_VT_USE_FRAC))// &
           "% of post-burn-in samples")
          call startSubsect(UNIT%cons+UNIT%log,"Evidences")
          do i=1,nNPlMds
            call printQty(UNIT%cons+UNIT%log,"Mode "//trim(niceFmt(i))//" ("//trim(niceFmt(nPl(i)))//" planets)", &
             mdlSelEvVT(i,m))
          end do
          call endSubsect(UNIT%cons+UNIT%log) ! Evidences
          call startSubsect(UNIT%cons+UNIT%log,"Bayes factors B_i,j = p(D|mode i)/p(D|mode j)")
          do i=nNPlMds,2,-1
            do j=i-1,1,-1
              call printQty(UNIT%cons+UNIT%log,"B_"//trim(niceFmt(i))//","//trim(niceFmt(j)),mdlSelBFVT(i,j,m))
            end do
          end do
          call endSubsect(UNIT%cons+UNIT%log)
          call endSubsect(UNIT%cons+UNIT%log) ! Using highest-posterior …

          call startSubsect(UNIT%cons+UNIT%log,"By subsampling of highest-posterior samples ("// &
           trim(niceFmt(mdlSelVTNSub))//" times each using "//trim(niceFmt(mdlSelVTSubNUsed))//" samples)")
          call startSubsect(UNIT%cons+UNIT%log,"Evidences")
          do i=1,nNPlMds
            call printQty(UNIT%cons+UNIT%log,"Mode "//trim(niceFmt(i))//" ("//trim(niceFmt(nPl(i)))//" planets)", &
             mdlSelEvSubVT(i,m),OPT_STDEV)
          end do
          call endSubsect(UNIT%cons+UNIT%log)
          call startSubsect(UNIT%cons+UNIT%log,"Bayes factors B_i,j = p(D|mode i)/p(D|mode j)")
          do i=nNPlMds,2,-1
            do j=i-1,1,-1
              call printQty(UNIT%cons+UNIT%log,"B_"//trim(niceFmt(i))//","//trim(niceFmt(j)),mdlSelBFSubVT(i,j,m), &
               OPT_STDEV)
            end do
          end do
          call endSubsect(UNIT%cons+UNIT%log)
          call endSubsect(UNIT%cons+UNIT%log) ! By subsampling …

          call endSubsect(UNIT%cons+UNIT%log) ! Integral-estimation variant …
        end do ! m=1,MDL_SEL_N_VT

        call startSubsect(UNIT%cons+UNIT%log,"Statistics over all integral-estimation variants and sets of samples used")
        call startSubsect(UNIT%cons+UNIT%log,"Evidences")
        do i=1,nNPlMds
          call printQty(UNIT%cons+UNIT%log,"Mode "//trim(niceFmt(i))//" ("//trim(niceFmt(nPl(i)))//" planets)",mdlSelEv(i), &
           OPT_STDEV)
        end do
        call endSubsect(UNIT%cons+UNIT%log)
        call startSubsect(UNIT%cons+UNIT%log,"Bayes factors B_i,j = p(D|mode i)/p(D|mode j)")
        do i=nNPlMds,2,-1
          do j=i-1,1,-1
            call printQty(UNIT%cons+UNIT%log,"B_"//trim(niceFmt(i))//","//trim(niceFmt(j)),mdlSelBF(i,j),OPT_STDEV)
          end do
        end do
        call endSubsect(UNIT%cons+UNIT%log)

        call endSubsect(UNIT%cons+UNIT%log) ! Statistics over all integral-estimation variants and sets of samples used
        call endSubsect(UNIT%cons+UNIT%log) ! Calculated by VTA

        call endSubsect(UNIT%cons+UNIT%log) ! Model selection based on final MCMC samples of all #-planets modes
      end if
      

      ! CSV output
      
      if (filOutput .AND. nNPlMds > 1) then
        
        ! evidences
        block
          character(STR_LEN_STD) :: &
           valArr(MDL_SEL_N_VT+1,3), &
           hdrArr(2,4), &
           leftArr(MDL_SEL_N_VT+1,1)
          
          hdrArr(1,1)="VTA variant"
          hdrArr(1,2)=trim(niceFmt(100.*MDL_SEL_VT_USE_FRAC))//"% of samples"
          hdrArr(1,3)="Subsampling"
          hdrArr(1,4)=""
          hdrArr(2,1)=""
          hdrArr(2,2)="Evidence"
          hdrArr(2,3)="Evidence"
          hdrArr(2,4)="Uncertainty"
          
          do i=1,MDL_SEL_N_VT
            leftArr(i,1)=trim(niceFmt(MDL_SEL_VT_N_PTS(i)))//" leaves"
          end do
          leftArr(MDL_SEL_N_VT+1,1)="Overall statistics"
          
          do i=1,nNPlMds
                  
            do m=1,MDL_SEL_N_VT
              valArr(m,1)=niceFmt(mdlSelEvVT(i,m))
              valArr(m,2)=niceFmt(mdlSelEvSubVT(i,m)%val)
              valArr(m,3)=niceFmt(mdlSelEvSubVT(i,m)%unc)
            end do
            valArr(MDL_SEL_N_VT+1,1:2)=""
            valArr(MDL_SEL_N_VT+1,2)=niceFmt(mdlSelEv(i)%val)
            valArr(MDL_SEL_N_VT+1,3)=niceFmt(mdlSelEv(i)%unc)

            call writeArrCSV(trim(outDir)//trim(fmtFilNmPlt(trim(MDL_SEL_FIL_NM%ev)))//"_"//trim(niceFmt(i))//FIL_EXT_CSV, &
             valArr,hdrArr=hdrArr,leftArr=leftArr)
            
          end do

        end block
        
        ! BFs
        block
          character(STR_LEN_STD) :: &
           valArr(MDL_SEL_N_VT+1,3), &
           hdrArr(2,4), &
           leftArr(MDL_SEL_N_VT+1,1)
          
          hdrArr(1,1)="VTA variant"
          hdrArr(1,2)=trim(niceFmt(100.*MDL_SEL_VT_USE_FRAC))//"% of samples"
          hdrArr(1,3)="Subsampling"
          hdrArr(1,4)=""
          hdrArr(2,1)=""
          hdrArr(2,2)="Bayes factor"
          hdrArr(2,3)="Bayes factor"
          hdrArr(2,4)="Uncertainty"
          
          do i=1,MDL_SEL_N_VT
            leftArr(i,1)=trim(niceFmt(MDL_SEL_VT_N_PTS(i)))//" leaves"
          end do
          leftArr(MDL_SEL_N_VT+1,1)="Overall statistics"
          
          do i=nNPlMds,2,-1
            do j=i-1,1,-1
                  
              do m=1,MDL_SEL_N_VT
                valArr(m,1)=niceFmt(mdlSelBFVT(i,j,m))
                valArr(m,2)=niceFmt(mdlSelBFSubVT(i,j,m)%val)
                valArr(m,3)=niceFmt(mdlSelBFSubVT(i,j,m)%unc)
              end do
              valArr(MDL_SEL_N_VT+1,1:2)=""
              valArr(MDL_SEL_N_VT+1,2)=niceFmt(mdlSelBF(i,j)%val)
              valArr(MDL_SEL_N_VT+1,3)=niceFmt(mdlSelBF(i,j)%unc)

              call writeArrCSV(trim(outDir)//trim(fmtFilNmPlt(trim(MDL_SEL_FIL_NM%bayesFac)))//"_"//trim(niceFmt(i))//","// &
               trim(niceFmt(j))//FIL_EXT_CSV,valArr,hdrArr=hdrArr,leftArr=leftArr)
            
            end do
          end do

        end block
      end if                         ! (filOutput)
      
    else
    
      call startSubsect(UNIT%cons+UNIT%log,"Posterior summary based on the final MCMC samples")

      if (usrPred) then
        call startSubsect(UNIT%cons+UNIT%log,"Extremal posterior model uncertainties")
        call printQty(UNIT%cons+UNIT%log,"Timespan of calculation [d]",shiftIntvl(pred1D%tRge(1,1),tOfs))


        ! AMAngPos

        if (dtProps%nSetsOfTy(DT_TY%AMAngPos) > 0) then
          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(DT_TY%AMAngPos))))
          call printQty(UNIT%cons+UNIT%log,"Time resolution [d]",pred2D%tResol(DT_TY%AMAngPos,1))
          do iSet=1,dtProps%nSetsOfTy(DT_TY%AMAngPos)
            if (.NOT. pred2D%skipSet(DT_TY%AMAngPos,iSet)) then
              call startSubsect(UNIT%cons+UNIT%log,"Set "//trim(niceFmt(iSet)))
              call startSubsect(UNIT%cons+UNIT%log,"Minimum uncertainty (smallest semi-minor axis of uncertainty ellipse)")
              call printQty(UNIT%cons+UNIT%log,fmtObsvble(DT_TY%AMAngPos,VAL_UNDEF_INT,"Value",.NOT. OPT_FOR_PLT), &
               pred2D%minUnc(DT_TY%AMAngPos,iSet))
              call printQty(UNIT%cons+UNIT%log,"Position angle of semi-minor axis w.r.t. North [deg]", &
              pred2D%phiMinUnc(DT_TY%AMAngPos,iSet))
              call printQty(UNIT%cons+UNIT%log,dateDescAndUnit,tOfs+pred2D%tMinUnc(DT_TY%AMAngPos,iSet))
              call endSubsect(UNIT%cons+UNIT%log)
              call startSubsect(UNIT%cons+UNIT%log,"Maximum uncertainty (largest semi-major axis of uncertainty ellipse)")
              call printQty(UNIT%cons+UNIT%log,fmtObsvble(DT_TY%AMAngPos,VAL_UNDEF_INT,"Value"),pred2D%maxUnc(DT_TY%AMAngPos,iSet))
              call printQty(UNIT%cons+UNIT%log,"Position angle of semi-major axis w.r.t. North [deg]", &
              pred2D%phiMaxUnc(DT_TY%AMAngPos,iSet))
              call printQty(UNIT%cons+UNIT%log,dateDescAndUnit,tOfs+pred2D%tMaxUnc(DT_TY%AMAngPos,iSet))
              call endSubsect(UNIT%cons+UNIT%log)
              call endSubsect(UNIT%cons+UNIT%log)
            end if
          end do
          call endSubsect(UNIT%cons+UNIT%log)
        end if


        ! RV

        if (dtProps%nSetsOfTy(DT_TY%RV) > 0) then
          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(DT_TY%RV))))
          call printQty(UNIT%cons+UNIT%log,"Time resolution [d]",pred1D%tResol(DT_TY%RV,1))
          do iSet=1,dtProps%nSetsOfTy(DT_TY%RV)
            if (.NOT. pred1D%skipSet(DT_TY%RV,iSet)) then
              call startSubsect(UNIT%cons+UNIT%log,"Set "//trim(niceFmt(iSet)))
              call startSubsect(UNIT%cons+UNIT%log,"Minimum uncertainty")
              call printQty(UNIT%cons+UNIT%log,fmtObsvble(DT_TY%RV,VAL_UNDEF_INT,"Value",.NOT. OPT_FOR_PLT), &
               pred1D%minUnc(DT_TY%RV,iSet))
              call printQty(UNIT%cons+UNIT%log,dateDescAndUnit,tOfs+pred1D%tMinUnc(DT_TY%RV,iSet))
              call endSubsect(UNIT%cons+UNIT%log)
              call startSubsect(UNIT%cons+UNIT%log,"Maximum uncertainty")
              call printQty(UNIT%cons+UNIT%log,fmtObsvble(DT_TY%RV,VAL_UNDEF_INT,"Value",.NOT. OPT_FOR_PLT), &
               pred1D%maxUnc(DT_TY%RV,iSet))
              call printQty(UNIT%cons+UNIT%log,dateDescAndUnit,tOfs+pred1D%tMaxUnc(DT_TY%RV,iSet))
              call endSubsect(UNIT%cons+UNIT%log)
              call endSubsect(UNIT%cons+UNIT%log)
            end if
          end do
          call endSubsect(UNIT%cons+UNIT%log)
        end if

        call endSubsect(UNIT%cons+UNIT%log) ! Minimum/maximum posterior model uncertainties

      end if                           ! (pred1D)


      call startSubsect(UNIT%cons+UNIT%log,"Summaries using MAP parameter estimates")

      if (logPostrEv > -huge(logPostrEv)) call printQty(UNIT%cons+UNIT%log, &
       "ln(prior*likelihood) = ln(posterior*evidence)",logPostrEv)
      call startSubsect(UNIT%cons+UNIT%log,"chi^2")
      if (isUndef(dtProps%onlyDtTy)) then
        call startSubsect(UNIT%cons+UNIT%log,"All data")
        call printQty(UNIT%cons+UNIT%log,uCaseFst(TTL_PART%woJitt),postrSumm(iNPlMd)%maxPostrChiSq%woJitt)
        call printQty(UNIT%cons+UNIT%log,uCaseFst(TTL_PART%wJitt),postrSumm(iNPlMd)%maxPostrChiSq%wJitt)
        call endSubsect(UNIT%cons+UNIT%log) ! All data
      end if
      do i=1,N_DT_TYPES
        if (dtProps%nSetsOfTy(i) > 0) then
          if (isUndef(dtProps%onlyDtTy)) call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(i))))
          call printQty(UNIT%cons+UNIT%log,uCaseFst(TTL_PART%woJitt),postrSumm(iNPlMd)%maxPostrChiSqByDtTy(i)%woJitt)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(TTL_PART%wJitt),postrSumm(iNPlMd)%maxPostrChiSqByDtTy(i)%wJitt)
          if (isUndef(dtProps%onlyDtTy)) call endSubsect(UNIT%cons+UNIT%log)
        end if
      end do
      call endSubsect(UNIT%cons+UNIT%log) ! chi^2

      if (residCalcd) then

        call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%distr))//" of "//TTL_PART%normResids)

        if (dtProps%nSetsOfTy(DT_TY%AMAngPos) > 0) then
          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(DT_TY%AMAngPos))))

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%woJitt)))
          call printQty(UNIT%cons+UNIT%log,"Mean",dtPropsAMAngPos%normResidMean%woJitt)
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",dtPropsAMAngPos%normResidStdDev%woJitt)
          call printQty(UNIT%cons+UNIT%log,"Kolmogorov-Smirnov p-value (comparison with N(0,1))", &
           dtPropsAMAngPos%normResidPVal%woJitt)
          call endSubsect(UNIT%cons+UNIT%log)

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%wJitt)))
          call printQty(UNIT%cons+UNIT%log,"Mean",dtPropsAMAngPos%normResidMean%wJitt)
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",dtPropsAMAngPos%normResidStdDev%wJitt)
          call printQty(UNIT%cons+UNIT%log,"Kolmogorov-Smirnov p-value (comparison with N(0,1))", &
           dtPropsAMAngPos%normResidPVal%wJitt)
          call endSubsect(UNIT%cons+UNIT%log)

          call endSubsect(UNIT%cons+UNIT%log)
        end if

        if (dtProps%nSetsOfTy(DT_TY%RV) > 0) then
          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(DT_TY%RV))))

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%woJitt)))
          call printQty(UNIT%cons+UNIT%log,"Mean",dtPropsRV%normResidMean%woJitt)
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",dtPropsRV%normResidStdDev%woJitt)
          call printQty(UNIT%cons+UNIT%log,"Kolmogorov-Smirnov p-value (comparison with N(0,1))",dtPropsRV%normResidPVal%woJitt)
          call endSubsect(UNIT%cons+UNIT%log)

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%wJitt)))
          call printQty(UNIT%cons+UNIT%log,"Mean",dtPropsRV%normResidMean%wJitt)
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",dtPropsRV%normResidStdDev%wJitt)
          call printQty(UNIT%cons+UNIT%log,"Kolmogorov-Smirnov p-value (comparison with N(0,1))",dtPropsRV%normResidPVal%wJitt)
          call endSubsect(UNIT%cons+UNIT%log)

          call endSubsect(UNIT%cons+UNIT%log)
        end if

        if (dtProps%nSetsOfTy(DT_TY%AMHipN) > 0) then
          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(DT_TY_DESC(DT_TY%AMHipN))))

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%woJitt)))
          call printQty(UNIT%cons+UNIT%log,"Mean",dtPropsAMHipN%normResidMean%woJitt)
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",dtPropsAMHipN%normResidStdDev%woJitt)
          call printQty(UNIT%cons+UNIT%log,"Kolmogorov-Smirnov p-value (comparison with N(0,1))",dtPropsAMHipN%normResidPVal%woJitt)
          call endSubsect(UNIT%cons+UNIT%log)

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%wJitt)))
          call printQty(UNIT%cons+UNIT%log,"Mean",dtPropsAMHipN%normResidMean%wJitt)
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",dtPropsAMHipN%normResidStdDev%wJitt)
          call printQty(UNIT%cons+UNIT%log,"Kolmogorov-Smirnov p-value (comparison with N(0,1))",dtPropsAMHipN%normResidPVal%wJitt)
          call endSubsect(UNIT%cons+UNIT%log)

          call endSubsect(UNIT%cons+UNIT%log)
        end if

        if (isUndef(dtProps%onlyDtTy)) then
          call startSubsect(UNIT%cons+UNIT%log,"All data types") ! always w/o jitter

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%woJitt)))
          call printQty(UNIT%cons+UNIT%log,"Mean",dtProps%normResidMean%woJitt)
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",dtProps%normResidStdDev%woJitt)
          call printQty(UNIT%cons+UNIT%log,"Kolmogorov-Smirnov p-value (comparison with N(0,1))",dtProps%normResidPVal%woJitt)
          call endSubsect(UNIT%cons+UNIT%log)

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%wJitt)))
          call printQty(UNIT%cons+UNIT%log,"Mean",dtProps%normResidMean%wJitt)
          call printQty(UNIT%cons+UNIT%log,"Standard deviation",dtProps%normResidStdDev%wJitt)
          call printQty(UNIT%cons+UNIT%log,"Kolmogorov-Smirnov p-value (comparison with N(0,1))",dtProps%normResidPVal%wJitt)
          call endSubsect(UNIT%cons+UNIT%log)

          call endSubsect(UNIT%cons+UNIT%log)
        end if

        call endSubsect(UNIT%cons+UNIT%log) ! Distribution of normalised residuals

      end if

      call endSubsect(UNIT%cons+UNIT%log) ! Summaries using MAP parameter estimates



      call startSubsect(UNIT%cons+UNIT%log,"Best estimates and uncertainties of parameters")


      if (curNParsInfc > 1) then

        ! correlation matrix

        call startSubsect(UNIT%cons+UNIT%log,"Correlation matrix")

        do i=1,curNPars
          where (isUndef(parEst(iNPlMd,i)%corrCoeff(:curNPars)))
            corrMx(i,:)=undefVal(corrMx(i,:))
          elsewhere
            corrMx(i,:)=real(parEst(iNPlMd,i)%corrCoeff(:curNPars),sp)
          end where
        end do

        hdr(1,:)=parNms(:curNPars)
        do i=1,curNPars
          hdr(2,i)="("//trim(idStr(i))//")"
        end do
        prefxLen=len_trim(hdr(2,curNPars))

        maxLen=0
        cCorrMx=fmtMatrix(2+curNPars,corrMx,hdr) ! arr, hdr ARE CONTIGUOUS
        do i=1,2
          cCorrMx(i)=repeat(" ",prefxLen+1)//cCorrMx(i)
          maxLen=max(maxLen,len_trim(cCorrMx(i)))
        end do
        do i=3,curNPars+2
          cCorrMx(i)="("//trim(idStr(i-2))//")"//repeat(" ",prefxLen-(len_trim(idStr(i))+2)+1)//cCorrMx(i)
          maxLen=max(maxLen,len_trim(cCorrMx(i)))
        end do

        if (sectDepth(I_UNIT%cons)*CONS_TAB_WIDTH+maxLen > consLineLen) then
          call printMsg(UNIT%cons,"(see log file)")
          unMx=UNIT%log
        else
          unMx=UNIT%cons+UNIT%log
        end if
        do i=1,curNPars+2
          call printMsg(unMx,cCorrMx(i),setFlg(FLG%simpPrfx))
        end do

        call endSubsect(UNIT%cons+UNIT%log) ! Correlation matrix


        ! covariance matrix

        call startSubsect(UNIT%cons+UNIT%log,"Covariance matrix")

        do i=1,curNPars
          where (isUndef(parEst(iNPlMd,i)%cov(:curNPars)))
            corrMx(i,:)=undefVal(corrMx(i,:))
          elsewhere
            corrMx(i,:)=real(parEst(iNPlMd,i)%cov(:curNPars),sp)
          end where
        end do

        hdr(1,:)=parNms(:curNPars)
        do i=1,curNPars
          hdr(2,i)="("//trim(idStr(i))//")"
        end do
        prefxLen=len_trim(hdr(2,curNPars))

        maxLen=0
        cCorrMx=fmtMatrix(2+curNPars,corrMx,hdr) ! arr, hdr ARE CONTIGUOUS
        do i=1,2
          cCorrMx(i)=repeat(" ",prefxLen+1)//cCorrMx(i)
          maxLen=max(maxLen,len_trim(cCorrMx(i)))
        end do
        do i=3,curNPars+2
          cCorrMx(i)="("//trim(idStr(i-2))//")"//repeat(" ",prefxLen-(len_trim(idStr(i))+2)+1)//cCorrMx(i)
          maxLen=max(maxLen,len_trim(cCorrMx(i)))
        end do

        if (sectDepth(I_UNIT%cons)*CONS_TAB_WIDTH+maxLen > consLineLen) then
          call printMsg(UNIT%cons,"(see log file)")
          unMx=UNIT%log
        else
          unMx=UNIT%cons+UNIT%log
        end if
        do i=1,curNPars+2
          call printMsg(unMx,cCorrMx(i),setFlg(FLG%simpPrfx))
        end do

        call endSubsect(UNIT%cons+UNIT%log) ! covariance matrix

      end if                           ! (curNParsInfc > 1)


      ! other summaries

      do i=1,curNPars
        call startSubsect(UNIT%cons+UNIT%log,fmtPar(i,OPT_FOR_STDOUT))

        if (parFixed(i)) then
          call printQty(UNIT%cons+UNIT%log,"Fixed value",priLBnd(i))
        else if (parInfc(i)) then
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%mode)),parEst(iNPlMd,i)%mode)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%median)),parEst(iNPlMd,i)%median)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%mean)),parEst(iNPlMd,i)%mean,OPT_STDEV)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%mMode)),parEst(iNPlMd,i)%mMode)
          if (isDef(parEst(iNPlMd,i)%bestSummTy)) &
           call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%bestSumm)),PAR_EST_SUMM_DESC(parEst(iNPlMd,i)%bestSummTy))
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%sampStdDev)),parEst(iNPlMd,i)%sampStdDev)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%nIndepSamp)),parEst(iNPlMd,i)%nIndepSamp)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%KDEBandw)),parEst(iNPlMd,i)%KDEBandw)

          call startSubsect(UNIT%cons+UNIT%log,"HPD intervals with given probability contents")
          do j=1,nHPDI
            if (isDef(HPDIRole(j)) .AND. len_trim(STD_HPDI_DESC(HPDIRole(j))) > 0) then
              curHPDIDesc=" ("//trim(STD_HPDI_DESC(HPDIRole(j)))//")"
            else
              curHPDIDesc=""
            end if
            call printQty(UNIT%cons+UNIT%log,trim(fmtHPDIProbCont(j))//trim(curHPDIDesc),parEst(iNPlMd,i)%HPDI(j))
          end do
          call endSubsect(UNIT%cons+UNIT%log) ! HPD intervals
        else
          call printMsg(UNIT%cons+UNIT%log,"(excluded from inference)")
        end if                         ! (parFixed(i))
        
        call printQty(UNIT%cons+UNIT%log,"Epoch "//trim(mrg("(RJD)","",dtProps%timesAreJD))//" [d]", &
         merge(mdlProps%epoch+tOfs,VAL_UNDEF_DBL,parHasEpoch(i)))

        call endSubsect(UNIT%cons+UNIT%log) ! parNms(i)

      end do
      call endSubsect(UNIT%cons+UNIT%log) ! Best estimates and uncertainties of parameters


      if (curNDQs > 0) then
        call startSubsect(UNIT%cons+UNIT%log,"Best estimates and uncertainties of derived quantities")
        do i=1,curNDQs
          call startSubsect(UNIT%cons+UNIT%log,fmtDQ(i,OPT_FOR_STDOUT))
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%mode)),parEstDQ(iNPlMd,i)%mode)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%median)),parEstDQ(iNPlMd,i)%median)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%mean)),parEstDQ(iNPlMd,i)%mean,OPT_STDEV)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%mMode)),parEstDQ(iNPlMd,i)%mMode)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%sampStdDev)),parEstDQ(iNPlMd,i)%sampStdDev)
          call printQty(UNIT%cons+UNIT%log,uCaseFst(trim(PAR_EST_DESC%KDEBandw)),parEstDQ(iNPlMd,i)%KDEBandw)
          if (DQSkipped(i)) then
            call printMsg(UNIT%cons+UNIT%log,"HPD intervals are undefined as all underlying parameters are fixed")
          else
            call startSubsect(UNIT%cons+UNIT%log,"HPD intervals with given probability contents")
            do j=1,nHPDI
              if (isDef(HPDIRole(j)) .AND. len_trim(STD_HPDI_DESC(HPDIRole(j))) > 0) then
                curHPDIDesc=" ("//trim(STD_HPDI_DESC(HPDIRole(j)))//")"
              else
                curHPDIDesc=""
              end if
              call printQty(UNIT%cons+UNIT%log,trim(fmtHPDIProbCont(j))//trim(curHPDIDesc),parEstDQ(iNPlMd,i)%HPDI(j))
            end do
            call endSubsect(UNIT%cons+UNIT%log) ! HPD intervals
          end if
          if (isDef(transTmStart) .AND. gIDQs(i) == G_I_DQ_ROLE%tT1) call printQty(UNIT%cons+UNIT%log, &
           "Probability that 1st transit is primary transit [%]",100.*prob1stTransIsPri)
          call endSubsect(UNIT%cons+UNIT%log) ! fmtDQ(i,OPT_FOR_STDOUT)
        end do
        call endSubsect(UNIT%cons+UNIT%log) ! Best estimates and uncertainties of DQs

      end if


      call endSubsect(UNIT%cons+UNIT%log) ! Posterior summary based on the final MCMC samples


      if (tmpNCalcProb(iNPlMd) > 0) then
        call startSubsect(UNIT%cons+UNIT%log,"Posterior probabilities for given parameters")
        do i=1,tmpNCalcProb(iNPlMd)
          if (isDef(calcdPostrProb(iNPlMd,i))) then
            call startSubsect(UNIT%cons+UNIT%log,trim(tmpCalcProbNms(iNPlMd,i)))
            call printQty(UNIT%cons+UNIT%log,"Requested range",tmpCalcProbRge(iNPlMd,i))
            call printQty(UNIT%cons+UNIT%log,"Posterior probability [%]",100._qp*calcdPostrProb(iNPlMd,i))
            call endSubsect(UNIT%cons+UNIT%log)
          end if
        end do
        call endSubsect(UNIT%cons+UNIT%log)
      end if

      if (tmpNCalcProbAllPars(iNPlMd) > 0) then
        call startSubsect(UNIT%cons+UNIT%log,"Posterior probabilities for given hypercubes")
        j=0
        do i=1,tmpNCalcProbAllPars(iNPlMd)
          if (isDef(calcdPostrProbAllPars(iNPlMd,i))) then
            call inc(j)
            call startSubsect(UNIT%cons+UNIT%log,"Hypercube #"//trim(niceFmt(j)))
            call startSubsect(UNIT%cons+UNIT%log,"Sides of requested hypercube")
            do k=1,curNPars
              if (isDef(tmpCalcProbAllParsRegion(iNPlMd,i,k))) then
                call printQty(UNIT%cons+UNIT%log,parNms(k),tmpCalcProbAllParsRegion(iNPlMd,i,k))
              else
                call printQty(UNIT%cons+UNIT%log,parNms(k),"<whole prior range>")
              end if
            end do
            call endSubsect(UNIT%cons+UNIT%log) ! Requested intervals
            call printQty(UNIT%cons+UNIT%log,"Posterior probability [%]",100._dp*calcdPostrProbAllPars(iNPlMd,i))
            call endSubsect(UNIT%cons+UNIT%log) ! Hypercube #…
          end if
        end do
        call endSubsect(UNIT%cons+UNIT%log)
      end if



      if (filOutput) then

        block

          integer :: &
           i, &
           j
          real(dp) :: &
           tmpMx(curNPars,curNPars)
          character(STR_LEN_STD) :: &
           tmpArrStr(9+2*nHPDI,curNPars+curNDQs), &
           tmpArrStrL(9+2*nHPDI,1)

          if (curNParsInfc > 1) then

          ! covariance matrix

            do i=1,curNPars
              do j=1,curNPars
                tmpMx(i,j)=parEst(iNPlMd,i)%cov(j)
              end do
            end do

            call writeArrCSV(trim(outDir)//trim(fmtFilNmPlt(trim(PAR_EST_FIL_NM%covM)))//FIL_EXT_CSV,tmpMx, &
            hdrArr=reshape([character(STR_LEN_STD) :: "",(parNms(i), i=1,curNPars)],[1,curNPars+1]), &
            leftArr=reshape(parNms(:curNPars),[curNPars,1])) ! arr IS CONTIGUOUS

            ! correlation matrix

            do i=1,curNPars
              do j=1,curNPars
                tmpMx(i,j)=parEst(iNPlMd,i)%corrCoeff(j)
              end do
            end do

            call writeArrCSV(trim(outDir)//trim(fmtFilNmPlt(trim(PAR_EST_FIL_NM%corrM)))//FIL_EXT_CSV,tmpMx, &
             hdrArr=reshape([character(STR_LEN_STD) :: "",(parNms(i), i=1,curNPars)],[1,curNPars+1]), &
             leftArr=reshape(parNms(:curNPars),[curNPars,1])) ! arr IS CONTIGUOUS

          end if


          ! scalars (mean, median, mode, best summ., sample stdev., independent samples, stdev. of mean, HPDI lbounds, HPDI ubounds)

          do i=1,curNPars
            tmpArrStr(1,i)=niceFmt(parEst(iNPlMd,i)%mode)
            tmpArrStr(2,i)=niceFmt(parEst(iNPlMd,i)%median)
            tmpArrStr(3,i)=niceFmt(parEst(iNPlMd,i)%mean%val)
            tmpArrStr(4,i)=niceFmt(parEst(iNPlMd,i)%mean%unc)
            tmpArrStr(5,i)=niceFmt(parEst(iNPlMd,i)%mMode%val)
            tmpArrStr(6,i)=niceFmt(parEst(iNPlMd,i)%mMode%unc)
            if (isDef(parEst(iNPlMd,i)%bestSummTy)) then
              tmpArrStr(7,i)=PAR_EST_SUMM_DESC(parEst(iNPlMd,i)%bestSummTy)
            else
              call undef(tmpArrStr(7,i))
            end if
            tmpArrStr(8,i)=niceFmt(parEst(iNPlMd,i)%sampStdDev)
            tmpArrStr(9,i)=niceFmt(parEst(iNPlMd,i)%nIndepSamp)
            do j=1,nHPDI
              tmpArrStr(9+(j-1)*2+1,i)=niceFmt(parEst(iNPlMd,i)%HPDI(j)%lBnd)
              tmpArrStr(9+(j-1)*2+2,i)=niceFmt(parEst(iNPlMd,i)%HPDI(j)%uBnd)
            end do
          end do

          do i=1,curNDQs
            tmpArrStr(1,curNPars+i)=niceFmt(parEstDQ(iNPlMd,i)%mode)
            tmpArrStr(2,curNPars+i)=niceFmt(parEstDQ(iNPlMd,i)%median)
            tmpArrStr(3,curNPars+i)=niceFmt(parEstDQ(iNPlMd,i)%mean%val)
            call undef(tmpArrStr(4,curNPars+i))
            tmpArrStr(5,curNPars+i)=niceFmt(parEstDQ(iNPlMd,i)%mMode%val)
            tmpArrStr(6,curNPars+i)=niceFmt(parEstDQ(iNPlMd,i)%mMode%unc)
            call undef(tmpArrStr(7,curNPars+i))
            tmpArrStr(8,curNPars+i)=niceFmt(parEstDQ(iNPlMd,i)%sampStdDev)
            call undef(tmpArrStr(9,curNPars+i))
            do j=1,nHPDI
              tmpArrStr(9+(j-1)*2+1,curNPars+i)=niceFmt(parEstDQ(iNPlMd,i)%HPDI(j)%lBnd)
              tmpArrStr(9+(j-1)*2+2,curNPars+i)=niceFmt(parEstDQ(iNPlMd,i)%HPDI(j)%uBnd)
            end do
          end do

          tmpArrStrL(1,1)=PAR_EST_DESC%mode
          tmpArrStrL(2,1)=PAR_EST_DESC%median
          tmpArrStrL(3,1)=PAR_EST_DESC%mean
          tmpArrStrL(4,1)=PAR_EST_DESC%stdDevMean
          tmpArrStrL(5,1)=PAR_EST_DESC%mMode
          tmpArrStrL(6,1)="Uncertainty of "//lCaseFst(PAR_EST_DESC%mMode)
          tmpArrStrL(7,1)=PAR_EST_DESC%bestSumm
          tmpArrStrL(8,1)=PAR_EST_DESC%sampStdDev
          tmpArrStrL(9,1)=PAR_EST_DESC%nIndepSamp
          do j=1,nHPDI
            tmpArrStrL(9+(j-1)*2+1,1)=trim(fmtHPDIProbCont(j))//" "//PAR_EST_DESC%HPDIL
            tmpArrStrL(9+(j-1)*2+2,1)=trim(fmtHPDIProbCont(j))//" "//PAR_EST_DESC%HPDIU
          end do

          call writeArrCSV(trim(outDir)//trim(fmtFilNmPlt(trim(PAR_EST_FIL_NM%scalar)))//FIL_EXT_CSV,tmpArrStr, &
           hdrArr=reshape([character(STR_LEN_STD) :: "",(parNms(i), i=1,curNPars),(DQNms(i), i=1,curNDQs)], &
           [1,curNPars+curNDQs+1]),leftArr=tmpArrStrL) ! arr IS CONTIGUOUS



        end block
      end if

    end if ! (present(final) .AND. final)


  end subroutine printSummaries




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine doPlotting
!
!   performs plotting
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     exitStat
    character(MAX_LEN_PATH) :: &
     filNmTex

    if (execGnuplt .AND. filOutputPlt) then
      call printMsg(UNIT%cons+UNIT%log,"Executing Gnuplot")
      call execute_command_line("cd "//trim(outDirCSV)//"; gnuplot "//trim(filNmGnupltCmdsNoDirs),exitstat=exitStat)
      if (exitStat /= 0) call warn("executing Gnuplot failed")
    end if


    ! write TeX file (if in PSTricks mode)

    if (pltPSTricks) then
      filNmTex=trim(fmtFilNmPlt("pstricks"))//SUFFX_TEX
      call addGnupltCmds("",trim(outDirCSV)//filNmTex,"","","",setFlg(FLG%mkTex))
      if (execPS4PDF) then
        call printMsg(UNIT%cons+UNIT%log,"Executing ps4pdf")
        call execute_command_line("cd "//trim(outDirCSV)//"; ps4pdf "//trim(filNmTex),exitstat=exitStat)
        if (exitStat /= 0) call warn("executing ps4pdf failed")
      end if
    end if

  end subroutine doPlotting




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine makePostrPlots(inclPri,befSort)
!
!   produces posterior-plot files
!
!   Arguments:
!     inclPri                        whether the prior plots should be included
!     befSort                        whether the plots produced before sorting the chain are requested
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     p, &
     p2, &
     q, &
     curStride, &
     allocStat, &
     iZm, &
     j
    integer(i8b) :: &
     curNPts, &
     i, &
     curSampNrs(:)
    real(dp) :: &
     curPtsX(:), &
     curPtsY(:), &
     covM(2,2), &
     contH(curNPars-1,curNPars,N_CONTOURS), &
     mul
    logical :: &
     befSort, &
     inclPri, &
     redoKDE
    character(MAX_LEN_PATH) :: &
     curFilNmCSV, &
     curFilNmPlt
    type(tyCoord3DDbl) :: &
     kerDensEst(pltNPtsKDE**2)
    type(tyIntvlDbl) :: &
     xRge1Zm(curNPars-1,curNPars), &
     xRge2Zm(curNPars-1,curNPars), &
     xRge1(curNPars-1,curNPars), &
     xRge2(curNPars-1,curNPars), &
     curZmRge1, &
     curZmRge2, &
     curPriRge
    allocatable :: &
     curSampNrs, &
     curPtsX, &
     curPtsY
    intent(in) :: &
     befSort, &
     inclPri

    call startSubsect(UNIT%cons+UNIT%log,"Calculating and writing plot points")

    if (befSort) then

        ! priors

      if (inclPri) then

        call startSubsect(UNIT%cons+UNIT%log,"Priors")

        curNPts=PLT_N_PTS%std
        allocate(curPtsX(curNPts),curPtsY(curNPts),stat=allocStat)
        call testAllocErr(allocStat,30)

        if (.NOT. singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(FIL_NM_PART%prior))//trim(pltFilExt)

        call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
        do p=1,curNPars

          if (.NOT. parFixed(p)) then

            ! calc. points
            if (priLogScl(p)) then
              curPriRge=pltRge(priRge(p),.TRUE.,priTypes(p))
              mul=intvlBoundsRatio(curPriRge)**(1._dp/curNPts)
              do i=1,curNPts
                curPtsX(i)=curPriRge%lBnd*mul**(i-.5)
                curPtsY(i)=singlePrior(p,curPtsX(i))
              end do
            else
              do i=1,curNPts
                curPtsX(i)=priLBnd(p)+(i-.5)/curNPts*priW(p)
                curPtsY(i)=singlePrior(p,curPtsX(i))
              end do
            end if

            ! write CSV output
            curFilNmCSV=trim(fmtFilNmPlt(FIL_NM_PART%prior,parNms(p)))
            if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
            curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),curPtsX,curPtsY)

            ! add plotting commands
            if (curPtsY(1) == curPtsY(curNPts)) then ! empty y range, Gnuplot will warn
              call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtPar(p,OPT_FOR_PLOT),trim(fmtPrior(parNmsPlt(p))), &
               uCaseFst(trim(TTL_PART%prior))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclX,priLogScl(p)), &
               rgeX=pltRge(priRge(p),priLogScl(p),priTypes(p)), &
               rgeY=tyIntvlDbl(.99*curPtsY(1),1.01*curPtsY(1)))
            else
              call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtPar(p,OPT_FOR_PLOT),trim(fmtPrior(parNmsPlt(p))), &
               uCaseFst(trim(TTL_PART%prior))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclX,priLogScl(p)), &
               rgeX=pltRge(priRge(p),priLogScl(p),priTypes(p)))
            end if

            call printPrgrs([VAL_UNDEF_INT],nTot=curNParsFree)

          end if

        end do

        deallocate(curPtsX,curPtsY)
        call endSubsect(UNIT%cons+UNIT%log) ! Priors

      end if



        ! traces

      call startSubsect(UNIT%cons+UNIT%log,"Traces")

      curStride=max(1,ceiling(1.*chn%length/PLT_N_PTS%std)) ! steps in chain because of limited # of points drawn
      curNPts=floor(1.*chn%length/curStride)
      if (allocated(curSampNrs)) deallocate(curSampNrs)
      allocate(curSampNrs(curNPts),stat=allocStat) ! numbers of displayed samples
      call testAllocErr(allocStat,20)
      curSampNrs=curStride*genSeq(1_i8b,curNPts)

          ! normal

      if (.NOT. singlePlts) &
       curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(FIL_NM_PART%trace))//trim(pltFilExt)
      call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
      do p=1,curNPars
        if (parInfc(p)) then
          ! write CSV output
          curFilNmCSV=trim(fmtFilNmPlt(FIL_NM_PART%trace,parNms(p)))
          if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
          curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
          call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),curSampNrs,chn%links(curSampNrs,p))
          ! add plotting commands
          call addGnupltCmds(curFilNmCSV,curFilNmPlt,uCaseFst(trim(TTL_PART%sampNo)),fmtPar(p,OPT_FOR_PLOT), &
           uCaseFst(trim(TTL_PART%trace))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclY,maPoLogScl(p)), &
           ticIncrX=merge(round(1._dp*chn%length/PLT_N_LBLS,1._dp,ROUND_MD%up),VAL_UNDEF_DBL,pltPSTricks), &
           rgeX=tyIntvlDbl(curStride,curNPts*curStride),rgeY=pltRge(priRge(p),maPoLogScl(p),priTypes(p)))
          call printPrgrs([VAL_UNDEF_INT],nTot=2*curNParsInfc)
        end if
      end do

          ! zoomed

      if (.NOT. singlePlts) &
       curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%trace)//"_"//FIL_NM_PART%zoom))//trim(pltFilExt)
      do p=1,curNPars
        if (parInfc(p)) then
          curFilNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%trace),parNms(p)))
          if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%trace)//"_"// &
           FIL_NM_PART%zoom,parNms(p)))//trim(pltFilExt)
          curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
          ! add plotting commands
          call addGnupltCmds(curFilNmCSV,curFilNmPlt,uCaseFst(trim(TTL_PART%sampNo)),fmtPar(p,OPT_FOR_PLOT), &
           uCaseFst(trim(TTL_PART%trace))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclY,maPoLogScl(p)), &
           ticIncrX=merge(round(1._dp*chn%length/PLT_N_LBLS,1._dp,ROUND_MD%up),VAL_UNDEF_DBL,pltPSTricks), &
           rgeX=tyIntvlDbl(curStride,curNPts*curStride), &
           rgeY=pltRge(tyIntvlDbl(minval(chn%links(curSampNrs,p)),maxval(chn%links(curSampNrs,p))), &
            maPoLogScl(p),priTypes(p)))
          call printPrgrs([VAL_UNDEF_INT],nTot=2*curNParsInfc)
        end if
      end do

      call endSubsect(UNIT%cons+UNIT%log) ! Traces


        ! -ln(prior·likelihood)+const.

      block
        real(dp) :: &
         pts(curNPts), &
         mi
        
        do i=1,curNPts
          pts(i)=real(-logPrior(chn%links(curSampNrs(i),:)),qp)-chn%logLikh(curSampNrs(i))
        end do
        mi=minval(pts)
        if (mi > 0._qp) then
          logPostrEvConst=0._qp
        else
          logPostrEvConst=1._qp-mi
          pts=pts+logPostrEvConst
        end if
      
        call startSubsect(UNIT%cons+UNIT%log,uCaseFst(TTL_PART%negLogPriLikh))

        curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(FIL_NM_PART%negLogPriLikh))//trim(pltFilExt)
        call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
        ! write CSV output
        curFilNmCSV=trim(fmtFilNmPlt(FIL_NM_PART%negLogPriLikh))//FIL_EXT_CSV
        call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),curSampNrs,pts)
        ! add plotting commands
        call addGnupltCmds(curFilNmCSV,curFilNmPlt,uCaseFst(trim(TTL_PART%sampNo)),trim(fmtPostr(VAL_UNDEF_CHR,OPT_TIMES_EV)), &
         uCaseFst(trim(TTL_PART%negLogPriLikh)//trim(mrg(""," plus const",logPostrEvConst == 0._qp))),setFlg(FLG%logSclY), &
         ticIncrX=merge(round(1._dp*chn%length/PLT_N_LBLS,1._dp,ROUND_MD%up),VAL_UNDEF_DBL,pltPSTricks), &
         rgeX=tyIntvlDbl(curStride,curNPts*curStride))
        call printPrgrs([VAL_UNDEF_INT],nTot=1)

        call endSubsect(UNIT%cons+UNIT%log) ! uCaseFst(TTL_PART%negLogPriLikh)
        
      end block

      deallocate(curSampNrs)


        ! scatter plots (only zoomed)

      if (usrScatterPlts(iNPlMd)) then

        call startSubsect(UNIT%cons+UNIT%log,"Scatter plots")

        curNPts=min(chn%length-chn%burnInLen,int(pltNPtsScatter,i8b))
        curStride=floor(real(chn%length-chn%burnInLen)/curNPts)
        allocate(curPtsX(curNPts),curPtsY(curNPts),stat=allocStat) ! numbers of displayed samples
        call testAllocErr(allocStat,21)

        if (.NOT. singlePlts) &
        curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt( &
        trim(FIL_NM_PART%scatter)//"_"//FIL_NM_PART%zoom))//trim(pltFilExt)
        call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
        do p=1,curNPars-1
          if (parInfc(p)) then
            curPtsX=chn%links(chn%burnInLen+curStride:chn%burnInLen+curNPts*curStride:curStride,p)
            do p2=p+1,curNPars
              if (parInfc(p2)) then
                curPtsY=chn%links(chn%burnInLen+curStride:chn%burnInLen+curNPts*curStride:curStride,p2)
                ! write CSV output
                curFilNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%scatter)//"_"//FIL_NM_PART%zoom,trim(parNms(p))//"-"// &
                 parNms(p2)))
                if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
                curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
                call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),curPtsX,curPtsY)
                ! add plotting commands
                call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtPar(p,OPT_FOR_PLOT),fmtPar(p2,OPT_FOR_PLOT), &
                 uCaseFst(trim(TTL_PART%scatter))//" of "//trim(parNmsPlt(p))//" and "//trim(parNmsPlt(p2)),NO_FLAGS, &
                 sym=PLT_SYM%point,styMod="pt 7 ps .25",rgeX=tyIntvlDbl(minval(curPtsX),maxval(curPtsX)), &
                 rgeY=tyIntvlDbl(minval(curPtsY),maxval(curPtsY)))
                call printPrgrs([VAL_UNDEF_INT],nTot=int(curNParsInfc*(curNParsInfc-1)/2.))
              end if
            end do
          end if
        end do
        deallocate(curPtsX,curPtsY)

        call endSubsect(UNIT%cons+UNIT%log) ! Scatter plots

      end if



        ! autocorrelation

      call startSubsect(UNIT%cons+UNIT%log,"Autocorrelation")

      if (autoCorrNLags(iNPlMd) < 2) then
        call printMsg(UNIT%cons+UNIT%log,"Skipped ("//trim(TTL_PART%autoCorr)//" has not been calculated more than once)")
      else
        if (.NOT. singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(FIL_NM_PART%autoCorr))//trim(pltFilExt)
        call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
        do p=1,curNPars
          if (parInfc(p)) then
            ! write CSV output
            curFilNmCSV=trim(fmtFilNmPlt(FIL_NM_PART%autoCorr,parNms(p)))
            if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
            curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),autoCorrLags(:autoCorrNLags(iNPlMd)), &
             abs(autoCorr(iNPlMd,p,:autoCorrNLags(iNPlMd))))
            ! add plotting commands
            call addGnupltCmds(curFilNmCSV,curFilNmPlt,uCaseFst(trim(TTL_PART%lag)),"Absolute value of "//trim(TTL_PART%autoCorr), &
             uCaseFst(trim(TTL_PART%autoCorr))//" of "//trim(parNmsPlt(p)),setFlg([FLG%logSclX,FLG%logSclY]))
            call printPrgrs([VAL_UNDEF_INT],nTot=curNParsInfc)
          end if
        end do
      end if

      call endSubsect(UNIT%cons+UNIT%log) ! Autocorrelation



        ! 2-d kernel-density-estimation contour plots

      if (usrJointMaPo(iNPlMd)) then

        call startSubsect(UNIT%cons+UNIT%log,"2-parameter joint marginal-posterior plots")

        curNPts=pltNPtsKDE

        call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
        do iZm=merge(1,0,usrZmRge(iNPlMd)),1 ! for normal and zoomed plots

          if (.NOT. singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%jointMaPo)// &
           trim(mrg("","_"//FIL_NM_PART%zoom,iZm == 0))))//trim(pltFilExt)

          do p=1,curNPars-1
            if (parInfc(p)) then
              do p2=p+1,curNPars
                if (parInfc(p2)) then

                  covM(1,1)=parEst(iNPlMd,p)%cov(p)
                  covM(1,2)=parEst(iNPlMd,p)%cov(p2)
                  covM(2,1)=covM(1,2)
                  covM(2,2)=parEst(iNPlMd,p2)%cov(p2)

                  if (iZm == 0) then ! normal

                    redoKDE=.FALSE.

                    call kerDensEst2D(dens2DBiWeight,I_KER_2D%biWeight, &
                     chn%links(chn%burnInLen+1:chn%length,p),chn%links(chn%burnInLen+1:chn%length,p2), &
                     covM,priRge(p),priRge(p2),int(curNPts,i8b),kerDensEst,.NOT. OPT_EXTEND_RGE, &
                     zmCont=PROB_CONT_CONTOUR(N_CONTOURS),xRge1Zm=xRge1Zm(p,p2),xRge2Zm=xRge2Zm(p,p2))

                    xRge1(p,p2)=tyIntvlDbl(kerDensEst(1)%x,kerDensEst(curNPts**2)%x)
                    xRge2(p,p2)=tyIntvlDbl(kerDensEst(1)%y,kerDensEst(curNPts**2)%y)

                  else               ! zoomed

                    redoKDE=isDef(xRge1Zm(p,p2)) .OR. isDef(xRge2Zm(p,p2)) .OR. usrZmRge(iNPlMd)

                    if (redoKDE) then

                      ! ranges

                      if (usrZmRge(iNPlMd)) then

                        curZmRge1=zmRgeAllPars(iNPlMd,p)
                        curZmRge2=zmRgeAllPars(iNPlMd,p2)

                      else

                        if (isDef(xRge1Zm(p,p2))) then
                          curZmRge1=xRge1Zm(p,p2)
                        else
                          curZmRge1=priRge(p)
                        end if

                        if (isDef(xRge2Zm(p,p2))) then
                          curZmRge2=xRge2Zm(p,p2)
                        else
                          curZmRge2=priRge(p2)
                        end if

                      end if


                      call kerDensEst2D(dens2DBiWeight,I_KER_2D%biWeight, &
                       chn%links(chn%burnInLen+1:chn%length,p),chn%links(chn%burnInLen+1:chn%length,p2), &
                       covM,curZmRge1,curZmRge2,int(curNPts,i8b),kerDensEst,.NOT. OPT_EXTEND_RGE)

                      xRge1(p,p2)=tyIntvlDbl(kerDensEst(1)%x,kerDensEst(curNPts**2)%x)
                      xRge2(p,p2)=tyIntvlDbl(kerDensEst(1)%y,kerDensEst(curNPts**2)%y)

                    end if

                  end if


                  if (iZm == 0 .OR. redoKDE) then

                    ! find contour heights

                    contH(p,p2,:)=dens2DContourHeight(PROB_CONT_CONTOUR/merge(1._dp,PROB_CONT_CONTOUR(N_CONTOURS), &
                     iZm == 0),int(curNPts,i8b),kerDensEst) ! correcting for the prob. content of the whole plot (PROB_CONT_CONTOUR(N_CONTOURS))

                    ! file names

                    curFilNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%jointMaPo)//trim(mrg("","_"//FIL_NM_PART%zoom,iZm == 0)), &
                     trim(parNms(p))//"-"//parNms(p2)))
                    if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%jointMaPo)// &
                     trim(mrg("","_"//FIL_NM_PART%zoom,iZm == 0)),trim(parNms(p))//"-"//parNms(p2)))//trim(pltFilExt)
                    curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV

                    call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),kerDensEst%x,kerDensEst%y,kerDensEst%z,vSpc=.TRUE.)

                    ! add plotting commands

                    call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtPar(p,OPT_FOR_PLOT),fmtPar(p2,OPT_FOR_PLOT), &
                     uCaseFst(trim(TTL_PART%jointMaPo))//" of "//trim(parNmsPlt(p))//" and "//trim(parNmsPlt(p2)), &
                     setFlg([FLG%noSty,FLG%splot]),rgeX=xRge1(p,p2),rgeY=xRge2(p,p2),levels=contH(p,p2,:))

                  end if


                  call printPrgrs([VAL_UNDEF_INT],nTot=int(curNParsFree*(curNParsFree-1)*merge(.5,1.,usrZmRge(iNPlMd))))

                end if               ! (.NOT. parFixed(p2))

              end do                 ! p2=p+1,curNPars
            end if                   !  (.NOT. parFixed(p))
          end do                     ! p=1,curNPars-1

        end do                       ! iZm=0,1

        call endSubsect(UNIT%cons+UNIT%log) ! 2-parameter joint marginal-posterior plots

      end if                         ! (usrJointMaPo(iNPlMd))



    else                             ! after sorting

        ! PSR

      call startSubsect(UNIT%cons+UNIT%log,uCaseFst(TTL_PART%PSR))

      if (nPSR < 2) then

        call printMsg(UNIT%cons+UNIT%log,"Skipped ("//trim(TTL_PART%PSR)//" has not been calculated more than once)")

      else

        if (.NOT. singlePlts) &
         curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(FIL_NM_PART%PSR))//trim(pltFilExt)
        call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
        do p=1,curNPars
          if (.NOT. parFixed(p)) then
            ! write CSV output
            curFilNmCSV=trim(fmtFilNmPlt(FIL_NM_PART%PSR,parNms(p)))
            if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
            curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),PSR(:nPSR,p)%x,PSR(:nPSR,p)%y-1._dp)
            ! add plotting commands
            call addGnupltCmds(curFilNmCSV,curFilNmPlt,uCaseFst(trim(TTL_PART%sampNo)),"PSR-1", &
             uCaseFst(trim(TTL_PART%PSR))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclY),sym=PLT_SYM%point, &
             ticIncrX=merge(round(1._dp*chn%length/PLT_N_LBLS,1._dp,ROUND_MD%up),VAL_UNDEF_DBL,pltPSTricks))
            call printPrgrs([VAL_UNDEF_INT],nTot=curNParsFree)
          end if
        end do

      end if

      call endSubsect(UNIT%cons+UNIT%log) ! PSR


        ! marginal posteriors

      if (usrMaPo(iNPlMd)) then

        call startSubsect(UNIT%cons+UNIT%log,"Marginal posteriors")

        call startSubsect(UNIT%cons+UNIT%log,"Parameters")


            ! normal

        if (.NOT. usrZmRge(iNPlMd)) then
          if (usrPeriodogram) then
            curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(FIL_NM_PART%maPo))//trim(pltFilExt)
            p=pRole%f
            
                  ! linear f scale

            ! write CSV output
            curFilNmCSV=trim(fmtFilNmPlt(FIL_NM_PART%maPo,parNms(p)))
            if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
            curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),periodogramPts(:nPtsPeriodogram)%x,periodogramPts(:nPtsPeriodogram)%y)
            ! add plotting commands
            call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtPar(p,OPT_FOR_PLOT),trim(fmtPostr(parNmsPlt(p))), &
             uCaseFst(trim(TTL_PART%maPo))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclX,maPoLogScl(p)), &
             rgeX=pltRge(shiftIntvl(priRge(p),maPoShft(p)),maPoLogScl(p),priTypes(p)))
            
                  ! logarithmic f scale

            curFilNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%maPo)//"_log",parNms(p)))
            if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
            curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),periodogramPtsLog(:nPtsPeriodogram)%x, &
             periodogramPtsLog(:nPtsPeriodogram)%y)
            ! add plotting commands
            call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtPar(p,OPT_FOR_PLOT),trim(fmtPostr(parNmsPlt(p))), &
             uCaseFst(trim(TTL_PART%maPo))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclX), &
             rgeX=pltRge(shiftIntvl(priRge(p),maPoShft(p)),.TRUE.,priTypes(p)))
            call printMsg(UNIT%cons+UNIT%log,"Done")
          else
            if (.NOT. singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(FIL_NM_PART%maPo))//trim(pltFilExt)
            call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
            do p=1,curNPars
              if (parInfc(p)) then
                ! write CSV output
                curFilNmCSV=trim(fmtFilNmPlt(FIL_NM_PART%maPo,parNms(p)))
                if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
                curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
                call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),maPoPts(p,:pltNPts)%x,maPoPts(p,:pltNPts)%y)
                ! add plotting commands
                call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtPar(p,OPT_FOR_PLOT),trim(fmtPostr(parNmsPlt(p))), &
                 uCaseFst(trim(TTL_PART%maPo))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclX,maPoLogScl(p)), &
                 rgeX=pltRge(shiftIntvl(priRge(p),maPoShft(p)),maPoLogScl(p),priTypes(p)))
                 call printPrgrs([VAL_UNDEF_INT],nTot=2*curNParsInfc)
              end if
            end do
          end if
        end if

            ! zoomed

        if (.NOT. usrPeriodogram) then
          if (.NOT. singlePlts) &
           curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%maPo)//"_"//FIL_NM_PART%zoom))//trim(pltFilExt)
          do p=1,curNPars
            if (parInfc(p)) then
              ! write CSV output
              curFilNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%maPo)//"_"//FIL_NM_PART%zoom,parNms(p)))
              if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
              curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
              call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),maPoPtsZm(p,:pltNPts)%x,maPoPtsZm(p,:pltNPts)%y)
              ! add plotting commands
              call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtPar(p,OPT_FOR_PLOT),trim(fmtPostr(parNmsPlt(p))), &
               uCaseFst(trim(TTL_PART%maPo))//" of "//trim(parNmsPlt(p)),setFlg(FLG%logSclX,maPoLogScl(p)), &
               rgeX=pltRge(merge(zmRgeAllPars(iNPlMd,p), &
               parEst(iNPlMd,p)%HPDI(iStdHPDI(I_STD_HPDI_MA_PO_ZM)),usrZmRge(iNPlMd)),maPoLogScl(p),priTypes(p)))
              call printPrgrs([VAL_UNDEF_INT],nTot=merge(1,2,usrZmRge(iNPlMd))*curNParsInfc)
            end if
          end do
        end if

        call endSubsect(UNIT%cons+UNIT%log) ! Parameters


          ! marginal posteriors (DQs)

        if (curNDQs > 0 .AND. .NOT. usrPeriodogram) then
          call startSubsect(UNIT%cons+UNIT%log,"Derived quantities")


              ! normal

          if (.NOT. usrZmRge(iNPlMd)) then
            if (.NOT. singlePlts) &
             curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%maPoDQ)))//trim(pltFilExt)
            call printPrgrs([VAL_UNDEF_INT],reset=.TRUE.)
            do q=1,curNDQs
              if (.NOT. DQSkipped(q)) then
                ! write CSV output
                curFilNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%maPoDQ),DQNms(q)))
                if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
                curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
                call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),maPoDQPts(q,:pltNPts)%x,maPoDQPts(q,:pltNPts)%y)
                ! add plotting commands
                call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtDQ(q,OPT_FOR_PLOT),trim(fmtPostr(DQNmsPlt(q))), &
                 uCaseFst(trim(TTL_PART%maPo))//" of "//trim(DQNmsPlt(q)),setFlg(FLG%logSclX,maPoLogSclDQ(q)), &
                 rgeX=pltRge(DQSampRge(q),maPoLogSclDQ(q)))
              end if
              call printPrgrs([1,q],[1,1],[2,curNDQs])
            end do
          end if


              ! zoomed

          if (.NOT. singlePlts) &
           curFilNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(FIL_NM_PART%maPoDQ)//"_"//FIL_NM_PART%zoom))//trim(pltFilExt)
          do q=1,curNDQs
            if (.NOT. DQSkipped(q)) then
              ! write CSV output
              curFilNmCSV=trim(fmtFilNmPlt(trim(FIL_NM_PART%maPoDQ)//"_"//FIL_NM_PART%zoom,DQNms(q)))
              if (singlePlts) curFilNmPlt=REL_DIR_PLT_CSV//trim(curFilNmCSV)//trim(pltFilExt)
              curFilNmCSV=trim(curFilNmCSV)//FIL_EXT_CSV
              call writeCSV(trim(outDirCSV)//trim(curFilNmCSV),maPoDQPtsZm(q,:pltNPts)%x,maPoDQPtsZm(q,:pltNPts)%y)
              ! add plotting commands
              call addGnupltCmds(curFilNmCSV,curFilNmPlt,fmtDQ(q,OPT_FOR_PLOT),trim(fmtPostr(DQNmsPlt(q))), &
               uCaseFst(trim(TTL_PART%maPo))//" of "//trim(DQNmsPlt(q)),setFlg(FLG%logSclX,maPoLogSclDQ(q)), &
               rgeX=DQZmRge(q))
            end if
            call printPrgrs([merge(1,2,usrZmRge(iNPlMd)),q],[1,1],[merge(1,2,usrZmRge(iNPlMd)),curNDQs])
          end do

          call endSubsect(UNIT%cons+UNIT%log) ! Derived quantities
        end if

        call endSubsect(UNIT%cons+UNIT%log) ! Marginal posteriors

      end if                         ! (usrMaPo(iNPlMd))


        ! data and model

      if (dtProps%nRec > 0 .AND. .NOT. usrPeriodogram) then

        block
          type(tyCoord2DDbl) :: &
           refPos
          
          if (isDef(pRole%deltaR) .AND. isDef(pRole%alphaR)) then
            refPos%x=parEst(iNPlMd,pRole%deltaR)%bestSumm
            refPos%y=parEst(iNPlMd,pRole%alphaR)%bestSumm*cos(refPos%x*DEG_IN_RAD)
          else
            call undef(refPos)
          end if
        
          call startSubsect(UNIT%cons+UNIT%log,"Data and model")

          call plotDataAndModel(NO_FLAGS,refPos)
          if (psdNPl == 1) call plotDataAndModel(setFlg(FLG%fold),refPos)

          call printMsg(UNIT%cons+UNIT%log,"Done")
          call endSubsect(UNIT%cons+UNIT%log) ! Data and model


            ! plot residuals

          call startSubsect(UNIT%cons+UNIT%log,uCaseFst(trim(TTL_PART%resid)))

          call plotDataAndModel(setFlg(FLG%resid),refPos)
          call plotDistrNormResid

          call printMsg(UNIT%cons+UNIT%log,"Done")
          call endSubsect(UNIT%cons+UNIT%log) ! Residuals
          
        end block

      end if


        ! predictions


      if (usrPred) then
        call startSubsect(UNIT%cons+UNIT%log,"Predictions")
        call plotPred(DT_TY%AMAngPos,pred2D=pred2D,pltNmPart=FIL_NM_PART%predict)
        call plotPred(DT_TY%RV,pred1D=pred1D,pltNmPart=FIL_NM_PART%predict)
        call printMsg(UNIT%cons+UNIT%log,"Done")
        call endSubsect(UNIT%cons+UNIT%log) ! Predictions
      end if

    end if                           ! (befSort)

    call endSubsect(UNIT%cons+UNIT%log) ! Calculating and writing plot points


    contains

    subroutine plotPred(dtTy,pred1D,pred2D,pltNmPart) ! plots uncertainty predictions
      implicit none
      integer :: &
       iSet, &
       dtTy
      character(STR_LEN_STD) :: &
       lblX, &
       lblY
      character(*) :: &
       pltNmPart
      character(MAX_LEN_PATH) :: &
       filNmCSV, &
       filNmCSVA, &
       filNmCSVB, &
       filNmPlt
      type(tyPred1D) :: &
       pred1D
      type(tyPred2D) :: &
       pred2D
      optional :: &
       pred1D, &
       pred2D, &
       pltNmPart
      intent(in) :: &
       dtTy, &
       pred1D, &
       pred2D, &
       pltNmPart

      select case (dtTy)
      case (DT_TY%AMAngPos)

        lblX=dateDescAndUnitPlt
        lblY=fmtObsvble(DT_TY%AMAngPos,VAL_UNDEF_INT,"Uncertainty")

        do iSet=1,dtProps%nSetsOfTy(dtTy)
          if (.NOT. pred2D%skipSet(dtTy,iSet)) then
            ! write CSV output

            filNmCSVA=trim(fmtFilNmPlt(trim(pltNmPart)//"_"//FIL_NM_PART%a//"_"//trim(DT_TY_NMS(dtTy))// &
            trim(fmtDtSet(DT_TY%AMAngPos,iSet))))//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(filNmCSVA),tOfs+pred2D%t(dtTy,iSet,:),pred2D%unc(dtTy,iSet,:)%a)

            filNmCSVB=trim(fmtFilNmPlt(trim(pltNmPart)//"_"//FIL_NM_PART%b//"_"//trim(DT_TY_NMS(dtTy))// &
            trim(fmtDtSet(DT_TY%AMAngPos,iSet))))//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(filNmCSVB),tOfs+pred2D%t(dtTy,iSet,:),pred2D%unc(dtTy,iSet,:)%b)

            filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(pltNmPart)//"_"//trim(DT_TY_NMS(dtTy))// &
            trim(fmtDtSet(DT_TY%AMAngPos,iSet))))//pltFilExt

            call addGnupltCmds("",filNmPlt,trim(lblX),trim(lblY),uCaseFst(trim(TTL_PART%predict))//" ("// &
             trim(TTL_PART%uncEllAx)//")",NO_FLAGS,filNmInArr=[filNmCSVA,filNmCSVB],sym=PLT_SYM%line)


            filNmCSV=trim(fmtFilNmPlt(trim(pltNmPart)//"_"//FIL_NM_PART%phi//"_"//trim(DT_TY_NMS(dtTy))// &
            trim(fmtDtSet(DT_TY%AMAngPos,iSet))))//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(filNmCSV),tOfs+pred2D%t(dtTy,iSet,:),pred2D%unc(dtTy,iSet,:)%phi)

            filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(pltNmPart)//"_"//trim(DT_TY_NMS(dtTy))//"_"//FIL_NM_PART%phi// &
            trim(fmtDtSet(DT_TY%AMAngPos,iSet))))//pltFilExt

            call addGnupltCmds(filNmCSV,filNmPlt,trim(lblX),"Position angle [deg]",uCaseFst(trim(TTL_PART%predict))//" ("// &
             trim(TTL_PART%uncEllPosAng)//")",NO_FLAGS,sym=PLT_SYM%line)
          end if                     ! (.NOT. pred2D%skipSet(iSet))

        end do


      case (DT_TY%RV)

        lblX=dateDescAndUnitPlt
        lblY=fmtObsvble(DT_TY%RV,1,"Uncertainty")

        do iSet=1,dtProps%nSetsOfTy(dtTy)
          if (.NOT. pred1D%skipSet(dtTy,iSet)) then

            ! write CSV output

            filNmCSV=trim(fmtFilNmPlt(trim(pltNmPart)//"_"//trim(DT_TY_NMS(dtTy))//trim(fmtDtSet(DT_TY%RV,iSet))))//FIL_EXT_CSV
            call writeCSV(trim(outDirCSV)//trim(filNmCSV),tOfs+pred1D%t(dtTy,iSet,:),pred1D%unc(dtTy,iSet,:))

            filNmPlt=REL_DIR_PLT_CSV//trim(fmtFilNmPlt(trim(pltNmPart)//"_"//trim(DT_TY_NMS(dtTy))// &
            trim(fmtDtSet(DT_TY%RV,iSet))))//pltFilExt

            call addGnupltCmds(filNmCSV,filNmPlt,trim(lblX),trim(lblY),uCaseFst(trim(TTL_PART%predict)),NO_FLAGS,sym=PLT_SYM%line)

          end if                     ! (.NOT. pred1D%skipSet(iSet))
        end do

      end select
    end subroutine

  end subroutine makePostrPlots




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtPostr(parNm,timesEv)
!
!   Returns:                         formatted string, depending on args., for:
!                                      * marginal posterior of a parameter
!                                      * posterior times evidence
!
!   Arguments:
!     parNm                          par. name (ignored if timesEv)
!     timesEv                        whether the second kind of result is requested
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     timesEv
    character(*) :: &
     parNm
    character(STR_LEN_STD) :: &
     fmtPostr, &
     cConst
    optional :: &
     timesEv
    intent(in) :: &
     parNm, &
     timesEv

    if (present(timesEv) .AND. timesEv) then
      cConst=mrg("","+"//trim(niceFmt(logPostrEvConst)),logPostrEvConst == 0._qp)
      if (pltPSTricks) then
        fmtPostr="$-ln(p(\mathbf\theta|D,\mathcal M,K)\cdot p(D|\mathcal{M,K}))"//trim(cConst)//"$"
      else
        fmtPostr="-ln(p(theta|D,M,K)*p(D|M,K))"//trim(cConst)
      end if
    else
      if (pltPSTricks) then
        fmtPostr="$p($"//trim(parNm)//"$|D,\mathcal{M,K})$"
      else
        fmtPostr="p("//trim(parNm)//"|D,M,K)"
      end if
    end if

  end function fmtPostr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtFilNmPlt(filNmPart,parNm)
!
!   Returns:                         formatted string to be used in file name for plotting
!
!   Arguments:
!     filNmPart                      first part of the file name after the run (and #-pl. mode) prefix
!     parNm                          par. name
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     filNmPart, &
     parNm
    character(MAX_LEN_PATH) :: &
     fmtFilNmPlt
    optional :: &
     parNm
    intent(in) :: &
     filNmPart, &
     parNm

    fmtFilNmPlt=trim(curFilNmPrefx)//"_"//fmtFilNm(trim(filNmPart)//trim(mrg("_"//parNm,"",present(parNm))))

  end function fmtFilNmPlt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine addGnupltCmds(filNmIn,filNmOut,lblX,lblY,pltTitle,flags, &
     filNmInArr,rgeX,rgeY,sym,symArr,titleArr,modArr,styMod,styModArr,ticIncrX,ticIncrY,levels)
!
!   adds Gnuplot commands to corresponding command file
!
!   Arguments:
!     filNmIn                        data file name
!     filNmOut                       name of Gnuplot command file
!     lblX                           x-axis label
!     lblY                           y-axis label
!     pltTitle                       plot title
!     flags                          option flags
!     filNmInArr                     data file names
!     rgeX                           x range
!     rgeY                           y range
!     sym                            plotting symbol
!     symArr                         plotting symbols
!     titleArr                       curve titles
!     modArr                         data modifiers
!     styMod                         style modifier
!     styModArr                      style modifiers
!     ticIncrX                       x-tic increment
!     ticIncrY                       y-tic increment
!     levels                         contour levels 
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i, &
     j, &
     flags, &
     sym, &
     symArr(:), &
     nFilNmSav=0
    real(dp) :: &
     ticIncrX, &
     ticIncrY, &
     levels(:)
    logical :: &
     addFilOut, &
     splot, &
     noSty, &
     fstCall=.TRUE., &
     lstSplot=.FALSE., &
     lstEqAxR=.FALSE.
    character(*) :: &
     filNmIn, &
     filNmOut, &
     lblX, &
     lblY, &
     pltTitle, &
     filNmInArr(:), &
     modArr(:), &
     styModArr(:), &
     styMod, &
     titleArr(:)
    character(MAX_LEN_PATH) :: &
     lstFilNmOut="", &
     cTmp, &
     filNmSav(N_FILS_OUT_MAX)=""
    character(STR_LEN_LONGEST) :: &
     cmdL, &
     tmpStrLvl
    type(tyIntvlDbl) :: &
     rgeX, &
     rgeY
    optional :: &
     filNmInArr, &
     modArr, &
     styModArr, &
     styMod, &
     titleArr, &
     rgeX, &
     rgeY, &
     sym, &
     symArr, &
     ticIncrX, &
     ticIncrY, &
     levels
    save :: &
     lstFilNmOut, &
     fstCall, &
     lstSplot, &
     lstEqAxR, &
     filNmSav, &
     nFilNmSav
    intent(in) :: &
     filNmIn, &
     filNmOut, &
     lblX, &
     lblY, &
     pltTitle, &
     flags, &
     filNmInArr, &
     rgeX, &
     rgeY, &
     sym, &
     symArr, &
     titleArr, &
     modArr, &
     styMod, &
     styModArr, &
     ticIncrX, &
     ticIncrY, &
     levels

    splot=getFlg(flags,FLG%splot)
    noSty=getFlg(flags,FLG%noSty)

    if (getFlg(flags,FLG%mkTex)) then ! produce TeX file for PSTricks output

      call tryOpenFilRW("LaTeX",filNmOut,UN_TEX,"replace","write")

      ! header
      do i=1,size(TEX_CMDS_PRE_ONCE)
        write(UN_TEX,"(A)") trim(TEX_CMDS_PRE_ONCE(i))
      end do

      ! include all plotted files
      do i=1,nFilNmSav
        cTmp=filNmSav(i)
        cTmp=cTmp(1:len_trim(cTmp)-len(SUFFX_TEX))
        do j=1,size(TEX_CMDS_PRE)
          write(UN_TEX,"(A)") trim(TEX_CMDS_PRE(j))
        end do
        write(UN_TEX,"(A)") "    \include{"//trim(cTmp)//"}"
        do j=1,size(TEX_CMDS_POST)
          write(UN_TEX,"(A)") trim(TEX_CMDS_POST(j))
        end do
      end do

      ! footer
      do i=1,size(TEX_CMDS_POST_ONCE)
        write(UN_TEX,"(A)") trim(TEX_CMDS_POST_ONCE(i))
      end do

      close(UN_TEX)

    else                             ! add Gnuplot plotting commands

      call inc(nFilNmSav)
      filNmSav(nFilNmSav)=filNmOut

      if (trim(adjustl(filNmOut)) == trim(adjustl(lstFilNmOut))) then
        addFilOut=.FALSE.
      else
        addFilOut=.TRUE.
        lstFilNmOut=filNmOut
      end if

      if (fstCall) then
        if (pltPSTricks) then
          do i=1,size(GNUPLT_STD_CMDS_ONCE_PSTRICKS)
            write(UN_GNUPLT,"(A)") trim(GNUPLT_STD_CMDS_ONCE_PSTRICKS(i))
          end do
        else
          do i=1,size(GNUPLT_STD_CMDS_ONCE_NORM)
            write(UN_GNUPLT,"(A)") trim(GNUPLT_STD_CMDS_ONCE_NORM(i))
          end do
        end if
        do i=1,size(GNUPLT_STD_CMDS_ONCE_ALL)
          write(UN_GNUPLT,"(A)") trim(GNUPLT_STD_CMDS_ONCE_ALL(i))
        end do
        write(UN_GNUPLT,*)
        fstCall=.FALSE.
      end if

      if (getFlg(flags,FLG%logSclX)) then
        write(UN_GNUPLT,"(A)") "set log x"
      else
        write(UN_GNUPLT,"(A)") "unset log x"
      end if
      if (getFlg(flags,FLG%logSclY)) then
        write(UN_GNUPLT,"(A)") "set log y"
      else
        write(UN_GNUPLT,"(A)") "unset log y"
      end if

      do i=1,size(GNUPLT_STD_CMDS)
        write(UN_GNUPLT,"(A)") trim(GNUPLT_STD_CMDS(i))
      end do

      if (getFlg(flags,FLG%eqAxR)) then
        write(UN_GNUPLT,"(A)") "set size ratio -1"
        lstEqAxR=.TRUE.
      else if (lstEqAxR) then
        write(UN_GNUPLT,"(A)") "set size noratio"
        lstEqAxR=.FALSE.
      end if

      if (splot) then                ! "3-d" plot
        if (.NOT. lstSplot) then
          do i=1,size(GNUPLT_CMDS_ONCE_SPLOT)
            write(UN_GNUPLT,"(A)") trim(GNUPLT_CMDS_ONCE_SPLOT(i))
          end do
          lstSplot=.TRUE.
        end if
        do i=1,size(GNUPLT_CMDS_SPLOT)
          write(UN_GNUPLT,"(A)") trim(GNUPLT_CMDS_SPLOT(i))
        end do
        if (present(levels) .AND. .NOT. any(isUndef(levels))) then
          tmpStrLvl="set cntrpar levels discrete"
          do i=1,size(levels)
            tmpStrLvl=trim(tmpStrLvl)//merge(" ",",",i == 1)//niceFmt(levels(i))
          end do
          write(UN_GNUPLT,"(A)") trim(tmpStrLvl)
        end if
      else if (lstSplot) then
        do i=1,size(GNUPLT_CMDS_POST_SPLOT)
          write(UN_GNUPLT,"(A)") trim(GNUPLT_CMDS_POST_SPLOT(i))
        end do
        lstSplot=.FALSE.
      end if

      write(UN_GNUPLT,"(A)") "set title '"//trim(pltTitle)//"'"
      if (present(rgeX)) write(UN_GNUPLT,"(A)") "set xrange ["//trim(niceFmt(rgeX%lBnd))//":"//trim(niceFmt(rgeX%uBnd))//"]"
      if (present(rgeY)) write(UN_GNUPLT,"(A)") "set yrange ["//trim(niceFmt(rgeY%lBnd))//":"//trim(niceFmt(rgeY%uBnd))//"]"
      write(UN_GNUPLT,"(A)") "set xrange [] "//trim(merge("  ","no",getFlg(flags,FLG%rvX)))//"reverse"
      write(UN_GNUPLT,"(A)") "set yrange [] "//trim(merge("  ","no",getFlg(flags,FLG%rvY)))//"reverse"
      if (present(ticIncrX) .AND. isDef(ticIncrX)) write(UN_GNUPLT,"(A)") "set xtics "//trim(niceFmt(ticIncrX))
      if (present(ticIncrY) .AND. isDef(ticIncrY)) write(UN_GNUPLT,"(A)") "set ytics "//trim(niceFmt(ticIncrY))
      write(UN_GNUPLT,"(A)") "set xlabel '"//trim(lblX)//"'"
      write(UN_GNUPLT,"(A)") "set ylabel '"//trim(lblY)//"'"
      if (addFilOut) write(UN_GNUPLT,"(A)") "set output '"//trim(filNmOut)//"'"

      if (present(filNmInArr)) then  ! plot several files
        if (present(titleArr)) write(UN_GNUPLT,"(A)") "set key box"
        cmdL=trim(merge("s"," ",splot))//"plot"
        do i=1,size(filNmInArr)
          cmdL=trim(cmdL)//merge(" ",",",i == 1)//"'"//trim(filNmInArr(i))//"'"
          if (present(modArr)) cmdL=trim(cmdL)//" "//trim(modArr(i))
          if (present(titleArr)) cmdL=trim(cmdL)//" title '"//trim(titleArr(i))//"'"
          if (.NOT. (noSty)) then
            cmdL=trim(cmdL)//" with"
            if (present(symArr)) then
              cmdL=trim(cmdL)//" "//PLT_SYM_NMS(symArr(i))
            else if (present(sym)) then
              cmdL=trim(cmdL)//" "//PLT_SYM_NMS(sym)
            else
              cmdL=trim(cmdL)//" "//PLT_SYM_NMS(1)
            end if
          end if
          if (present(styModArr)) then
            cmdL=trim(cmdL)//" "//styModArr(i)
          else if (present(styMod)) then
            cmdL=trim(cmdL)//" "//styMod
          end if
        end do
        write(UN_GNUPLT,"(A)") trim(cmdL)
        if (present(titleArr)) write(UN_GNUPLT,"(A)") "unset key"
      else
        cmdL=trim(merge("s"," ",splot))//"plot '"//trim(filNmIn)//"'"
        if (.NOT. (noSty)) then
          if (present(sym)) then
            cmdL=trim(cmdL)//" w "//trim(PLT_SYM_NMS(sym))
          else
            cmdL=trim(cmdL)//" w "//trim(PLT_SYM_NMS(1))
          end if
        end if
        if (present(styMod)) cmdL=trim(cmdL)//" "//styMod
        write(UN_GNUPLT,"(A)") trim(cmdL)
      end if
      write(UN_GNUPLT,*)

    end if                           ! (getFlg(flags,FLG%mkTex))

  end subroutine addGnupltCmds




end module
